#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)assembler_sparc.cpp	1.207 07/07/19 19:08:28 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_assembler_mips.cpp.incl"

// Implementation of Address

address Assembler::locate_next_instruction(address inst)
{
  // Note this doesn't work for branches
  assert(sizeof(*inst) == 1, "otherwise need to redo the addition");
  return inst + BytesPerInstWord;
}

Address Argument::as_callee_address() const {
  return Address(FP, (number() - n_register_parameters) * BytesPerLong);
}

Address Argument::as_caller_address() const {
  return Address(SP, (number() - n_register_parameters) * BytesPerLong);
}

static const char* argumentNames[] = {
  "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8", "A9",
  "A(n>9)"
};

const char* Argument::name() const {
  int nofArgs = sizeof argumentNames / sizeof argumentNames[0];
  int num = number();
  if (num >= nofArgs)  num = nofArgs - 1;
  return argumentNames[num];
}

void Assembler::print_instruction(int inst) {
  const char* s;
  switch (inv_op(inst)) {
  case op_beq:         s = "beq";    break;
  case op_bgtz:        s = "bgtz";   break;
  case op_blez:        s = "blez";   break;
  case op_bne:         s = "bne";    break;
  case op_regimm:
    switch (inv_rt(inst)) {
    case rtop_bgez:    s = "bgez";   break;
    case rtop_bgezal:  s = "bgezal"; break;
    case rtop_bltz:    s = "bltz";   break;
    case rtop_bltzal:  s = "bltzal"; break;
    default:           s = "????";   break;
    }
    break;
  default:             s = "????";   break;
  }
  ::tty->print("%s", s);
}


int AbstractAssembler::code_fill_byte() {
  // We'd like to return a byte that implies an illegal instruction,
  // like on SPARC, but the encodings don't make that easy.
  // Breakpoints need the first 6 bits zero and the last 5 bits 13.
  return 0x00;			// nop
}


// Implementation of MacroAssembler

void MacroAssembler::null_check(Register reg, int offset) {
  if (needs_explicit_null_check((intptr_t)offset)) {
    // provoke OS NULL exception if reg = NULL by
    // accessing M[reg] w/o changing any registers
    ld_ptr(R0, 0, reg);
  }
  else {
    // nothing to do, (later) access of M[reg + offset]
    // will provoke OS NULL exception if reg = NULL
  }
}


Assembler::Condition Assembler::negateCondition(Condition cond) {
  switch (cond) {
  case never:
    return always;
    break;
  case equal:
    return notEqual;
    break;
  case less:
    return greaterEqual;
    break;
  case lessUnsigned:
    return greaterEqualUnsigned;
    break;
  case lessEqual:
    return greater;
    break;
  case lessEqualUnsigned:
    return greaterUnsigned;
    break;
  case always:
    return never;
    break;
  case notEqual:
    return equal;
    break;
  case greater:
    return lessEqual;
    break;
  case greaterUnsigned:
    return lessEqualUnsigned;
    break;
  case greaterEqual:
    return less;
    break;
  case greaterEqualUnsigned:
    return lessUnsigned;
    break;
  default:
    ShouldNotReachHere();
    break;
  }
}

void Assembler::pd_patch_instruction(address branch, address target) {
  jint& stub_inst = *(jint*) branch;

#ifdef ASSERT
  bool ok_low16 = false;
  int op = inv_op(stub_inst);
  if (op == op_regimm) {
    switch (inv_rt(stub_inst)) {
    case rtop_bgez:
    case rtop_bgezal:
    case rtop_bltz:
    case rtop_bltzal:
      ok_low16 = true;
      break;
    default:
      // ok_low16 remains false
      break;
    }
  } else if (op == op_cop1) {
    if (inv_rs(stub_inst) == rsop_bc) {
      ok_low16 = true;
    }
  } else {
    switch (op) {
    case op_beq:
    case op_bgtz:
    case op_blez:
    case op_bne:
      ok_low16 = true;
      break;
    default:
      // ok_low16 remains false
      break;
    }
  }
  assert(ok_low16, "pd_patch_instruction applied to illegal instruction");
#endif

  int field = (target - branch - BytesPerInstWord) >> 2;
  assert(fits_in_s_field(field, 15, 0), "offset too large for branch");
  stub_inst = (stub_inst & ~fmask(15, 0)) | ximmed(field);
}

#ifndef PRODUCT
void Assembler::pd_print_patched_instruction(address branch) {
  jint stub_inst = *(jint*) branch;
  print_instruction(stub_inst);
  ::tty->print("%s", " (unresolved)");
}
#endif

// synthetic instructions

void MacroAssembler::addiu(Register rd, Register rs, int imm32) {
  if (fits_in_s_field(imm32, 15, 0)) {
    Assembler::addiu(rd, rs, imm32);
  } else {
    assert_not_delayed();
    li(AT, imm32);
    addu(rd, rs, AT);
  }
}

void MacroAssembler::daddiu(Register rd, Register rs, long imm64) {
  if (fits_in_s_lfield(imm64, 15, 0)) {
    Assembler::daddiu(rd, rs, imm64);
  } else {
    assert_not_delayed();
    dli(AT, imm64);
    daddu(rd, rs, AT);
  }
}

void MacroAssembler::andi(Register rd, Register rs, long imm32_64) {
  if (fits_in_u_lfield(imm32_64, 15, 0)) {
    Assembler::andi(rd, rs, imm32_64);
  } else {
    assert_not_delayed();
    li(AT, imm32_64);
    and_(rd, rd, AT);
  }
}

static int mask16(int x) {
  return x & 0xFFFF;
}

// MUST NOT USE AT
// Note this is what is commonly referred to as dli in MIPS asm.
// YYY Should fix this name at some point to avoid confusion.
// YYY But there are a lot of references to li that should be dli.

#ifdef _LP64
void MacroAssembler::li(Register rd, long imm) {
  if (imm <= max_jint && imm >= min_jint) {
    li32(rd, (int)imm);
  } else if (julong(imm) <= 0xFFFFFFFF) {
    assert_not_delayed();
    // lui sign-extends, so we can't use that.
    ori(rd, R0, julong(imm) >> 16);
    dsll(rd, rd, 16);
    ori(rd, rd, mask16(imm));
  } else {
    assert_not_delayed();
    lui(rd, imm >> 48);
    ori(rd, rd, mask16(imm >> 32));
    dsll(rd, rd, 16);
    ori(rd, rd, mask16(imm >> 16));
    dsll(rd, rd, 16);
    ori(rd, rd, mask16(imm));
  }
}
#else
void MacroAssembler::li(Register rd, long imm) {
  li32(rd, (int)imm);
}
#endif

// MUST NOT USE AT
// YYY merge with li - no need to publish an helper function, and this
// isn't a MIPS mneumonic that I see in Sweetman
void MacroAssembler::li32(Register rd, int imm32) {
  if (imm32 >= 0 && imm32 <= 65535) {
    ori(rd, R0, imm32);
  } else if (imm32 < 0 && imm32 >= -32768) {
    addiu(rd, R0, imm32);
  } else {
    assert_not_delayed();
    lui(rd, imm32 >> 16);
    ori(rd, rd, mask16(imm32));
  }
}

// A function defined with bLongOrOffset chooses at compile-time between
// a pc-relative branch or a sequence to handle arbitrary addresses.

#define bLongOrOffset(bXX, bNotXX, uOrB, asmPfx)                       \
                                                                       \
void MacroAssembler::bXX(bLO_##uOrB##_FormalRegs , address target,     \
			 relocInfo::relocType rtype)                   \
{                                                                      \
  intptr_t delayslot = intptr_t(locate_next_instruction(pc()));        \
  intptr_t offset = intptr_t(target) - delayslot;                      \
  assert((offset & 3) == 0, "supposed to be aligined");                \
  int int_offset = (int)(offset >> 2);                                 \
  if (fits_in_s_lfield(offset, 15, 0)) {                               \
    asmPfx bXX(bLO_##uOrB##_RegArgs, int_offset, rtype);               \
  } else {                                                             \
    Label skip;                                                        \
    bNotXX(bLO_##uOrB##_RegArgs, skip);                                \
    delayed()->nop();                                                  \
    jump_to(target, rtype);                                            \
    delayed()->nop();                                                  \
    bind(skip);                                                        \
  }                                                                    \
}

#define bLO_unary_FormalRegs   Register rs
#define bLO_binary_FormalRegs  Register rs, Register rt
#define bLO_unary_RegArgs   rs
#define bLO_binary_RegArgs  rs, rt

// Here are the functions we define.
// The first argument is the function name, the second is the logical opposite.

bLongOrOffset(beq,   bne,   binary, Assembler::)
bLongOrOffset(bge,   blt,   binary, )
bLongOrOffset(bgeu,  bltu,  binary, )
bLongOrOffset(bgez,  bltz,  unary,  Assembler::)
bLongOrOffset(bgezal,bltzal,unary,  Assembler::)
bLongOrOffset(bgt,   ble,   binary, )
bLongOrOffset(bgtu,  bleu,  binary, )
bLongOrOffset(bgtz,  blez,  unary,  Assembler::)
bLongOrOffset(ble,   bgt,   binary, )
bLongOrOffset(bleu,  bgtu,  binary, )
bLongOrOffset(blez,  bgtz,  unary,  Assembler::)
bLongOrOffset(blt,   bge,   binary, )
bLongOrOffset(bltu,  bgeu,  binary, )
bLongOrOffset(bltz,  bgez,  unary,  Assembler::)
bLongOrOffset(bltzal,bgezal,unary,  Assembler::)
bLongOrOffset(bne,   beq,   binary, Assembler::)

; // fixup emacs indentation

#ifdef RESERVOIR
/* for etags:
void MacroAssembler::beq()   { see bLongOrOffset; }
void MacroAssembler::bge()   { see bLongOrOffset; }
void MacroAssembler::bgeu()  { see bLongOrOffset; }
void MacroAssembler::bgez()  { see bLongOrOffset; }
void MacroAssembler::bgezal(){ see bLongOrOffset; }
void MacroAssembler::bgt()   { see bLongOrOffset; }
void MacroAssembler::bgtu()  { see bLongOrOffset; }
void MacroAssembler::bgtz()  { see bLongOrOffset; }
void MacroAssembler::ble()   { see bLongOrOffset; }
void MacroAssembler::bleu()  { see bLongOrOffset; }
void MacroAssembler::blez()  { see bLongOrOffset; }
void MacroAssembler::blt()   { see bLongOrOffset; }
void MacroAssembler::bltu()  { see bLongOrOffset; }
void MacroAssembler::bltz()  { see bLongOrOffset; }
void MacroAssembler::bltzal(){ see bLongOrOffset; }
void MacroAssembler::bne()   { see bLongOrOffset; }
*/
#endif

// bx() simply translates the Condition code into the corresponding branch

void MacroAssembler::bx(Condition cond, Register r1, Register r2, address addr,
			relocInfo::relocType rtype) {
  switch (cond) {
  case equal:
    beq(r1, r2, addr, rtype);
    break;
  case less:
    blt(r1, r2, addr, rtype);
    break;
  case lessUnsigned:
    bltu(r1, r2, addr, rtype);
    break;
  case lessEqual:
    ble(r1, r2, addr, rtype);
    break;
  case lessEqualUnsigned:
    bleu(r1, r2, addr, rtype);
    break;
  case always:
    b(addr, rtype);
    break;
  case notEqual:
    bne(r1, r2, addr, rtype);
    break;
  case greater:
    bgt(r1, r2, addr, rtype);
    break;
  case greaterUnsigned:
    bgtu(r1, r2, addr, rtype);
    break;
  case greaterEqual:
    bge(r1, r2, addr, rtype);
    break;
  case greaterEqualUnsigned:
    bgeu(r1, r2, addr, rtype);
    break;
  case never:
  default:
    ShouldNotReachHere();
    break;
  }
}

void MacroAssembler::bx(Condition cond, Register r1, Register r2, Label& L,
			relocInfo::relocType rtype) {
  switch (cond) {
  case equal:
    beq(r1, r2, L, rtype);
    break;
  case less:
    blt(r1, r2, L, rtype);
    break;
  case lessUnsigned:
    bltu(r1, r2, L, rtype);
    break;
  case lessEqual:
    ble(r1, r2, L, rtype);
    break;
  case lessEqualUnsigned:
    bleu(r1, r2, L, rtype);
    break;
  case always:
    b(L, rtype);
    break;
  case notEqual:
    bne(r1, r2, L, rtype);
    break;
  case greater:
    bgt(r1, r2, L, rtype);
    break;
  case greaterUnsigned:
    bgtu(r1, r2, L, rtype);
    break;
  case greaterEqual:
    bge(r1, r2, L, rtype);
    break;
  case greaterEqualUnsigned:
    bgeu(r1, r2, L, rtype);
    break;
  case never:
  default:
    ShouldNotReachHere();
    break;
  }
}

void MacroAssembler::ld_ptr(Register rt, Register offset, Register base) {
  addu_long(AT, base, offset);
  ld_ptr(rt, 0, AT);
}

void MacroAssembler::st_ptr(Register rt, Register offset, Register base) {
  addu_long(AT, base, offset);
  st_ptr(rt, 0, AT);
}

void MacroAssembler::ld_long(Register rt, Register offset, Register base) {
  addu_long(AT, base, offset);
  ld_long(rt, 0, AT);
}

void MacroAssembler::st_long(Register rt, Register offset, Register base) {
  addu_long(AT, base, offset);
  st_long(rt, 0, AT);
}

static Thread* verify_thread_subroutine(Thread* gthread_value) {
  Thread* correct_value = ThreadLocalStorage::thread();
  guarantee(gthread_value == correct_value, "Rthread value must be the thread");
  return correct_value;
}

void MacroAssembler::verify_thread() {
  if (VerifyThread) {
    save_frame(1);
    st_ptr(Rcall, wordSize, FP);
    li(Rcall, CAST_FROM_FN_PTR(address, verify_thread_subroutine));
    jalr(Rcall, relocInfo::runtime_call_type);
    delayed()->move(A0, Rthread);
    ld_ptr(Rcall, wordSize, FP);
    restore_frame(1);
  }
}


// %%% maybe get rid of [re]set_last_Java_frame
void MacroAssembler::set_last_Java_frame(Register last_java_sp,
					 Register last_java_fp,
					 Register last_java_pc)
{
  assert_not_delayed();
  const int last_Java_sp_offset =
    in_bytes(JavaThread::last_Java_sp_offset());
  const int last_Java_fp_offset =
    in_bytes(JavaThread::last_Java_fp_offset());
  const int last_Java_pc_offset =
    in_bytes(JavaThread::last_Java_pc_offset());

#ifdef ASSERT
  // Verify that pc was zeroed on return to Java
  Label PcOk;
  ld_ptr(Rscratch0, last_Java_pc_offset, Rthread);
  beq(Rscratch0, R0, PcOk);
  delayed()->nop();
  stop("last_Java_pc not zeroed before leaving Java");
  bind(PcOk);
#endif /* ASSERT */
  //
  // When returning from calling out from Java mode the frame anchor's last_Java_pc
  // will always be set to NULL. It is set here so that if we are doing a call to
  // native (not VM) that we capture the known pc and don't have to rely on the
  // native call having a standard frame linkage where we can find the pc.

  // Always set pc and fp first, because once last_Java_sp is visible
  // has_last_Java_frame is true and users will look at the rest of the fields.

  if (last_java_pc->is_valid()) {
    st_ptr(last_java_pc, last_Java_pc_offset, Rthread);
  }
#if 1  // Just an idea, to get a valid PC.
  else {
    Label target;
    move(AT, RA);
    bgezal(R0, target);
    delayed()->move(Rscratch0, RA);
    bind(target);
    move(RA, AT);
    st_ptr(Rscratch0, last_Java_pc_offset, Rthread);
  }
#endif // 1

  if (last_java_fp->is_valid()) {
    st_ptr(last_java_fp, last_Java_fp_offset, Rthread);
  }

  assert(last_java_sp->is_valid(), "invalid sp");
  st_ptr(last_java_sp, last_Java_sp_offset, Rthread);
}

void MacroAssembler::reset_last_Java_frame(void) {
  assert_not_delayed();
  const int last_Java_sp_offset =
    in_bytes(JavaThread::last_Java_sp_offset());
  const int last_Java_fp_offset =
    in_bytes(JavaThread::last_Java_fp_offset());    
  const int last_Java_pc_offset =
    in_bytes(JavaThread::last_Java_pc_offset());

#if 0  // not clear this is actually an invariant
#ifdef ASSERT
  // Verify that sp has some value
  Label ok;
  ld_ptr(Rscratch0, last_Java_sp_offset, Rthread);
  bne(Rscratch0, R0, ok);
  delayed()->nop();
  stop("last_Java_sp not set");
  bind(ok);
#endif // ASSERT
#endif

  // Null last_Java* and flags after return to Java
  st_ptr(R0, last_Java_sp_offset, Rthread);
  st_ptr(R0, last_Java_fp_offset, Rthread);
  st_ptr(R0, last_Java_pc_offset, Rthread);
}


void MacroAssembler::call_VM_base(
  Register        oop_result,
  Register        thread_cache,
  Register        last_java_sp,
  address         entry_point,
  int             number_of_arguments,
  bool            check_exceptions)
{
  assert_not_delayed();

  // determine last_java_sp register
  if (!last_java_sp->is_valid()) {
    last_java_sp = SP;
  }
  // debugging support
  assert(number_of_arguments >= 0   , "cannot have negative number of arguments");

  set_last_Java_frame(last_java_sp, FP, noreg);
  jal_to_C(entry_point, relocInfo::runtime_call_type);
  delayed()->move(A0, Rthread);  // pass thread as first argument
  reset_last_Java_frame();

  // check for pending exceptions.
  if (check_exceptions) {
    check_and_forward_exception(Rscratch1);
  }
  
  // get oop result if there is one and reset the value in the thread
  if (oop_result->is_valid()) {
    get_vm_result(oop_result);
  }
}

void MacroAssembler::check_and_forward_exception(Register scratch_reg)
{
  Label L;

  check_and_handle_popframe(scratch_reg);
  check_and_handle_earlyret(scratch_reg);

  ld_ptr(scratch_reg, in_bytes(Thread::pending_exception_offset()), Rthread);
  beq(scratch_reg, R0, L);
  delayed()->nop();
  call(StubRoutines::forward_exception_entry(), relocInfo::runtime_call_type);
  delayed()->nop();
  bind(L);
}


void MacroAssembler::check_and_handle_popframe(Register scratch_reg) {
}


void MacroAssembler::check_and_handle_earlyret(Register scratch_reg) {
}


void MacroAssembler::call_VM(Register oop_result, address entry_point, int number_of_arguments, bool check_exceptions) {
  call_VM_base(oop_result, noreg, noreg, entry_point, number_of_arguments, check_exceptions);
}


void MacroAssembler::call_VM(Register oop_result, address entry_point, Register arg_1, bool check_exceptions) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  call_VM(oop_result, entry_point, 1, check_exceptions);
}


void MacroAssembler::call_VM(Register oop_result, address entry_point, Register arg_1, Register arg_2, bool check_exceptions) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  move(A2, arg_2); assert(arg_2 != A1, "smashed argument");
  call_VM(oop_result, entry_point, 2, check_exceptions);
}


void MacroAssembler::call_VM(Register oop_result, address entry_point, Register arg_1, Register arg_2, Register arg_3, bool check_exceptions) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  move(A2, arg_2); assert(arg_2 != A1,                "smashed argument");
  move(A3, arg_3); assert(arg_3 != A1 && arg_3 != A2, "smashed argument");
  call_VM(oop_result, entry_point, 3, check_exceptions);
}


// patterned after the above call_VMs

void MacroAssembler::call_VM(Register oop_result, Register last_java_sp, address entry_point) { // 0 args
  call_VM_base(oop_result, noreg, last_java_sp, entry_point, 0, true);
}


void MacroAssembler::call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  call_VM_base(oop_result, noreg, last_java_sp, entry_point, 1, true);
}


void MacroAssembler::call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1, Register arg_2) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  move(A2, arg_2); assert(arg_2 != A1, "smashed argument");
  call_VM_base(oop_result, noreg, last_java_sp, entry_point, 2, true);
}


void MacroAssembler::call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1, Register arg_2, Register arg_3) {
  // A0 is reserved for the thread
  move(A1, arg_1);
  move(A2, arg_2); assert(arg_2 != A1,                "smashed argument");
  move(A3, arg_3); assert(arg_3 != A1 && arg_3 != A2, "smashed argument");
  call_VM_base(oop_result, noreg, last_java_sp, entry_point, 3, true);
}



// YYY thread_cache is vestigial, and can probably be removed later.
// YYY We are not using it currently.
void MacroAssembler::call_VM_leaf_base(Register thread_cache,
				       address entry_point,
				       int number_of_arguments) {
  assert_not_delayed();
  jal_to_C(entry_point, relocInfo::runtime_call_type);
  delayed()->nop();
}


void MacroAssembler::call_VM_leaf(Register thread_cache, address entry_point, int number_of_arguments) {
  call_VM_leaf_base(thread_cache, entry_point, number_of_arguments);
}


void MacroAssembler::call_VM_leaf(Register thread_cache, address entry_point, Register arg_1) {
  move(A0, arg_1);
  call_VM_leaf(thread_cache, entry_point, 1);
}


void MacroAssembler::call_VM_leaf(Register thread_cache, address entry_point, Register arg_1, Register arg_2) {
  move(A0, arg_1);
  move(A1, arg_2); assert(arg_2 != A0, "smashed argument");
  call_VM_leaf(thread_cache, entry_point, 2);
}


void MacroAssembler::call_VM_leaf(Register thread_cache, address entry_point, Register arg_1, Register arg_2, Register arg_3) {
  move(A0, arg_1);
  move(A1, arg_2); assert(arg_2 != A0,                "smashed argument");
  move(A2, arg_3); assert(arg_3 != A0 && arg_3 != A1, "smashed argument");
  call_VM_leaf(thread_cache, entry_point, 3);
}


void MacroAssembler::get_vm_result(Register oop_result) {
  verify_thread();
  ld_ptr(oop_result, in_bytes(JavaThread::vm_result_offset()), Rthread);
  st_ptr(R0,         in_bytes(JavaThread::vm_result_offset()), Rthread);
  verify_oop(oop_result);
}


void MacroAssembler::get_vm_result_2(Register oop_result) {
  verify_thread();
  ld_ptr(oop_result, in_bytes(JavaThread::vm_result_2_offset()), Rthread);
  st_ptr(R0,         in_bytes(JavaThread::vm_result_2_offset()), Rthread);
  verify_oop(oop_result);
}


// We require that C code which does not return a value in vm_result will
// leave it undisturbed.
void MacroAssembler::set_vm_result(Register oop_result) {
  verify_thread();
  verify_oop(oop_result);

# ifdef ASSERT
  // Check that we are not overwriting any other oop.
  // YYY
# endif

  st_ptr(oop_result, in_bytes(JavaThread::vm_result_offset()), Rthread);
}


void MacroAssembler::store_check(Register tmp, Register obj) {
  BarrierSet* bs = Universe::heap()->barrier_set();
  assert(bs->kind() == BarrierSet::CardTableModRef, "Wrong barrier set kind");
  CardTableModRefBS* ct = (CardTableModRefBS*)bs;
  assert(sizeof(*ct->byte_map_base) == sizeof(jbyte), "adjust this code");
  srl_long(obj, obj, CardTableModRefBS::card_shift);
  assert( tmp != obj, "need separate temp reg");
  li(tmp, (address)ct->byte_map_base);
  addu_long(tmp, tmp, obj);
  sb(R0, 0, tmp);
}

void MacroAssembler::store_check(Register tmp, Register obj, Register offset) {
  store_check(tmp, obj);
}


void MacroAssembler::align(int modulus) {
  while (offset() % modulus != 0) nop();
}


void RegistersForDebugging::print(outputStream* s) {
  int j;
  for (j = 0; j < 32; j++) {
    if (j == K0->encoding() || j == K1->encoding())
      continue;   // kernel regs are not interesting
    s->print_cr("%s = %0x%.16lx", as_Register(j)->name(), r[j]);
  }
  for (j = 0; j < 32; ) {
    s->print_cr("%s = %g", as_FloatRegister(j)->name(), f[j]);
  }
}

// Space for a RegistersForDebugging object must have been allocated
// on the stack at the top of the current frame.  We furthermore
// expect that the saved registers have just been saved in this frame.
void RegistersForDebugging::save_registers(MacroAssembler* a) {
  a->addu_long(S0, FP, - ::round_to(sizeof(RegistersForDebugging), sizeof(jdouble)));
  assert(S0->encoding() == 16 &&
	 S1->encoding() == 17 &&
	 SP->encoding() == 29 &&
	 FP->encoding() == 30,
	 "switch cases need updated constants");
  int i;
  for (i = 0; i < 32; i++) {
    switch(i) {
    case 16:  // S0.encoding()
    case 17:  // S1.encoding()
    case 30:  // FP.encoding()
      // Get the value from the save area of the enclosing save_frame
      a->ld_ptr(S1, as_Register(i)->sp_offset_in_save_area(), SP);
      a->st_ptr(S1, r_offset(i), S0);
      break;
    case 29:  // SP.encoding()
      // Calculate the previous value of SP and store that
      a->addu_long(S1, SP,
		   - (::round_to(sizeof(RegistersForDebugging), sizeof(jdouble)) +
		      frame::register_save_space));
      a->st_ptr(S1, r_offset(i), S0);
      break;
    default:
      a->st_ptr(as_Register(i), r_offset(i), S0);
      break;
    }
  }
  for (i = 0; i < 32; i++) {
    a->sdc1(as_FloatRegister(i), f_offset(i), S0);
  }
}

void RegistersForDebugging::restore_registers(MacroAssembler* a) {
  // We just need to restore the volatiles,
  // excluding: SP, K0, K1, and RA.
  // The rest will be handled by the following restore_frame.
  a->addu_long(S0, FP, - ::round_to(sizeof(RegistersForDebugging), sizeof(jdouble)));
  int i;
  for (i = 1; i <= 15; i++) {
    a->ld_ptr(as_Register(i), r_offset(i), S0);
  }
  for (i = 24; i <= 25; i++) {
    a->ld_ptr(as_Register(i), r_offset(i), S0);
  }
  for (i = 0; i < 32; i++) {
    a->ldc1(as_FloatRegister(i), f_offset(i), S0);
  }
}


void MacroAssembler::_verify_oop(Register reg, const char* msg, const char * file, int line) {
  // plausibility check for oops
  if (!VerifyOops) return;

  if (reg == R0)  return;	// always NULL, which is always an oop

  char buffer[16];
  sprintf(buffer, "%d", line);
  int len = strlen(file) + strlen(msg) + 1 + 4 + strlen(buffer);
  char * real_msg = new char[len];
  sprintf(real_msg, "%s (%s:%d)", msg, file, line);

  move(A0, reg);
  li(A1, (intptr_t)real_msg);
  li(AT, (address)StubRoutines::verify_oop_subroutine_entry_address());
  jalr(AT);
  delayed()->nop();
}

void MacroAssembler::_verify_oop_addr(Address addr, const char* msg, const char * file, int line) {
  // plausibility check for oops
  if (!VerifyOops) return;

  char buffer[64];
  sprintf(buffer, "%d", line);
  int len = strlen(file) + strlen(msg) + 1 + 4 + strlen(buffer);
  sprintf(buffer, " at SP+%d ", addr.disp());
  len += strlen(buffer);
  char * real_msg = new char[len];
  sprintf(real_msg, "%s at %s+%d (%s:%d)", msg, addr.base()->name(), addr.disp(), file, line);

  ld_ptr(A0, addr.disp(), addr.base());
  li(A1, (intptr_t)real_msg);
  li(AT, (address)StubRoutines::verify_oop_subroutine_entry_address());
  jalr(AT);
  delayed()->nop();
}

// This macro is expanded just once; it creates shared code.  Contract:
// receives an oop in A0 and a description in A1.  Acts as a 'leaf'
// call.
void MacroAssembler::verify_oop_subroutine() {
  // YYY unimplemented
  j(RA);
  delayed()->nop();
}


// Generate short sequence to call a routine to print a message and
// halt the VM.
void MacroAssembler::stop(const char* msg) {
  // pass message pointer to stop_subroutine.
  li(A0, (long)msg);

  // factor long stop-sequence into subroutine to save space
  assert(StubRoutines::Mips::stop_subroutine_entry_address(), "hasn't been generated yet");
  li(Rcall, (address)StubRoutines::Mips::stop_subroutine_entry_address());

  // call indirectly to solve generation ordering problem
  ld_ptr(Rcall, 0, Rcall);

  // YYY This clobbers RA - we ought to allocate frame and save RA here.
  // YYY Prefer not to do the whole save_frame for space reasons.  We
  // can save the rest of the registers in stop_subroutine.
  jalr(Rcall);
  delayed()->nop();

  break_();   // make stop actually stop
}


static void MacroAssembler_debug_print(const char* msg, long value) {
  tty->print_cr("%s0x%lx", msg, value);
}


// Generate short sequence to call a routine to print a message.
void MacroAssembler::warn(const char* msg, Register Rval) {
  save_frame(::round_to(sizeof(RegistersForDebugging) / BytesPerWord, 2));
  RegistersForDebugging::save_registers(this);
  li(A0, (long)msg);
  move(A1, Rval);
  // could also call warning() [debug.cpp], but note it is varargs
  jal_to_C(CAST_FROM_FN_PTR(address, MacroAssembler_debug_print));
  delayed()->nop();
  RegistersForDebugging::restore_registers(this);
  restore_frame(::round_to(sizeof(RegistersForDebugging) / BytesPerWord, 2));
}


void MacroAssembler::untested(const char* what) {
  // We must be able to turn interactive prompting off
  // in order to run automated test scripts on the VM
  // Use the flag ShowMessageBoxOnError
  
  char* b = new char[1024];
  sprintf(b, "untested: %s", what);  

  if ( ShowMessageBoxOnError )   stop(b);
  else                           warn(b, R0);
}


// support for the workaround below
static void MacroAssembler_debug(char* msg, RegistersForDebugging* regs);


void MacroAssembler::stop_subroutine() {
  save_frame(::round_to(sizeof(RegistersForDebugging) / BytesPerWord, 2));
  RegistersForDebugging::save_registers(this);  // sets S0

  move(A1, S0);
  // workaround for compile error: was MacroAssembler::debug, but compiler complained (also complained when I added an '&')
  jal_to_C(CAST_FROM_FN_PTR(address, MacroAssembler_debug), relocInfo::runtime_call_type);
  delayed()->nop();

  RegistersForDebugging::restore_registers(this);

  save_frame(1);
  st_ptr(Rcall, wordSize, FP);
  jal_to_C(CAST_FROM_FN_PTR(address,breakpoint), relocInfo::runtime_call_type);
  delayed()->nop();
  ld_ptr(Rcall, wordSize, FP);
  restore_frame(1);

  restore_frame_and_return(::round_to(sizeof(RegistersForDebugging) / BytesPerWord, 2));
}

void MacroAssembler::debug(char* msg, RegistersForDebugging* regs) {
  MacroAssembler_debug(msg, regs);
}

static void MacroAssembler_debug(char* msg, RegistersForDebugging* regs) {
  if ( ShowMessageBoxOnError ) {
      JavaThreadState saved_state = JavaThread::current()->thread_state();
      JavaThread::current()->set_thread_state(_thread_in_vm);
      {
        // In order to get locks work, we need to fake a in_VM state
        ttyLocker ttyl;
        ::tty->print_cr("EXECUTION STOPPED: %s\n", msg);
        if (CountBytecodes || TraceBytecodes || StopInterpreterAt) {
	  ::tty->print_cr("Interpreter::bytecode_counter = %d", BytecodeCounter::counter_value());
        }
        if (os::message_box(msg, "Execution stopped, print registers?"))
          regs->print(::tty);
      }
      ThreadStateTransition::transition(JavaThread::current(), _thread_in_vm, saved_state);
  }
  else
     ::tty->print_cr("=============== DEBUG MESSAGE: %s ================\n", msg);
  assert(false, "error");
}


#ifndef PRODUCT
void MacroAssembler::test() {
  ResourceMark rm;

  CodeBuffer cb("test", 10000, 10000);
  MacroAssembler* a = new MacroAssembler(&cb);

  StubRoutines::Mips::test_stop_entry()();
}
#endif


void MacroAssembler::save_registers(Register save_base) {
  st_long(RA, RA->sp_offset_in_save_area(), save_base);
  st_long(GP, GP->sp_offset_in_save_area(), save_base);
  Register r = S0;
  for (;
       r->encoding() <= S7_RegisterEnumValue;
       r = r->successor()) {
    st_long(r, r->sp_offset_in_save_area(), save_base);
  }
  st_long(FP, FP->sp_offset_in_save_area(), save_base);
  FloatRegister f = F23;
  while (f->encoding() < F31_FloatRegisterEnumValue) {
    f = f->successor();
    sdc1(f, f->sp_offset_in_save_area(), save_base);
  } 
}

void MacroAssembler::restore_registers(Register save_base) {
  // Note: Must not touch AT, if restore_frame is using it.
  ld_long(RA, RA->sp_offset_in_save_area(), save_base);
  ld_long(GP, GP->sp_offset_in_save_area(), save_base);
  Register r = S0;
  for (;
       r->encoding() <= S7_RegisterEnumValue;
       r = r->successor()) {
    ld_long(r, r->sp_offset_in_save_area(), save_base);
  }
  ld_long(FP, FP->sp_offset_in_save_area(), save_base);
  FloatRegister f = F23;
  while (f->encoding() < F31_FloatRegisterEnumValue) {
    f = f->successor();
    ldc1(f, f->sp_offset_in_save_area(), save_base);
  } 
}

void MacroAssembler::save_frame(int extraWords) {
  int frame_size
    = total_frame_size_in_bytes(extraWords);
  addu_long(SP, SP, - frame_size);
  save_registers();
  addu_long(FP, SP,   frame_size);
}

#define RESTORE_WITH_FRAME_POINTER

void MacroAssembler::restore_frame(int extraWords) {
#ifdef RESTORE_WITH_FRAME_POINTER
  // This takes an extra instruction, but is robust in case caller
  // gets extraWords wrong.
  move(AT, FP);
  restore_registers();
  move(SP, AT);
#else
  assert(extraWords >= 0, "must specify valid extraWords");
  int frame_size
    = total_frame_size_in_bytes(extraWords);
  restore_registers();
  addu_long(SP, SP, frame_size);
#endif
}

void MacroAssembler::restore_frame_and_return(int extraWords) {
#ifdef RESTORE_WITH_FRAME_POINTER
  move(AT, FP);
  restore_registers();
  jr(RA);
  delayed()->move(SP, AT);
#else
  assert(extraWords >= 0, "must specify valid extraWords");
  int frame_size
    = total_frame_size_in_bytes(extraWords);
  restore_registers();
  jr(RA);
  delayed()->addu_long(SP, SP, frame_size);
#endif
}

int MacroAssembler::total_frame_size_in_bytes(int extraWords) {
  return ::round_to(frame::register_save_space + extraWords * wordSize, 2 * wordSize);
}

void MacroAssembler::save_minimal_frame(int extraWords) {
  assert(extraWords >= 0, "must specify valid extraWords");
  int frame_size
    = ::round_to((2 + extraWords) * wordSize, 2 * wordSize);
  addu_long(SP, SP, - frame_size);
  st_long(FP, wordSize, SP);
  st_long(RA, 0,        SP);
  addu_long(FP, SP,   frame_size);
}

void MacroAssembler::restore_minimal_frame(int extraWords) {
#ifdef RESTORE_WITH_FRAME_POINTER
  move(AT, FP);
  ld_long(RA, 0,        SP);
  ld_long(FP, wordSize, SP);
  move(SP, AT);
#else
  assert(extraWords >= 0, "must specify valid extraWords");
  int frame_size
    = ::round_to((2 + extraWords) * wordSize, 2 * wordSize);
  ld_long(RA, 0,        SP);
  ld_long(FP, wordSize, SP);
  addu_long(SP, SP, frame_size);
#endif
}

void MacroAssembler::restore_minimal_frame_and_return(int extraWords) {
#ifdef RESTORE_WITH_FRAME_POINTER
  move(AT, FP);
  ld_long(RA, 0,        SP);
  ld_long(FP, wordSize, SP);
  jr(RA);
  delayed()->move(SP, AT);
#else
  assert(extraWords >= 0, "must specify valid extraWords");
  int frame_size
    = ::round_to((2 + extraWords) * wordSize, 2 * wordSize);
  ld_long(RA, 0,        SP);
  ld_long(FP, wordSize, SP);
  jr(RA);
  delayed()->addu_long(SP, SP, frame_size);
#endif
}

void MacroAssembler::bang_stack_with_offset(int offset) {
  // stack grows down, caller passes positive offset that is subtracted
  assert(offset > 0, "must bang with positive offset");
  offset = -offset;
  if (fits_in_s_field(offset, 15, 0)) {
    sw(R0, offset, SP);
  } else {
    li(AT, offset);
    addu_long(AT, SP, AT);
    sw(R0, 0, AT);
  }
}

void MacroAssembler::membar() {
  // YYY ought to work, but may be able to be more subtle
  sync();
}

void MacroAssembler::verify_FPU(int stack_depth, const char* s) {
  // %%%%% need to implement this
}

SkipIfEqual::SkipIfEqual(
    MacroAssembler* masm, Register temp, Register r2, const bool* flag_addr,
    Assembler::Condition condition) {
  _masm = masm;
  _masm->li(temp, (long)flag_addr);
  _masm->lbu(temp, 0, temp);
  _masm->bx(condition, temp, r2, _label);
  _masm->delayed()->nop();
}

SkipIfEqual::~SkipIfEqual() {
  _masm->bind(_label);
}

#ifdef ASSERT
// like assert, but also capture info from lexical scope
#define a(p, msg) do {							\
  if (!(p)) {								\
    warning("%s:%d: %s(%d, %d, %d):\n%s:%d: %s, %s\n",			\
	    __FILE__, __LINE__, __FUNCTION__, x, hi_bit, lo_bit, 	\
	    file, line, XSTR(p), msg);					\
  }									\
} while (0)
int Assembler::check_u_field(int x, int hi_bit, int lo_bit,
			     const char *file, int line)
{
  a(fits_in_u_field(x, hi_bit, lo_bit), "value out of range");
  int r = U_FIELD(x, hi_bit, lo_bit);
  a(inv_u_field(r, hi_bit, lo_bit) == x, "just checking");

  return r;
}
int Assembler::check_s_field(int x, int hi_bit, int lo_bit,
			     const char *file, int line)
{
  a(fits_in_s_field(x, hi_bit, lo_bit), "value out of range");
  int r = S_FIELD(x, hi_bit, lo_bit);
  a(inv_s_field(r, hi_bit, lo_bit) == x, "just checking");

  int mask = ~fmask(hi_bit-1, lo_bit);
  a(((x & mask) == 0) || ((x & mask) == mask), "value out of range");

  return r;
}
#undef a
#endif // ASSERT
