#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)assembler_sparc.hpp	1.179 07/05/17 15:47:51 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// This register always holds the current JavaThread pointer:
REGISTER_DECLARATION(Register, Rthread, S7);

// Used in Linux calling convention to hold call target
REGISTER_DECLARATION(Register, Rcall, T9);

// Exception processing
// These registers are passed into exception handlers.
// All exception handlers require the exception object being thrown.
// In addition, an nmethod's exception handler must be passed
// the address of the call site within the nmethod, to allow
// proper selection of the applicable catch block.
// (Interpreter frames use their own bcp() for this purpose.)
//
// The Rissuing_pc value is not always needed.  When jumping to a
// handler that is known to be interpreted, the Rissuing_pc value can be
// omitted.  An actual catch block in compiled code receives (from its
// nmethod's exception handler) the thrown exception in the Rexception,
// but it doesn't need the Oissuing_pc.
//
// If an exception handler (either interpreted or compiled)
// discovers there is no applicable catch block, it updates
// the Rissuing_pc to the continuation PC of its own caller,
// pops back to that caller's stack frame, and executes that
// caller's exception handler.  Obviously, this process will
// iterate until the control stack is popped back to a method
// containing an applicable catch block.  A key invariant is
// that the Rissuing_pc value is always a value local to
// the method whose exception handler is currently executing.
//
REGISTER_DECLARATION(Register, Rexception,   A0);
REGISTER_DECLARATION(Register, Rissuing_pc,  A1);

// Used in Java interpreter method calls.
// We used to use T8 and T9 (Rscratch4, Rscratch5).  This
// was ok except for a conflict in generate_fixed_frame
// (May also have been conflicts in untested bytecodes
// invokevirtual and invokeinterface.)
// YYY When we implement the compiler, I'm not sure but that
// YYY we may want to use the A registers for arguments and
// YYY have some conflict there.  May have to revisit this.
REGISTER_DECLARATION(Register, Rargs,        A6);
REGISTER_DECLARATION(Register, Rmethod_call, A7);
REGISTER_DECLARATION(Register, Rlast_SP,     S0); // == Rsave0

#ifndef DONT_USE_REGISTER_DEFINES
#define Rthread       AS_REGISTER(Register, Rthread)
#define Rcall         AS_REGISTER(Register, Rcall)
#define Rexception    AS_REGISTER(Register, Rexception)
#define Rissuing_pc   AS_REGISTER(Register, Rissuing_pc)
#define Rmethod_call  AS_REGISTER(Register, Rmethod_call)
#define Rargs         AS_REGISTER(Register, Rargs)
#endif

REGISTER_DECLARATION(Register, Rsave0,      S0);
REGISTER_DECLARATION(Register, Rmethod,     S1); // method
REGISTER_DECLARATION(Register, Rlocals,     S2); // local vars
REGISTER_DECLARATION(Register, Rmonitors,   S3); // monitors
REGISTER_DECLARATION(Register, Resp,        S4); // expr stack ptr
REGISTER_DECLARATION(Register, Rbcp,        S5); // ptr to next bytecode
REGISTER_DECLARATION(Register, RcpoolCache, S6); // cp cache pointer
REGISTER_DECLARATION(Register, Rbyte_code,  T8); // next bytecode

REGISTER_DECLARATION(Register, Rscratch0, T0);
REGISTER_DECLARATION(Register, Rscratch1, T1);
REGISTER_DECLARATION(Register, Rscratch2, T2);
REGISTER_DECLARATION(Register, Rscratch3, T3);
REGISTER_DECLARATION(Register, Rscratch4, T8);
REGISTER_DECLARATION(Register, Rscratch5, T9); // last resort scratch

REGISTER_DECLARATION(FloatRegister, Fscratch0, F1);
REGISTER_DECLARATION(FloatRegister, Fscratch1, F3);
REGISTER_DECLARATION(FloatRegister, Fscratch2, F4);
REGISTER_DECLARATION(FloatRegister, Fscratch3, F5);


// Address is an abstraction used to represent a memory location.
//
// Note: A register location is represented via a Register, not
//       via an address for efficiency & simplicity reasons.

class Address VALUE_OBJ_CLASS_SPEC {
 private:
  Register         _base;
  int              _disp;  // signed 16-bit displacement
  RelocationHolder _rspec;

 public:
  // creation
  Address()
  : _base (noreg)
  , _disp (0)
  {}

  Address(Register base, int disp = 0)
  : _base (base)
  , _disp (disp)
  {}

  Address(int disp, relocInfo::relocType rtype);

  // accessors
  bool        uses(Register reg) const { return _base == reg; }
  Register    base()             const { return _base; }
  int         disp()             const { return _disp; }

  friend class Assembler;
  friend class MacroAssembler;
};


// Argument is an abstraction used to represent an outgoing
// actual argument or an incoming formal parameter, whether
// it resides in memory or in a register, in a manner consistent
// with the MIPS n64 Application Binary Interface, or ABI.  This is
// often referred to as the native or C calling convention.

class Argument VALUE_OBJ_CLASS_SPEC {
 private:
  int _number;

 public:
  enum {
    n_register_parameters = 8, 		// up to 8 integer register args
    n_float_register_parameters = 8	// up to 8 floating register args
  };

  // creation
  Argument(int number) : _number(number) {}

  int  number() const  { return _number;  }
  Argument successor() const  { return Argument(number() + 1); }

  // locating register-based arguments:
  bool is_register() const { return _number < n_register_parameters; }
  bool is_float_register() const { return _number < n_float_register_parameters; }

  Register as_register() const {
    assert(is_register(), "must be a register argument");
    return as_Register(A0->encoding() + number());
  }
  FloatRegister as_float_register() const {
    assert(is_float_register(), "must be a register argument");
    return as_FloatRegister(F12->encoding() + number());
  }

  // locating memory-based arguments
  Address as_callee_address() const;
  Address as_caller_address() const;

  // debugging
  const char* name() const;

  friend class Assembler;
};


// The MIPS Assembler: Pure assembler doing NO optimizations on the instruction
// level; i.e., what you write is what you get.
// The Assembler is generating code into a CodeBuffer.

// See the MIPS 5Kf manual.

class Assembler : public AbstractAssembler  {
  friend class AbstractAssembler; // for the non-virtual hack
  friend class NativeInstruction; // for inv_u_field()

 protected:
   
  static void print_instruction(int inst);

  // fields: note bits numbered from LSB = 0,
  //  fields known by inclusive bit range
  // NB use 'l'-labeled versions (below) for long int values
  // (even when dealing with the first 32 bits).

  // Create a bit mask that selects bits from 0 to (hi_bit - lo_bit)
  static int fmask(juint hi_bit, juint lo_bit) {
    assert( hi_bit >= lo_bit  &&  0 <= lo_bit  &&  hi_bit < 32, "bad bits");
    return (1 << ( hi_bit-lo_bit + 1 )) - 1;
  }
  
  // Extract a bit range from a word
  static int inv_u_field(int x, int hi_bit, int lo_bit) {
    juint r = juint(x) >> lo_bit;
    r &= fmask( hi_bit, lo_bit);
    return int(r);
  }

  // given a field that ranges from hi_bit to lo_bit (inclusive,
  // LSB = 0), and an unsigned value for the field,
  // shift it into the field
#ifdef ASSERT
#define u_field(x, hi_bit, lo_bit) \
  check_u_field(x, hi_bit, lo_bit, __FILE__, __LINE__)
  static int check_u_field(int x, int hi_bit, int lo_bit,
			   const char *file, int line);
#else
  // make sure this is inlined as it will reduce code size significantly
#define u_field U_FIELD
#endif
#define U_FIELD(x, hi_bit, lo_bit) ((x) << (lo_bit))

  static int inv_s_field(int x, int hi_bit, int lo_bit) {
    int sign_shift = 31 - hi_bit + lo_bit;
    return (inv_u_field(x, hi_bit, lo_bit) << sign_shift) >> sign_shift;
  }

#ifdef ASSERT
#define s_field(x, hi_bit, lo_bit) \
  check_s_field(x, hi_bit, lo_bit, __FILE__, __LINE__)
  static int check_s_field(int x, int hi_bit, int lo_bit,
			   const char *file, int line);
#else
  // make sure this is inlined as it will reduce code size significantly
#define s_field S_FIELD
#endif
#define S_FIELD(x, hi_bit, lo_bit) ((x & fmask(hi_bit, lo_bit)) << (lo_bit))

  static bool fits_in_u_field(int x, int hi_bit, int lo_bit) {
    return (x & ~fmask(hi_bit, lo_bit)) == 0;
  }

  static bool fits_in_s_field(int x, int hi_bit, int lo_bit) {
    int sign_only = x >> (hi_bit - lo_bit);
    return sign_only == 0 || sign_only == -1;
  }

  static int inv_op(int x)    { return inv_u_field(x, 31, 26); }
  static int inv_rs(int x)    { return inv_u_field(x, 25, 21); }
  static int inv_rt(int x)    { return inv_u_field(x, 20, 16); }
  static int inv_funct(int x) { return inv_u_field(x, 5, 0); }

  // Create a 'long' bit mask that selects bits from 0 to (hi_bit - lo_bit)
  static julong lfmask(int hi_bit, int lo_bit) {
    assert( hi_bit >= lo_bit  &&  0 <= lo_bit  &&  hi_bit < 32, "bad bits");
    return (1 << ( hi_bit-lo_bit + 1 )) - 1;
  }

  static long inv_lu_field(long x, int hi_bit, int lo_bit) {
    julong r = julong(x) >> lo_bit;
    r &= lfmask(hi_bit, lo_bit);
    return long(r);
  }

  static bool fits_in_u_lfield(unsigned long x, int hi_bit, int lo_bit) {
    return (x & ~lfmask(hi_bit, lo_bit)) == 0;
  }

  static bool fits_in_s_lfield(long x, int hi_bit, int lo_bit) {
    long sign_only = x >> (hi_bit - lo_bit);
    return sign_only == 0 || sign_only == -1L;
  }

  static int xop(int x)        { return u_field(x, 31, 26); }
  static int xrs(Register r)   { return xrsop(r->encoding()); }
  static int xrsop(int x)      { return u_field(x, 25, 21); }
  static int xrt(Register r)   { return xrtop(r->encoding()); }
  static int xrt(FloatRegister r) { return xrtop(r->encoding()); }
  static int xrtop(int x)      { return u_field(x, 20, 16); }
  static int xrd(Register r)   { return xrdop(r->encoding()); }
  static int xrd(FloatRegister r) { return xrdop(r->encoding()); }
  static int xrdop(int x)      { return u_field(x, 15, 11); }
  static int xsa(int x)        { return u_field(x, 10, 6); }
  static int xsafr(Register r) { return xsa(r->encoding()); }
  static int xsafr(FloatRegister r) { return xsa(r->encoding()); }
  static int xfunc(int x)      { return u_field(x, 5, 0); }
  static int ximmed(int x)     { return s_field(x, 15, 0); }
  static int ximmedu(int x)    { return u_field(x, 15, 0); }
  static int xtarget(int x)    { return u_field(x, 25, 0); }

  static int xfmt(int x)          { return u_field(x, 25, 21); }
  static int xfr(FloatRegister f) { return u_field(f->encoding(), 25, 21); }
  static int xft(FloatRegister f) { return u_field(f->encoding(), 20, 16); }
  static int xftop(int x)         { return u_field(x, 20, 16); }
  static int xfs(FloatRegister f) { return u_field(f->encoding(), 15, 11); }
  static int xfd(FloatRegister f) { return u_field(f->encoding(), 10, 6); }

  // Simple delay-slot scheme:
  // In order to check the programmer, the assembler keeps track of deley slots.
#ifdef ASSERT
  #define CHECK_DELAY
#endif
#ifdef CHECK_DELAY
  enum Delay_state { no_delay, at_delay_slot, filling_delay_slot } delay_state;
#endif

 public:
  // Tells assembler next instruction must NOT be in delay slot.
  // Use at start of multinstruction macros.
  void assert_not_delayed() {
    // This is a separate overloading to avoid creation of string constants
    // in non-asserted code--with some compilers this pollutes the object code.
#ifdef CHECK_DELAY
    assert_not_delayed("next instruction should not be a delay slot");
#endif
  }
  void assert_not_delayed(const char* msg) {
#ifdef CHECK_DELAY
    guarantee( delay_state == no_delay, msg );
#endif
  }

 protected:
  void has_delay_slot() { 
#ifdef CHECK_DELAY
    assert_not_delayed("just checking");
    delay_state = at_delay_slot;
#endif
  }

public:
  // YYY delayed() used to be here.

  void flush() {
#ifdef CHECK_DELAY
    guarantee( delay_state == no_delay, "ending code with a delay slot");
#endif
    AbstractAssembler::flush();
  }

  inline void emit_long(int);  // shadows AbstractAssembler::emit_long
  inline void emit_rdat(int x, relocInfo::relocType rtype);
  inline void emit_rdat(int x, RelocationHolder const& rspec);

 public:

  enum ops {
    op_addi      =  8, // 001000
    op_addiu     =  9, // 001001
    op_andi      = 12, // 001100
    op_beq       =  4, // 000100
    op_bgtz      =  7, // 000111
    op_blez      =  6, // 000110
    op_bne       =  5, // 000101
    op_cache     = 47, // 101111
    op_cop0      = 16,
    op_cop1      = 17, // 010001
    op_cop1x     = 19, // 010011
    op_cop2      = 18, // 010010
    op_daddi     = 24, // 011000
    op_daddiu    = 25, // 011001
    op_j         =  2, // 000010
    op_jal       =  3, // 000011
    op_lb        = 32, // 100000
    op_lbu       = 36, // 100100
    op_ld        = 55, // 110111
    op_ldc1      = 53, // 110101
    op_ldc2      = 54, // 110110
    op_ldl       = 26, // 011010
    op_ldr       = 27, // 011011
    op_lh        = 33, // 100001
    op_lhu       = 37, // 100101
    op_ll        = 48, // 110000
    op_lld       = 52, // 110100
    op_lui       = 15, // 001111
    op_lw        = 35, // 100011
    op_lwc1      = 49, // 110001
    op_lwc2      = 50, // 110010
    op_lwl       = 34, // 100010
    op_lwr       = 38, // 100110
    op_lwu       = 39, // 100111
    op_ori       = 13, // 001101
    op_pref      = 51, // 110011
    op_regimm    =  1,
    op_sb        = 40, // 101000
    op_sc        = 56, // 111000
    op_scd       = 60, // 111100
    op_sd        = 63, // 111111
    op_sdc1      = 61, // 111101
    op_sdc2      = 62, // 111110
    op_sdl       = 44, // 101100
    op_sdr       = 45, // 101101
    op_sh        = 41, // 101001
    op_slti      = 10, // 001010
    op_sltiu     = 11, // 001011
    op_special   =  0,
    op_special2  = 28, // 011100
    op_sw        = 43, // 101011
    op_swc1      = 57, // 111001
    op_swc2      = 58, // 111010
    op_swl       = 42, // 101010
    op_swr       = 46, // 101110
    op_xori      = 14, // 001110
  };

  enum funcs {
    func_add     = 32, // 100000
    func_addu    = 33, // 100001
    func_and     = 36, // 100100
    func_break   = 13, // 001101
    func_clo     = 33, // 100001
    func_clz     = 32, // 100000
    func_dadd    = 44, // 101100
    func_daddu   = 45, // 101101
    func_dclo    = 37, // 100101
    func_dclz    = 36, // 100100
    func_ddiv    = 30, // 011110
    func_ddivu   = 31, // 011111
    func_deret   = 31, // 011111 COP0
    func_div     = 26, // 011010
    func_divu    = 27, // 011011
    func_dmult   = 28, // 011100
    func_dmultu  = 29, // 011101
    func_dsll    = 56, // 111000
    func_dsll32  = 60, // 111100
    func_dsllv   = 20, // 010100
    func_dsra    = 59, // 111011
    func_dsra32  = 63, // 111111
    func_dsrav   = 23, // 010111
    func_dsrl    = 58, // 111010
    func_dsrl32  = 62, // 111110
    func_dsrlv   = 22, // 010110
    func_dsub    = 46, // 101110
    func_dsubu   = 47, // 101111
    func_eret    = 24, // 011000 COP0
    func_jalr    =  9, // 001001
    func_jr      =  8, // 001000
    func_madd    =  0,
    func_maddu   =  1, // 000001
    func_mfhi    = 16, // 010000
    func_mflo    = 18, // 010010
    func_movci   =  1, // 000001
    func_movn    = 11, // 001011
    func_movz    = 10, // 001010
    func_msub    =  4, // 000100
    func_msubu   =  5, // 000101
    func_mthi    = 17, // 010001
    func_mtlo    = 19, // 010011
    func_mul     =  2, // 000010
    func_mult    = 24, // 011000
    func_multu   = 25, // 011001
    func_nor     = 39, // 100111
    func_or      = 37, // 100101
    func_prefx   = 15,
    func_sdbbp   = 63, // 111111
    func_sll     =  0, // 000000
    func_sllv    =  4, // 000100
    func_slt     = 42, // 101010
    func_sltu    = 43, // 101011
    func_sra     =  3, // 000011
    func_srav    =  7, // 000111
    func_srl     =  2, // 000010
    func_srlv    =  6, // 000110
    func_sub     = 34, // 100010
    func_subu    = 35, // 100011
    func_sync    = 15, // 001111
    func_syscall = 12, // 001100
    func_teq     = 52, // 110100
    func_tge     = 48, // 110000
    func_tgeu    = 49, // 110001
    func_tlbp    =  8, // 001000 COP0
    func_tlbr    =  1, // 000001 COP0
    func_tlbwi   =  2, // 000010 COP0
    func_tlbwr   =  6, // 000110 COP0
    func_tlt     = 50, // 110010
    func_tltu    = 51, // 110011
    func_tne     = 54, // 110110
    func_wait    = 32, // 100000 COP0
    func_xor     = 38, // 100110
  };

  enum func1s {
    func1_abs     =  9, // 000101 COP1
    func1_add     =  0, // 000000 COP1
    func1_ceil_l  = 10, // 001010 COP1
    func1_ceil_w  = 14, // 001110 COP1
    func1_cvt_d   = 33, // 100001 COP1
    func1_cvt_l   = 37, // 100101 COP1
    func1_cvt_s   = 32, // 100000 COP1
    func1_cvt_w   = 36, // 100100 COP1
    func1_div     =  3, // 000011 COP1
    func1_floor_l = 11, // 001011 COP1
    func1_floor_w = 15, // 001111 COP1
    func1_ldxc1   =  1, // 000001 COP1X
    func1_luxc1   =  5, // 000101 COP1X
    func1_lwxc1   =  0, // 000000 COP1X
    func1_madd    = 32, // 100xxx COP1X - must "or" with 3-bit format
    func1_mov     =  6, // 000110 COP1
    func1_movcf   = 17, // 010001 COP1
    func1_movn    = 19, // 010011 COP1
    func1_movz    = 18, // 010010 COP1
    func1_msub    = 40, // 101xxx COP1X - must "or" with 3-bit format
    func1_mul     =  2, // 000010 COP1
    func1_neg     =  7, // 000111 COP1
    func1_nmadd   = 48, // 110xxx COP1X - must "or" with 3-bit format
    func1_nmsub   = 56, // 111xxx COP1X - must "or" with 3-bit format
    func1_prefx   = 15, // 001111 COP1X
    func1_recip   = 21, // 010101 COP1
    func1_round_l =  8, // 001000 COP1
    func1_round_w = 12, // 001100 COP1
    func1_rsqrt   = 22, // 010110 COP1
    func1_sdxc1   =  9, // 001001 COP1X
    func1_sqrt    =  4, // 000100 COP1
    func1_sub     =  1, // 000001 COP1
    func1_suxc1   = 13, // 001101 COP1X
    func1_swxc1   =  8, // 001000 COP1X
    func1_trunc_l =  9, // 001001 COP1
    func1_trunc_w = 13, // 001101 COP1
  };

  // Used in many FP (COP1) instructions
  enum fmts {
    fmt_s = 16,  // single-precision
    fmt_d = 17,  // double-precision
    fmt_w = 20,  // word (4 bytes)
    fmt_l = 21   // long word (8 bytes)
  };

  // rs field used by cop0, cop1, cop2
  enum rsop {
    rsop_bc     =  8, // 01000
    rsop_cf     =  2, // 00010
    rsop_ct     =  6, // 00110
    rsop_dmf    =  1, // 00001 COP0, COP1, COP2
    rsop_dmt    =  5, // 00101 COP0, COP1, COP2
    rsop_mf     =  0, // 00000 COP0, COP1, COP2
    rsop_mt     =  4, // 00100 COP0, COP1, COP2
  };

  // rt field used by the REGIMM op
  enum rtop {
    rtop_bgez   =  1, // 00001
    rtop_bgezal = 17, // 10001
    rtop_bltz   =  0, // 00000
    rtop_bltzal = 16, // 10000
    rtop_teqi   = 12, // 01100
    rtop_tgei   =  8, // 01000
    rtop_tgeiu  =  9, // 01001
    rtop_tlti   = 10, // 01010
    rtop_tltiu  = 11, // 01011
    rtop_tnei   = 14, // 01110
  };

  // Used for the c.cond.fmt instructions
  enum c_conds {
    cond_f   =  0,  // false
    cond_un  =  1,  // unordered
    cond_eq  =  2,  // equal
    cond_ueq =  3,  // uordered or equal
    cond_olt =  4,  // ordered or less than
    cond_ult =  5,  // unordered or less than
    cond_ole =  6,  // ordered or less than or equal
    cond_ule =  7,  // unordered or less than or equal
    cond_sf  =  8,  // signaling false
    cond_ngle=  9,  // not greater than or less than or equal
    cond_seq = 10,  // signalling equal
    cond_ngl = 11,  // not greater than or less than
    cond_lt  = 12,  // less than
    cond_nge = 13,  // not greater than or equal
    cond_le  = 14,  // less than or equal
    cond_ngt = 15   // not greater than
  };

  enum cp1crs {
    cp1_fir  =  0,  // capabilities (read-only)
    cp1_fcsr = 31,  // all info
    cp1_fccr = 25,  // condition codes
    cp1_fexr = 26,  // exceptional-condition info
    cp1_fenr = 28   // exceptional-condition enables
  };

  enum fcsr_fields {
    fcsr_cause_invalidop = 0x10000
    // There are many more here; add them as needed
  };

  enum Condition {
    never                 =  0,
    equal                 =  1,
    less                  =  2,
    lessUnsigned          =  3,
    lessEqual             =  4,
    lessEqualUnsigned     =  5,
    always                =  8,
    notEqual              =  9,
    greater               =  10,
    greaterUnsigned       =  11,
    greaterEqual          =  12, 
    greaterEqualUnsigned  =  13
  };

  // Creation
  Assembler(CodeBuffer* code) : AbstractAssembler(code) {
#ifdef CHECK_DELAY
    delay_state = no_delay; 
#endif
  }

  static address locate_next_instruction(address inst);

  Condition negateCondition(Condition cond);

  // Required platform-specific helpers for Label::patch_instructions.
  // They _shadow_ the declarations in AbstractAssembler, which are undefined.
  void pd_patch_instruction(address branch, address target);
#ifndef PRODUCT
  static void pd_print_patched_instruction(address branch);
#endif

  // Core instructions
  // Operands are in assembler order
  // Underscore suffixes indicate name conflict with C++ keyword

  inline void add(Register rd, Register rs, Register rt);
  inline void addi(Register rt, Register rs, int imm16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void addiu(Register rt, Register rs, int imm16,
		    relocInfo::relocType rtype = relocInfo::none);
  inline void addu(Register rd, Register rs, Register rt);
  inline void and_(Register rd, Register rs, Register rt);
  inline void andi(Register rt, Register rs, int imm16);

  inline void beq(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  // beql deprecated
  inline void bgez(Register rs, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bgezal(Register rs, int offset16,
		     relocInfo::relocType rtype = relocInfo::none);
  // bgezall deprecated
  // bgezl deprecated
  inline void bgtz(Register rs, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // bgtzl deprecated
  inline void blez(Register rs, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // blezl deprecated
  inline void bltz(Register rs, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bltzal(Register rs, int offset16,
		     relocInfo::relocType rtype = relocInfo::none);
  // bltzall deprecated
  // bltzl deprecated
  inline void bne(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  // bnel deprecated

  inline void break_(int code20 = 0);
  inline void cache(int op5, int offset16, Register base,
		    relocInfo::relocType rtype = relocInfo::none);

  inline void cl0(Register rd, Register rs);
  inline void clz(Register rd, Register rs);

  inline void dadd(Register rd, Register rs, Register rt);
  inline void daddi(Register rt, Register rs, int imm16,
		    relocInfo::relocType rtype = relocInfo::none);
  inline void daddiu(Register rt, Register rs, int imm16,
		     relocInfo::relocType rtype = relocInfo::none);
  inline void daddu(Register rd, Register rs, Register rt);
  inline void dcl0(Register rd, Register rs);
  inline void dclz(Register rd, Register rs);
  inline void ddiv(Register rs, Register rt);
  inline void ddivu(Register rs, Register rt);
  inline void deret();
  inline void div(Register rs, Register rt);
  inline void divu(Register rs, Register rt);
  inline void dmult(Register rs, Register rt);
  inline void dmultu(Register rs, Register rt);
  inline void dsll(Register rd, Register rt, int sa5);
  inline void dsll32(Register rd, Register rt, int sa5);
  inline void dsllv(Register rd, Register rt, Register rs);
  inline void dsra(Register rd, Register rt, int sa5);
  inline void dsra32(Register rd, Register rt, int sa5);
  inline void dsrav(Register rd, Register rt, Register rs);
  inline void dsrl(Register rd, Register rt, int sa5);
  inline void dsrl32(Register rd, Register rt, int sa5);
  inline void dsrlv(Register rd, Register rt, Register rs);
  inline void dsub(Register rd, Register rs, Register rt);
  inline void dsubu(Register rd, Register rs, Register rt);

  inline void eret();

  inline void j(int index26,
		relocInfo::relocType rtype = relocInfo::none);
  inline void jal(int index26,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void jalr(Register rs, int hint5 = 0);  // assume rd = $31
  inline void jalr(Register rd, Register rs, int hint5);
  inline void jr(Register rs, int hint5 = 0);

  inline void lb(Register rt, int offset16, Register base);
  inline void lbu(Register rt, int offset16, Register base);
  inline void ld(Register rt, int offset16, Register base);
  inline void ldl(Register rt, int offset16, Register base);
  inline void ldr(Register rt, int offset16, Register base);
  inline void lh(Register rt, int offset16, Register base);
  inline void lhu(Register rt, int offset16, Register base);
  inline void ll(Register rt, int offset16, Register base);
  inline void lld(Register rt, int offset16, Register base);
  inline void lui(Register rt, int imm16);
  inline void lw(Register rt, int offset16, Register base);
  inline void lwl(Register rt, int offset16, Register base);
  inline void lwr(Register rt, int offset16, Register base);
  inline void lwu(Register rt, int offset16, Register base);

  inline void madd(Register rs, Register rt);
  inline void maddu(Register rs, Register rt);
  inline void mfhi(Register rd);
  inline void mflo(Register rd);
  inline void movf(Register rd, Register rs, int cc3);
  inline void movn(Register rd, Register rs, Register rt);
  inline void movt(Register rd, Register rs, int cc3);
  inline void movz(Register rd, Register rs, Register rt);
  inline void msub(Register rs, Register rt);
  inline void msubu(Register rs, Register rt);
  inline void mthi(Register rs);
  inline void mtlo(Register rs);
  inline void mul(Register rd, Register rs, Register rt);
  inline void mult(Register rs, Register rt);
  inline void multu(Register rs, Register rt);

  inline void nor(Register rd, Register rs, Register rt);
  inline void or_(Register rd, Register rs, Register rt);
  inline void ori(Register rt, Register rs, int imm16);

  inline void pref(int hint5, int offset16, Register base);
  inline void prefx(int hint5, Register index, Register base);

  inline void sb(Register rt, int offset16, Register base);
  inline void sc(Register rt, int offset16, Register base);
  inline void scd(Register rt, int offset16, Register base);
  inline void sd(Register rt, int offset16, Register base);
  inline void sdbbp(int code20);
  inline void sdl(Register rt, int offset16, Register base);
  inline void sdr(Register rt, int offset16, Register base);
  inline void sh(Register rt, int offset16, Register base);
  inline void sll(Register rd, Register rt, int sa5);
  inline void sllv(Register rd, Register rt, Register rs);
  inline void slt(Register rd, Register rs, Register rt);
  inline void slti(Register rt, Register rs, int imm16);
  inline void sltiu(Register rt, Register rs, int imm16);
  inline void sltu(Register rd, Register rs, Register rt);
  inline void sra(Register rd, Register rt, int sa5);
  inline void srav(Register rd, Register rt, Register rs);
  inline void srl(Register rd, Register rt, int sa5);
  inline void srlv(Register rd, Register rt, Register rs);
  inline void ssnop();
  inline void sub(Register rd, Register rs, Register rt);
  inline void subu(Register rd, Register rs, Register rt);
  inline void sw(Register rt, int offset16, Register base);
  inline void swl(Register rt, int offset16, Register base);
  inline void swr(Register rt, int offset16, Register base);
  inline void sync();
  inline void syscall(int code20);

  inline void teq(Register rs, Register rt, int code10 = 0);
  inline void teqi(Register rs, int imm16);
  inline void tge(Register rs, Register rt, int code10 = 0);
  inline void tgei(Register rs, int imm16);
  inline void tgeiu(Register rs, int imm16);
  inline void tgeu(Register rs, Register rt, int code10 = 0);
  inline void tlbp();
  inline void tlbr();
  inline void tlbwi();
  inline void tlbwr();
  inline void tlt(Register rs, Register rt, int code10 = 0);
  inline void tlti(Register rs, int imm16);
  inline void tltiu(Register rs, int imm16);
  inline void tltu(Register rs, Register rt, int code10 = 0);
  inline void tne(Register rs, Register rt, int code10 = 0);
  inline void tnei(Register rs, int imm16);
  inline void wait();
  inline void xor_(Register rd, Register rs, Register rt);
  inline void xori(Register rt, Register rs, int imm16);


  // coprocessor instructions

  inline void bc1f(int cc3, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // bc1fl deprecated
  inline void bc1t(int cc3, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // bc1tl deprecated

  inline void bc2f(int cc3, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // bc2fl deprecated
  inline void bc2t(int cc3, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  // bc2tl deprecated

  inline void cfc1(Register rt, int fs5);  // see cp1crs enum
  inline void cfc2(Register rt, int rd5);
  inline void ctc1(Register rt, int fs5);  // see cp1crs enum
  inline void ctc2(Register rt, int rd5);

  inline void dmfc0(Register rt, int rd5, int sel3);
  inline void dmfc1(Register rt, FloatRegister fs);
  inline void dmfc2(Register rt, int rd5, int sel3);
  inline void dmtc0(Register rt, int rd5, int sel3);
  inline void dmtc1(Register rt, FloatRegister fs);
  inline void dmtc2(Register rt, int rd5, int sel3);

  inline void ldc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void ldc2(int rt5, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void ldxc1(FloatRegister fd, Register index, Register base);
  inline void luxc1(FloatRegister fd, Register index, Register base);
  inline void lwc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void lwc2(int rt5, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void lwxc1(FloatRegister fd, Register index, Register base);

  inline void mfc0(Register rt, int rd5, int sel3);
  inline void mfc1(Register rt, FloatRegister fs);
  inline void mfc2(Register rt, int rd5, int sel3);
  inline void mtc0(Register rt, int rd5, int sel3);
  inline void mtc1(Register rt, FloatRegister fs);
  inline void mtc2(Register rt, int rd5, int sel3);

  inline void sdc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void sdc2(int rt5, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void sdxc1(FloatRegister fs, Register index, Register base);
  inline void suxc1(FloatRegister fs, Register index, Register base);
  inline void swc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void swc2(int rt5, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void swxc1(FloatRegister ft, Register index, Register base);


  // COP1 instructions parameterized by type (see enum fmts)

  inline void abs_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs);
  inline void add_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void c_cond_fmt(int fmt5, c_conds cond4,
			 int cc3, FloatRegister fs, FloatRegister ft);
  inline void c_cond_fmt(int fmt5, c_conds cond4,
			 FloatRegister fs, FloatRegister ft);
  inline void ceil_l_fmt(int fmt5,
			 FloatRegister fd, FloatRegister fs);
  inline void ceil_w_fmt(int fmt5,
			 FloatRegister fd, FloatRegister fs);
  inline void cvt_d_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void cvt_l_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void cvt_s_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void cvt_w_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void div_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void floor_l_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);
  inline void floor_w_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);
  inline void madd_fmt(int fmt3,
		       FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void mov_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs);
  inline void movf_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs, int cc3);
  inline void movn_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs, Register rt);
  inline void movt_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs, int cc3);
  inline void movz_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs, Register rt);
  inline void msub_fmt(int fmt3,
		       FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void mul_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void neg_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs);
  inline void nmadd_fmt(int fmt3,
			FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void nmsub_fmt(int fmt3,
			FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void recip_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void round_l_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);
  inline void round_w_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);
  inline void rsqrt_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs);
  inline void sub_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void trunc_l_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);
  inline void trunc_w_fmt(int fmt5,
			  FloatRegister fd, FloatRegister fs);

protected:
  inline void check_delay();
};


class RegistersForDebugging : public StackObj {
 public:
  intptr_t r[32];
  double f[32];

  static int r_offset(int j) { return offset_of(RegistersForDebugging, r[j]); }
  static int f_offset(int j) { return offset_of(RegistersForDebugging, f[j]); }

  // gen asm code to save regs
  static void save_registers(MacroAssembler* a);

  // restore global registers in case C code disturbed them
  static void restore_registers(MacroAssembler* a);

  void print(outputStream* s);
};


class MacroAssembler: public Assembler {

public:
  MacroAssembler(CodeBuffer* code) : Assembler(code) {}

  // Tells assembler you know that next instruction is delayed
  // YYY add checks for multi-instruction expansion in delay slot
  MacroAssembler* delayed() {
#ifdef CHECK_DELAY
    guarantee( delay_state == at_delay_slot, "delayed instruction is not in delay slot");
    delay_state = filling_delay_slot;
#endif
    return this;
  }

protected:

  // Support for VM calls
  // It is imperative that all calls into the VM are handled via the
  // call_VM macros.  They make sure that the stack linkage is setup
  // correctly. call_VM's correspond to ENTRY/ENTRY_X entry points
  // while call_VM_leaf's correspond to LEAF entry points.

  // This is the base routine called by the different versions of
  // call_VM_leaf. The interpreter may customize this version by
  // overriding it for its purposes (e.g., to save/restore additional
  // registers when doing a VM call).

  void call_VM_leaf_base(Register thread_cache,
			 address entry_point,
			 int number_of_arguments);

  // This is the base routine called by the different versions of
  // call_VM. The interpreter may customize this version by overriding
  // it for its purposes (e.g., to save/restore additional registers
  // when doing a VM call).
  //
  // A non-volatile java_thread_cache register should be specified so
  // that it will be preserved across the call.  (If java_thread_cache
  // is noreg, then a slow get_thread call will re-initialize
  // Rthread.) call_VM_base returns the register that contains the
  // thread.
  //
  // If no last_java_sp is specified (noreg) then SP will be used
  // instead.

  virtual void call_VM_base(
    Register  oop_result,           // where an oop-result ends up if any; use noreg otherwise
    Register  thread_cache,         // the thread if computed before     ; use noreg otherwise
    Register  last_java_sp,         // to set up last_Java_frame in stubs; use noreg otherwise
    address   entry_point,          // the entry point
    int       number_of_arguments,  // the number of arguments (w/o thread) to pop after call
    bool      check_exception=true  // whether to check for pending exceptions after return
  );


public:

  // Debugging
  void _verify_oop(Register reg, const char * msg, const char * file, int line);
  void _verify_oop_addr(Address addr, const char * msg, const char * file, int line);

#define verify_oop(reg) _verify_oop(reg, "broken oop " #reg, __FILE__, __LINE__)
#define verify_oop_addr(addr) _verify_oop_addr(addr, "broken oop addr ", __FILE__, __LINE__)

  // factor out part of verify_oop into subroutine to save space
  void verify_oop_subroutine();

  void stop(const char* msg);                          // prints msg, dumps registers and stops execution
#define STOP(msg) stop(__FILE__ ":" XSTR(__LINE__) ": " msg)
#define UNIMPLEMENTED() STOP("unimplemented")
  void warn(const char* msg, Register Rval);           // prints msg and value
  // factor out part of stop into subroutine to save space
  void stop_subroutine();
  void debug(char* msg, RegistersForDebugging* regs);
  void untested(const char* what);
  void unimplemented(const char* what = "")  { char* b = new char[1024];  sprintf(b, "unimplemented: %s", what);  stop(b); }
  void should_not_reach_here()               { stop("should not reach here"); }
#ifndef PRODUCT
  void test();
#endif

  // Standard synthetic instructions

  inline void nop();

  void addiu(Register rd, Register rs, int imm32);
  void daddiu(Register rd, Register rs, long imm64);
  void andi(Register rd, Register rs, long imm32_64);

  inline void b(int offset16, relocInfo::relocType rtype = relocInfo::none);
  inline void bge(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bgeu(Register rs, Register rt, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bgt(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bgtu(Register rs, Register rt, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void ble(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bleu(Register rs, Register rt, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void blt(Register rs, Register rt, int offset16,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bltu(Register rs, Register rt, int offset16,
		   relocInfo::relocType rtype = relocInfo::none);

  void li(Register rd, long imm);
  void li(Register rd, address addr) { li(rd, (long)addr); }
  void li32(Register rd, int imm32); // sign-extends to 64 bits on mips64
#ifdef _LP64
  void dli(Register rd, long imm) { li(rd, imm); }
#endif

  inline void move(Register rd, Register rs);

  // Extended load instructions

  inline void lb(Register rt, Register offset, Register base);
  inline void lbu(Register rt, Register offset, Register base);
  inline void lh(Register rt, Register offset, Register base);
  inline void lhu(Register rt, Register offset, Register base);
  inline void lw(Register rt, Register offset, Register base);

  inline void sb(Register rt, Register offset, Register base);
  inline void sh(Register rt, Register offset, Register base);
  inline void sw(Register rt, Register offset, Register base);

  inline void swc1(FloatRegister ft, Register offset, Register base);
  inline void sdc1(FloatRegister ft, Register offset, Register base);

  // more?

  // Overloadings for visibility
  inline void lb(Register rt, int offset16, Register base);
  inline void lbu(Register rt, int offset16, Register base);
  inline void lh(Register rt, int offset16, Register base);
  inline void lhu(Register rt, int offset16, Register base);
  inline void lw(Register rt, int offset16, Register base);

  inline void sb(Register rt, int offset16, Register base);
  inline void sh(Register rt, int offset16, Register base);
  inline void sw(Register rt, int offset16, Register base);

  inline void swc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void sdc1(FloatRegister ft, int offset16, Register base,
		   relocInfo::relocType rtype = relocInfo::none);

  // Standard synthetic FP instructions

  inline void abs_s(FloatRegister fd, FloatRegister fs);
  inline void abs_d(FloatRegister fd, FloatRegister fs);
  inline void add_s(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void add_d(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void c_cond_s(c_conds cond4, int cc3, FloatRegister fs, FloatRegister ft);
  inline void c_cond_d(c_conds cond4, int cc3, FloatRegister fs, FloatRegister ft);
  inline void c_cond_s(c_conds cond4, FloatRegister fs, FloatRegister ft);
  inline void c_cond_d(c_conds cond4, FloatRegister fs, FloatRegister ft);
  inline void ceil_l_s(FloatRegister fd, FloatRegister fs);
  inline void ceil_l_d(FloatRegister fd, FloatRegister fs);
  inline void ceil_w_s(FloatRegister fd, FloatRegister fs);
  inline void ceil_w_d(FloatRegister fd, FloatRegister fs);
  inline void cvt_d_s(FloatRegister fd, FloatRegister fs);
  inline void cvt_d_w(FloatRegister fd, FloatRegister fs);
  inline void cvt_d_l(FloatRegister fd, FloatRegister fs);
  inline void cvt_l_s(FloatRegister fd, FloatRegister fs);
  inline void cvt_l_d(FloatRegister fd, FloatRegister fs);
  inline void cvt_s_d(FloatRegister fd, FloatRegister fs);
  inline void cvt_s_w(FloatRegister fd, FloatRegister fs);
  inline void cvt_s_l(FloatRegister fd, FloatRegister fs);
  inline void cvt_w_s(FloatRegister fd, FloatRegister fs);
  inline void cvt_w_d(FloatRegister fd, FloatRegister fs);
  inline void div_s(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void div_d(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void floor_l_s(FloatRegister fd, FloatRegister fs);
  inline void floor_l_d(FloatRegister fd, FloatRegister fs);
  inline void floor_w_s(FloatRegister fd, FloatRegister fs);
  inline void floor_w_d(FloatRegister fd, FloatRegister fs);
  inline void madd_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void madd_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void mov_s(FloatRegister fd, FloatRegister fs);
  inline void mov_d(FloatRegister fd, FloatRegister fs);
  inline void movf_s(FloatRegister fd, FloatRegister fs, int cc3);
  inline void movf_d(FloatRegister fd, FloatRegister fs, int cc3);
  inline void movn_s(FloatRegister fd, FloatRegister fs, Register rt);
  inline void movn_d(FloatRegister fd, FloatRegister fs, Register rt);
  inline void movt_s(FloatRegister fd, FloatRegister fs, int cc3);
  inline void movt_d(FloatRegister fd, FloatRegister fs, int cc3);
  inline void movz_s(FloatRegister fd, FloatRegister fs, Register rt);
  inline void movz_d(FloatRegister fd, FloatRegister fs, Register rt);
  inline void msub_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void msub_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void mul_s(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void mul_d(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void neg_s(FloatRegister fd, FloatRegister fs);
  inline void neg_d(FloatRegister fd, FloatRegister fs);
  inline void nmadd_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void nmadd_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void nmsub_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void nmsub_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft);
  inline void recip_s(FloatRegister fd, FloatRegister fs);
  inline void recip_d(FloatRegister fd, FloatRegister fs);
  inline void round_l_s(FloatRegister fd, FloatRegister fs);
  inline void round_l_d(FloatRegister fd, FloatRegister fs);
  inline void round_w_s(FloatRegister fd, FloatRegister fs);
  inline void round_w_d(FloatRegister fd, FloatRegister fs);
  inline void rsqrt_s(FloatRegister fd, FloatRegister fs);
  inline void rsqrt_d(FloatRegister fd, FloatRegister fs);
  inline void sub_s(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void sub_d(FloatRegister fd, FloatRegister fs, FloatRegister ft);
  inline void trunc_l_s(FloatRegister fd, FloatRegister fs);
  inline void trunc_l_d(FloatRegister fd, FloatRegister fs);
  inline void trunc_w_s(FloatRegister fd, FloatRegister fs);
  inline void trunc_w_d(FloatRegister fd, FloatRegister fs);

  // Other helper functions

  inline int target_to_offset16(address target); // with error checking
  inline int target_to_offset16(Label& L);

  // Branches using address
  inline void b(address, relocInfo::relocType = relocInfo::none);
  void beq(Register rs, Register rt, address addr,
	   relocInfo::relocType rtype = relocInfo::none);
  void bge(Register rs, Register rt, address,
	   relocInfo::relocType = relocInfo::none);
  void bgeu(Register rs, Register rt, address,
	    relocInfo::relocType = relocInfo::none);
  void bgez(Register rs, address addr,
	    relocInfo::relocType rtype = relocInfo::none);
  void bgezal(Register rs, address addr,
	      relocInfo::relocType rtype = relocInfo::none);
  void bgt(Register rs, Register rt, address,
	   relocInfo::relocType = relocInfo::none);
  void bgtu(Register rs, Register rt, address,
	    relocInfo::relocType = relocInfo::none);
  void bgtz(Register rs, address addr,
	    relocInfo::relocType rtype = relocInfo::none);
  void ble(Register rs, Register rt, address,
	   relocInfo::relocType = relocInfo::none);
  void bleu(Register rs, Register rt, address,
	    relocInfo::relocType = relocInfo::none);
  void blez(Register rs, address addr,
	    relocInfo::relocType rtype = relocInfo::none);
  void blt(Register rs, Register rt, address,
	   relocInfo::relocType = relocInfo::none);
  void bltu(Register rs, Register rt, address,
	    relocInfo::relocType = relocInfo::none);
  void bltz(Register rs, address addr,
	    relocInfo::relocType rtype = relocInfo::none);
  void bltzal(Register rs, address addr,
	      relocInfo::relocType rtype = relocInfo::none);
  void bne(Register rs, Register rt, address addr,
	   relocInfo::relocType rtype = relocInfo::none);

  // Branches using Label
  inline void b(Label&, relocInfo::relocType = relocInfo::none);
  inline void beq(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bgez(Register rs, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bgezal(Register rs, Label& L,
		     relocInfo::relocType rtype = relocInfo::none);
  inline void bgtz(Register rs, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void blez(Register rs, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bltz(Register rs, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bltzal(Register rs, Label& L,
		     relocInfo::relocType rtype = relocInfo::none);
  inline void bne(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);

  inline void bge(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bgeu(Register rs, Register rt, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void bgt(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bgtu(Register rs, Register rt, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void ble(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bleu(Register rs, Register rt, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);
  inline void blt(Register rs, Register rt, Label& L,
		  relocInfo::relocType rtype = relocInfo::none);
  inline void bltu(Register rs, Register rt, Label& L,
		   relocInfo::relocType rtype = relocInfo::none);

  inline void bc1f(int cc3, Label& L);
  inline void bc1t(int cc3, Label& L);
  inline void bc1f(Label& L) { bc1f(0, L); }
  inline void bc1t(Label& L) { bc1t(0, L); }

  // Miscellaneous branches, jumps
  void bx(Condition cond, Register r1, Register r2, address addr,
	  relocInfo::relocType rtype = relocInfo::none);
  void bx(Condition cond, Register r1, Register r2, Label& L,
	  relocInfo::relocType rtype = relocInfo::none);

  inline void j(Register rs) { jr(rs); }
  // these load the target address into AT; have unfilled delay slot
  inline void jump_to(address a, relocInfo::relocType rtype = relocInfo::none);
  inline void jal_to(address a, relocInfo::relocType rtype = relocInfo::none);
  // YYY redundant with above -- remove?
  inline void call(address target, relocInfo::relocType rtype);
  inline void jal_to_C(address target, relocInfo::relocType rtype = relocInfo::none);

  // Argument ops
  inline void store_int_argument(Register s, Argument& a);
  inline void store_long_argument(Register s, Argument& a);
  inline void store_float_argument(FloatRegister s, Argument& a);
  inline void store_double_argument(FloatRegister s, Argument& a);
  inline void store_ptr_argument(Register s, Argument& a);

  // Functions for isolating 64 bit ops for LP64

  // Regular vs. d* versions
  inline void addu_long(Register rd, Register rs, Register rt);
  inline void addu_long(Register rd, Register rs, long imm32_64);
  inline void sll_long(Register rd, Register rt, int sa5);
  inline void sllv_long(Register rd, Register rt, Register rs);
  inline void sra_long(Register rd, Register rt, int sa5);
  inline void srav_long(Register rd, Register rt, Register rs);
  inline void srl_long(Register rd, Register rt, int sa5);
  inline void srlv_long(Register rd, Register rt, Register rs);
  inline void subu_long(Register rd, Register rs, Register rt);

  // ld_ptr will perform lw for 32 bit VMs and ld for 64 bit VMs
  // st_ptr will perform sw for 32 bit VMs and sd for 64 bit VMs
  inline void ld_ptr(Register rt, int offset16, Register base);
  inline void st_ptr(Register rt, int offset16, Register base);

  void ld_ptr(Register rt, Register offset, Register base);
  void st_ptr(Register rt, Register offset, Register base);

  // ld_long will perform lw for 32 bit VMs and ld for 64 bit VMs
  // st_long will perform sw for 32 bit VMs and sd for 64 bit VMs
  inline void ld_long(Register rt, int offset16, Register base);
  inline void st_long(Register rt, int offset16, Register base);

  void ld_long(Register rt, Register offset, Register base);
  void st_long(Register rt, Register offset, Register base);

  inline void ll_long(Register rt, Register offset, Register base);
  inline void ll_long(Register rt, int offset16, Register base);
  inline void sc_long(Register rt, Register offset, Register base);
  inline void sc_long(Register rt, int offset16, Register base);

  // common JVM operations

  void save_registers(Register save_base = SP);
  void restore_registers(Register save_base = SP);

  void save_frame(int extraWords = 0);
  void restore_frame(int extraWords = 0);
  void restore_frame_and_return(int extraWords = 0);  // use delay slot
  int total_frame_size_in_bytes(int extraWords = 0);

  void save_minimal_frame(int extraWords = 0);
  void restore_minimal_frame(int extraWords = 0);
  void restore_minimal_frame_and_return(int extraWords = 0);  // use delay slot

  void get_thread(Register thread = Rthread);
  void verify_thread();

  void set_last_Java_frame(Register last_java_sp, Register last_java_fp, Register last_java_pc);
  void reset_last_Java_frame(void);

  // Call into the VM.
  // Passes the thread pointer (in A0) as a prepended argument.
  // Makes sure oop return values are visible to the GC.
  void call_VM(Register oop_result, address entry_point, int number_of_arguments = 0, bool check_exceptions = true);
  void call_VM(Register oop_result, address entry_point, Register arg_1, bool check_exceptions = true);
  void call_VM(Register oop_result, address entry_point, Register arg_1, Register arg_2, bool check_exceptions = true);
  void call_VM(Register oop_result, address entry_point, Register arg_1, Register arg_2, Register arg_3, bool check_exceptions = true);

  // Same as above, but specifies Register last_java_sp and
  // not check_exceptions (used in templateTable.cpp).
  void call_VM(Register oop_result, Register last_java_sp, address entry_point); // 0 args
  void call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1);
  void call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1, Register arg_2);
  void call_VM(Register oop_result, Register last_java_sp, address entry_point, Register arg_1, Register arg_2, Register arg_3);

// YYY thread_cache is vestigial, and can probably be removed later.
// YYY We are not using it currently.
  void call_VM_leaf(Register thread_cache, address entry_point, int number_of_arguments = 0);
  void call_VM_leaf(Register thread_cache, address entry_point, Register arg_1);
  void call_VM_leaf(Register thread_cache, address entry_point, Register arg_1, Register arg_2);
  void call_VM_leaf(Register thread_cache, address entry_point, Register arg_1, Register arg_2, Register arg_3);

  void call_VM_leaf(address entry_point, int number_of_arguments = 0) {
    call_VM_leaf(noreg, entry_point, number_of_arguments);
  }
  void call_VM_leaf(address entry_point, Register arg_1) {
    call_VM_leaf(noreg, entry_point, arg_1);
  }
  void call_VM_leaf(address entry_point, Register arg_1, Register arg_2) {
    call_VM_leaf(noreg, entry_point, arg_1, arg_2);
  }
  void call_VM_leaf(address entry_point, Register arg_1, Register arg_2, Register arg_3) {
    call_VM_leaf(noreg, entry_point, arg_1, arg_2, arg_3);
  }

  void get_vm_result  (Register oop_result);
  void get_vm_result_2(Register oop_result);
  void set_vm_result(Register oop_result);

  void check_and_forward_exception(Register scratch_reg);
  void check_and_handle_popframe(Register scratch_reg);
  void check_and_handle_earlyret(Register scratch_reg);

  void null_check(Register reg, int offset = -1);
  static bool needs_explicit_null_check(int offset);

  void store_check(Register tmp, Register obj);                  // store check for obj - register is destroyed afterwards
  void store_check(Register tmp, Register obj, Register offset); // store check for obj - register is destroyed afterwards

  void verify_FPU(int stack_depth, const char* s = "illegal FPU state");
	// only if +VerifyFPU

  // nop padding
  void align(int modulus);

  inline void round_to(Register r, int modulus) {
    assert_not_delayed();
    assert(modulus > 0, "modulus must be positive"); // should check power of 2
    addu_long(r, r, modulus - 1);
    li(AT, ~(modulus - 1));
    and_(r, r, AT);  // andi won't work because it zero-extends
  }

  inline void round_down_to(Register r, int modulus) {
    li(AT, ~(modulus - 1));
    and_(r, r, AT);  // andi won't work because it zero-extends
  }

  void bang_stack_with_offset(int offset);

  void membar();
};


/**
 * class SkipIfEqual:
 *
 * Instantiating this class will result in assembly code being output that will
 * jump around any code emitted between the creation of the instance and its
 * automatic destruction at the end of a scope block, depending on the value of
 * the flag passed to the constructor, which will be checked at run-time.
 */
class SkipIfEqual : public StackObj {
 private:
  MacroAssembler* _masm;
  Label _label;

 public:
  // 'temp' is a temp register that this object can use (and trash).
  // r2 is a register that the loaded value will be compared to.
  SkipIfEqual(MacroAssembler*, Register temp, Register r2,
              const bool* flag_addr, Assembler::Condition condition);
  ~SkipIfEqual();
};

#ifdef ASSERT
// No benefit -- all our instructions are the same size.
inline bool AbstractAssembler::pd_check_instruction_mark() { return false; }
#endif
