#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)assembler_sparc.inline.hpp	1.79 07/05/05 17:04:24 JVM"
#endif
/*
 * Copyright 1997-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

inline void Assembler::check_delay() {
# ifdef CHECK_DELAY
  guarantee( delay_state != at_delay_slot, "must say delayed() when filling delay slot");
  delay_state = no_delay;
# endif
}

// a misnomer on 64-bit platforms: this is really always a 32-bit emit
inline void Assembler::emit_long(int x) {
  check_delay();
  AbstractAssembler::emit_long(x);
}

inline void Assembler::emit_rdat(int x, relocInfo::relocType rtype) {
  relocate(rtype);
  emit_long(x);
}

inline void Assembler::emit_rdat(int x, RelocationHolder const& rspec) {
  relocate(rspec);
  emit_long(x);
}


// Core instructions

inline void Assembler::add(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_add) );
}

inline void Assembler::addi(Register rt, Register rs, int imm16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_addi) | xrs(rs) | xrt(rt) | ximmed(imm16), rtype );
}

inline void Assembler::addiu(Register rt, Register rs, int imm16,
			     relocInfo::relocType rtype) {
  emit_rdat( xop(op_addiu) | xrs(rs) | xrt(rt) | ximmed(imm16), rtype );
}

inline void Assembler::addu(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_addu) );
}

inline void Assembler::and_(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_and) );
}

inline void Assembler::andi(Register rt, Register rs, int imm16) {
  emit_long( xop(op_andi) | xrs(rs) | xrt(rt) | ximmedu(imm16) );
}

inline void Assembler::beq(Register rs, Register rt, int offset16,
			   relocInfo::relocType rtype) {
  emit_rdat( xop(op_beq) | xrs(rs) | xrt(rt) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::bgez(Register rs, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_regimm) | xrs(rs) | xrtop(rtop_bgez) | ximmed(offset16),
	     rtype );
  has_delay_slot();
}

inline void Assembler::bgezal(Register rs, int offset16,
			      relocInfo::relocType rtype) {
  emit_rdat( xop(op_regimm) | xrs(rs) | xrtop(rtop_bgezal) | ximmed(offset16),
	     rtype );
  has_delay_slot();
}

inline void Assembler::bgtz(Register rs, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_bgtz) | xrs(rs) | xrtop(0) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::blez(Register rs, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_blez) | xrs(rs) | xrtop(0) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::bltz(Register rs, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_regimm) | xrs(rs) | xrtop(rtop_bltz) | ximmed(offset16),
	     rtype );
  has_delay_slot();
}

inline void Assembler::bltzal(Register rs, int offset16,
			      relocInfo::relocType rtype) {
  emit_rdat( xop(op_regimm) | xrs(rs) | xrtop(rtop_bltzal) | ximmed(offset16),
	     rtype );
  has_delay_slot();
}

inline void Assembler::bne(Register rs, Register rt, int offset16,
			   relocInfo::relocType rtype) {
  emit_rdat( xop(op_bne) | xrs(rs) | xrt(rt) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::break_(int code20) {
  emit_long( xop(op_special) | u_field(code20, 25, 6) | xfunc(func_break) );
}

inline void Assembler::cache(int op5, int offset16, Register base,
			     relocInfo::relocType rtype) {
  emit_rdat( xop(op_cache) | xrs(base) | xrtop(op5) | ximmed(offset16),
	     rtype );
}

inline void Assembler::cl0(Register rd, Register rs) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rd) | xrd(rd) | xsa(0) | xfunc(func_clo) );
}

inline void Assembler::clz(Register rd, Register rs) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rd) | xrd(rd) | xsa(0) | xfunc(func_clz) );
}

inline void Assembler::dadd(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dadd) );
}

inline void Assembler::daddi(Register rt, Register rs, int imm16,
			     relocInfo::relocType rtype) {
  emit_rdat( xop(op_daddi) | xrs(rs) | xrt(rt) | ximmed(imm16), rtype );
}

inline void Assembler::daddiu(Register rt, Register rs, int imm16,
			      relocInfo::relocType rtype) {
  emit_rdat( xop(op_daddiu) | xrs(rs) | xrt(rt) | ximmed(imm16), rtype );
}

inline void Assembler::daddu(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_daddu) );
}

inline void Assembler::dcl0(Register rd, Register rs) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rd) | xrd(rd) | xsa(0) | xfunc(func_dclo) );
}

inline void Assembler::dclz(Register rd, Register rs) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rd) | xrd(rd) | xsa(0) | xfunc(func_dclz) );
}

inline void Assembler::ddiv(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_ddiv) );
}

inline void Assembler::ddivu(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_ddivu) );
}

inline void Assembler::deret() {
  // YYY undefined in the delay slot.  Check?
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_deret) );
}

inline void Assembler::div(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_div) );
}

inline void Assembler::divu(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_divu) );
}

inline void Assembler::dmult(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_dmult) );
}

inline void Assembler::dmultu(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_dmultu) );
}

inline void Assembler::dsll(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsll) );
}

inline void Assembler::dsll32(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsll32) );
}

inline void Assembler::dsllv(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dsllv) );
}

inline void Assembler::dsra(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsra) );
}

inline void Assembler::dsra32(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsra32) );
}

inline void Assembler::dsrav(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dsrav) );
}

inline void Assembler::dsrl(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsrl) );
}

inline void Assembler::dsrl32(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_dsrl32) );
}

inline void Assembler::dsrlv(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dsrlv) );
}

inline void Assembler::dsub(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dsub) );
}

inline void Assembler::dsubu(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_dsubu) );
}

inline void Assembler::eret() {
  // YYY undefined in the delay slot.  Check?
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_eret) );
}

inline void Assembler::j(int index26,
			 relocInfo::relocType rtype) {
  emit_rdat( xop(op_j) | xtarget(index26), rtype );
  has_delay_slot();
}

inline void Assembler::jal(int index26,
			   relocInfo::relocType rtype) {
  emit_rdat( xop(op_jal) | xtarget(index26), rtype );
  has_delay_slot();
}

inline void Assembler::jalr(Register rs, int hint5) {
  // assume rd = $31
  emit_long( xop(op_special) | xrs(rs) | xrtop(0) | xrdop(31) | xsa(hint5) |
	     xfunc(func_jalr) );
  has_delay_slot();
}

inline void Assembler::jalr(Register rd, Register rs, int hint5) {
  emit_long( xop(op_special) | xrs(rs) | xrtop(0) | xrd(rd) | xsa(hint5) | xfunc(func_jalr) );
  has_delay_slot();
}

inline void Assembler::jr(Register rs, int hint5) {
  emit_long( xop(op_special) | xrs(rs) | xrtop(0) | xrdop(0) | xsa(hint5) | xfunc(func_jr) );
  has_delay_slot();
}

inline void Assembler::lb(Register rt, int offset16, Register base) {
  emit_long( xop(op_lb) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lbu(Register rt, int offset16, Register base) {
  emit_long( xop(op_lbu) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::ld(Register rt, int offset16, Register base) {
  emit_long( xop(op_ld) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::ldl(Register rt, int offset16, Register base) {
  emit_long( xop(op_ldl) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::ldr(Register rt, int offset16, Register base) {
  emit_long( xop(op_ldr) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lh(Register rt, int offset16, Register base) {
  emit_long( xop(op_lh) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lhu(Register rt, int offset16, Register base) {
  emit_long( xop(op_lhu) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::ll(Register rt, int offset16, Register base) {
  emit_long( xop(op_ll) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lld(Register rt, int offset16, Register base) {
  emit_long( xop(op_lld) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lui(Register rt, int imm16) {
  emit_long( xop(op_lui) | xrsop(0) | xrt(rt) | ximmed(imm16) );
}

inline void Assembler::lw(Register rt, int offset16, Register base) {
  emit_long( xop(op_lw) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lwl(Register rt, int offset16, Register base) {
  emit_long( xop(op_lwl) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lwr(Register rt, int offset16, Register base) {
  emit_long( xop(op_lwr) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::lwu(Register rt, int offset16, Register base) {
  emit_long( xop(op_lwu) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::madd(Register rs, Register rt) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_madd) );
}

inline void Assembler::maddu(Register rs, Register rt) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_maddu) );
}

inline void Assembler::mfhi(Register rd) {
  emit_long( xop(op_special) | u_field(0, 25, 16) | xrd(rd) | xsa(0) | xfunc(func_mfhi) );
}

inline void Assembler::mflo(Register rd) {
  emit_long( xop(op_special) | u_field(0, 25, 16) | xrd(rd) | xsa(0) | xfunc(func_mflo) );
}

inline void Assembler::movf(Register rd, Register rs, int cc3) {
  emit_long( xop(op_special) | xrs(rs) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(0, 16, 16) | xrd(rd) | xsa(0) | xfunc(func_movci) );
}

inline void Assembler::movn(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_movn) );
}

inline void Assembler::movt(Register rd, Register rs, int cc3) {
  emit_long( xop(op_special) | xrs(rs) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(1, 16, 16) | xrd(rd) | xsa(0) | xfunc(func_movci) );
}

inline void Assembler::movz(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_movz) );
}

inline void Assembler::msub(Register rs, Register rt) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_msub) );
}

inline void Assembler::msubu(Register rs, Register rt) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rt) | xrdop(0) | xsa(0) | xfunc(func_msubu) );
}

inline void Assembler::mthi(Register rs) {
  emit_long( xop(op_special) | xrs(rs) | u_field(0, 20, 6) | xfunc(func_mthi) );
}

inline void Assembler::mtlo(Register rs) {
  emit_long( xop(op_special) | xrs(rs) | u_field(0, 20, 6) | xfunc(func_mtlo) );
}

inline void Assembler::mul(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special2) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_mul) );
}

inline void Assembler::mult(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(0, 15, 6) | xfunc(func_mult) );
}

inline void Assembler::multu(Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(0, 15, 6) | xfunc(func_multu) );
}

inline void Assembler::nor(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_nor) );
}

inline void Assembler::or_(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_or) );
}

inline void Assembler::ori(Register rt, Register rs, int imm16) {
  emit_long( xop(op_ori) | xrs(rs) | xrt(rt) | ximmedu(imm16) );
}

inline void Assembler::pref(int hint5, int offset16, Register base) {
  emit_long( xop(op_pref) | xrs(base) | xrtop(hint5) | ximmed(offset16) );
}

inline void Assembler::prefx(int hint5, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrdop(hint5) | xsa(0) | xfunc(func_prefx) );
}

inline void Assembler::sb(Register rt, int offset16, Register base) {
  emit_long( xop(op_sb) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sc(Register rt, int offset16, Register base) {
  emit_long( xop(op_sc) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::scd(Register rt, int offset16, Register base) {
  emit_long( xop(op_scd) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sd(Register rt, int offset16, Register base) {
  emit_long( xop(op_sd) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sdbbp(int code20) {
  emit_long( xop(op_special2) | u_field(code20, 25, 6) | xfunc(func_sdbbp) );
}

inline void Assembler::sdl(Register rt, int offset16, Register base) {
  emit_long( xop(op_sdl) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sdr(Register rt, int offset16, Register base) {
  emit_long( xop(op_sdr) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sh(Register rt, int offset16, Register base) {
  emit_long( xop(op_sh) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sll(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) | xfunc(func_sll) );
}

inline void Assembler::sllv(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) |
	     xfunc(func_sllv) );
}

inline void Assembler::slt(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_slt) );
}

inline void Assembler::slti(Register rt, Register rs, int imm16) {
  emit_long( xop(op_slti) | xrs(rs) | xrt(rt) | ximmed(imm16) );
}

inline void Assembler::sltiu(Register rt, Register rs, int imm16) {
  emit_long( xop(op_sltiu) | xrs(rs) | xrt(rt) | ximmed(imm16) );
}

inline void Assembler::sltu(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_sltu) );
}

inline void Assembler::sra(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) |
	     xfunc(func_sra) );
}

inline void Assembler::srav(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_srav) );
}

inline void Assembler::srl(Register rd, Register rt, int sa5) {
  emit_long( xop(op_special) | xrsop(0) | xrt(rt) | xrd(rd) | xsa(sa5) |
	     xfunc(func_srl) );
}

inline void Assembler::srlv(Register rd, Register rt, Register rs) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_srlv) );
}

inline void Assembler::ssnop() {
  emit_long( xop(op_special) | xrsop(0) | xrtop(0) | xrdop(0) | xsa(1) | xfunc(func_sll) );
}

inline void Assembler::sub(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_sub) );
}

inline void Assembler::subu(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_subu) );
}

inline void Assembler::sw(Register rt, int offset16, Register base) {
  emit_long( xop(op_sw) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::swl(Register rt, int offset16, Register base) {
  emit_long( xop(op_swl) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::swr(Register rt, int offset16, Register base) {
  emit_long( xop(op_swr) | xrs(base) | xrt(rt) | ximmed(offset16) );
}

inline void Assembler::sync() {
  emit_long( xop(op_special) | u_field(0, 25, 11) | xsa(0) | xfunc(func_sync) );
}

inline void Assembler::syscall(int code20) {
  emit_long( xop(op_special) | u_field(code20, 25, 6) | xfunc(func_syscall) );
}

inline void Assembler::teq(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_teq) );
}

inline void Assembler::teqi(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_teqi) | ximmed(imm16) );
}

inline void Assembler::tge(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_tge) );
}

inline void Assembler::tgei(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_tgei) | ximmed(imm16) );
}

inline void Assembler::tgeiu(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_tgeiu) | ximmed(imm16) );
}

inline void Assembler::tgeu(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_tgeu) );
}

inline void Assembler::tlbp() {
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_tlbp) );
}

inline void Assembler::tlbr() {
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_tlbr) );
}

inline void Assembler::tlbwi() {
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_tlbwi) );
}

inline void Assembler::tlbwr() {
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_tlbwr) );
}

inline void Assembler::tlt(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_tlt) );
}

inline void Assembler::tlti(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_tlti) | ximmed(imm16) );
}

inline void Assembler::tltiu(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_tltiu) | ximmed(imm16) );
}

inline void Assembler::tltu(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_tltu) );
}

inline void Assembler::tne(Register rs, Register rt, int code10) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | u_field(code10, 15, 6) | xfunc(func_tne) );
}

inline void Assembler::tnei(Register rs, int imm16) {
  emit_long( xop(op_regimm) | xrs(rs) | xrtop(rtop_tnei) | ximmed(imm16) );
}

inline void Assembler::wait() {
  emit_long( xop(op_cop0) | u_field(1, 25, 25) | u_field(0, 24, 6) | xfunc(func_wait) );
}

inline void Assembler::xor_(Register rd, Register rs, Register rt) {
  emit_long( xop(op_special) | xrs(rs) | xrt(rt) | xrd(rd) | xsa(0) | xfunc(func_xor) );
}

inline void Assembler::xori(Register rt, Register rs, int imm16) {
  emit_long( xop(op_xori) | xrs(rs) | xrt(rt) | ximmedu(imm16) );
}


// Misc coprocessor instructions

inline void Assembler::bc1f(int cc3, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_cop1) | xrsop(rsop_bc) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(0, 16, 16) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::bc1t(int cc3, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_cop1) | xrsop(rsop_bc) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(1, 16, 16) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::bc2f(int cc3, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_cop2) | xrsop(rsop_bc) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(0, 16, 16) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::bc2t(int cc3, int offset16,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_cop2) | xrsop(rsop_bc) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(1, 16, 16) | ximmed(offset16), rtype );
  has_delay_slot();
}

inline void Assembler::cfc1(Register rt, int fs5) {
  emit_long( xop(op_cop1) | xrsop(rsop_cf) | xrt(rt) | xrdop(fs5) | u_field(0, 10, 0) );
}

inline void Assembler::cfc2(Register rt, int rd5) {
  emit_long( xop(op_cop2) | xrsop(rsop_cf) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 0) );
}

inline void Assembler::ctc1(Register rt, int fs5) {
  emit_long( xop(op_cop1) | xrsop(rsop_ct) | xrt(rt) | xrdop(fs5) | u_field(0, 10, 0) );
}

inline void Assembler::ctc2(Register rt, int rd5) {
  emit_long( xop(op_cop2) | xrsop(rsop_ct) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 0) );
}

inline void Assembler::dmfc0(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop0) | xrsop(rsop_dmf) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::dmfc1(Register rt, FloatRegister fs) {
  emit_long( xop(op_cop1) | xrsop(rsop_dmf) | xrt(rt) | xrd(fs) | u_field(0, 10, 0) );
}

inline void Assembler::dmfc2(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop2) | xrsop(rsop_dmf) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::dmtc0(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop0) | xrsop(rsop_dmt) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::dmtc1(Register rt, FloatRegister fs) {
  emit_long( xop(op_cop1) | xrsop(rsop_dmt) | xrt(rt) | xrd(fs) | u_field(0, 10, 0) );
}

inline void Assembler::dmtc2(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop2) | xrsop(rsop_dmt) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::ldc1(FloatRegister ft, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_ldc1) | xrs(base) | xrt(ft) | ximmed(offset16) );
}

inline void Assembler::ldc2(int rt5, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_ldc2) | xrs(base) | xrtop(rt5) | ximmed(offset16) );
}

inline void Assembler::ldxc1(FloatRegister fd, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrdop(0) | xsafr(fd) | xfunc(func1_ldxc1) );
}

inline void Assembler::luxc1(FloatRegister fd, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrdop(0) | xsafr(fd) | xfunc(func1_luxc1) );
}

inline void Assembler::lwc1(FloatRegister ft, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_lwc1) | xrs(base) | xrt(ft) | ximmed(offset16), rtype );
}

inline void Assembler::lwc2(int rt, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_rdat( xop(op_lwc2) | xrs(base) | xrtop(rt) | ximmed(offset16), rtype );
}

inline void Assembler::lwxc1(FloatRegister fd, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrdop(0) | xsafr(fd) | xfunc(func1_lwxc1) );
}

inline void Assembler::mfc0(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop0) | xrsop(rsop_mf) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::mfc1(Register rt, FloatRegister fs) {
  emit_long( xop(op_cop1) | xrsop(rsop_mf) | xrt(rt) | xrd(fs) | u_field(0, 10, 0) );
}

inline void Assembler::mfc2(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop2) | xrsop(rsop_mf) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::mtc0(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop0) | xrsop(rsop_mt) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::mtc1(Register rt, FloatRegister fs) {
  emit_long( xop(op_cop1) | xrsop(rsop_mt) | xrt(rt) | xrd(fs) | u_field(0, 10, 0) );
}

inline void Assembler::mtc2(Register rt, int rd5, int sel3) {
  emit_long( xop(op_cop2) | xrsop(rsop_mt) | xrt(rt) | xrdop(rd5) | u_field(0, 10, 3) | u_field(sel3, 2, 0) );
}

inline void Assembler::sdc1(FloatRegister ft, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_sdc1) | xrs(base) | xrt(ft) | ximmed(offset16) );
}

inline void Assembler::sdc2(int rt5, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_sdc2) | xrs(base) | xrtop(rt5) | ximmed(offset16) );
}

inline void Assembler::sdxc1(FloatRegister fs, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrd(fs) | xsa(0) | xfunc(func1_sdxc1) );
}

inline void Assembler::suxc1(FloatRegister fs, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrd(fs) | xsa(0) | xfunc(func1_suxc1) );
}

inline void Assembler::swc1(FloatRegister ft, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_swc1) | xrs(base) | xrt(ft) | ximmed(offset16) );
}

inline void Assembler::swc2(int rt5, int offset16, Register base,
			    relocInfo::relocType rtype) {
  emit_long( xop(op_swc2) | xrs(base) | xrtop(rt5) | ximmed(offset16) );
}

inline void Assembler::swxc1(FloatRegister fs, Register index, Register base) {
  emit_long( xop(op_cop1x) | xrs(base) | xrt(index) | xrd(fs) | xsa(0) | xfunc(func1_swxc1) );
}


// COP1 instructions parameterized by type (see enum fmts)

inline void Assembler::abs_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_abs) );
}

inline void Assembler::add_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_add) );
}

inline void Assembler::c_cond_fmt(int fmt5, c_conds cond4,
		       int cc3, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | u_field(cc3, 10, 8) | u_field(0, 7, 7) | u_field(0, 6, 6) | u_field(3, 5, 4) | u_field((int)cond4, 3, 0) );
}

inline void Assembler::c_cond_fmt(int fmt5, c_conds cond4,
		       FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | u_field(0, 10, 8) | u_field(0, 7, 7) | u_field(0, 6, 6) | u_field(3, 5, 4) | u_field((int)cond4, 3, 0) );
}

inline void Assembler::ceil_l_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_ceil_l) );
}

inline void Assembler::ceil_w_fmt(int fmt5,
		       FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_ceil_w) );
}

inline void Assembler::cvt_d_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_cvt_d) );
}

inline void Assembler::cvt_l_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_cvt_l) );
}

inline void Assembler::cvt_s_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_cvt_s) );
}

inline void Assembler::cvt_w_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_cvt_w) );
}

inline void Assembler::div_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_div) );
}

inline void Assembler::floor_l_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_floor_l) );
}

inline void Assembler::floor_w_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_floor_w) );
}

inline void Assembler::madd_fmt(int fmt3,
		     FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1x) | xfr(fr) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_madd | fmt3) );
}

inline void Assembler::mov_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_mov) );
}

inline void Assembler::movf_fmt(int fmt5,
		     FloatRegister fd, FloatRegister fs, int cc3) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(0, 16, 16) | xfs(fs) | xfd(fd) | xfunc(func1_movcf) );
}

inline void Assembler::movn_fmt(int fmt5,
		     FloatRegister fd, FloatRegister fs, Register rt) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xrt(rt) | xfs(fs) | xfd(fd) | xfunc(func1_movn) );
}

inline void Assembler::movt_fmt(int fmt5,
		     FloatRegister fd, FloatRegister fs, int cc3) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | u_field(cc3, 20, 18) | u_field(0, 17, 17) | u_field(0, 16, 16) | xfs(fs) | xfd(fd) | xfunc(func1_movcf) );
}

inline void Assembler::movz_fmt(int fmt5,
		     FloatRegister fd, FloatRegister fs, Register rt) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xrt(rt) | xfs(fs) | xfd(fd) | xfunc(func1_movz) );
}

inline void Assembler::msub_fmt(int fmt3,
		     FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1x) | xfr(fr) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_msub | fmt3) );
}

inline void Assembler::mul_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_mul) );
}

inline void Assembler::neg_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_neg) );
}

inline void Assembler::nmadd_fmt(int fmt3,
		      FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1x) | xfr(fr) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_nmadd | fmt3) );
}

inline void Assembler::nmsub_fmt(int fmt3,
		      FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1x) | xfr(fr) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_nmsub | fmt3) );
}

inline void Assembler::recip_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_recip) );
}

inline void Assembler::round_l_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_round_l) );
}

inline void Assembler::round_w_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_round_w) );
}

inline void Assembler::rsqrt_fmt(int fmt5,
		      FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_rsqrt) );
}

inline void Assembler::sub_fmt(int fmt5,
		    FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xft(ft) | xfs(fs) | xfd(fd) | xfunc(func1_sub) );
}

inline void Assembler::trunc_l_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_trunc_l) );
}

inline void Assembler::trunc_w_fmt(int fmt5,
			FloatRegister fd, FloatRegister fs) {
  emit_long( xop(op_cop1) | xfmt(fmt5) | xftop(0) | xfs(fs) | xfd(fd) | xfunc(func1_trunc_w) );
}


// Synthetic instructions

inline void MacroAssembler::b(int offset16, relocInfo::relocType rtype) {
  Assembler::beq(R0, R0, offset16, rtype);
}

inline void MacroAssembler::bge(Register rs, Register rt, int offset16,
				relocInfo::relocType rtype) {
  slt(AT, rs, rt);
  Assembler::beq(AT, R0, offset16, rtype);
}

inline void MacroAssembler::bgeu(Register rs, Register rt, int offset16,
				 relocInfo::relocType rtype) {
  sltu(AT, rs, rt);
  Assembler::beq(AT, R0, offset16, rtype);
}

inline void MacroAssembler::bgt(Register rs, Register rt, int offset16,
				relocInfo::relocType rtype) {
  slt(AT, rt, rs);
  Assembler::bne(AT, R0, offset16, rtype);
}

inline void MacroAssembler::bgtu(Register rs, Register rt, int offset16,
				 relocInfo::relocType rtype) {
  sltu(AT, rt, rs);
  Assembler::bne(AT, R0, offset16, rtype);
}

inline void MacroAssembler::ble(Register rs, Register rt, int offset16,
				relocInfo::relocType rtype) {
  slt(AT, rt, rs);
  Assembler::beq(AT, R0, offset16, rtype);
}

inline void MacroAssembler::bleu(Register rs, Register rt, int offset16,
				 relocInfo::relocType rtype) {
  sltu(AT, rt, rs);
  Assembler::beq(AT, R0, offset16, rtype);
}

inline void MacroAssembler::blt(Register rs, Register rt, int offset16,
				relocInfo::relocType rtype) {
  slt(AT, rs, rt);
  Assembler::bne(AT, R0, offset16, rtype);
}

inline void MacroAssembler::bltu(Register rs, Register rt, int offset16,
				 relocInfo::relocType rtype) {
  sltu(AT, rs, rt);
  Assembler::bne(AT, R0, offset16, rtype);
}


inline void MacroAssembler::move(Register rd, Register rs) {
  or_(rd, rs, R0);
}

inline void MacroAssembler::nop() {
  sll(R0, R0, 0);
}


inline void MacroAssembler::lb(Register rt, Register offset, Register base) {
  addu_long(rt, offset, base);
  lb(rt, 0, rt);
}

inline void MacroAssembler::lbu(Register rt, Register offset, Register base) {
  addu_long(rt, offset, base);
  lbu(rt, 0, rt);
}

inline void MacroAssembler::lh(Register rt, Register offset, Register base) {
  addu_long(rt, offset, base);
  lh(rt, 0, rt);
}

inline void MacroAssembler::lhu(Register rt, Register offset, Register base) {
  addu_long(rt, offset, base);
  lhu(rt, 0, rt);
}

inline void MacroAssembler::lw(Register rt, Register offset, Register base) {
  addu_long(rt, offset, base);
  lw(rt, 0, rt);
}


inline void MacroAssembler::sb(Register rt, Register offset, Register base) {
  addu_long(AT, offset, base);
  sb(rt, 0, AT);
}

inline void MacroAssembler::sh(Register rt, Register offset, Register base) {
  addu_long(AT, offset, base);
  sh(rt, 0, AT);
}

inline void MacroAssembler::sw(Register rt, Register offset, Register base) {
  addu_long(AT, offset, base);
  sw(rt, 0, AT);
}

inline void MacroAssembler::swc1(FloatRegister ft, Register offset, Register base) {
  addu_long(AT, offset, base);
  swc1(ft, 0, AT);
}

inline void MacroAssembler::sdc1(FloatRegister ft, Register offset, Register base) {
  addu_long(AT, offset, base);
  sdc1(ft, 0, AT);
}


// Overloadings

inline void MacroAssembler::lb(Register rt, int offset16, Register base) {
  Assembler::lb(rt, offset16, base);
}

inline void MacroAssembler::lbu(Register rt, int offset16, Register base) {
  Assembler::lbu(rt, offset16, base);
}

inline void MacroAssembler::lh(Register rt, int offset16, Register base) {
  Assembler::lh(rt, offset16, base);
}

inline void MacroAssembler::lhu(Register rt, int offset16, Register base) {
  Assembler::lhu(rt, offset16, base);
}

inline void MacroAssembler::lw(Register rt, int offset16, Register base) {
  Assembler::lw(rt, offset16, base);
}


inline void MacroAssembler::sb(Register rt, int offset16, Register base) {
  Assembler::sb(rt, offset16, base);
}

inline void MacroAssembler::sh(Register rt, int offset16, Register base) {
  Assembler::sh(rt, offset16, base);
}

inline void MacroAssembler::sw(Register rt, int offset16, Register base) {
  Assembler::sw(rt, offset16, base);
}

inline void MacroAssembler::swc1(FloatRegister ft, int offset16, Register base,
				 relocInfo::relocType rtype) {
  Assembler::swc1(ft, offset16, base, rtype);
}

inline void MacroAssembler::sdc1(FloatRegister ft, int offset16, Register base,
				 relocInfo::relocType rtype) {
  Assembler::sdc1(ft, offset16, base, rtype);
}


// Synthetic FP

inline void MacroAssembler::abs_s(FloatRegister fd, FloatRegister fs) {
  abs_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::abs_d(FloatRegister fd, FloatRegister fs) {
  abs_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::add_s(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  add_fmt(fmt_s, fd, fs, ft);
}
inline void MacroAssembler::add_d(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  add_fmt(fmt_d, fd, fs, ft);
}

inline void MacroAssembler::c_cond_s(c_conds cond4, int cc3, FloatRegister fs, FloatRegister ft) {
  c_cond_fmt(fmt_s, cond4, cc3, fs, ft);
}
inline void MacroAssembler::c_cond_d(c_conds cond4, int cc3, FloatRegister fs, FloatRegister ft) {
  c_cond_fmt(fmt_d, cond4, cc3, fs, ft);
}

inline void MacroAssembler::c_cond_s(c_conds cond4, FloatRegister fs, FloatRegister ft) {
  c_cond_fmt(fmt_s, cond4, fs, ft);
}
inline void MacroAssembler::c_cond_d(c_conds cond4, FloatRegister fs, FloatRegister ft) {
  c_cond_fmt(fmt_d, cond4, fs, ft);
}

inline void MacroAssembler::ceil_l_s(FloatRegister fd, FloatRegister fs) {
  ceil_l_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::ceil_l_d(FloatRegister fd, FloatRegister fs) {
  ceil_l_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::ceil_w_s(FloatRegister fd, FloatRegister fs) {
  ceil_w_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::ceil_w_d(FloatRegister fd, FloatRegister fs) {
  ceil_w_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::cvt_d_s(FloatRegister fd, FloatRegister fs) {
  cvt_d_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::cvt_d_w(FloatRegister fd, FloatRegister fs) {
  cvt_d_fmt(fmt_w, fd, fs);
}
inline void MacroAssembler::cvt_d_l(FloatRegister fd, FloatRegister fs) {
  cvt_d_fmt(fmt_l, fd, fs);
}

inline void MacroAssembler::cvt_l_s(FloatRegister fd, FloatRegister fs) {
  cvt_l_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::cvt_l_d(FloatRegister fd, FloatRegister fs) {
  cvt_l_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::cvt_s_d(FloatRegister fd, FloatRegister fs) {
  cvt_s_fmt(fmt_d, fd, fs);
}
inline void MacroAssembler::cvt_s_w(FloatRegister fd, FloatRegister fs) {
  cvt_s_fmt(fmt_w, fd, fs);
}
inline void MacroAssembler::cvt_s_l(FloatRegister fd, FloatRegister fs) {
  cvt_s_fmt(fmt_l, fd, fs);
}

inline void MacroAssembler::cvt_w_s(FloatRegister fd, FloatRegister fs) {
  cvt_w_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::cvt_w_d(FloatRegister fd, FloatRegister fs) {
  cvt_w_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::div_s(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  div_fmt(fmt_s, fd, fs, ft);
}
inline void MacroAssembler::div_d(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  div_fmt(fmt_d, fd, fs, ft);
}

inline void MacroAssembler::floor_l_s(FloatRegister fd, FloatRegister fs) {
  floor_l_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::floor_l_d(FloatRegister fd, FloatRegister fs) {
  floor_l_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::floor_w_s(FloatRegister fd, FloatRegister fs) {
  floor_w_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::floor_w_d(FloatRegister fd, FloatRegister fs) {
  floor_w_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::madd_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  madd_fmt(fmt_s, fd, fr, fs, ft);
}
inline void MacroAssembler::madd_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  madd_fmt(fmt_d, fd, fr, fs, ft);
}

inline void MacroAssembler::mov_s(FloatRegister fd, FloatRegister fs) {
  mov_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::mov_d(FloatRegister fd, FloatRegister fs) {
  mov_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::movf_s(FloatRegister fd, FloatRegister fs, int cc3) {
  movf_fmt(fmt_s, fd, fs, cc3);
}
inline void MacroAssembler::movf_d(FloatRegister fd, FloatRegister fs, int cc3) {
  movf_fmt(fmt_d, fd, fs, cc3);
}

inline void MacroAssembler::movn_s(FloatRegister fd, FloatRegister fs, Register rt) {
  movn_fmt(fmt_s, fd, fs, rt);
}
inline void MacroAssembler::movn_d(FloatRegister fd, FloatRegister fs, Register rt) {
  movn_fmt(fmt_d, fd, fs, rt);
}

inline void MacroAssembler::movt_s(FloatRegister fd, FloatRegister fs, int cc3) {
  movt_fmt(fmt_s, fd, fs, cc3);
}
inline void MacroAssembler::movt_d(FloatRegister fd, FloatRegister fs, int cc3) {
  movt_fmt(fmt_d, fd, fs, cc3);
}

inline void MacroAssembler::movz_s(FloatRegister fd, FloatRegister fs, Register rt) {
  movz_fmt(fmt_s, fd, fs, rt);
}
inline void MacroAssembler::movz_d(FloatRegister fd, FloatRegister fs, Register rt) {
  movz_fmt(fmt_d, fd, fs, rt);
}

inline void MacroAssembler::msub_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  msub_fmt(fmt_s, fd, fr, fs, ft);
}
inline void MacroAssembler::msub_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  msub_fmt(fmt_d, fd, fr, fs, ft);
}

inline void MacroAssembler::mul_s(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  mul_fmt(fmt_s, fd, fs, ft);
}
inline void MacroAssembler::mul_d(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  mul_fmt(fmt_d, fd, fs, ft);
}

inline void MacroAssembler::neg_s(FloatRegister fd, FloatRegister fs) {
  neg_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::neg_d(FloatRegister fd, FloatRegister fs) {
  neg_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::nmadd_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  nmadd_fmt(fmt_s, fd, fr, fs, ft);
}
inline void MacroAssembler::nmadd_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  nmadd_fmt(fmt_d, fd, fr, fs, ft);
}

inline void MacroAssembler::nmsub_s(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  nmsub_fmt(fmt_s, fd, fr, fs, ft);
}
inline void MacroAssembler::nmsub_d(FloatRegister fd, FloatRegister fr, FloatRegister fs, FloatRegister ft) {
  nmsub_fmt(fmt_d, fd, fr, fs, ft);
}

inline void MacroAssembler::recip_s(FloatRegister fd, FloatRegister fs) {
  recip_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::recip_d(FloatRegister fd, FloatRegister fs) {
  recip_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::round_l_s(FloatRegister fd, FloatRegister fs) {
  round_l_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::round_l_d(FloatRegister fd, FloatRegister fs) {
  round_l_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::round_w_s(FloatRegister fd, FloatRegister fs) {
  round_w_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::round_w_d(FloatRegister fd, FloatRegister fs) {
  round_w_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::rsqrt_s(FloatRegister fd, FloatRegister fs) {
  rsqrt_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::rsqrt_d(FloatRegister fd, FloatRegister fs) {
  rsqrt_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::sub_s(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  sub_fmt(fmt_s, fd, fs, ft);
}
inline void MacroAssembler::sub_d(FloatRegister fd, FloatRegister fs, FloatRegister ft) {
  sub_fmt(fmt_d, fd, fs, ft);
}

inline void MacroAssembler::trunc_l_s(FloatRegister fd, FloatRegister fs) {
  trunc_l_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::trunc_l_d(FloatRegister fd, FloatRegister fs) {
  trunc_l_fmt(fmt_d, fd, fs);
}

inline void MacroAssembler::trunc_w_s(FloatRegister fd, FloatRegister fs) {
  trunc_w_fmt(fmt_s, fd, fs);
}
inline void MacroAssembler::trunc_w_d(FloatRegister fd, FloatRegister fs) {
  trunc_w_fmt(fmt_d, fd, fs);
}


// Helper instruction forms

inline int MacroAssembler::target_to_offset16(address target) {
  intptr_t delayslot = intptr_t(locate_next_instruction(pc()));
  intptr_t offset = intptr_t(target) - delayslot;
  assert((offset & 3) == 0, "supposed to be aligined");
  offset = offset >> 2;
  // This function is intended to be used for small offsets only
  // because it guarantees (in debug mode, at least) that
  // a pc-relative branch is safe to use.
  assert(fits_in_s_lfield(offset, 15, 0), "offset overflows - do not use");
  return (int)offset;
}

inline int MacroAssembler::target_to_offset16(Label& L) {
  return target_to_offset16(target(L));
}

inline void MacroAssembler::b(
              address addr,
	      relocInfo::relocType rtype) {
  beq(R0, R0, addr, rtype); // defined in .cpp
}

inline void MacroAssembler::b(
              Label& L,
	      relocInfo::relocType rtype) {
  beq(R0, R0, L, rtype);  // defined below
}

inline void MacroAssembler::beq(
              Register rs, Register rt, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::beq(rs, rt, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgez(
	      Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bgez(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgezal(
              Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bgezal(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgtz(
              Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bgtz(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::blez(
              Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::blez(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bltz(
	      Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bltz(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bltzal(
	      Register rs, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bltzal(rs, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bne(
              Register rs, Register rt, Label& L,
	      relocInfo::relocType rtype) {
  Assembler::bne(rs, rt, target_to_offset16(L), rtype);
}


inline void MacroAssembler::bge(Register rs, Register rt, Label& L,
				relocInfo::relocType rtype) {
  // Code duplicated from non-Label version so CodeSection::target
  // will be done at the right instruction.
  slt(AT, rs, rt);
  Assembler::beq(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgeu(Register rs, Register rt, Label& L,
				 relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  sltu(AT, rs, rt);
  Assembler::beq(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgt(Register rs, Register rt, Label& L,
				relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  slt(AT, rt, rs);
  Assembler::bne(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bgtu(Register rs, Register rt, Label& L,
				 relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  sltu(AT, rt, rs);
  Assembler::bne(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::ble(Register rs, Register rt, Label& L,
				relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  slt(AT, rt, rs);
  Assembler::beq(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bleu(Register rs, Register rt, Label& L,
				 relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  sltu(AT, rt, rs);
  Assembler::beq(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::blt(Register rs, Register rt, Label& L,
				relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  slt(AT, rs, rt);
  Assembler::bne(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bltu(Register rs, Register rt, Label& L,
				 relocInfo::relocType rtype) {
  // Code duplicated from non-Label version (see above)
  sltu(AT, rs, rt);
  Assembler::bne(AT, R0, target_to_offset16(L), rtype);
}

inline void MacroAssembler::bc1f(int cc3, Label& L) {
  Assembler::bc1f(cc3, target_to_offset16(L));
}

inline void MacroAssembler::bc1t(int cc3, Label& L) {
  Assembler::bc1t(cc3, target_to_offset16(L));
}


// NOTE: has an unfilled delay slot
inline void MacroAssembler::jump_to(address target, relocInfo::relocType rt) {
  // In the future, might try a faster sequence with the option
  // for later patching.
  li(AT, target);
  jr(AT);
}

// NOTE: has an unfilled delay slot
inline void MacroAssembler::jal_to(address target, relocInfo::relocType rt) {
  li(AT, target);
  jalr(AT);
}

// YYY redundant with above -- remove?
void MacroAssembler::call(address target, relocInfo::relocType rtype) {
  li(AT, target);
  jalr(AT);
}

// YYY We should probably always set T9 for jal.  The trouble is that T9
// YYY has already been used for Rmethod_call, and occasionally as a
// YYY temp.  Need to go back and ensure no conflicting uses.
// For this function, though, it's fine, since T9 is caller-save anyway
void MacroAssembler::jal_to_C(address target, relocInfo::relocType rtype) {
  li(Rcall, target);
  jalr(Rcall);
}

inline void MacroAssembler::store_int_argument(Register s, Argument& a) {
  if (a.is_register()) {
    move(a.as_register(), s);
  } else {
    sw(s, a.as_caller_address().disp(), a.as_caller_address().base());
  }
}

inline void MacroAssembler::store_long_argument(Register s, Argument& a) {
  if (a.is_register()) {
    move(a.as_register(), s);
  } else {
    sd(s, a.as_caller_address().disp(), a.as_caller_address().base());
  }
}

inline void MacroAssembler::store_float_argument(FloatRegister s, Argument& a) {
  if (a.is_register()) {
    mov_s(a.as_float_register(), s);
  } else {
    swc1(s, a.as_caller_address().disp(), a.as_caller_address().base());
  }
}

inline void MacroAssembler::store_double_argument(FloatRegister s, Argument& a) {
  if (a.is_register()) {
    mov_d(a.as_float_register(), s);
  } else {
    sdc1(s, a.as_caller_address().disp(), a.as_caller_address().base());
  }
}

inline void MacroAssembler::store_ptr_argument(Register s, Argument& a) {
  if (a.is_register()) {
    move(a.as_register(), s);
  } else {
    st_ptr(s, a.as_caller_address().disp(), a.as_caller_address().base());
  }
}


inline void MacroAssembler::addu_long(Register rd, Register rs, Register rt) {
#ifdef _LP64
  daddu(rd, rs, rt);
#else
  addu(rd, rs, rt);
#endif
}

inline void MacroAssembler::addu_long(Register rd, Register rs, long imm32_64) {
#ifdef _LP64
  daddiu(rd, rs, imm32_64);
#else
  addiu(rd, rs, imm32_64);
#endif
}

inline void MacroAssembler::sll_long(Register rd, Register rt, int sa5) {
#ifdef _LP64
  dsll(rd, rt, sa5);
#else
  sll(rd, rt, sa5);
#endif
}

inline void MacroAssembler::sllv_long(Register rd, Register rt, Register rs) {
#ifdef _LP64
  dsllv(rd, rt, rs);
#else
  sllv(rd, rt, rs);
#endif
}

inline void MacroAssembler::sra_long(Register rd, Register rt, int sa5) {
#ifdef _LP64
  dsra(rd, rt, sa5);
#else
  sra(rd, rt, sa5);
#endif
}

inline void MacroAssembler::srav_long(Register rd, Register rt, Register rs) {
#ifdef _LP64
  dsrav(rd, rt, rs);
#else
  srav(rd, rt, rs);
#endif
}

inline void MacroAssembler::srl_long(Register rd, Register rt, int sa5) {
#ifdef _LP64
  dsrl(rd, rt, sa5);
#else
  srl(rd, rt, sa5);
#endif
}

inline void MacroAssembler::srlv_long(Register rd, Register rt, Register rs) {
#ifdef _LP64
  dsrlv(rd, rt, rs);
#else
  srlv(rd, rt, rs);
#endif
}

inline void MacroAssembler::subu_long(Register rd, Register rs, Register rt) {
#ifdef _LP64
  dsubu(rd, rs, rt);
#else
  subu(rd, rs, rt);
#endif
}

inline void MacroAssembler::ld_ptr(Register rt, int offset16, Register base) {
#ifdef _LP64
  ld(rt, offset16, base);
#else
  lw(rt, offset16, base);
#endif
}

inline void MacroAssembler::st_ptr(Register rt, int offset16, Register base) {
#ifdef _LP64
  sd(rt, offset16, base);
#else
  sw(rt, offset16, base);
#endif
}

inline void MacroAssembler::ld_long(Register rt, int offset16, Register base) {
#ifdef _LP64
  ld(rt, offset16, base);
#else
  lw(rt, offset16, base);
#endif
}

inline void MacroAssembler::st_long(Register rt, int offset16, Register base) {
#ifdef _LP64
  sd(rt, offset16, base);
#else
  sw(rt, offset16, base);
#endif
}

inline void MacroAssembler::ll_long(Register rt, int offset16, Register base) {
#ifdef _LP64
  lld(rt, offset16, base);
#ifdef SICORTEX_ERRATA
  // YYY SiCortex errata: use two ll or lld in a row for correct operation.
  lld(rt, offset16, base);
#endif
#else
  ll(rt, offset16, base);
#ifdef SICORTEX_ERRATA
  ll(rt, offset16, base);
#endif
#endif
}

inline void MacroAssembler::sc_long(Register rt, int offset16, Register base) {
#ifdef _LP64
  scd(rt, offset16, base);
#else
  sc(rt, offset16, base);
#endif
}
