#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)dump_sparc.cpp	1.12 07/05/05 17:04:28 JVM"
#endif
/*
 * Copyright 2004-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_dump_mips.cpp.incl"



// Generate the self-patching vtable method:
//
// This method will be called (as any other Klass virtual method) with
// the Klass itself as the first argument.  Example:
//
// 	oop obj;
// 	int size = obj->klass()->klass_part()->oop_size(this);
//
// for which the virtual method call is Klass::oop_size();
//
// The dummy method is called with the Klass object as the first
// operand, and an object as the second argument.
//

//=====================================================================

// All of the dummy methods in the vtable are essentially identical,
// differing only by an ordinal constant, and they bear no releationship
// to the original method which the caller intended. Also, there needs
// to be 'vtbl_list_size' instances of the vtable in order to
// differentiate between the 'vtable_list_size' original Klass objects.

#define __ masm->

void CompactingPermGenGen::generate_vtable_methods(void** vtbl_list,
                                                   void** vtable,
                                                   char** md_top,
                                                   char* md_end,
                                                   char** mc_top,
                                                   char* mc_end) {

  intptr_t vtable_bytes = (num_virtuals * vtbl_list_size) * sizeof(void*);
  *(intptr_t *)(*md_top) = vtable_bytes;
  *md_top += sizeof(intptr_t);
  void** dummy_vtable = (void**)*md_top;
  *vtable = dummy_vtable;
  *md_top += vtable_bytes;

  // Get ready to generate dummy methods.

  CodeBuffer cb((unsigned char*)*mc_top, mc_end - *mc_top);
  MacroAssembler* masm = new MacroAssembler(&cb);

  Label common_code;
  for (int i = 0; i < vtbl_list_size; ++i) {
    for (int j = 0; j < num_virtuals; ++j) {
      dummy_vtable[num_virtuals * i + j] = (void*)masm->pc();
      // YYY work with temp regs only to avoid register saves
      __ b(common_code);

      // Load Rscratch0 with a value indicating vtable/offset pair.
      // -- bits[ 7..0]  (8 bits) which virtual method in table?
      // -- bits[12..8]  (5 bits) which virtual method table?
      // -- must fit in 16-bit instruction immediate field.
      __ delayed()->ori(Rscratch0, R0, (i << 8) + j);
    }
  }

  __ bind(common_code);

  // Expecting to be called with the "this" pointer in A0 (where
  // "this" is a Klass object).  In addition, Rscratch0 (T0) was set
  // (above) to identify the method and table.

  // Look up the correct vtable pointer.

  __ li(Rscratch2, (intptr_t)vtbl_list);         // address of new vtable list
  __ srl_long(Rscratch3, Rscratch0, 8);          // vtable identifier
  __ sll_long(Rscratch3, Rscratch3, LogBytesPerWord);
  __ ld_ptr(Rscratch2, Rscratch3, Rscratch3);    // new (correct) vtable pointer
  __ st_ptr(Rscratch3, 0, A0);                   // Save correct vtable ptr in entry

  __ andi(Rscratch4, Rscratch0, 255);            // Isolate method offset
  __ sll_long(Rscratch4, Rscratch4, LogBytesPerWord);
  __ addu_long(Rscratch3, Rscratch4, Rscratch3);
  __ ld_ptr(Rscratch4, 0, Rscratch3);            // Get address of virtual method
  __ jr(Rscratch4);
  __ delayed()->nop();

  __ flush();
  *mc_top = (char*)__ pc();

  guarantee(*mc_top <= mc_end, "Insufficient space for method wrappers.");
}

