#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)frame_sparc.cpp	1.174 07/05/05 17:04:27 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_frame_mips.cpp.incl"


#ifdef ASSERT
void RegisterMap::check_location_valid() {
  Unimplemented();
  //register_map_init();
  //assert((_location_valid[0] & bad_mask) == 0, "cannot have special locations for SP,FP,TLS,etc.");
}
#endif


  // for Profiling - acting on another frame. walks sender frames
  // if valid.
  // frame profile_find_Java_sender_frame(JavaThread *thread);
bool frame::safe_for_sender(JavaThread *thread) {
  address   sp = (address)_sp;
  if (sp != NULL && 
      (sp <= thread->stack_base() && sp >= thread->stack_base() - thread->stack_size())) {
      // Unfortunately we can only check frame complete for runtime stubs and nmethod
      // other generic buffer blobs are more problematic so we just assume they are
      // ok. adapter blobs never have a frame complete and are never ok.
      if (_cb != NULL && !_cb->is_frame_complete_at(_pc)) {
        if (_cb->is_nmethod() || _cb->is_adapter_blob() || _cb->is_runtime_stub()) {
          return false;
        }
      }
      return true;
  }
  return false;
}


frame frame::sender_for_entry_frame(RegisterMap *map) const {
  assert(map != NULL, "map must be set");
  // Java frame called from C; skip all C frames and return top C
  // frame of that chunk as the sender
  JavaFrameAnchor* jfa = entry_frame_call_wrapper()->anchor();
  assert(!entry_frame_is_first(), "next Java fp must be non zero");
  assert(jfa->last_Java_sp() > _sp, "must be above this frame on stack");
  map->clear();
  assert(map->include_argument_oops(), "should be set by clear");
  return frame(jfa->last_Java_sp(), jfa->last_Java_fp(), jfa->last_Java_pc());
}

frame frame::sender_for_interpreter_frame(RegisterMap *map) const {
  ShouldNotCallThis();
  return sender(map);
}

frame frame::sender_for_compiled_frame(RegisterMap *map) const {  
  ShouldNotCallThis();
  return sender(map);
}

// returns the sending frame
frame frame::sender(RegisterMap* map) const {
  assert(map != NULL, "map must be set");

  assert(CodeCache::find_blob_unsafe(_pc) == _cb, "inconsistent");

  // Default is not to follow arguments; update it accordingly below
  map->set_include_argument_oops(false);

  if (is_entry_frame()) return sender_for_entry_frame(map);

  if (_cb != NULL) {
    if (UseCompiler) {		// YYY unimplemented
    return sender_for_compiled_frame(map);
    }
  }

  intptr_t* older_sp = sender_sp();
  intptr_t* older_fp = link();
  address   older_pc = sender_pc();

  if (is_interpreted_frame()) {
    intptr_t* unextended_older_sp
      = older_sp + older_fp[interpreter_frame_args_size_offset];
    return frame(older_sp, unextended_older_sp, older_fp, older_pc);
  } else {
    return frame(older_sp, older_fp, older_pc);
  }
}


void frame::patch_pc(Thread* thread, address pc) {
  if (TracePcPatching) {
    Unimplemented();
  }
  _cb = CodeCache::find_blob(pc);
  Unimplemented();
  _cb = CodeCache::find_blob(_pc);
  if (_cb != NULL && _cb->is_nmethod() && ((nmethod*)_cb)->is_deopt_pc(_pc)) {
    address orig = ((nmethod*)_cb)->get_original_pc(this);
    assert(orig == _pc, "expected original to be stored before patching");
    _deopt_state = is_deoptimized;
  } else {
    _deopt_state = not_deoptimized;
  }
}


void frame::pd_gc_epilog() {
  if (is_interpreted_frame()) {
    // set constant pool cache entry for interpreter
    methodOop m = interpreter_frame_method();
    *interpreter_frame_cpoolcache_addr() = m->constants()->cache();
  }
}


bool frame::is_interpreted_frame_valid() const {
  assert(is_interpreted_frame(), "Not an interpreted frame");
  // These are reasonable sanity checks
  if (fp() == 0 || (intptr_t(fp()) & (2*wordSize-1)) != 0) {
    return false;
  }
  if (sp() == 0 || (intptr_t(sp()) & (2*wordSize-1)) != 0) {
    return false;
  }
  // These are hacks to keep us out of trouble.
  // The problem with these is that they mask other problems
  if (fp() <= sp()) {        // this attempts to deal with unsigned comparison above
    return false;
  }
  // Stack frames shouldn't be too large.  Note we could have 64k
  // locals plus 64k bytecode stack slots.
  if (fp() - sp() > 132000) {
    return false;
  }
  return true;
}


  // Return/result value from this interpreter frame
  // If the method return type is T_OBJECT or T_ARRAY populates oop_result
  // For other (non-T_VOID) the appropriate field in the jvalue is populated
  // with the result value.
  // Should only be called when at method exit when the method is not 
  // exiting due to an exception.
BasicType frame::interpreter_frame_result(oop* oop_result, jvalue* value_result) {
  assert(is_interpreted_frame(), "interpreted frame expected");
  methodOop method = interpreter_frame_method();
  BasicType type = method->result_type();

  if (method->is_native()) {
    // Prior to notifying the runtime of the method_exit the possible result
    // value is saved to l_scratch and d_scratch.

    intptr_t* l_scratch = fp() + interpreter_frame_l_scratch_offset;
    intptr_t* d_scratch = fp() + interpreter_frame_d_scratch_offset;
    address l_addr = (address)l_scratch;

    switch (type) {
      case T_OBJECT:
      case T_ARRAY: {
	oop obj = (oop) at(interpreter_frame_oop_temp_offset);
	assert(obj == NULL || Universe::heap()->is_in(obj), "sanity check");
	*oop_result = obj;
	break;
      }

      case T_BOOLEAN : { jint* p = (jint*)l_addr; value_result->z = (jboolean)((*p) & 0x1); break; }
      case T_BYTE    : { jint* p = (jint*)l_addr; value_result->b = (jbyte)((*p) & 0xff); break; }
      case T_CHAR    : { jint* p = (jint*)l_addr; value_result->c = (jchar)((*p) & 0xffff); break; }
      case T_SHORT   : { jint* p = (jint*)l_addr; value_result->s = (jshort)((*p) & 0xffff); break; }
      case T_INT     : value_result->i = *(jint*)l_addr; break;
      case T_LONG    : value_result->j = *(jlong*)l_scratch; break;
      case T_FLOAT   : value_result->f = *(jfloat*)d_scratch; break;
      case T_DOUBLE  : value_result->d = *(jdouble*)d_scratch; break;
      case T_VOID    : /* Nothing to do */ break;
      default        : ShouldNotReachHere();
    }

  } else {  
    intptr_t* tos_addr = interpreter_frame_tos_address();

    switch(type) {
      case T_OBJECT:
      case T_ARRAY: {
	oop obj = (oop)*tos_addr;
 	assert(obj == NULL || Universe::heap()->is_in(obj), "sanity check");	
	*oop_result = obj;
	break;
      }
      case T_BOOLEAN : { jint* p = (jint*)tos_addr; value_result->z = (jboolean)((*p) & 0x1); break; }
      case T_BYTE    : { jint* p = (jint*)tos_addr; value_result->b = (jbyte)((*p) & 0xff); break; }
      case T_CHAR    : { jint* p = (jint*)tos_addr; value_result->c = (jchar)((*p) & 0xffff); break; }
      case T_SHORT   : { jint* p = (jint*)tos_addr; value_result->s = (jshort)((*p) & 0xffff); break; }
      case T_INT     : value_result->i = *(jint*)tos_addr; break;
      case T_LONG    : value_result->j = *(jlong*)tos_addr; break;
      case T_FLOAT   : value_result->f = *(jfloat*)tos_addr; break;
      case T_DOUBLE  : value_result->d = *(jdouble*)tos_addr; break;
      case T_VOID    : /* Nothing to do */ break;
      default        : ShouldNotReachHere();
    }
  };

  return type;
} 

intptr_t* frame::interpreter_frame_tos_at(jint offset) const {
  int index = (Interpreter::expr_offset_in_bytes(offset)/wordSize) - 1;
  return &interpreter_frame_tos_address()[index];
}

intptr_t* frame::interpreter_frame_mdx_addr() const {
  Unimplemented();
  //// %%%%% reinterpreting ImethodDataPtr as a mdx
  //return (intptr_t*) sp_addr_at( ImethodDataPtr->sp_offset_in_saved_window());
}
