#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)frame_sparc.hpp	1.73 07/05/05 17:04:27 JVM"
#endif
/*
 * Copyright 1997-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// A frame represents a physical stack frame (an activation).  Frames can be
// C or Java frames, and the Java frames can be interpreted or compiled.
// In contrast, vframes represent source-level activations, so that one physical frame
// can correspond to multiple source level frames because of inlining.
// A frame is comprised of {pc, fp, sp}

// included by ../../../share/vm/runtime/frame.hpp

// Layout of interpreter frame:
//
// [highest address]
// ...
// [argument Nargs-1]           \
// ...                          |--- only if Nargs > 7 (0-based counting)
// [argument 8]                 / <-- FP
// [frame locals (maybe empty)]
// [local 0]                    <-- Rlocals
// ...
// [local Nlocals-1]
// [first monitor]
// ...
// [last monitor]               <-- Rmonitors (points to last alloc monitor)
// [bottom of expression stack]
// ...                          <-- Resp (points to first free entry)
// [top of expression stack]
// [FP/S8 save]
// [S7 save]
// ...
// [S0 save]
// [GP save]
// [RA save]
// [F31 save]
// ...
// [F24 save]
// [outgoing argument Nargs'-1] \
// ...                          |--- only if Nargs > 7 (0-based counting)
// [outgoing argument 8]        / <-- SP

// State registers (see assembler_mips.hpp for mapping to physical
// registers):
//
//    Resp         first free element of expression stack
//    Rbcp         address of bytecode to execute
//                 It may at times (during GC) be an index instead.
//    Rmethod      the method being interpreted
//    Rlocals      the base pointer for accessing the locals array
//                 (lower-numbered locals have lower addresses)
//    Rmonitors    the base pointer for accessing active monitors
//    Rcache       a saved pointer to the method's constant pool cache



// All frames: 

 public:

  enum {
    // Note: pc_return_offset is 0, since jal[r] adds it for us.
    pc_return_offset = 0,

    // size of each block, in order of increasing address:
    outgoing_args_space               = 0  * BytesPerLong, // none by default
    FPR_saves_space                   = 8  * BytesPerLong, // F24 - F31
    GPR_saves_space                   = 11 * BytesPerLong, // RA, GP, S0 - S8
    register_save_space               = FPR_saves_space + GPR_saves_space,

    // offset of each block, in order of increasing address:
    outgoing_args_offset              = 0,
    FPR_saves_offset                  = outgoing_args_offset + outgoing_args_space,
    GPR_saves_offset                  = FPR_saves_offset + FPR_saves_space,
    expr_stack_offset                 = GPR_saves_offset + GPR_saves_space,
  };

 private:
  // any additional field beyond _sp and _pc:
  intptr_t* _fp;                // frame pointer
  intptr_t* _unextended_sp;     // original sp, before call args pushed

 public:
  // Accessors

  intptr_t* fp() const {
    return _fp;
  }

  // Constructors

  frame(intptr_t* sp, intptr_t* fp);
  frame(intptr_t* sp, intptr_t* fp, address pc);
  frame(intptr_t* sp, intptr_t* unextended_sp, intptr_t* fp, address pc);

 public:
  // accessors for the instance variables

  // All frames

  intptr_t*  fp_addr_at(int index) const   { return &fp()[index];    }
  // Expecting to want unextended_sp in almost all cases.
  // But be careful of any exceptions to this rule!
  intptr_t*  sp_addr_at(int index) const   { return &unextended_sp()[index]; }
  intptr_t   fp_at(     int index) const   { return *fp_addr_at(index); }
  intptr_t   sp_at(     int index) const   { return *sp_addr_at(index); }


  // Interpreter frames

 public:
  enum interpreter_frame_vm_locals {
    // Both kinds, native and bytecode, have this first set.
    interpreter_frame_args_size_offset = -1,
    interpreter_frame_Rmethod_offset   = -2,
    interpreter_frame_Resp_offset      = -3,
    interpreter_frame_Rmonitors_offset = -4,
    interpreter_frame_Rlocals_offset   = -5,
    interpreter_frame_Rbcp_offset      = -6,
    interpreter_frame_RcpoolCache_offset = -7,

    // Only bytecode methods have these
    interpreter_frame_vm_locals_words  =  7,

    // Only native methods have these
    interpreter_frame_oop_temp_offset  = -8,
    interpreter_frame_l_scratch_offset = -9,
    interpreter_frame_d_scratch_offset = -10,
    interpreter_frame_vm_native_locals_words = 10,

    // Interpreter frame set-up needs 2 extra words in outgoing param area
    // for class and jnienv arguments for native stubs
    // (see nativeStubGen_sparc.cpp).
    interpreter_frame_extra_outgoing_argument_words = 2
  };

  // YYY These should possibly be unified with above offset constants
  // YYY Or, modify things like entry_frame_argument_at so they don't
  // YYY share code.
  enum call_stub_locals {
    call_stub_args_size_offset = -1,
    call_stub_wrapper_offset   = -2,
    call_stub_Resp_offset      = -3,  // must be same as interp frame!
    call_stub_locals_words     =  3
  };

  // the compiler frame has many of the same fields as the interpreter frame
  // %%%%% factor out declarations of the shared fields
  enum compiler_frame_fixed_locals {
    // YYY unimplemented
  };

 private:

  // where RcpoolCache is saved:
  constantPoolCacheOop* interpreter_frame_cpoolcache_addr() const { 
    return (constantPoolCacheOop*) fp_addr_at(interpreter_frame_RcpoolCache_offset);
  }

  // where Rmonitors is saved:
  BasicObjectLock**  interpreter_frame_monitors_addr() const { 
    return (BasicObjectLock**) fp_addr_at(interpreter_frame_Rmonitors_offset);
  }
  intptr_t** interpreter_frame_esp_addr() const {
    return (intptr_t**) fp_addr_at(interpreter_frame_Resp_offset);
  }

  // monitors:

 private:
  BasicObjectLock* interpreter_frame_monitors()           const  { return *interpreter_frame_monitors_addr(); }



  // Compiled frames

 public:
  // YYY unimplemented
