#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)frame_sparc.inline.hpp	1.78 07/05/05 17:04:28 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// Inline functions for MIPS frames:

// Constructors

inline frame::frame() {
  _pc = NULL;
  _sp = NULL;
  _unextended_sp = NULL;
  _fp = NULL;
}

inline frame::frame(intptr_t* sp, intptr_t* fp) { 
  _sp = sp;
  _unextended_sp = sp;
  _fp = fp;
  // YYY this isn't quite right -- it's the pc in the caller frame
  // YYY assert to check if this path is ever actually used
  assert(0, "pc maybe not right");
  _pc = (address)sp[RA->sp_index_in_save_area()];
  assert(_pc != NULL, "no pc?");
  _cb = CodeCache::find_blob(_pc);
}

inline frame::frame(intptr_t* sp, intptr_t* fp, address pc) { 
  _sp = sp;
  _unextended_sp = sp;
  _fp = fp;
  _pc = pc;
  // pc may be NULL, e.g., os::current_frame
  _cb = CodeCache::find_blob(pc);
}

inline frame::frame(intptr_t* sp, intptr_t* unextended_sp, intptr_t* fp, address pc) {
  _sp = sp;
  _unextended_sp = unextended_sp;
  _fp = fp;
  _pc = pc;
  assert(_pc != NULL, "no pc?");
  _cb = CodeCache::find_blob(_pc);
}


inline bool frame::is_interpreted_frame() const {
  return AbstractInterpreter::contains(pc());
}

// Accessors:

// Return unique id for this frame. The id must have a value where we can distinguish
// identity and younger/older relationship. NULL represents an invalid (incomparable)
// frame.
inline intptr_t* frame::id(void) const { return unextended_sp(); }

// Relationals on frames based 
// Return true if the frame is younger (more recent activation) than the frame represented by id
inline bool frame::is_younger(intptr_t* id) const { assert(this->id() != NULL && id != NULL, "NULL frame id");
					            return this->id() < id ; }

// Return true if the frame is older (less recent activation) than the frame represented by id
inline bool frame::is_older(intptr_t* id) const   { assert(this->id() != NULL && id != NULL, "NULL frame id");
					            return this->id() > id ; }

inline int frame::frame_size() const { return sender_sp() - sp(); }

inline intptr_t* frame::link() const {
  return (intptr_t *)(unextended_sp()[FP->sp_index_in_save_area()]);
}

inline void frame::set_link(intptr_t* addr) { assert(link()==addr, "frame nesting is controlled by hardware"); }

inline intptr_t* frame::unextended_sp() const {
  int args_offset = fp()[interpreter_frame_args_size_offset];
  assert(_unextended_sp == sp() + args_offset, "_unextended_sp doesn't match frame");
  return _unextended_sp;
}

inline intptr_t* frame::entry_frame_argument_at(int offset) const {
  // Since an entry frame always calls the interpreter first,
  // the format of the call is always compatible with the interpreter.
  return interpreter_frame_tos_at(offset);
}

// return address:

inline address frame::sender_pc() const {
  return (address)(unextended_sp()[RA->sp_index_in_save_area()]);
}

inline intptr_t*    frame::sender_sp() const  { return fp(); }

inline intptr_t** frame::interpreter_frame_locals_addr() const {
  return (intptr_t**) fp_addr_at(interpreter_frame_Rlocals_offset);
}

inline intptr_t* frame::interpreter_frame_bcx_addr() const {
  // %%%%% reinterpreting Rbcp as a bcx
  return (intptr_t*) fp_addr_at(interpreter_frame_Rbcp_offset);
}

inline jint frame::interpreter_frame_expression_stack_direction() { return -1; }

// bottom(base) of the expression stack (highest address)
inline intptr_t* frame::interpreter_frame_expression_stack() const {
  return (intptr_t*)interpreter_frame_monitors() - 1;
}

// top of expression stack (lowest address)
inline intptr_t* frame::interpreter_frame_tos_address() const {
  return *interpreter_frame_esp_addr() + 1;
}


  // BasicObjectLocks:
  //
  // interpreter_frame_monitor_begin is higher in memory than interpreter_frame_monitor_end
  // Interpreter_frame_monitor_begin points to one element beyond the oldest one,
  // interpreter_frame_monitor_end   points to the youngest one, or if there are none,
  //                                 it points to one beyond where the first element will be.
  // interpreter_frame_monitor_size  reports the allocation size of a monitor in the interpreter stack.
  //                                 this value is >= BasicObjectLock::size(), and may be rounded up

// monitor elements

// in keeping with Intel side: end is lower in memory than begin;
// and beginning element is oldest element
// Also begin is one past last monitor.

inline BasicObjectLock* frame::interpreter_frame_monitor_begin() const {
  methodOop method = interpreter_frame_method();
  return (BasicObjectLock *)(fp()
    - (method->is_native() ?
       frame::interpreter_frame_vm_native_locals_words :
       frame::interpreter_frame_vm_locals_words)
    - method->max_locals());
  // remember, this just points to the word above the first monitor
}

inline BasicObjectLock* frame::interpreter_frame_monitor_end() const {
  return interpreter_frame_monitors();
}


inline int frame::interpreter_frame_monitor_size() {
  return round_to(BasicObjectLock::size(), WordsPerLong);
}

inline methodOop* frame::interpreter_frame_method_addr() const {
  return (methodOop*)fp_addr_at(interpreter_frame_Rmethod_offset);
}



// Constant pool cache

inline constantPoolCacheOop* frame::interpreter_frame_cache_addr() const {
  return (constantPoolCacheOop*)fp_addr_at(interpreter_frame_RcpoolCache_offset);
}


inline JavaCallWrapper* frame::entry_frame_call_wrapper() const {
  return (JavaCallWrapper*)fp()[call_stub_wrapper_offset];
}

inline oop  frame::saved_oop_result(RegisterMap* map) const      {
  Unimplemented();
  return 0;
}

inline void frame::set_saved_oop_result(RegisterMap* map, oop obj) {
  Unimplemented();
}

