#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)icache_sparc.cpp	1.29 07/05/05 17:04:28 JVM"
#endif
/*
 * Copyright 1997-2004 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_icache_mips.cpp.incl"
#include <sys/cachectl.h>

#define __ _masm->

void ICacheStubGenerator::generate_icache_flush(
  ICache::flush_icache_stub_t* flush_icache_stub
) {
  StubCodeMark mark(this, "ICache", "flush_icache_stub");
  address start = __ pc();

  // MIPS64 r1 doesn't have a way to flush the icache in user mode.
  // Thus the code below relies on a Linux system call, cacheflush.
  // In principle, this code maybe should be moved to linux_mips.
  // YYY On MIPS64 r2, just use synci.
#ifndef LINUX
#error code relies on Linux system call
#endif

  __ save_frame();
  __ move(S4, A2);  // save magic

#if 0

  // convert lines to bytes
  __ sll_long(A1, A1, ICache::log2_line_size);

  __ move(S0, A0);        // start address
  __ move(S1, A1);        // number of bytes
  __ jal_to_C(cacheflush);
  __ delayed()->li(A2, DCACHE);
  __ move(A0, S0);
  __ move(A1, S1);
  __ jal_to_C(cacheflush);
  __ delayed()->li(A2, ICACHE);

#else
  // Rumors are that the cacheflush call doesn't work properly if it's
  // not given a page-sized range to work on.  This version does that.

  Label loop;

  __ sll_long(A1, A1, ICache::log2_line_size);
  __ addu_long(S3, A0, A1);
  __ li(S1, os::vm_page_size());
  // page_base = addr & ~(os::page_size() - 1)
  __ addu_long(T0, S1, -1);
  __ nor(T0, T0, R0);
  __ and_(S0, A0, T0);

  __ bind(loop);
  __ move(A0, S0);
  __ move(A1, S1);
  __ jal_to_C(CAST_FROM_FN_PTR(address, cacheflush));
  __ delayed()->li(A2, DCACHE);

  __ move(A0, S0);
  __ move(A1, S1);
  __ jal_to_C(CAST_FROM_FN_PTR(address, cacheflush));
  __ delayed()->li(A2, ICACHE);

  // The ranges we get should always fit on a single page.  But if
  // not, keep going.
  __ addu_long(T0, S0, S1);
  __ bltu(T0, S3, loop);
  __ delayed()->addu_long(S0, S0, S1);

#endif

  __ move(V0, S4);
  __ restore_frame_and_return();

  // Must be set here so StubCodeMark destructor can call the flush stub.
  *flush_icache_stub = (ICache::flush_icache_stub_t)start;

  // How can we execute the stub if it hasn't been flushed?  Do it here.
  cacheflush(start, ICache::stub_size, DCACHE);
  cacheflush(start, ICache::stub_size, ICACHE);
};

#undef __
