#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)interp_masm_sparc.cpp	1.198 07/05/17 15:48:06 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_interp_masm_mips.cpp.incl"

#ifndef FAST_DISPATCH
#define FAST_DISPATCH 1
#endif
#undef FAST_DISPATCH

// Implementation of InterpreterMacroAssembler

// This file specializes the assembler with interpreter-specific macros


// Dispatch code executed in the prolog of a bytecode which does not do its
// own dispatch.
void InterpreterMacroAssembler::dispatch_prolog(TosState state, int bcp_incr) {
  // do it all in the epilog
}


// Dispatch code executed in the epilog of a bytecode which does not do its
// own dispatch.
void InterpreterMacroAssembler::dispatch_epilog(TosState state, int bcp_incr) {
  assert_not_delayed();
  verify_FPU(1, state);
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);

  lbu(Rscratch0, bcp_incr, Rbcp);   // Rscratch0 = bytecode
  // Increment Rbcp in the delay slot below
  sll_long(Rscratch0, Rscratch0, LogBytesPerWord);    // multiply by wordSize
  addu_long(Rscratch0, Rscratch0, (long)Interpreter::dispatch_table(state));  // add to table base
  ld(Rscratch2, 0, Rscratch0);  // Rscratch2 = target address
  jr(Rscratch2, 0);
  if (bcp_incr != 0)  delayed()->addu_long(Rbcp, Rbcp, bcp_incr);
  else                delayed()->nop();
}


void InterpreterMacroAssembler::dispatch_next(TosState state, int bcp_incr) {
  // %%%% consider branching to a single shared dispatch stub (for each bcp_incr)
  assert_not_delayed();
  lbu(Rbyte_code, bcp_incr, Rbcp);               // load next bytecode
  dispatch_Rbyte_code(state, Interpreter::dispatch_table(state), bcp_incr);
}


void InterpreterMacroAssembler::dispatch_next_noverify_oop(TosState state, int bcp_incr) {
  // %%%% consider branching to a single shared dispatch stub (for each bcp_incr)
  assert_not_delayed();
  lbu(Rbyte_code, bcp_incr, Rbcp);               // load next bytecode
  dispatch_Rbyte_code(state, Interpreter::dispatch_table(state), bcp_incr, false);
}


void InterpreterMacroAssembler::dispatch_via(TosState state, address* table) {
  // load current bytecode       
  assert_not_delayed();
  lbu(Rbyte_code, 0, Rbcp);               // load next bytecode
  dispatch_base(state, table);   
}                                


void InterpreterMacroAssembler::call_VM_leaf_base(
  Register java_thread,
  address  entry_point,
  int      number_of_arguments
) {
  // We basically ignore java_thread; it should always be valid.
  // super call
  MacroAssembler::call_VM_leaf_base(java_thread, entry_point, number_of_arguments);
}
 

void InterpreterMacroAssembler::call_VM_base(
  Register        oop_result,
  Register        java_thread,
  Register        last_java_sp,
  address         entry_point,
  int             number_of_arguments,
  bool		  check_exception
) {
  // We basically ignore java_thread; it should always be valid.

  // See class ThreadInVMfromInterpreter, which assumes that the interpreter
  // takes responsibility for setting its own thread-state on call-out.
  // However, ThreadInVMfromInterpreter resets the state to "in_Java".

  // Need to save interpreter state that could change (e.g., Rbcp might
  // be updated due to GC).
  save_dynamic_state();
  //save_bcp();                                  // save bcp
  MacroAssembler::call_VM_base(oop_result, java_thread, last_java_sp, entry_point, number_of_arguments, check_exception);
  //restore_bcp();                               // restore bcp
  //restore_locals();                            // restore locals pointer
  restore_dynamic_state();
}


void InterpreterMacroAssembler::check_and_handle_popframe(Register scratch_reg) {
  if (JvmtiExport::can_pop_frame()) {
    Label L;

    // Check the "pending popframe condition" flag in the current thread
    Unimplemented();

    // Initiate popframe handling only if it is not already being processed.  If the flag
    // has the popframe_processing bit set, it means that this code is called *during* popframe
    // handling - we don't want to reenter.
    Unimplemented();

    // Call Interpreter::remove_activation_preserving_args_entry() to get the
    // address of the same-named entrypoint in the generated interpreter code.
    call_VM_leaf(noreg, CAST_FROM_FN_PTR(address, Interpreter::remove_activation_preserving_args_entry));

    // Jump to Interpreter::_remove_activation_preserving_args_entry
    Unimplemented();
    bind(L);
  }
}


void InterpreterMacroAssembler::load_earlyret_value(TosState state) {
  Register thr_state = Rscratch4;
  ld_ptr(thr_state, in_bytes(JavaThread::jvmti_thread_state_offset()), Rthread);
  switch (state) {
  case ltos:
    ld_long(Rtos, in_bytes(JvmtiThreadState::earlyret_value_offset()), thr_state);
    break;
  case atos:
    ld_ptr(Rtos, in_bytes(JvmtiThreadState::earlyret_oop_offset()), thr_state);
    st_ptr(R0,   in_bytes(JvmtiThreadState::earlyret_oop_offset()), thr_state);
    break;
  case btos:                                           // fall through
  case ctos:                                           // fall through
  case stos:                                           // fall through
  case itos:
    lw(Rtos, in_bytes(JvmtiThreadState::earlyret_value_offset()), thr_state);
    break;
  case ftos:
    lwc1(Ftos, in_bytes(JvmtiThreadState::earlyret_value_offset()), thr_state);
    break;
  case dtos:
    ldc1(Ftos, in_bytes(JvmtiThreadState::earlyret_value_offset()), thr_state);
    break;
  case vtos: /* nothing to do */                          break;
  default  : ShouldNotReachHere();
  }
  // Clean up tos value in the jvmti thread state
  li(Rscratch3, ilgl);
  sw(Rscratch3, in_bytes(JvmtiThreadState::earlyret_tos_offset()), thr_state);
  st_long(R0, in_bytes(JvmtiThreadState::earlyret_value_offset()), thr_state);
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
}


void InterpreterMacroAssembler::check_and_handle_earlyret(Register scratch_reg) {
  if (JvmtiExport::can_force_early_return()) {
    Label L;
    Unimplemented();
    delayed()->nop();

    // Initiate earlyret handling only if it is not already being processed.
    // If the flag has the earlyret_processing bit set, it means that this code
    // is called *during* earlyret handling - we don't want to reenter.
    Unimplemented();
    delayed()->nop();

    // Call Interpreter::remove_activation_early_entry() to get the address of the
    // same-named entrypoint in the generated interpreter code
    Unimplemented();

    // Jump to Interpreter::_remove_activation_early_entry
    Unimplemented();
    delayed()->nop();
    bind(L);
  }
}


void InterpreterMacroAssembler::super_call_VM_leaf(Register thread_cache, address entry_point, Register arg_1) {
  move(A0, arg_1);
  MacroAssembler::call_VM_leaf_base(thread_cache, entry_point, 1);
}


void InterpreterMacroAssembler::dispatch_base(TosState state, address* table) {
  assert_not_delayed();
  dispatch_Rbyte_code(state, table);
}


void InterpreterMacroAssembler::dispatch_normal(TosState state) {
  dispatch_base(state, Interpreter::normal_table(state));
}


void InterpreterMacroAssembler::dispatch_only(TosState state) {
  dispatch_base(state, Interpreter::dispatch_table(state));
}

 
// common code to dispatch and dispatch_only
// dispatch value in Rbyte_code and increment Rbcp

void InterpreterMacroAssembler::dispatch_Rbyte_code(TosState state, address* table, int bcp_incr, bool verify) {
  verify_FPU(1, state);
  // %%%%% maybe implement +VerifyActivationFrameSize here
  //verify_thread(); //too slow; we will just verify on method entry & exit
  if (verify) interp_verify_oop(Rtos, state, __FILE__, __LINE__);
#ifdef FAST_DISPATCH
  if (table == Interpreter::dispatch_table(state)) {
    // use IdispatchTables
    Unimplemented();
  } else {
#endif
    // dispatch table to use
    sll_long(Rbyte_code, Rbyte_code, LogBytesPerWord);
    li(Rscratch3, (address)table);
    ld_ptr(Rscratch3, Rbyte_code, Rscratch3);
#ifdef FAST_DISPATCH
  }
#endif
  jr(Rscratch3);
  if (bcp_incr != 0)  delayed()->addu_long(Rbcp, Rbcp, bcp_incr);
  else                delayed()->nop();
}


// Helpers for expression stack

// Longs and doubles are Category 2 computational types in the
// JVM specification (section 3.11.1) and take 2 expression stack or
// local slots.
// Aligning them on 32 bit with tagged stacks is hard because the code generated
// for the dup* bytecodes depends on what types are already on the stack.
// If the types are split into the two stack/local slots, that is much easier
// (and we can use 0 for non-reference tags).

// Known good alignment in _LP64 but unknown otherwise
void InterpreterMacroAssembler::load_unaligned_double(FloatRegister ft, int offset, Register base) {
  assert_not_delayed();

#ifdef _LP64
  ldc1(ft, offset, base);
#else
#error unimplemented
#endif
}

// Known good alignment in _LP64 but unknown otherwise
void InterpreterMacroAssembler::store_unaligned_double(FloatRegister ft, int offset, Register base) {
  assert_not_delayed();

#ifdef _LP64
  sdc1(ft, offset, base);
#else
#error unimplemented
#endif
}


// Known good alignment in _LP64 but unknown otherwise
void InterpreterMacroAssembler::load_unaligned_long(Register rt, int offset, Register base) {
  assert_not_delayed();
#ifdef _LP64
  ld(rt, offset, base);
#else
#error unimplemented
#endif
}

// Known good alignment in _LP64 but unknown otherwise
void InterpreterMacroAssembler::store_unaligned_long(Register rt, int offset, Register base) {
  assert_not_delayed();

#ifdef _LP64
  sd(rt, offset, base);
#else
#error unimplemented
#endif
}

#ifdef ASSERT
void InterpreterMacroAssembler::verify_stack_tag(frame::Tag t,
                                                 Register r,
                                                 Register scratch) {
  if (TaggedStackInterpreter) {
    Label ok, long_ok;
    ld_ptr(r, Interpreter::expr_tag_offset_in_bytes(0), Resp);
    if (t == frame::TagCategory2) {
      beq(r, R0, long_ok);
      delayed()->ld_ptr(r, Interpreter::expr_tag_offset_in_bytes(1), Resp);
      stop("stack long/double tag value bad");
      bind(long_ok);
    } else if (t == frame::TagValue) {
      beq(r, R0, ok);
      delayed()->nop();
    } else {
      assert_different_registers(r, scratch);
      li(scratch, t);
      beq(r, scratch, ok);
      delayed()->nop();
    }
    // Also check if the stack value is zero; the tag might not have
    // been set coming from deopt.
    ld_ptr(r, Interpreter::expr_offset_in_bytes(0), Resp);
    beq(r, R0, ok);
    delayed()->nop();
    stop("Stack tag value is bad");
    bind(ok);
  }
}
#endif // ASSERT

void InterpreterMacroAssembler::pop_i(Register r) { 
  assert_not_delayed();
  // Uses destination register r for scratch
  debug_only(verify_stack_tag(frame::TagValue, r));
  ld_long(r, Interpreter::expr_offset_in_bytes(0), Resp);
  addu_long(Resp, Resp, Interpreter::stackElementSize());
  debug_only(verify_esp(Resp));
}

void InterpreterMacroAssembler::pop_ptr(Register r, Register scratch) {
  assert_not_delayed();
  // Uses destination register r for scratch
  debug_only(verify_stack_tag(frame::TagReference, r, scratch));
  ld_ptr(r, Interpreter::expr_offset_in_bytes(0), Resp);
  addu_long(Resp, Resp, Interpreter::stackElementSize());
  debug_only(verify_esp(Resp));
}

void InterpreterMacroAssembler::pop_l(Register r) { 
  assert_not_delayed();
  // Uses destination register r for scratch
  debug_only(verify_stack_tag(frame::TagCategory2, r));
  load_unaligned_long(r, Interpreter::expr_offset_in_bytes(0), Resp);
  addu_long(Resp, Resp, 2*Interpreter::stackElementSize());
  debug_only(verify_esp(Resp));
}


void InterpreterMacroAssembler::pop_f(FloatRegister f, Register scratch) {
  assert_not_delayed();
  debug_only(verify_stack_tag(frame::TagValue, scratch));
  lwc1(f, Interpreter::expr_offset_in_bytes(0), Resp);
  addu_long(Resp, Resp, Interpreter::stackElementSize());
  debug_only(verify_esp(Resp));
}


void InterpreterMacroAssembler::pop_d(FloatRegister r, Register scratch) {
  assert_not_delayed();
  debug_only(verify_stack_tag(frame::TagCategory2, scratch));
  load_unaligned_double(r, Interpreter::expr_offset_in_bytes(0), Resp);
  addu_long(Resp, Resp, 2*Interpreter::stackElementSize());
  debug_only(verify_esp(Resp));
}


void InterpreterMacroAssembler::tag_stack(Register r) {
  if (TaggedStackInterpreter) {
    st_ptr(r, Interpreter::tag_offset_in_bytes(), Resp);
  }
}

void InterpreterMacroAssembler::tag_stack(frame::Tag t, Register r) {
  if (TaggedStackInterpreter) {
    assert (frame::TagValue == 0, "TagValue must be zero");
    if (t == frame::TagValue) {
      st_ptr(R0, Interpreter::tag_offset_in_bytes(), Resp);
    } else if (t == frame::TagCategory2) {
      // Tag both slots
      st_ptr(R0, Interpreter::tag_offset_in_bytes(), Resp);
      st_ptr(R0, Interpreter::tag_offset_in_bytes() - Interpreter::stackElementSize(), Resp);
    } else {
      assert_different_registers(r, A3);
      li(A3, t);
      st_ptr(A3, Interpreter::tag_offset_in_bytes(), Resp);
    }
  }
}

void InterpreterMacroAssembler::push_i(Register r) {
  assert_not_delayed();
  debug_only(verify_esp(Resp));
  tag_stack(frame::TagValue, r);
  st_long(r, Interpreter::value_offset_in_bytes(), Resp);
  addu_long(Resp, Resp, -Interpreter::stackElementSize()); 
}

void InterpreterMacroAssembler::push_ptr(Register r) {
  assert_not_delayed();
  debug_only(verify_esp(Resp));
  tag_stack(frame::TagReference, r);
  st_ptr(r, Interpreter::value_offset_in_bytes(), Resp);
  addu_long(Resp, Resp, -Interpreter::stackElementSize()); 
}

void InterpreterMacroAssembler::push_ptr(Register r, Register tag) {
  assert_not_delayed();
  tag_stack(tag);
  st_ptr(r, Interpreter::value_offset_in_bytes(), Resp);
  addu_long(Resp, Resp, -Interpreter::stackElementSize()); 
}

void InterpreterMacroAssembler::push_l(Register r) {
  assert_not_delayed();
  debug_only(verify_esp(Resp));
  tag_stack(frame::TagCategory2, r);
  store_unaligned_long(r, Interpreter::value_offset_in_bytes() - Interpreter::stackElementSize(), Resp);
  addu_long(Resp, Resp, 2 * -Interpreter::stackElementSize());
}


void InterpreterMacroAssembler::push_f(FloatRegister f) {
  assert_not_delayed();
  debug_only(verify_esp(Resp));
  tag_stack(frame::TagValue, Rtos);  // Rtos doesn't really matter here
  swc1(f, Interpreter::value_offset_in_bytes(), Resp);
  addu_long(Resp, Resp, -Interpreter::stackElementSize()); 
}


void InterpreterMacroAssembler::push_d(FloatRegister d) {
  assert_not_delayed();
  debug_only(verify_esp(Resp));
  tag_stack(frame::TagCategory2, Rtos);  // Rtos doesn't really matter here
  store_unaligned_double(d, Interpreter::value_offset_in_bytes() - Interpreter::stackElementSize(), Resp);
  addu_long(Resp, Resp, 2 * -Interpreter::stackElementSize());
}


void InterpreterMacroAssembler::push(TosState state) {
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
  switch (state) {
    case atos: push_ptr();	      break;
    case btos: push_i();              break;  
    case ctos:
    case stos: push_i();              break;  
    case itos: push_i();              break;  
    case ltos: push_l();              break;
    case ftos: push_f();              break;
    case dtos: push_d();              break;
    case vtos: /* nothing to do */    break;
    default  : ShouldNotReachHere();
  }
}


void InterpreterMacroAssembler::pop(TosState state) {
  switch (state) {
    case atos: pop_ptr();	     break;
    case btos: pop_i();              break;
    case ctos:
    case stos: pop_i();              break;
    case itos: pop_i();              break;
    case ltos: pop_l();              break;
    case ftos: pop_f();              break;
    case dtos: pop_d();              break;
    case vtos: /* nothing to do */   break;
    default  : ShouldNotReachHere();
  }
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
}


// Tagged stack helpers for swap and dup
void InterpreterMacroAssembler::load_ptr_and_tag(int n, Register val,
                                                 Register tag) {
  ld_ptr(val, Interpreter::expr_offset_in_bytes(n), Resp);
  if (TaggedStackInterpreter) {
    ld_ptr(tag, Interpreter::expr_tag_offset_in_bytes(n), Resp);
  }
}
void InterpreterMacroAssembler::store_ptr_and_tag(int n, Register val,
						  Register tag) {
  st_ptr(val, Interpreter::expr_offset_in_bytes(n), Resp);
  if (TaggedStackInterpreter) {
    st_ptr(tag, Interpreter::expr_tag_offset_in_bytes(n), Resp);
  }
}


void InterpreterMacroAssembler::load_receiver(Register recv,
					      Register param_count) {
  sll(param_count, param_count, Interpreter::logStackElementSize());
  if (TaggedStackInterpreter) {
    addu_long(param_count, param_count, Interpreter::value_offset_in_bytes());
				// get obj address
  }
  ld_ptr(recv, param_count, Resp); // get receiver Oop
}

void InterpreterMacroAssembler::empty_expression_stack() {
  // Reset Resp
  addu_long(Resp, Rmonitors, -wordSize);

  // YYY SPARC goes to the trouble of resetting Lesp to a value that
  // YYY is max stack lower, and it also adjusts SP by this amount.
  // YYY Why?

  // YYY On amd64, it writes a zero to interpreter_frame_last_sp_offset.
  // YYY That makes a little more sense.

  // YYY This function is poorly named/explained, as it does more
  // YYY than just empty the expression stack, but what is not clear.
}


#ifdef ASSERT
void InterpreterMacroAssembler::verify_sp(Register Rsp, Register Rtemp) {
  Label Bad, OK;

  // Saved SP must be aligned.
#ifdef _LP64
  li(Rtemp, 2*BytesPerWord-1);
#else
  li(Rtemp, BytesPerWord-1);
#endif
  and_(Rtemp, Rtemp, Rsp);
  bne(Rtemp, R0, Bad);
  delayed()->nop();

  // Saved SP, plus register saves, must not be above FP.
  addu_long(Rtemp, Rsp, frame::register_save_space);
  bgtu(Rtemp, FP, Bad);
  delayed()->nop();

  b(OK);
  delayed()->nop();

  bind(Bad);
  stop("on return to interpreted call, restored SP is corrupted");

  bind(OK);
}


void InterpreterMacroAssembler::verify_esp(Register Resp) {
  // about to read or write Resp[0]
  // make sure it is not in the monitors or the register save area
  Label OK1, OK2;

  // This isn't exact; need to account for number of words in monitor area
  bltu(Resp, Rmonitors, OK1);
  delayed()->nop();
  stop("too many pops: Resp points into monitor area");

  bind(OK1);
  addu_long(Resp, Resp, - frame::register_save_space);
  // This isn't exact; need to account for alignment
  bgeu(Resp, SP, OK2);
  delayed()->addu_long(Resp, Resp, frame::register_save_space);
  stop("too many pushes: Resp points into register save area");

  bind(OK2);
}
#endif // ASSERT

// Load compiled (i2c) or interpreter entry when calling from interpreted and
// do the call. Centralized so that all interpreter calls will do the same actions.
// If jvmti single stepping is on for a thread we must not call compiled code.
void InterpreterMacroAssembler::call_from_interpreter(Register target, Register scratch, Register Rret) {

  // Assume we want to go compiled if available

  save_dynamic_state();

  ld_ptr(target, in_bytes(methodOopDesc::from_interpreted_offset()), Rmethod_call);

  if (JvmtiExport::can_post_interpreter_events()) {
    // JVMTI events, such as single-stepping, are implemented partly by avoiding running
    // compiled code in threads for which the event is enabled.  Check here for
    // interp_only_mode if these events CAN be enabled.
    verify_thread();
    Label skip_compiled_code;

    lwu(scratch, in_bytes(JavaThread::interp_only_mode_offset()), Rthread);
    beq(scratch, R0, skip_compiled_code);
    delayed()->nop();
    ld_ptr(target, in_bytes(methodOopDesc::interpreter_entry_offset()), Rmethod_call);

    bind(skip_compiled_code);
  }

  // the i2c_adapters need methodOop in Rmethod_call
#ifdef ASSERT
  {
    Label ok;
    bne(target, R0, ok);
    delayed()->nop();
    stop("null entry point");
    bind(ok);
  }
#endif // ASSERT

  //addu_long(Rargs, Resp, BytesPerWord);
  move(Rargs, Resp);

  // YYY More possibly needed here: save SP and/or adjust Rret.
  // YYY Coordinate with other end of call, generate_* code.

  // do the call
  move(Rscratch0, Rret);  // N.B.: Rret == Rlast_SP == S0.  Should juggle registers to save this move.
  move(Rlast_SP, SP);
  j(target);
  delayed()->move(RA, Rscratch0);
}


// Get constants in Java bytecode

void InterpreterMacroAssembler::get_2_byte_integer_at_base(
				  Register    Rbase,
                                  int         offset, 
                                  Register    Rtmp, 
				  Register    Rd,
                                  bool        is_signed) {
  assert_different_registers(Rtmp, Rd);
  assert_not_delayed();
  // first get high byte
  if (is_signed) {
    lb(Rd, offset, Rbase);
  } else {
    lbu(Rd, offset, Rbase);
  }
  // then shift it up and or in the low byte
  sll(Rd, Rd, 8);
  lbu(Rtmp, offset+1, Rbase);
  or_(Rd, Rd, Rtmp);
}

void InterpreterMacroAssembler::get_4_byte_integer_at_base(
				  Register   Rbase,
                                  int        offset,
                                  Register   Rtmp, 
				  Register   Rd,
				  bool       is_signed) {
  get_2_byte_integer_at_base(Rbase, offset, Rtmp, Rd, is_signed);

  sll(Rd, Rd, 8);
  lbu(Rtmp, offset+2, Rbase);
  or_(Rd, Rd, Rtmp);

  sll(Rd, Rd, 8);
  lbu(Rtmp, offset+3, Rbase);
  or_(Rd, Rd, Rtmp);
}

void InterpreterMacroAssembler::get_2_byte_integer_at_bcp(
                                  int         bcp_offset, 
                                  Register    Rtmp, 
				  Register    Rd,
                                  bool        is_signed) {
  get_2_byte_integer_at_base(Rbcp, bcp_offset, Rtmp, Rd, is_signed);
}

void InterpreterMacroAssembler::get_4_byte_integer_at_bcp(
                                  int        bcp_offset,
                                  Register   Rtmp, 
				  Register   Rd) {
  get_4_byte_integer_at_base(Rbcp, bcp_offset, Rtmp, Rd, true);
}

void InterpreterMacroAssembler::get_2_byte_integer(
                                  Register    Rbase,
                                  Register    Roffset,
                                  Register    Rtmp,
                                  Register    Rtmp2,
				  Register    Rd,
                                  bool        is_signed) {
  assert_different_registers(Rtmp, Rd);
  assert_not_delayed();
  addu_long(Rtmp, Rbase, Roffset);
#ifdef VM_LITTLE_ENDIAN
  // first get high byte
  if (is_signed) {
    lb(Rd, 0, Rtmp);
  } else {
    lbu(Rd, 0, Rtmp);
  }
  // then shift it up and or in the low byte
  sll(Rd, Rd, 8);
  addu_long(Rtmp, Rtmp, 1);
  lbu(Rtmp2, 0, Rtmp);
  or_(Rd, Rd, Rtmp2);
#else
  if (is_signed) {
    lh(Rd, 0, Rtmp);
  } else {
    lhu(Rd, 0, Rtmp);
  }
#endif
}

void InterpreterMacroAssembler::get_4_byte_integer(
                                  Register    Rbase,
                                  Register    Roffset,
                                  Register    Rtmp, 
                                  Register    Rtmp2, 
				  Register    Rd,
                                  bool        is_signed) {
#ifdef VM_LITTLE_ENDIAN
  get_2_byte_integer(Rbase, Roffset, Rtmp, Rtmp2, Rd, is_signed);

  addu_long(Rtmp, Rtmp, 1);
  sll(Rd, Rd, 8);
  lbu(Rtmp2, 0, Rtmp);
  or_(Rd, Rd, Rtmp2);

  addu_long(Rtmp, Rtmp, 1);
  sll(Rd, Rd, 8);
  lbu(Rtmp2, 0, Rtmp);
  or_(Rd, Rd, Rtmp2);
#else
  addu_long(Rtmp, Rbase, Roffset);
  if (is_signed) {
    lw(Rd, 0, Rtmp);
  } else {
    lwu(Rd, 0, Rtmp);
  }
#endif
}


void InterpreterMacroAssembler::get_cache_and_index_at_bcp(Register cache, Register tmp, int bcp_offset) {
  assert(bcp_offset > 0, "bcp is still pointing to start of bytecode");
  assert_different_registers(cache, tmp);
  assert_not_delayed();
    // methodrefs and fieldrefs have been converted to cpCache indices
    // as a part of loading.  These indices are stored in native format
    // (not Java format).  So we need to use a regular load here.
    //get_2_byte_integer_at_bcp(bcp_offset, cache, tmp, false);
  lhu(tmp, bcp_offset, Rbcp);
    // convert from field index to ConstantPoolCacheEntry index
    // and from word index to byte offset
  sll_long(tmp, tmp, exact_log2(in_words(ConstantPoolCacheEntry::size()) * BytesPerWord));
  addu_long(cache, RcpoolCache, tmp);
}


void InterpreterMacroAssembler::get_cache_entry_pointer_at_bcp(Register cache, Register tmp, int bcp_offset) {
  assert(bcp_offset > 0, "bcp is still pointing to start of bytecode");
  assert_different_registers(cache, tmp);
  assert_not_delayed();
  get_2_byte_integer_at_bcp(bcp_offset, cache, tmp, Unsigned);
  Unimplemented();
              // convert from field index to ConstantPoolCacheEntry index
              // and from word index to byte offset
              // skip past the header
              // construct pointer to cache entry
}


// Generate a subtype check: branch to ok_is_subtype if sub_klass is
// a subtype of super_klass.  Blows registers Rsub_klass, tmp1, tmp2.
void InterpreterMacroAssembler::gen_subtype_check(Register Rsub_klass,
                                                  Register Rsuper_klass,
                                                  Register Rtmp1,
                                                  Register Rtmp2,
                                                  Register Rtmp3,
                                                  Label &ok_is_subtype ) {
  Label not_subtype, loop;

  // Profile the not-null value's klass.
  profile_typecheck(Rsub_klass, Rtmp1);

  // Load the super-klass's check offset (juint) into Rtmp1
  lwu(Rtmp1, sizeof(oopDesc) + Klass::super_check_offset_offset_in_bytes(), Rsuper_klass);

  // Load from the sub-klass's super-class display list, or a 1-word cache of
  // the secondary superclass list, or a failing value with a sentinel offset
  // if the super-klass is an interface or exceptionally deep in the Java
  // hierarchy and we have to scan the secondary superclass list the hard way.
  ld_ptr(Rtmp2, Rtmp1, Rsub_klass);

  // See if we get an immediate positive hit
  beq(Rtmp2, Rsuper_klass, ok_is_subtype);
  delayed()->nop();

  // Check for immediate negative
  li(Rtmp3, sizeof(oopDesc) + Klass::secondary_super_cache_offset_in_bytes());
  bne(Rtmp1, Rtmp3, not_subtype);
  delayed()->nop();

  // check for self
  beq(Rsub_klass, Rsuper_klass, ok_is_subtype);

  // Now do a linear scan of the secondary super-klass chain.
  delayed()->ld_ptr(Rtmp2, sizeof(oopDesc) + Klass::secondary_supers_offset_in_bytes(), Rsub_klass);
  // Rtmp2 holds the objArrayOop of secondary supers.

  lwu(Rtmp1, arrayOopDesc::length_offset_in_bytes(), Rtmp2);
  // Rtmp1 holds the length of the array

  bind(loop);
  beq(Rtmp1, R0, not_subtype);
  delayed()->nop();
  ld_ptr(Rtmp3, arrayOopDesc::base_offset_in_bytes(T_OBJECT), Rtmp2);
  addu_long(Rtmp2, Rtmp2, wordSize);
  bne(Rtmp3, Rsuper_klass, loop);
  delayed()->addu_long(Rtmp1, Rtmp1, -1);
  b(ok_is_subtype);
  delayed()->st_ptr(Rsuper_klass, sizeof(oopDesc) + Klass::secondary_super_cache_offset_in_bytes(), Rsub_klass);

  bind(not_subtype);
  profile_typecheck_failed(Rtmp1);
}



#if 0
// unused
void InterpreterMacroAssembler::jump_entry_point(address entry_point,
						 Register Rtmp) {
  assert(entry_point != NULL, "entry_point must be generated by now");
  Address dest(Rtmp, entry_point);
  jump_to(dest);
  delayed()->nop();
}
#endif



void InterpreterMacroAssembler::jump_if(Condition cond,
					Register r1,
					Register r2,
					address entry_point) {
  jump_if_not(negateCondition(cond), r1, r2, entry_point);
}

void InterpreterMacroAssembler::jump_if_not(Condition cond,
					    Register r1,
					    Register r2,
					    address entry_point) {
  if (cond != always) {
    Label skip;
    bx(cond, r1, r2, skip);
    delayed()->nop();
    jump_to(entry_point);
    delayed()->nop();
    bind(skip);
  }
}

// Check that index is in range for array, then shift index by index_shift, and put arrayOop + shifted_index into res
// Note: res is still shy of address by array offset into object.

void InterpreterMacroAssembler::index_check_without_pop(Register array, Register index, int index_shift, Register tmp, Register res) {
  assert_not_delayed();

  verify_oop(array);
  // YYY SPARC impl. sign-extends here, but we always extend upon load

  // check for null array
  Label ptr_ok;
  bne(array, R0, ptr_ok);
  delayed()->ld_long(tmp, arrayOopDesc::length_offset_in_bytes(), array);
  jump_to(Interpreter::_throw_NullPointerException_entry);
  delayed()->nop();
  bind(ptr_ok);

  // check index bounds
  Label index_ok;
  bltu(index, tmp, index_ok);
  if (index_shift > 0) {
    delayed()->sll_long(index, index, index_shift);
  } else {
    delayed()->addu_long(res, array, index);
  }
  // convention: move index into V0 for exception message
  move(V0, index);
  jump_to(Interpreter::_throw_ArrayIndexOutOfBoundsException_entry);
  delayed()->nop();

  bind(index_ok);
  if (index_shift > 0) {
    addu_long(res, array, index);
  }
}

// Get array off the stack into "array" and put pointer to element in "res".
void InterpreterMacroAssembler::index_check(Register array, Register index, int index_shift, Register tmp, Register res) {
  assert_not_delayed();

  // pop array
  pop_ptr(array);

  // check array
  index_check_without_pop(array, index, index_shift, tmp, res);
}


void InterpreterMacroAssembler::get_constant_pool(Register Rdst) {
  ld_ptr(Rdst, in_bytes(methodOopDesc::constants_offset()), Rmethod);
}


void InterpreterMacroAssembler::get_constant_pool_cache(Register Rdst) {
  get_constant_pool(Rdst);
  ld_ptr(Rdst, constantPoolOopDesc::cache_offset_in_bytes(), Rdst);
}


void InterpreterMacroAssembler::get_cpool_and_tags(Register Rcpool, Register Rtags) {
  get_constant_pool(Rcpool);
  ld_ptr(Rtags, constantPoolOopDesc::tags_offset_in_bytes(), Rcpool);
}



// unlock if synchronized method
//
// Unlock the receiver if this is a synchronized method.
// Unlock any Java monitors from syncronized blocks.
//
// If there are locked Java monitors
//    If throw_monitor_exception
//       throws IllegalMonitorStateException
//    Else if install_monitor_exception
//       installs IllegalMonitorStateException
//    Else
//       no error processing
void InterpreterMacroAssembler::unlock_if_synchronized_method(TosState state,
							      bool throw_monitor_exception,
							      bool install_monitor_exception) {
  Label unlocked, unlock, no_unlock;

  // get the value of _do_not_unlock_if_synchronized
  lb(Rscratch1, in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);
  sb(R0,        in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);

  // check if synchronized method
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
  push(state); // save tos
  lw(Rscratch3, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);
  andi(Rscratch3, Rscratch3, JVM_ACC_SYNCHRONIZED);
  beq(Rscratch3, R0, unlocked);
  delayed()->nop();

  // Don't unlock anything if the _do_not_unlock_if_synchronized flag
  // is set.
  bne(Rscratch1, R0, no_unlock);
  delayed()->nop();

  // BasicObjectLock will be first in list, since this is a synchronized method. However, need
  // to check that the object has not been unlocked by an explicit monitorexit bytecode.  

  //Intel: if (throw_monitor_exception) ... else ...
  // Entry already unlocked, need to throw exception
  //...

  // pass top-most monitor elem
  get_top_most_monitor(A1);

  ld_ptr(Rscratch3, BasicObjectLock::obj_offset_in_bytes(), A1);
  bne(Rscratch3, R0, unlock);
  delayed()->nop();

  if (throw_monitor_exception) {
    // Entry already unlocked need to throw an exception
    MacroAssembler::call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_illegal_monitor_state_exception));
    should_not_reach_here();
  } else {
    // Monitor already unlocked during a stack unroll. 
    // If requested, install an illegal_monitor_state_exception.
    // Continue with stack unrolling.
    if (install_monitor_exception) {
      MacroAssembler::call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::new_illegal_monitor_state_exception));
    }
    b(unlocked);
    delayed()->nop();
  }

  bind(unlock);

  unlock_object(A1);

  bind(unlocked);

  // Check that all monitors are unlocked
  { Label loop, exception, entry, restart;

    Register Rmptr   = A0;
    Register Rtemp   = A1;
    Register Rlimit  = Rmonitors;
    const jint delta = frame::interpreter_frame_monitor_size() * wordSize;
    assert( (delta & LongAlignmentMask) == 0, 
	    "sizeof BasicObjectLock must be even number of doublewords");

    bind(restart);
    get_top_most_monitor(Rmptr);
    // We add a bias of one element to make the loop more natural.
    // This means that Rmptr == Rmonitors when there are no monitors at all.
    addu_long(Rmptr, Rmptr, delta);
#ifdef ASSERT
    { Label L;
      // ensure that Rmptr starts out above (or at) Rlimit
      bgeu(Rmptr, Rlimit, L);
      delayed()->nop();
      stop("monitor stack has negative size");
      bind(L);
    }
#endif
    b(entry);
    delayed()->nop();

    // Entry is still locked, need to throw exception
    bind(exception);
    if (throw_monitor_exception) {
      MacroAssembler::call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_illegal_monitor_state_exception));
      should_not_reach_here();
    } else {
      // Stack unrolling. Unlock object and if requested, install illegal_monitor_exception.
      // Unlock does not block, so don't have to worry about the frame
      unlock_object(Rmptr);
      if (install_monitor_exception) {
        MacroAssembler::call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::new_illegal_monitor_state_exception));
      }
      b(restart);
      delayed()->nop();
    }

    bind(loop);
    bne(Rtemp, R0, exception);            // check if current entry is used
    delayed()->
      addu_long(Rmptr, Rmptr, -delta);    // otherwise advance to next entry
#ifdef ASSERT
    { Label L;
      // ensure that Rmptr has not somehow stepped below Rlimit
      bgeu(Rmptr, Rlimit, L);
      delayed()->nop();
      stop("ran off the end of the monitor stack");
      bind(L);
    }
#endif
    bind(entry);
    bne(Rmptr, Rlimit, loop);   // if not at bottom then check this entry
    delayed()->
      ld_ptr(Rtemp, BasicObjectLock::obj_offset_in_bytes() - delta, Rmptr);
  }

  bind(no_unlock);
  pop(state);
  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
}


// remove activation
// ** Sets Rscratch5 to old FP **
//
// Unlock the receiver if this is a synchronized method.
// Unlock any Java monitors from syncronized blocks.
// Remove the activation from the stack.
//
// If there are locked Java monitors
//    If throw_monitor_exception
//       throws IllegalMonitorStateException
//    Else if install_monitor_exception
//       installs IllegalMonitorStateException
//    Else
//       no error processing
void InterpreterMacroAssembler::remove_activation(TosState state,
                                                  bool throw_monitor_exception,
                                                  bool install_monitor_exception) {

  unlock_if_synchronized_method(state, throw_monitor_exception, install_monitor_exception);

  // save result (push state before jvmti call and pop it afterwards) and notify jvmti
  notify_method_exit(false, state, NotifyJVMTI); 

  interp_verify_oop(Rtos, state, __FILE__, __LINE__);
  verify_oop(Rmethod);
  verify_thread();

  // return tos
#ifndef _LP64
  assert(Rtos_l1 == Rtos, "adjust code below");
#endif
  switch (state) {
#ifdef _LP64
  case ltos:  // fall through
#else
#error unimplemented
#endif
  case btos:  // fall through
  case ctos:  // fall through
  case stos:  // fall through
  case atos:  // fall through
  case itos:
    assert(V0 == Rtos, "must move to return register");
    break;
  case ftos:  // fall through
  case dtos:  // fall through
    assert(F0 == Ftos, "must move to return register");
    break;
  case vtos:
    // nothing to do -- only used for void return
    break;
  default:
    ShouldNotReachHere();
  }

  move(Rscratch5, FP);  // save FP so we can reset SP
  restore_registers();
}


// Lock object
//
// Argument - lock_reg points to the BasicObjectLock to be used for locking,
//            it must be initialized with the object to lock
void InterpreterMacroAssembler::lock_object(Register lock_reg, Register Object) {
  if (UseHeavyMonitors) {
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::monitorenter), lock_reg);
  }
  else {
    Register obj_reg = Object;
    Register mark_reg = Rscratch4;
    Register temp_reg = Rscratch1;
    Label    done;

    Label slow_case;

    assert_different_registers(lock_reg, obj_reg, mark_reg, temp_reg);

    // load markOop from object into mark_reg
    ld_ptr(mark_reg, oopDesc::mark_offset_in_bytes(), obj_reg);

    if (UseBiasedLocking) {
      assert(0, "biased locking not yet implemented");
      //biased_locking_enter(obj_reg, mark_reg, temp_reg, done, &slow_case);
    }
    
    // set mark reg to be (markOop of object | UNLOCK_VALUE)
    ori(mark_reg, mark_reg, markOopDesc::unlocked_value);
    // initialize the box  (Must happen before we update the object mark!)
    st_ptr(mark_reg, 
	   BasicLock::displaced_header_offset_in_bytes() + 
	   BasicObjectLock::lock_offset_in_bytes(),
	   lock_reg);

    // compare and exchange object_addr, markOop | 1, stack address of basicLock
#ifdef _LP64
    Label value_mismatch;
    lld(temp_reg, oopDesc::mark_offset_in_bytes(), obj_reg);
#ifdef SICORTEX_ERRATA
    // YYY temporary fix for SiCortex errata: doubled lld
    lld(temp_reg, oopDesc::mark_offset_in_bytes(), obj_reg);
#endif
    bne(temp_reg, mark_reg, value_mismatch);
    delayed()->move(Rscratch0, lock_reg);
    scd(Rscratch0, oopDesc::mark_offset_in_bytes(), obj_reg);
    bne(Rscratch0, R0, done);
    delayed()->nop();
#else
#error unimplemented
#endif

    bind(value_mismatch);
    // We did not see an unlocked object or the atomic op failed,
    // so try the fast recursive case.
    // Check if owner is self by comparing the value in the markOop of object
    // with the stack pointer
    subu_long(temp_reg, temp_reg, SP);
    assert(os::vm_page_size() > 0xfff, "page size too small - change the constant");
    addu_long(Rscratch0, R0, 0xF003);  // load signed immediate
    and_(temp_reg, temp_reg, Rscratch0);
    bne(temp_reg, R0, slow_case);
    delayed()->nop();
    // if same, then we can store 0 in the displaced header indicating
    // it is a recursive lock and be done
    st_ptr(R0,
	   BasicLock::displaced_header_offset_in_bytes() +
	   BasicObjectLock::lock_offset_in_bytes(),
	   lock_reg);
    b(done);
    delayed()->nop();

    // none of the above fast optimizations worked so we have to get into the
    // slow case of monitor enter
    bind(slow_case);
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::monitorenter), lock_reg);

    bind(done);
  }
}
    
// Unlocks an object. Used in monitorexit bytecode and remove_activation.
//
// Argument - lock_reg points to the BasicObjectLock for lock
// Throw IllegalMonitorException if object is not locked by current thread
void InterpreterMacroAssembler::unlock_object(Register lock_reg) {
  if (UseHeavyMonitors) {
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::monitorexit), lock_reg);
  } else {
    Register obj_reg = Rscratch3;
    Register mark_reg = Rscratch4;
    Register displaced_header_reg = Rscratch1;
    Label    done, free_entry;

    if (UseBiasedLocking) {
      // load the object out of the BasicObjectLock
      ld_ptr(obj_reg, BasicObjectLock::obj_offset_in_bytes(), lock_reg);
      guarantee(0, "biased locking not yet implemented");
      // biased_locking_exit ...
      st_ptr(R0, BasicObjectLock::obj_offset_in_bytes(), lock_reg);  // free entry
    }

    // Test first if we are in the fast recursive case
    ld_ptr(displaced_header_reg,
	   BasicLock::displaced_header_offset_in_bytes() + 
	   BasicObjectLock::lock_offset_in_bytes(),
	   lock_reg);
    beq(displaced_header_reg, R0, free_entry);
    delayed()->nop();

    // See if it is still a light weight lock, if so we just unlock
    // the object and we are done

    if (!UseBiasedLocking) {
      // load the object out of the BasicObjectLock
      ld_ptr(obj_reg, BasicObjectLock::obj_offset_in_bytes(), lock_reg);
    }

    // we have the displaced header in displaced_header_reg
    // we expect to see the stack address of the basicLock in case the
    // lock is still a light weight lock (lock_reg)
#ifdef _LP64
    Label value_mismatch;
    lld(Rscratch0, oopDesc::mark_offset_in_bytes(), obj_reg);
#ifdef SICORTEX_ERRATA
    // YYY temporary fix for SiCortex errata: doubled lld
    lld(Rscratch0, oopDesc::mark_offset_in_bytes(), obj_reg);
#endif
    bne(Rscratch0, lock_reg, value_mismatch);
    delayed()->move(Rscratch0, displaced_header_reg);
    scd(Rscratch0, oopDesc::mark_offset_in_bytes(), obj_reg);
    bne(Rscratch0, R0, free_entry);
    delayed()->nop();
#else
#error unimplemented
#endif

    bind(value_mismatch);
    // The lock has been converted into a heavy lock and hence
    // we need to get into the slow case

    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::monitorexit), lock_reg);
    b(done);
    delayed()->nop();

    bind(free_entry);
    st_ptr(R0, BasicObjectLock::obj_offset_in_bytes(), lock_reg);  // free entry

    bind(done);
  }
}

// Get the method data pointer from the methodOop and set the
// specified register to its value.

void InterpreterMacroAssembler::set_method_data_pointer_offset(Register Roff) {
  Unimplemented();
}

// Set the method data pointer for the current bcp.

void InterpreterMacroAssembler::set_method_data_pointer_for_bcp() {
  Unimplemented();
}

// Test ImethodDataPtr.  If it is null, continue at the specified label

void InterpreterMacroAssembler::test_method_data_pointer(Label& zero_continue) {
  assert(ProfileInterpreter, "must be profiling interpreter");
#ifdef _LP64
  Unimplemented();
#else
#error unimplemented
#endif
  delayed()->nop();
}

void InterpreterMacroAssembler::verify_method_data_pointer() {
  assert(ProfileInterpreter, "must be profiling interpreter");
#ifdef ASSERT
  Label verify_continue;
  Unimplemented();
  bind(verify_continue);
#endif // ASSERT
}

void InterpreterMacroAssembler::test_invocation_counter_for_mdp(Register invocation_count, 
								Register cur_bcp, 
								Register Rtmp,
								Label &profile_continue) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  // Control will flow to "profile_continue" if the counter is less than the 
  // limit or if we call profile_method()

  Label done;

  Unimplemented();
  delayed()->nop();
  bind(done);
}

// Store a value at some constant offset from the method data pointer.

void InterpreterMacroAssembler::set_mdp_data_at(int constant, Register value) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

void InterpreterMacroAssembler::increment_mdp_data_at(Address counter,
						      Register bumped_count,
                                                      bool decrement) {
  assert(ProfileInterpreter, "must be profiling interpreter");

  Unimplemented();
}

// Increment the value at some constant offset from the method data pointer.

void InterpreterMacroAssembler::increment_mdp_data_at(int constant,
						      Register bumped_count,
                                                      bool decrement) {
  // Locate the counter at a fixed offset from the mdp:
  Unimplemented();
}

// Increment the value at some non-fixed (reg + constant) offset from
// the method data pointer.

void InterpreterMacroAssembler::increment_mdp_data_at(Register reg,
                                                      int constant,
                                                      Register bumped_count,
						      Register scratch2,
                                                      bool decrement) {
  // Add the constant to reg to get the offset.
  Unimplemented();
}

// Set a flag value at the current method data pointer position.
// Updates a single byte of the header, to avoid races with other header bits.

void InterpreterMacroAssembler::set_mdp_flag_at(int flag_constant,
						Register scratch) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

// Test the location at some offset from the method data pointer.
// If it is not equal to value, branch to the not_equal_continue Label.
// Set condition codes to match the nullness of the loaded value.

void InterpreterMacroAssembler::test_mdp_data_at(int offset,
                                                 Register value,
                                                 Label& not_equal_continue,
						 Register scratch) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

// Update the method data pointer by the displacement located at some fixed
// offset from the method data pointer.

void InterpreterMacroAssembler::update_mdp_by_offset(int offset_of_disp,
						     Register scratch) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

// Update the method data pointer by the displacement located at the
// offset (reg + offset_of_disp).

void InterpreterMacroAssembler::update_mdp_by_offset(Register reg,
						     int offset_of_disp,
						     Register scratch) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

// Update the method data pointer by a simple constant displacement.

void InterpreterMacroAssembler::update_mdp_by_constant(int constant) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  Unimplemented();
}

// Update the method data pointer for a _ret bytecode whose target
// was not among our cached targets.

void InterpreterMacroAssembler::update_mdp_for_ret(TosState state,
						   Register return_bci) {
  assert(ProfileInterpreter, "must be profiling interpreter");
  push(state);
  Unimplemented();
  pop(state);
}

// Count a taken branch in the bytecodes.

void InterpreterMacroAssembler::profile_taken_branch(Register scratch, Register bumped_count) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    // We are taking a branch.  Increment the taken count.
    increment_mdp_data_at(in_bytes(JumpData::taken_offset()), bumped_count);

    // The method data pointer needs to be updated to reflect the new target.
    update_mdp_by_offset(in_bytes(JumpData::displacement_offset()), scratch);
    bind (profile_continue);
  }
}


// Count a not-taken branch in the bytecodes.

void InterpreterMacroAssembler::profile_not_taken_branch(Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    // We are taking a branch.  Increment the not taken count.
    increment_mdp_data_at(in_bytes(BranchData::not_taken_offset()), scratch);

    // The method data pointer needs to be updated to correspond to the
    // next bytecode.
    update_mdp_by_constant(in_bytes(BranchData::branch_data_size()));
    bind (profile_continue);
  }
}


// Count a non-virtual call in the bytecodes.

void InterpreterMacroAssembler::profile_call(Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    // We are making a call.  Increment the count.
    increment_mdp_data_at(in_bytes(CounterData::count_offset()), scratch);

    // The method data pointer needs to be updated to reflect the new target.
    update_mdp_by_constant(in_bytes(CounterData::counter_data_size()));
    bind (profile_continue);
  }
}


// Count a final call in the bytecodes.

void InterpreterMacroAssembler::profile_final_call(Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    // We are making a call.  Increment the count.
    increment_mdp_data_at(in_bytes(CounterData::count_offset()), scratch);

    // The method data pointer needs to be updated to reflect the new target.
    update_mdp_by_constant(in_bytes(VirtualCallData::virtual_call_data_size()));
    bind (profile_continue);
  }
}


// Count a virtual call in the bytecodes.

void InterpreterMacroAssembler::profile_virtual_call(Register receiver,
						     Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    Unimplemented();

    bind (profile_continue);
  }
}

void InterpreterMacroAssembler::record_klass_in_profile_helper(
                                        Register receiver, Register scratch,
                                        int start_row, Label& done) {
  if (UseCompiler) {		// YYY unimplemented
  int last_row = VirtualCallData::row_limit() - 1;
  assert(start_row <= last_row, "must be work left to do");
  // Test this row for both the receiver and for null.
  // Take any of three different outcomes:
  //   1. found receiver => increment count and goto done
  //   2. found null => keep looking for case 1, maybe allocate this cell
  //   3. found something else => keep looking for cases 1 and 2
  // Case 3 is handled by a recursive call.
  for (int row = start_row; row <= last_row; row++) {
    Label next_test;
    bool test_for_null_also = (row == start_row);

    // See if the receiver is receiver[n].
    int recvr_offset = in_bytes(VirtualCallData::receiver_offset(row));
    test_mdp_data_at(recvr_offset, receiver, next_test, scratch);

    // The receiver is receiver[n].  Increment count[n].
    int count_offset = in_bytes(VirtualCallData::receiver_count_offset(row));
    increment_mdp_data_at(count_offset, scratch);
    b(done);
    delayed()->nop();
    bind(next_test);

    if (test_for_null_also) {
      // Failed the equality check on receiver[n]...  Test for null.
      if (start_row == last_row) {
        // The only thing left to do is handle the null case.
	Unimplemented();
        //brx(Assembler::notZero, false, Assembler::pt, done);
        delayed()->nop();
        break;
      }
      // Since null is rare, make it be the branch-taken case.
      Label found_null;
      Unimplemented();
      //brx(Assembler::zero, false, Assembler::pn, found_null);
      delayed()->nop();

      // Put all the "Case 3" tests here.
      record_klass_in_profile_helper(receiver, scratch, start_row + 1, done);

      // Found a null.  Keep searching for a matching receiver,
      // but remember that this is an empty (unused) slot.
      bind(found_null);
    }
  }

  // In the fall-through case, we found no matching receiver, but we
  // observed the receiver[start_row] is NULL.

  // Fill in the receiver field and increment the count.
  int recvr_offset = in_bytes(VirtualCallData::receiver_offset(start_row));
  set_mdp_data_at(recvr_offset, receiver);
  int count_offset = in_bytes(VirtualCallData::receiver_count_offset(start_row));
  Unimplemented();
  //mov(DataLayout::counter_increment, scratch);
  set_mdp_data_at(count_offset, scratch);
  b(done);
  delayed()->nop();
  }
}

void InterpreterMacroAssembler::record_klass_in_profile(Register receiver,
                                                        Register scratch) {
  assert(ProfileInterpreter, "must be profiling");
  Label done;

  record_klass_in_profile_helper(receiver, scratch, 0, done);

  bind (done);
}


// Count a ret in the bytecodes.

void InterpreterMacroAssembler::profile_ret(TosState state,
					    Register return_bci,
					    Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;
    uint row;
    
    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);
    
    // Update the total ret count.
    increment_mdp_data_at(in_bytes(CounterData::count_offset()), scratch);
    
    for (row = 0; row < RetData::row_limit(); row++) {
      Label next_test;
      
      // See if return_bci is equal to bci[n]:
      test_mdp_data_at(in_bytes(RetData::bci_offset(row)),
		       return_bci, next_test, scratch);
      
      // return_bci is equal to bci[n].  Increment the count.
      increment_mdp_data_at(in_bytes(RetData::bci_count_offset(row)), scratch);
      
      // The method data pointer needs to be updated to reflect the new target.
      update_mdp_by_offset(in_bytes(RetData::bci_displacement_offset(row)), scratch);
      b(profile_continue);
      delayed()->nop();
      bind(next_test);
    }
    
    update_mdp_for_ret(state, return_bci);
    
    bind (profile_continue);
  }
}

// Profile an unexpected null in the bytecodes.
void InterpreterMacroAssembler::profile_null_seen(Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    set_mdp_flag_at(BitData::null_seen_byte_constant(), scratch);

    // The method data pointer needs to be updated.
    int mdp_delta = in_bytes(BitData::bit_data_size());
    if (TypeProfileCasts) {
      mdp_delta = in_bytes(VirtualCallData::virtual_call_data_size());
    }
    update_mdp_by_constant(mdp_delta);

    bind (profile_continue);
  }
}

void InterpreterMacroAssembler::profile_typecheck(Register klass,
						  Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    // If no method data exists, go to profile_continue.
    test_method_data_pointer(profile_continue);

    int mdp_delta = in_bytes(BitData::bit_data_size());
    if (TypeProfileCasts) {
      mdp_delta = in_bytes(VirtualCallData::virtual_call_data_size());

      // Record the object type.
      record_klass_in_profile(klass, scratch);
    }

    // The method data pointer needs to be updated.
    update_mdp_by_constant(mdp_delta);

    bind (profile_continue);
  }
}

void InterpreterMacroAssembler::profile_typecheck_failed(Register scratch) {
  if (ProfileInterpreter && TypeProfileCasts) {
    Label profile_continue;

    Unimplemented();

    bind (profile_continue);
  }
}

// Count the default case of a switch construct.

void InterpreterMacroAssembler::profile_switch_default(Register scratch) {
  if (ProfileInterpreter) {
    Label profile_continue;

    Unimplemented();

    bind (profile_continue);
  }
}

// Count the index'th case of a switch construct.

void InterpreterMacroAssembler::profile_switch_case(Register index,
						    Register scratch,
						    Register scratch2,
						    Register scratch3) {
  if (ProfileInterpreter) {
    Label profile_continue;

    Unimplemented();

    bind (profile_continue);
  }
}

// add a InterpMonitorElem to stack (see frame_mips.hpp)
void InterpreterMacroAssembler::add_monitor_to_stack(Register Rtemp,
						     Register Rtemp2)
{
  const jint delta = frame::interpreter_frame_monitor_size() * wordSize;
  assert( (delta & LongAlignmentMask) == 0, 
	  "sizeof BasicObjectLock must be even number of doublewords");

  addu_long(SP, SP, - delta);
  addu_long(Resp, Resp, - delta);
  addu_long(Rmonitors, Rmonitors, - delta);

  // must copy stack contents down
  Label start_copying, next;

  b(start_copying);
  delayed()->move(Rtemp, SP);
  // YYY sparc version starts at Resp + wordSize, which I don't understand.
  // YYY SP seems right to me, so I'm using that.

  // Copies from low addresses upwards, moving everything down one
  // slot (delta) to make room for another monitor.

  bind(next);
  st_ptr(Rtemp2, 0, Rtemp);
  addu_long(Rtemp, Rtemp, wordSize);

  bind(start_copying);
  bltu(Rtemp, Rmonitors, next);
  delayed()->ld_ptr(Rtemp2, delta, Rtemp);

  st_ptr(Rmonitors, frame::interpreter_frame_Rmonitors_offset * BytesPerWord, FP);
  st_ptr(Resp, frame::interpreter_frame_Resp_offset * BytesPerWord, FP);
}

// Locals
#ifdef ASSERT
void InterpreterMacroAssembler::verify_local_tag(frame::Tag t,
                                                 Register base,
                                                 Register scratch, 
                                                 int n) {
  if (TaggedStackInterpreter) {
    Label ok, long_ok;
    assert_different_registers(base, scratch);
    ld_ptr(scratch, Interpreter::local_tag_offset_in_bytes(n), base);
    if (t == frame::TagCategory2) {
      beq(scratch, R0, long_ok);
      delayed()->ld_ptr(scratch, Interpreter::local_tag_offset_in_bytes(n+1), base);
      stop("local long/double tag value bad");
      bind(long_ok);
      beq(scratch, R0, ok);
    } else if (t == frame::TagValue) {
      beq(scratch, R0, ok);
    } else {
      li(AT, t);
      beq(AT, scratch, ok);
    }
    delayed()->nop();
    // Also compare if the local value is zero, then the tag might
    // not have been set coming from deopt.
    ld_ptr(scratch, Interpreter::local_offset_in_bytes(n), base);
    beq(scratch, R0, ok);
    delayed()->nop();
    stop("Local tag value is bad");
    bind(ok);
  }
}
#endif // ASSERT

void InterpreterMacroAssembler::access_local_ptr(Register dst, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagReference, index, dst));
  ld_ptr(dst, Interpreter::value_offset_in_bytes(), index);
}

// Just like access_local_ptr but the tag is a returnAddress
void InterpreterMacroAssembler::access_local_returnAddress(Register dst,
							   Register index)
{
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagValue, index, dst));
  ld_ptr(dst, Interpreter::value_offset_in_bytes(), index);
}

void InterpreterMacroAssembler::access_local_int(Register dst, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagValue, index, dst));
  lw(dst, Interpreter::value_offset_in_bytes(), index);
  // Note:  index must hold the effective address--the iinc template uses it
}


void InterpreterMacroAssembler::access_local_long(Register dst,
						  Register index)
{
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagCategory2, index, dst));
  load_unaligned_long(dst, Interpreter::local_offset_in_bytes(1), index); // reads from low to high
}


void InterpreterMacroAssembler::access_local_float(FloatRegister dst,
						   Register index)
{
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagValue, index, Rscratch1));
  lwc1(dst, Interpreter::value_offset_in_bytes(), index);
}


void InterpreterMacroAssembler::access_local_double(FloatRegister dst,
						    Register index)
{
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  debug_only(verify_local_tag(frame::TagCategory2, index, Rscratch1));
  load_unaligned_double(dst, Interpreter::local_offset_in_bytes(1), index); // reads from low to high
}


#ifdef ASSERT
void InterpreterMacroAssembler::check_for_regarea_stomp(Register index,
							int offset,
							Register limit,
							Register scratch,
							Register scratch1)
{
  Label L;

  assert(index != scratch, "Registers cannot be same");
  assert(index != scratch1, "Registers cannot be same");
  assert(limit != scratch, "Registers cannot be same");
  assert(limit != scratch1, "Registers cannot be same");
  assert(scratch1 != scratch, "Registers cannot be same");

  // untested("reg area corruption");
  Unimplemented();
  delayed()->nop();
  stop("regsave area is being clobbered");
  bind(L);
}
#endif // ASSERT

void InterpreterMacroAssembler::tag_local(frame::Tag t, 
                                          Register base,
                                          Register tmp,
                                          int n) {
  if (TaggedStackInterpreter) {
    // have to store zero because local slots can be reused (rats!)
    if (t == frame::TagValue) {
      st_ptr(R0, Interpreter::local_tag_offset_in_bytes(n), base);
    } else if (t == frame::TagCategory2) {
      st_ptr(R0, Interpreter::local_tag_offset_in_bytes(n), base);
      st_ptr(R0, Interpreter::local_tag_offset_in_bytes(n+1), base);
    } else {
      // assert that we don't stomp the value in 'src'
      assert_different_registers(base, tmp);
      li(tmp, t);
      st_ptr(tmp, Interpreter::local_tag_offset_in_bytes(n), base);
    }
  }
}


void InterpreterMacroAssembler::store_local_int(Register src, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  // YYY fix this
  // debug_only(check_for_regarea_stomp(index, Interpreter::value_offset_in_bytes(), ...);)
  assert_different_registers(src, Rscratch3);
  tag_local(frame::TagValue, index, Rscratch3);
  sw(src, Interpreter::value_offset_in_bytes(), index);
}

void InterpreterMacroAssembler::store_local_ptr(Register src,
						Register tag,
						Register index)
{
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  // YYY fix this
  // debug_only(check_for_regarea_stomp(index, Interpreter::value_offset_in_bytes(), ...);)
  st_ptr(src, Interpreter::value_offset_in_bytes(), index);
  // Store tag register directly
  if (TaggedStackInterpreter) {
    st_ptr(tag, Interpreter::tag_offset_in_bytes(), index);
  }
}



void InterpreterMacroAssembler::store_local_ptr(Register src,
						Register tag,
						int n)
{
  st_ptr(src, Interpreter::local_offset_in_bytes(n), Rlocals);
  if (TaggedStackInterpreter) {
    st_ptr(tag, Interpreter::local_tag_offset_in_bytes(n), Rlocals);
  }
}

void InterpreterMacroAssembler::store_local_long(Register src, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  // YYY fix this
  // debug_only(check_for_regarea_stomp(index, Interpreter::value_offset_in_bytes(), ...);)
  assert_different_registers(src, Rscratch3);
  tag_local(frame::TagCategory2, index, Rscratch3);
  store_unaligned_long(src, Interpreter::local_offset_in_bytes(1), index); // reads from low to high
}


void InterpreterMacroAssembler::store_local_float(FloatRegister src, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  // YYY fix this
  // debug_only(check_for_regarea_stomp(index, Interpreter::value_offset_in_bytes(), ...);)
  assert_different_registers(src, Rscratch3);
  tag_local(frame::TagValue, index, Rscratch3);
  swc1(src, Interpreter::value_offset_in_bytes(), index);
}


void InterpreterMacroAssembler::store_local_double(FloatRegister src, Register index) {
  assert_not_delayed();
  sll_long(index, index, Interpreter::logStackElementSize());
  subu_long(index, Rlocals, index);
  // YYY fix this
  // debug_only(check_for_regarea_stomp(index, Interpreter::value_offset_in_bytes(), ...);)
  assert_different_registers(src, Rscratch3);
  tag_local(frame::TagCategory2, index, Rscratch3);
  store_unaligned_double(src, Interpreter::local_offset_in_bytes(1), index); // reads from low to high
}


// Set pointer to the topmost monitor slot, i.e. the oldest one.
// If it ends up that Rtopmonitor < Rmonitors, then there are no
// monitors and the pointer should not be used.
void InterpreterMacroAssembler::get_top_most_monitor(Register Rtopmonitor) {
  lhu(Rtopmonitor, in_bytes(methodOopDesc::size_of_locals_offset()), Rmethod);
  sll_long(Rtopmonitor, Rtopmonitor, Interpreter::logStackElementSize());
  subu_long(Rtopmonitor, Rlocals, Rtopmonitor);
  const jint delta = frame::interpreter_frame_monitor_size() * wordSize;
  if (delta != wordSize) {
    addu_long(Rtopmonitor, Rtopmonitor, - (delta - wordSize));
  }
}





void InterpreterMacroAssembler::increment_invocation_counter( Register Rtmp, Register Rtmp2 ) {
  assert(UseCompiler, "incrementing must be useful");
  int invocation_counter_method_offset =
    in_bytes(methodOopDesc::invocation_counter_offset() +
	     InvocationCounter::counter_offset());
  int backedge_counter_method_offset =
    in_bytes(methodOopDesc::backedge_counter_offset() +
	     InvocationCounter::counter_offset());
  int delta = InvocationCounter::count_increment;

  // Load each counter in a register
  lw(Rtmp, invocation_counter_method_offset, Rmethod);
  lw(Rtmp2, backedge_counter_method_offset, Rmethod);

  // Add the delta to the invocation counter and store the result
  addiu(Rtmp, Rtmp, delta);

  // Mask the backedge counter
  andi(Rtmp2, Rtmp2, InvocationCounter::count_mask_value);

  // Store value
  sw(Rtmp, invocation_counter_method_offset, Rmethod);

  // Add invocation counter + backedge counter
  addu_long(Rtmp, Rtmp, Rtmp2);

  // Note that this macro must leave the backedge_count + invocation_count in Rtmp!
}

void InterpreterMacroAssembler::increment_backedge_counter( Register Rtmp, Register Rtmp2 ) {
  assert(UseCompiler, "incrementing must be useful");
  int backedge_counter_method_offset =
    in_bytes(methodOopDesc::backedge_counter_offset() +
	     InvocationCounter::counter_offset());
  int invocation_counter_method_offset =
    in_bytes(methodOopDesc::invocation_counter_offset() +
	     InvocationCounter::counter_offset());
  int delta = InvocationCounter::count_increment;
  // Load each counter in a register
  lw(Rtmp, backedge_counter_method_offset, Rmethod);
  lw(Rtmp2, invocation_counter_method_offset, Rmethod);

  // Add the delta to the backedge counter
  addiu(Rtmp, Rtmp, delta);

  // Mask the invocation counter, add to backedge counter
  andi(Rtmp2, Rtmp2, InvocationCounter::count_mask_value);

  // and store the result to memory
  sw(Rtmp, backedge_counter_method_offset, Rmethod);

  // Add backedge + invocation counter
  addu_long(Rtmp, Rtmp, Rtmp2);

  // Note that this macro must leave backedge_count + invocation_count in Rtmp!
}

void InterpreterMacroAssembler::test_backedge_count_for_osr( Register backedge_count,
                                                             Register branch_bcp,
                                                             Register Rtmp ) {
  Label did_not_overflow;
  Label overflow_with_error;
  assert_different_registers(backedge_count, Rtmp, branch_bcp);
  assert(UseOnStackReplacement,"Must UseOnStackReplacement to test_backedge_count_for_osr");

  Unimplemented();

  // When ProfileInterpreter is on, the backedge_count comes from the 
  // methodDataOop, which value does not get reset on the call to 
  // frequency_counter_overflow().  To avoid excessive calls to the overflow
  // routine while the method is being compiled, add a second test to make sure 
  // the overflow function is called only once every overflow_frequency.
  if (ProfileInterpreter) {
    const int overflow_frequency = 1024;
    Unimplemented();
    bne(Rtmp, R0, did_not_overflow);
    delayed()->nop();
  }

  // overflow in loop, pass branch bytecode
  Unimplemented();
  call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::frequency_counter_overflow), branch_bcp, Rtmp);

  // Was an OSR adapter generated?
  // A0 = osr nmethod
  // [YYY?]
  beq(A0, R0, overflow_with_error);
  delayed()->nop();

  // Has the nmethod been invalidated already?
  ld(Rtmp, nmethod::entry_bci_offset(), A0);
  addiu(Rtmp, Rtmp, -InvalidOSREntryBci); // "subiu"
  beq(Rtmp, R0, overflow_with_error); // ==
  delayed()->nop();

  // migrate the interpreter frame off of the stack

  // save nmethod
  Unimplemented();
  call_VM_leaf(noreg, CAST_FROM_FN_PTR(address, SharedRuntime::OSR_migration_begin), Rthread);
  reset_last_Java_frame();

  Unimplemented();

  // remove the interpreter frame
  Unimplemented();

  // Jump to the osr code.
  Unimplemented();

  bind(overflow_with_error);

  bind(did_not_overflow);
}



void InterpreterMacroAssembler::interp_verify_oop(Register reg, TosState state, const char * file, int line) {
  if (state == atos) { MacroAssembler::_verify_oop(reg, "broken oop ", file, line); }
}


// local helper function for the verify_oop_or_return_address macro
static bool verify_return_address(methodOopDesc* m, int bci) {
#ifndef PRODUCT
  address pc = (address)(m->constMethod())
    + in_bytes(constMethodOopDesc::codes_offset()) + bci;
  // assume it is a valid return address if it is inside m and is preceded by a jsr
  if (!m->contains(pc))                                          return false;
  address jsr_pc;
  jsr_pc = pc - Bytecodes::length_for(Bytecodes::_jsr);
  if (*jsr_pc == Bytecodes::_jsr   && jsr_pc >= m->code_base())    return true;
  jsr_pc = pc - Bytecodes::length_for(Bytecodes::_jsr_w);
  if (*jsr_pc == Bytecodes::_jsr_w && jsr_pc >= m->code_base())    return true;
#endif // PRODUCT
  return false;
}


// for astore
void InterpreterMacroAssembler::verify_oop_or_return_address(Register reg, Register Rtmp) {
  if (!VerifyOops)  return;
  // the VM documentation for the astore[_wide] bytecode allows
  // the TOS to be not only an oop but also a return address
  Label test;
  Label skip;
  // See if it is an address (in the current method):

  move(Rtmp, reg);
  const int log2_bytecode_size_limit = 16;
  srl_long(Rtmp, Rtmp, log2_bytecode_size_limit);
  bne(Rtmp, R0, test);
  delayed()->nop();

  // %%% should use call_VM_leaf here?
  move(A0, Rmethod);
  move(A1, reg);
  save_frame(0);
  jal_to_C(CAST_FROM_FN_PTR(address,verify_return_address), relocInfo::none);
  delayed()->nop();
  restore_frame(0);
  bne(V0, R0, skip);
  delayed()->nop();

  // Perform a more elaborate out-of-line call
  // Not an address; verify it:
  bind(test);
  verify_oop(reg);
  bind(skip);
}


void InterpreterMacroAssembler::verify_FPU(int stack_depth, TosState state) {
  // only if +VerifyFPU  && (state == ftos || state == dtos)
  // For now, do nothing.
}


// Inline assembly for:
//
// if (thread is in interp_only_mode) {
//   InterpreterRuntime::post_method_entry();
// }
// if (DTraceMethodProbes) {
//   SharedRuntime::dtrace_method_entry(method, reciever);
// }

void InterpreterMacroAssembler::notify_method_entry() {
  // Whenever JVMTI puts a thread in interp_only_mode, method
  // entry/exit events are sent for that thread to track stack
  // depth.  If it is possible to enter interp_only_mode we add
  // the code to check if the event should be sent.
  Register tempreg = Rscratch0;
  if (JvmtiExport::can_post_interpreter_events()) {
    Label L;
    lw(tempreg, in_bytes(JavaThread::interp_only_mode_offset()), Rthread);
    beq(tempreg, R0, L);
    delayed()->nop();
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::post_method_entry));
    bind(L);
  }

  {
    SkipIfEqual skip_if(this, tempreg, R0, &DTraceMethodProbes, equal);
    call_VM_leaf(noreg,
		 CAST_FROM_FN_PTR(address, SharedRuntime::dtrace_method_entry),
		 Rthread,
		 Rmethod);
  }
}

// Inline assembly for:
//
// if (thread is in interp_only_mode) {
//   // save result
//   InterpreterRuntime::post_method_exit();
//   // restore result
// }
// if (DTraceMethodProbes) {
//   SharedRuntime::dtrace_method_exit(thread, method);
// }
// 
// Native methods have their result stored in d_tmp and l_tmp
// Java methods have their result stored in the expression stack

void InterpreterMacroAssembler::notify_method_exit(
    bool is_native_method, TosState state, NotifyMethodExitMode mode) {
  // Whenever JVMTI puts a thread in interp_only_mode, method
  // entry/exit events are sent for that thread to track stack
  // depth.  If it is possible to enter interp_only_mode we add
  // the code to check if the event should be sent.
  Register tempreg = Rscratch0;
  if (mode == NotifyJVMTI && JvmtiExport::can_post_interpreter_events()) {
    Label skip;
    lw(tempreg, in_bytes(JavaThread::interp_only_mode_offset()), Rthread);
    beq(tempreg, R0, skip);
    delayed()->nop();

    // Note: frame::interpreter_frame_result has a dependency on how the 
    // method result is saved across the call to post_method_exit. For
    // native methods it assumes the result registers are saved to
    // l_scratch and d_scratch. If this changes then the interpreter_frame_result
    // implementation will need to be updated too.
    save_return_value(state, is_native_method);
    call_VM(noreg,
	    CAST_FROM_FN_PTR(address, InterpreterRuntime::post_method_exit));
    restore_return_value(state, is_native_method);
    bind(skip);
  }

  {
    // Dtrace notification
    SkipIfEqual skip_if(this, tempreg, R0, &DTraceMethodProbes, equal);
    save_return_value(state, is_native_method);
    call_VM_leaf(
		 noreg,
		 CAST_FROM_FN_PTR(address, SharedRuntime::dtrace_method_exit),
		 Rthread, Rmethod);
    restore_return_value(state, is_native_method);
  }
}

void InterpreterMacroAssembler::save_return_value(
    TosState state, bool is_native_call) {
  if (is_native_call) {
    // save any potential method result value
#ifdef _LP64
    sd(V0, frame::interpreter_frame_l_scratch_offset * wordSize, FP);
    sdc1(F0, frame::interpreter_frame_d_scratch_offset * wordSize, FP);
#else
#error must save V0 and V1
#endif
  } else {
    push(state);
  }
}

void InterpreterMacroAssembler::restore_return_value(
    TosState state, bool is_native_call) {
  if (is_native_call) {
    // Restore any method result value
#ifdef _LP64
    ld(V0, frame::interpreter_frame_l_scratch_offset * wordSize, FP);
    ldc1(F0, frame::interpreter_frame_d_scratch_offset * wordSize, FP);
#else
#error must restore V0 and V1
#endif
  } else {
    pop(state);
  }
}
