#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)interp_masm_sparc.hpp	1.104 07/05/17 15:48:09 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// This file specializes the assembler with interpreter-specific macros

REGISTER_DECLARATION(Register, Rtos,  V0); // top of stack for ints, etc.
REGISTER_DECLARATION(Register, Rtos2, V1); // 2nd part of longs or tag
REGISTER_DECLARATION(FloatRegister, Ftos, F0); // tos for floats

#ifndef DONT_USE_REGISTER_DEFINES
#ifndef _LP64
REGISTER_DECLARATION(Register, Rtos_l1, V0);
REGISTER_DECLARATION(Register, Rtos_l2, V1);
#endif // _LP64
#endif // DONT_USE_REGISTER_DEFINES

class InterpreterMacroAssembler: public MacroAssembler {
 protected:
  // Interpreter specific version of call_VM_base
  virtual void call_VM_leaf_base(
    Register java_thread,
    address  entry_point,
    int      number_of_arguments
  );

  virtual void call_VM_base(
    Register        oop_result,
    Register        java_thread,
    Register        last_java_sp,
    address         entry_point,
    int             number_of_arguments,
    bool	    check_exception=true
  );

  virtual void check_and_handle_popframe(Register java_thread);
  virtual void check_and_handle_earlyret(Register java_thread);

  // base routine for all dispatches
  void dispatch_base(TosState state, address* table);

 public:
  InterpreterMacroAssembler(CodeBuffer* c)
    : MacroAssembler(c) {}

  virtual void load_earlyret_value(TosState state);

  static const Address l_tmp ;
  static const Address d_tmp ;

  // Save dynamic interpreter state from registers to the stack.
  // Note that the primary purpose is not to preserve the registers
  // (e.g. across calls), since they are already callee-save registers.
  // The point is to allow frame:: to be able to access them and get
  // the correct current values.
  void save_dynamic_state() {
    st_ptr(Rbcp, frame::interpreter_frame_Rbcp_offset * BytesPerWord, FP);
    st_ptr(Resp, frame::interpreter_frame_Resp_offset * BytesPerWord, FP);
  }

  void restore_dynamic_state() {
    ld_ptr(Rbcp, frame::interpreter_frame_Rbcp_offset * BytesPerWord, FP);
    // This one can't change, so no point in restoring it.
    //ld_ptr(Resp, frame::interpreter_frame_Resp_offset * BytesPerWord, FP);
  }

  // helper routine for frame allocation/deallocation
  // compute the delta by which the caller's SP has to
  // be adjusted to accomodate for the non-argument
  // locals
  void compute_extra_locals_size_in_bytes(Register args_size, Register locals_size, Register delta);

  // dispatch routines
  void dispatch_prolog(TosState state, int step = 0);
  void dispatch_epilog(TosState state, int step = 0);
  void dispatch_only(TosState state);
  void dispatch_normal(TosState state);
  void dispatch_next(TosState state, int step = 0);
  void dispatch_next_noverify_oop(TosState state, int step = 0);
  void dispatch_via (TosState state, address* table);


  // Removes the current activation (incl. unlocking of monitors).
  // Additionally this code is used for earlyReturn in which case we
  // want to skip throwing an exception and installing an exception.
  void remove_activation(TosState state, 
                         bool throw_monitor_exception = true, 
                         bool install_monitor_exception = true);

 protected:
  void dispatch_Rbyte_code(TosState state, address* table, int bcp_incr = 0, bool verify = true);

 public:
  // Super call_VM calls - correspond to MacroAssembler::call_VM(_leaf) calls
  void super_call_VM_leaf(Register thread_cache, address entry_point, Register arg_1);

  // Generate a subtype check: branch to ok_is_subtype if sub_klass is
  // a subtype of super_klass.  Blows registers tmp1, tmp2 and tmp3.
  void gen_subtype_check( Register sub_klass, Register super_klass, Register tmp1, Register tmp2, Register tmp3, Label &ok_is_subtype );

  void jump_if_not(Condition cond,
		   Register r1,
		   Register r2,
		   address entry_point);

  void jump_if(Condition cond,
	       Register r1,
	       Register r2,
	       address entry_point);

  // helpers for expression stack

  void pop_i(     Register r = Rtos);
  void pop_ptr(   Register r = Rtos, Register scratch = Rscratch0);
  void pop_l(     Register r = Rtos);
  void pop_f(FloatRegister f = Ftos, Register scratch = Rscratch0);
  void pop_d(FloatRegister f = Ftos, Register scratch = Rscratch0);
  
  void push_i(     Register r = Rtos);
  void push_ptr(   Register r = Rtos);
  void push_ptr(   Register r, Register tag);
  void push_l(     Register r = Rtos);
  void push_f(FloatRegister f = Ftos);
  void push_d(FloatRegister f = Ftos);


  void pop (TosState state);           // transition vtos -> state
  void push(TosState state);           // transition state -> vtos
  void empty_expression_stack();       // resets both Resp and SP

  // Support for Tagged Stacks
  void tag_stack(frame::Tag t, Register r);
  void tag_stack(Register tag);
  void tag_local(frame::Tag t, Register base, Register tmp, int n = 0);

#ifdef ASSERT
  void verify_sp(Register Rsp, Register Rtemp);
  void verify_esp(Register Resp);      // verify that Resp points to a word in the temp stack

  void verify_stack_tag(frame::Tag t,
			Register r,
			Register scratch = Rscratch0);
  void verify_local_tag(frame::Tag t, Register base, Register scr, int n = 0);
#endif // ASSERT

 public:
  // Load values from bytecode stream:

  enum signedOrNot { Signed, Unsigned };
  enum setCCOrNot  { set_CC,  dont_set_CC };

  void get_2_byte_integer_at_base(Register   Rbase,
				  int        offset,
                                  Register   Rtmp,
				  Register   Rd,
                                  bool       is_signed);

  void get_4_byte_integer_at_base(Register   Rbase,
				  int        offset,
                                  Register   Rtmp,
				  Register   Rd,
                                  bool       is_signed);

  void get_2_byte_integer_at_bcp( int        bcp_offset, 
                                  Register   Rtmp, 
				  Register   Rd,
                                  bool       is_signed);

  void get_4_byte_integer_at_bcp( int        bcp_offset,
                                  Register   Rtmp, 
				  Register   Rd);

  void get_2_byte_integer(Register    Rbase,
			  Register    Roffset,
			  Register    Rtmp,
			  Register    Rtmp2,
			  Register    Rd,
			  bool        is_signed);

  void get_4_byte_integer(Register    Rbase,
			  Register    Roffset,
			  Register    Rtmp,
			  Register    Rtmp2,
			  Register    Rd,
			  bool        is_signed);

  void get_cache_and_index_at_bcp(Register cache, Register tmp, int bcp_offset);
  void get_cache_entry_pointer_at_bcp(Register cache, Register tmp, int bcp_offset);


  // common code

  void field_offset_at(int n, Register tmp, Register dest, Register base);
  int  field_offset_at(Register object, address bcp, int offset);
  void fast_iaaccess(int n, address bcp);
  void fast_iagetfield(address bcp);
  void fast_iaputfield(address bcp, bool do_store_check );
  
  void index_check(Register array, Register index, int index_shift, Register tmp, Register res);
  void index_check_without_pop(Register array, Register index, int index_shift, Register tmp, Register res);

  void get_constant_pool(Register Rdst);
  void get_constant_pool_cache(Register Rdst);
  void get_cpool_and_tags(Register Rcpool, Register Rtags);
  void is_a(Label& L);

  // Load compiled (i2c) or interpreter entry and call from interpreted
  void call_from_interpreter(Register target, Register scratch, Register Rret);

  // --------------------------------------------------

  void unlock_if_synchronized_method(TosState state, bool throw_monitor_exception = true, bool install_monitor_exception = true);

  void add_monitor_to_stack( Register Rtemp,
			     Register Rtemp2 );

  // Load/store aligned in _LP64 but unaligned otherwise
  // These only apply to the Interpreter expression stack and locals!
  void  load_unaligned_double(FloatRegister ft, int offset, Register base);
  void store_unaligned_double(FloatRegister ft, int offset, Register base);

  // Load/store aligned in _LP64 but unaligned otherwise
  void  load_unaligned_long(Register rt, int offset, Register base);
  void store_unaligned_long(Register rt, int offset, Register base);

  void access_local_int(Register dst, Register index);
  void access_local_ptr(Register dst, Register index);
  void access_local_returnAddress(Register dst, Register index);
  void access_local_long(Register dst, Register index);
  void access_local_float(FloatRegister dst, Register index);
  void access_local_double(FloatRegister dst, Register index);
#ifdef ASSERT
  void check_for_regarea_stomp( Register index, int offset, Register limit, Register scratch, Register scratch1);
#endif // ASSERT
  void store_local_int(Register src, Register index);
  void store_local_ptr(Register src, Register tag, Register index);
  void store_local_ptr(Register src, Register tag, int n);
  void store_local_long(Register src, Register index);
  void store_local_float(FloatRegister src, Register index);
  void store_local_double(FloatRegister src, Register index);

  // Tagged stack helpers for swap and dup
  void load_ptr_and_tag(int n, Register val, Register tag);
  void store_ptr_and_tag(int n, Register val, Register tag);

  // Tagged stack helper for getting receiver in register.
  void load_receiver(Register param_count, Register recv);

  void get_top_most_monitor(Register Rtopmonitor);

  void increment_invocation_counter( Register Rtmp, Register Rtmp2 );
  void increment_backedge_counter( Register Rtmp, Register Rtmp2 );
  void test_backedge_count_for_osr( Register backedge_count, Register branch_bcp, Register Rtmp );

  void record_static_call_in_profile( Register Rentry, Register Rtmp );
  void record_receiver_call_in_profile( Register Rklass, Register Rentry, Register Rtmp );

  // Object locking
  void lock_object  (Register lock_reg, Register obj_reg);
  void unlock_object(Register lock_reg);

  // Interpreter profiling operations
  void set_method_data_pointer() { set_method_data_pointer_offset(noreg); }
  void set_method_data_pointer_for_bcp();
  void set_method_data_pointer_offset(Register mdi_reg);
  void test_method_data_pointer(Label& zero_continue);
  void verify_method_data_pointer();
  void test_invocation_counter_for_mdp(Register invocation_count, Register cur_bcp, Register Rtmp, Label &profile_continue);

  void set_mdp_data_at(int constant, Register value);
  void increment_mdp_data_at(Address counter, Register bumped_count,
                             bool decrement = false);
  void increment_mdp_data_at(int constant, Register bumped_count,
                             bool decrement = false);
  void increment_mdp_data_at(Register reg, int constant,
			     Register bumped_count, Register scratch2,
                             bool decrement = false);
  void set_mdp_flag_at(int flag_constant, Register scratch);
  void test_mdp_data_at(int offset, Register value, Label& not_equal_continue,
			Register scratch);

  void record_klass_in_profile(Register receiver, Register scratch);
  void record_klass_in_profile_helper(Register receiver, Register scratch,
                                      int start_row, Label& done);

  void update_mdp_by_offset(int offset_of_disp, Register scratch);
  void update_mdp_by_offset(Register reg, int offset_of_disp,
			    Register scratch);
  void update_mdp_by_constant(int constant);
  void update_mdp_for_ret(TosState state, Register return_bci);

  void profile_taken_branch(Register scratch, Register bumped_count);
  void profile_not_taken_branch(Register scratch);
  void profile_call(Register scratch);
  void profile_final_call(Register scratch);
  void profile_virtual_call(Register receiver, Register scratch);
  void profile_ret(TosState state, Register return_bci, Register scratch);
  void profile_null_seen(Register scratch);
  void profile_typecheck(Register klass, Register scratch);
  void profile_typecheck_failed(Register scratch);
  void profile_switch_default(Register scratch);
  void profile_switch_case(Register index,
			   Register scratch1,
			   Register scratch2,
			   Register scratch3);

  // Debugging
  void interp_verify_oop(Register reg, TosState state, const char * file, int line);    // only if +VerifyOops && state == atos
  void verify_oop_or_return_address(Register reg, Register rtmp); // for astore
  void verify_FPU(int stack_depth, TosState state = ftos); // only if +VerifyFPU  && (state == ftos || state == dtos)

  // support for JVMTI/Dtrace
  typedef enum { NotifyJVMTI, SkipNotifyJVMTI } NotifyMethodExitMode;
  void notify_method_entry();
  void notify_method_exit(
    bool save_result, TosState state, NotifyMethodExitMode mode);

  void save_return_value(TosState state, bool is_native_call);
  void restore_return_value(TosState state, bool is_native_call);
};
