#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)interpreterRT_sparc.cpp	1.72 07/05/05 17:04:29 JVM"
#endif
/*
 * Copyright 1998-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_interpreterRT_mips.cpp.incl"
#include <sys/cachectl.h>


#define __ _masm->


// Implementation of SignatureHandlerGenerator

void InterpreterRuntime::SignatureHandlerGenerator::pass_int()
{
  Argument jni_arg(jni_offset());
  Register Rtmp = V0;
  __ lw(Rtmp, Interpreter::local_offset_in_bytes(offset()), T2);
  __ store_int_argument(Rtmp, jni_arg);
  if (jni_arg.is_register()) {
    __ addu_long(T1, T1, 1);
  }
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_long()
{
  Argument jni_arg(jni_offset());
  Register Rtmp = V0;

#ifdef ASSERT
  if (TaggedStackInterpreter) {
    // check at least one tag is okay
    Label ok;
    __ ld_ptr(Rtmp, Interpreter::local_tag_offset_in_bytes(offset() + 1), T2);
    __ beq(Rtmp, R0, ok);
    __ delayed()->nop();
    __ stop("Native object has bad tag value");
    __ bind(ok);
  }
#endif // ASSERT

#ifdef _LP64
  __ ld(Rtmp, Interpreter::local_offset_in_bytes(offset() + 1), T2);
  __ store_long_argument(Rtmp, jni_arg);
  if (jni_arg.is_register()) {
    __ addu_long(T1, T1, 1);
  }
#else
#error unimplemented
#endif
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_float()
{
  Argument jni_arg(jni_offset());
  FloatRegister Ftmp = F0;
  __ lwc1(Ftmp, Interpreter::local_offset_in_bytes(offset()), T2);
  __ store_float_argument(Ftmp, jni_arg);
  if (jni_arg.is_register()) {
    __ addu_long(T1, T1, 1);
  }
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_double()
{
  Argument jni_arg(jni_offset());
  FloatRegister Ftmp = F0;
#ifdef _LP64
  __ ldc1(Ftmp, Interpreter::local_offset_in_bytes(offset() + 1), T2);
#else
#error unimplemented
#endif
  __ store_double_argument(Ftmp, jni_arg);
  if (jni_arg.is_register()) {
    __ addu_long(T1, T1, 1);
  }
}

void InterpreterRuntime::SignatureHandlerGenerator::pass_object()
{
  Argument jni_arg(jni_offset());

  Register Rtmp1 = V0;
  Register Rtmp2 = jni_arg.is_register() ?  jni_arg.as_register() : Rtmp1;

  // the handle for a receiver will never be null
  bool do_NULL_check = offset() != 0 || is_static();

  __ ld_ptr(Rtmp1, Interpreter::local_offset_in_bytes(offset()), T2);

#ifdef ASSERT
  Register Rtmp3 = Rscratch3;
  if (TaggedStackInterpreter) {
    // check we have the obj and not the tag
    Label ok;
    __ li(Rtmp3, frame::TagReference);
    __ bne(Rtmp1, Rtmp3, ok);
    __ delayed()->nop();
    __ stop("Native object passed tag by mistake");
    __ bind(ok);
  }
#endif // ASSERT

  if (!do_NULL_check) {
    __ addu_long(Rtmp2, T2, Interpreter::local_offset_in_bytes(offset()));
  } else {
    if (Rtmp1 != Rtmp2) {
      __ move(Rtmp2, Rtmp1);
    }
    Label is_null;
    __ beq(Rtmp2, R0, is_null);
    __ delayed()->nop();
    __ addu_long(Rtmp2, T2, Interpreter::local_offset_in_bytes(offset()));
    __ bind(is_null);
  }
  __ store_ptr_argument(Rtmp2, jni_arg);
  if (jni_arg.is_register()) {
    __ addu_long(T1, T1, 1);
  }
}

void InterpreterRuntime::SignatureHandlerGenerator::generate(
  uint64_t fingerprint) {
  address entry = __ pc();

  __ lhu(T2, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod);
  __ sll_long(T2, T2, Interpreter::logStackElementSize());
  __ ld_ptr(T0, frame::interpreter_frame_l_scratch_offset * wordSize, FP);  // Rargs cache
  __ addu_long(T2, T0, T2);
  // save the effective value of Rlocals for stack walkers
  __ st_ptr(T2, frame::interpreter_frame_Rlocals_offset * BytesPerWord, FP);

  // We count the number of register parameters so we can remove them
  // later.  We start at 1 to account for the JNIEnv, plus an
  // additional 1 if it is static for the mirror arg.
  if (is_static()) {
    __ li(T1, 2);
  } else {
    __ li(T1, 1);
  }

  // generate code to handle arguments
  iterate(fingerprint);

  // Pop stack space for args that were handled in registers.
  // The rounding is subtle: if it is n_register_parameters, then
  // it's already rounded and it's harmless.  If not, then we want
  // to remove all of the space we added anyway, including any
  // padding, so it's ok.
  __ sll_long(T1, T1, LogBytesPerWord);
  __ round_to(T1, 2 * BytesPerWord);
  __ addu_long(SP, SP, T1);

#ifdef ASSERT
  { // We stashed the original SP before arguments, so we can check
    // that we haven't exceeded it now.
    Label ok;
    __ ld_ptr(T0, frame::interpreter_frame_d_scratch_offset * wordSize, FP);
    __ bleu(SP, T0, ok);
    __ delayed()->nop();
    __ stop("Popped too much stack in signature handler");
    __ bind(ok);
  }
#endif

  // We take the amount that we saved originally, subtract the amount we
  // just eliminated (T1), and the result is how much we have left for
  // arguments.  We write that back.
  __ ld_long(T2, frame::interpreter_frame_args_size_offset * wordSize, FP);
  __ subu_long(T2, T2, T1);
  __ st_long(T2, frame::interpreter_frame_args_size_offset * wordSize, FP);

#ifdef ASSERT
  { // Check that "args SP" + args_size = "orig SP"
    Label ok;
    __ addu_long(T2, SP, T2);
    __ beq(T2, T0, ok);         // note T0 was loaded in previous assert
    __ delayed()->nop();
    __ stop("args_size didn't agree with origSP");
    __ bind(ok);
  }
#endif

  // return result handler in Rsave0
  __ li(Rsave0, AbstractInterpreter::result_handler(method()->result_type()));
  __ j(RA);
  __ delayed()->nop();

  cacheflush(entry, __ pc() - entry, DCACHE);
  cacheflush(entry, __ pc() - entry, ICACHE);
}


// Implementation of SignatureHandlerLibrary

void SignatureHandlerLibrary::pd_set_handler(address handler) {}


class SlowSignatureHandler: public NativeSignatureIterator {
 private:
  address   _from;
  intptr_t* _to;
  intptr_t* _RegArgSignature;			// Signature of first Arguments to be passed in Registers
  uint      _argcount;

  enum {					// We need to differenciate float from non floats in reg args
    int_sig    = 0,
    float_sig  = 1,
    double_sig = 2,
    long_sig   = 3
  };

#ifdef ASSERT
  void verify_tag(frame::Tag t) {
    assert(!TaggedStackInterpreter ||
           *(intptr_t*)(_from+Interpreter::local_tag_offset_in_bytes(0)) == t, "wrong tag");
  }
#endif // ASSERT

  virtual void pass_int()
  {
    *_to++ = *(jint *)(_from+Interpreter::local_offset_in_bytes(0)); 
    debug_only(verify_tag(frame::TagValue));
    _from -= Interpreter::stackElementSize();
    add_signature(int_sig);
  }

  virtual void pass_object() { 
    // pass address of from
    intptr_t *from_addr = (intptr_t*)(_from + Interpreter::local_offset_in_bytes(0));
    *_to++ = (*from_addr == 0) ? NULL : (intptr_t) from_addr;
    debug_only(verify_tag(frame::TagReference));
    _from -= Interpreter::stackElementSize();
#ifdef _LP64
    add_signature(long_sig);
#else
    add_signature(int_sig);
#endif
   }

  virtual void pass_float()  { 
    *_to++ = *(jint *)(_from+Interpreter::local_offset_in_bytes(0)); 
    debug_only(verify_tag(frame::TagValue));
    _from -= Interpreter::stackElementSize();
    add_signature(float_sig);
   }

  virtual void pass_double() { 
    *_to++ = *(intptr_t*)(_from+Interpreter::local_offset_in_bytes(1)); 
    debug_only(verify_tag(frame::TagValue));
    _from -= 2*Interpreter::stackElementSize();
   add_signature(double_sig);
   }

  virtual void pass_long() { 
    _to[0] = *(intptr_t*)(_from+Interpreter::local_offset_in_bytes(1));
    debug_only(verify_tag(frame::TagValue));
    _to += 1; 
    _from -= 2*Interpreter::stackElementSize();
    add_signature(long_sig);
  }

  virtual void add_signature( intptr_t sig_type ) { 
    if ( _argcount < (sizeof (intptr_t))*4 ) { 
      *_RegArgSignature |= (sig_type << (_argcount*2) );
      _argcount++;
    }
  }

 public:
  SlowSignatureHandler(methodHandle method, address from, intptr_t* to, intptr_t *RegArgSig) : NativeSignatureIterator(method) {
    _from = from;
    _to   = to;
    _RegArgSignature = RegArgSig;	
    *_RegArgSignature = 0;
    _argcount = method->is_static() ? 2 : 1;
  }
};


IRT_ENTRY(address, InterpreterRuntime::slow_signature_handler(
                                                    JavaThread* thread,
                                                    methodOopDesc* method,
                                                    intptr_t* from,
                                                    intptr_t* to ))
  methodHandle m(thread, method);
  assert(m->is_native(), "sanity check");
  // handle arguments
  // Warning: We use V0 temporarily to return the RegArgSignature
  // back to the code that pops the arguments into the CPU registers
  SlowSignatureHandler(m, (address)from, m->is_static() ? to+2 : to+1, to).iterate(UCONST64(-1));
  // return result handler
  return AbstractInterpreter::result_handler(m->result_type());
IRT_END
