#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)interpreter_sparc.cpp	1.262 07/08/07 09:07:51 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_interpreter_mips.cpp.incl"

#ifndef FAST_DISPATCH
#define FAST_DISPATCH 1
#endif
#undef FAST_DISPATCH


// Generation of Interpreter
//
// The InterpreterGenerator generates the interpreter into Interpreter::_code.


#define __ _masm->


//----------------------------------------------------------------------------------------------------


address AbstractInterpreterGenerator::generate_exception_handler_common(const char* name, const char* message, bool pass_oop) {
  assert(!pass_oop || message == NULL, "either oop or message but not both");
  address entry = __ pc();
  // expression stack must be empty before entering the VM if an exception happened
  __ empty_expression_stack();
  // load exception object
  __ li(Rscratch3, (intptr_t)name);
  if (pass_oop) {
    __ call_VM(Rexception, CAST_FROM_FN_PTR(address, InterpreterRuntime::create_klass_exception), Rscratch3, Rtos);
  } else {
    __ li(Rscratch4, (intptr_t)message);
    __ call_VM(Rexception, CAST_FROM_FN_PTR(address, InterpreterRuntime::create_exception), Rscratch3, Rscratch4);
  }
  // throw exception
  assert(Interpreter::throw_exception_entry() != NULL, "generate it first");
  __ li(Rscratch3, Interpreter::throw_exception_entry());
  __ jr(Rscratch3);
  __ delayed()->nop();
  return entry;
}

address AbstractInterpreterGenerator::generate_ClassCastException_handler() {
  address entry = __ pc();
  // expression stack must be empty before entering the VM if an exception
  // happened
  __ empty_expression_stack();
  // load exception object
  __ call_VM(Rexception, 
             CAST_FROM_FN_PTR(address, 
                              InterpreterRuntime::throw_ClassCastException), 
             Rtos);
  __ should_not_reach_here();
  return entry;
}


address AbstractInterpreterGenerator::generate_ArrayIndexOutOfBounds_handler(const char* name) {
  address entry = __ pc();
  // expression stack must be empty before entering the VM if an exception happened
  __ empty_expression_stack();
  // convention: expect aberrant index in register Rscratch3, then shuffle the 
  // index to Rscratch4 for the VM call
  // YYY This swapping is probably pointless here.  In fact, it seems
  // YYY pointless in the original code, which is why I'm leaving it.
  __ move(Rscratch4, Rscratch3);
  __ li(Rscratch3, (intptr_t)name);
  __ call_VM(Rexception, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_ArrayIndexOutOfBoundsException), Rscratch3, Rscratch4);
  __ should_not_reach_here();
  return entry;
}


address AbstractInterpreterGenerator::generate_StackOverflowError_handler() {
  address entry = __ pc();
  // expression stack must be empty before entering the VM if an exception happened
  __ empty_expression_stack();
  __ call_VM(Rexception, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_StackOverflowError));
  __ should_not_reach_here();
  return entry;
}


address AbstractInterpreterGenerator::generate_return_entry_for(TosState state, int step) {
  address compiled_entry = __ pc();
  Label cont;

  address entry = __ pc();
#if !defined(_LP64) && defined(COMPILER2) 
#error unimplemented
#endif /* !_LP64 && COMPILER2 */


  __ bind(cont);

  // SPARC has special handling for longs here, which can be ignored
  // on MIPS64.  But may have to revisit this.  YYY

  // Restore from Rlast_SP, in case we adjust SP in adapters.
  __ move(SP, Rlast_SP);

  // All interpreter local registers are untouched. Any result is passed back
  // in the V0 or F0 registers. Before continuing, the arguments must be
  // popped from the java expression stack; i.e., Resp must be adjusted. 

  const Register cache = Rscratch3;
  const Register size  = Rscratch1;
  __ get_cache_and_index_at_bcp(cache, Rscratch1, 1);
  __ ld_ptr(size,
	    in_bytes(constantPoolCacheOopDesc::base_offset()) +
	    in_bytes(ConstantPoolCacheEntry::flags_offset()),
	    cache);
  __ andi(size, size, 0xFF);                   // argument size in words
  __ sll_long(size, size, Interpreter::logStackElementSize());
  __ addu_long(Resp, Resp, size);              // pop arguments
  __ dispatch_next(state, step);

  return entry;
}
  

address AbstractInterpreterGenerator::generate_deopt_entry_for(TosState state, int step) {
  address entry = __ pc();
  __ get_constant_pool_cache(RcpoolCache); // load RcpoolCache
  { Label L;
    __ ld_ptr(Rscratch0, in_bytes(Thread::pending_exception_offset()), Rthread);
    __ bne(Rscratch0, R0, L);
    __ delayed()->nop();
    // The following line has been deleted from the Intel version:
    __ stop("new exception propagation across deoptimization has not been tested yet");
    __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_pending_exception));
    __ should_not_reach_here();
    __ bind(L);
  }
  __ dispatch_next(state, step);
  return entry;
}


int AbstractInterpreter::BasicType_as_index(BasicType type) {
  int i = 0;
  switch (type) {
    case T_BOOLEAN: i = 0; break;
    case T_CHAR   : i = 1; break;
    case T_BYTE   : i = 2; break;
    case T_SHORT  : i = 3; break;
    case T_INT    : i = 4; break;
    case T_LONG   : i = 5; break;
    case T_VOID   : i = 6; break;
    case T_FLOAT  : i = 7; break;
    case T_DOUBLE : i = 8; break;
    case T_OBJECT : i = 9; break;
    case T_ARRAY  : i = 9; break;
    default       : ShouldNotReachHere();
  }
  assert(0 <= i && i < AbstractInterpreter::number_of_result_handlers, "index out of bounds");
  return i;
}


// A result handler converts/unboxes a native call result into
// a java interpreter/compiler result. The current frame is an
// interpreter frame. The activation frame unwind code must be
// consistent with that of TemplateTable::_return(...). In the
// case of native methods, the caller's SP was not modified.
address AbstractInterpreterGenerator::generate_result_handler_for(BasicType type) {
  assert(Rtos == V0 && Ftos == F0, "or fix the code below");
  address entry = __ pc();
  switch (type) {
    case T_BOOLEAN:
      __ li(V1, 1);
      __ movn(V0, V1, V0);  // !0 -> 1, 0 -> 0
      break;
    case T_CHAR   :
      __ andi(V0, V0, 0xFFFF);
      break;
    case T_BYTE   :
      // This sign-extends properly to 64 bits on MIPS64
      __ sll(V0, V0, 24);
      __ sra(V0, V0, 24);
      break;
    case T_SHORT  :
      // This sign-extends properly to 64 bits on MIPS64
      __ sll(V0, V0, 16);
      __ sra(V0, V0, 16);
    case T_LONG   :
#ifndef _LP64
#error unimplemented		// move other half of long
#endif              // ifdef or no ifdef, fall through to the T_INT case
    case T_INT    : /* nothing to do */ break;
    case T_VOID   : /* nothing to do */ break;
    case T_FLOAT  : /* nothing to do */ break;
    case T_DOUBLE : /* nothing to do */ break;
    case T_OBJECT : 
      // load from temp storage
      __ ld_ptr(Rtos, frame::interpreter_frame_oop_temp_offset*wordSize, FP);
      __ verify_oop(Rtos);
      break;
    default       : ShouldNotReachHere();
  }
  // Remove interpreter frame added in generate_fixed_frame.
  // First, remove arguments from SP, to more easily restore registers.
  __ ld_long(Rscratch0, frame::interpreter_frame_args_size_offset * wordSize, FP);
  __ addu_long(SP, SP, Rscratch0);
  __ move(Rscratch0, FP);
  __ restore_registers(SP);
  __ jr(RA);                        // return from interpreter activation
  __ delayed()->move(SP, Rscratch0);
  NOT_PRODUCT(__ emit_long(0);)       // marker for disassembly
  return entry;
}

#ifndef _LP64
#error unimplemented
#else
address AbstractInterpreterGenerator::generate_slow_signature_handler() {

  enum {
    int_sig    = 0,
    float_sig  = 1,
    double_sig = 2,
    long_sig   = 3,
    sig_mask   = 3
  };

  address entry = __ pc();

  __ set_last_Java_frame(SP, FP, noreg);
  __ move(A3, SP);  // setting it early, but must survive to the call below
  __ save_minimal_frame();
  __ move(A1, Rmethod);
  __ lhu(A2, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod);
  __ sll_long(A2, A2, Interpreter::logStackElementSize());
  __ ld_ptr(Rscratch0, frame::interpreter_frame_l_scratch_offset * wordSize, FP);  // Rargs cache
  __ addu_long(A2, Rscratch0, A2);
  // save the effective value of Rlocals for stack walkers
  __ st_ptr(A2, frame::interpreter_frame_Rlocals_offset * BytesPerWord, FP);
  __ jal_to_C(CAST_FROM_FN_PTR(address, InterpreterRuntime::slow_signature_handler), relocInfo::runtime_call_type);
  __ delayed()->move(A0, Rthread);
  __ restore_minimal_frame();
  __ move(Rsave0, V0);
  __ reset_last_Java_frame();

  const Register Rlongcount = Rscratch2;

  __ ld_ptr(Rscratch3, 0, SP);
  __ move(Rscratch4, Rscratch3);
  __ srl_long(Rscratch4, Rscratch4, 2);  // skip arg 0
  __ li32(Rlongcount, 0);
  for (int arg_index = 1;  // we can skip the first one (0)
       arg_index < Argument::n_register_parameters;
       arg_index++) {
    Label NotInt;
    Label NotLong;
    Label NotFloat;
    Label NextArg;
    const int SP_offset = arg_index * BytesPerWord;
    __ andi(Rscratch3, Rscratch4, sig_mask);

    __ bne(Rscratch3, R0, NotInt);
    __ delayed()->nop();
    // it's an int
    __ b(NextArg);
    __ delayed()->lw(as_Register(arg_index), SP_offset, SP);

    __ bind(NotInt);
    __ li(Rscratch0, long_sig);
    __ bne(Rscratch3, Rscratch0, NotLong);
    __ delayed()->nop();
    // it's a long
    __ addu_long(Rlongcount, Rlongcount, 1);
    __ b(NextArg);
    __ delayed()->ld(as_Register(arg_index), SP_offset, SP);

    __ bind(NotLong);
    __ li(Rscratch0, float_sig);
    __ bne(Rscratch3, Rscratch0, NotFloat);
    __ delayed()->nop();
    // it's a float
    __ lwc1(as_FloatRegister(arg_index), SP_offset, SP);
    __ b(NextArg);
    __ delayed()->nop();

    __ bind(NotFloat);
    // must be a double
    __ addu_long(Rlongcount, Rlongcount, 1);
    __ ldc1(as_FloatRegister(arg_index), SP_offset, SP);

    __ bind(NextArg);
    __ srl_long(Rscratch4, Rscratch4, 2);
  }

  // Now adjust SP back to where it ought to be for this call.
  Label not_static;
  Label lessThan8;  // n_register_parameters == 8 on N64, but it's abstracted below
  Label done;

  // We need to calculate how many actual arguments were put on the C
  // stack.  This is 
  //   the number of bytecode stack entries (size_of_parameters)
  //   - 1 for every long or double, which takes extra bytecode stack space
  //   + 1 for the JNIEnv
  //   + 1 for the mirror argument if the callee is static
  // We also need to account for alignment.

  __ subu_long(T0, Rlocals, Rlongcount);
  __ lwu(T2, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);
  __ andi(T2, T2, JVM_ACC_STATIC);
  __ beq(T2, R0, not_static);
  __ delayed()->addu_long(T0, T0, 1);
  __ addu_long(T0, T0, 1);

  __ bind(not_static);
  __ li(T1, Argument::n_register_parameters);
  __ bltu(T0, T1, lessThan8);
  __ delayed()->sll_long(T1, T0, LogBytesPerWord);
  __ b(done);
  __ delayed()->li(T1, Argument::n_register_parameters * BytesPerWord);

  __ bind(lessThan8);
  __ round_to(T1, 2 * wordSize);              // keep stack aligned

  __ bind(done);
  __ addu_long(SP, SP, T1);

#ifdef ASSERT
  { // We stashed the original SP before arguments, so we can check
    // that we haven't exceeded it now.
    Label ok;
    __ ld_ptr(T0, frame::interpreter_frame_d_scratch_offset * wordSize, FP);
    __ bleu(SP, T0, ok);
    __ delayed()->nop();
    __ stop("Popped too much stack in signature handler");
    __ bind(ok);
  }
#endif

  // We take the amount that we saved originally, subtract the amount we
  // just eliminated (T1), and the result is how much we have left for
  // arguments.  We write that back.
  __ ld_long(T2, frame::interpreter_frame_args_size_offset * wordSize, FP);
  __ subu_long(T2, T2, T1);
  __ st_long(T2, frame::interpreter_frame_args_size_offset * wordSize, FP);


#ifdef ASSERT
  { // Check that "args SP" + args_size = "orig SP"
    Label ok;
    __ addu_long(T2, SP, T2);
    __ beq(T2, T0, ok);         // note T0 was loaded in previous assert
    __ delayed()->nop();
    __ stop("args_size didn't agree with origSP");
    __ bind(ok);
  }
#endif

  __ j(RA);
  __ delayed()->
     move(Rbcp, R0);  // !!! aliases RorigSP -- but we're done with that

  return entry;
}
#endif


address AbstractInterpreterGenerator::generate_safept_entry_for(TosState state, address runtime_entry) {
  address entry = __ pc();
  __ push(state);
  __ call_VM(noreg, runtime_entry);
  __ dispatch_via(vtos, Interpreter::normal_table(vtos));
  return entry;
}


address AbstractInterpreterGenerator::generate_continuation_for(TosState state) {
  address entry = __ pc();
  __ dispatch_next(state);
  return entry;
}

//
// Helpers for commoning out cases in the various type of method entries.
//

// increment invocation count & check for overflow
//
// Note: checking for negative value instead of overflow
//       so we have a 'sticky' overflow test
//
// Rmethod: method
// ??: invocation counter
//
void InterpreterGenerator::generate_counter_incr(Label* overflow, Label* profile_method, Label* profile_method_continue) {
  // Update standard invocation counters
  __ increment_invocation_counter(Rscratch0, Rscratch1);
  if (ProfileInterpreter) {  // %%% Merge this into methodDataOop
    int i = in_bytes(methodOopDesc::interpreter_invocation_counter_offset());
    __ lw(Rscratch1, i, Rmethod);
    __ addiu(Rscratch1, Rscratch1, 1);
    __ sw(Rscratch1, i, Rmethod);
  }

  if (ProfileInterpreter && profile_method != NULL) {
    // Test to see if we should create a method data oop
    __ li(Rscratch1, (long)&InvocationCounter::InterpreterProfileLimit);
    __ lw(Rscratch1, 0, Rscratch1);
    __ bltu(Rscratch0, Rscratch1, *profile_method_continue);
    __ delayed()->nop();

    // if no method data exists, go to profile_method
    __ test_method_data_pointer(*profile_method);
  }

  __ li(Rscratch1, (long)&InvocationCounter::InterpreterInvocationLimit);
  __ lw(Rscratch1, 0, Rscratch1);
  __ bgeu(Rscratch0, Rscratch1, *overflow);
  __ delayed()->nop();

}

void InterpreterGenerator::generate_counter_overflow(Label* Lcontinue) {

  // Generate code to initiate compilation on the counter overflow.

  // InterpreterRuntime::frequency_counter_overflow takes two arguments,
  // the first indicates if the counter overflow occurs at a backwards branch (NULL bcp)
  // and the second is only used when the first is true.  We pass zero for both.
  // The call returns the address of the verified entry point for the method or NULL
  // if the compilation did not complete (either went background or bailed out).
  __ call_VM(noreg,
	     CAST_FROM_FN_PTR(address,
			      InterpreterRuntime::frequency_counter_overflow),
	     R0,
	     R0,
	     true);
  // returns verified_entry_point or NULL
  // we ignore it in any case

  __ b(*Lcontinue);
  __ delayed()->nop();

}


// Allocate monitor and lock method (asm interpreter)
//
void InterpreterGenerator::lock_method(bool native) {
  __ lwu(A0, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);

#ifdef ASSERT
  { Label ok;
    __ andi(Rscratch3, A0, JVM_ACC_SYNCHRONIZED);
    __ bne(Rscratch3, R0, ok);
    __ delayed()->nop();
    __ stop("method doesn't need synchronization");
    __ bind(ok);
  }
#endif // ASSERT

  // get synchronization object to A0
  { Label done;
    Label nonstatic;
    const int mirror_offset = klassOopDesc::klass_part_offset_in_bytes() + Klass::java_mirror_offset_in_bytes();
    __ andi(Rscratch3, A0, JVM_ACC_STATIC);
    __ beq(Rscratch3, R0, nonstatic);
    __ delayed()->nop();

    // static case
    __ ld_ptr(A0, in_bytes(methodOopDesc::constants_offset()), Rmethod);
    __ ld_ptr(A0, constantPoolOopDesc::pool_holder_offset_in_bytes(), A0);

    // lock the mirror, not the klassOop
    __ ld_ptr(A0, mirror_offset, A0);

#ifdef ASSERT
    // breakpoint_trap if !A0
#endif // ASSERT

    __ b(done);
    __ delayed()->nop();

    __ bind(nonstatic);
    // get receiver for not-static case
    if (native) {
	// see also InterpreterRuntime::SignatureHandlerGenerator::generate
	__ lhu(T2, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod);
	__ sll_long(T2, T2, Interpreter::logStackElementSize());
	__ ld_ptr(T1, frame::interpreter_frame_l_scratch_offset * wordSize, FP);
	__ addu_long(T1, T1, T2);  // address of first parameter
	__ ld_ptr(A0, Interpreter::value_offset_in_bytes(), T1);
    } else {
	__ ld_ptr(A0, Interpreter::local_offset_in_bytes(0), Rlocals);
    }

    __ bind(done);

#ifdef ASSERT
    // We should have something to lock now
    __ teq(A0, R0);
#endif // ASSERT
  }

  // Monitor space was preallocated in generate_fixed_frame, so we don't
  // need to call add_monitor_to_stack, just adjust Rmonitors.
  __ addu_long(Rmonitors, Rmonitors, - frame::interpreter_frame_monitor_size() * wordSize);
  __ st_ptr(Rmonitors, frame::interpreter_frame_Rmonitors_offset * BytesPerWord, FP);
  __ st_ptr(A0, BasicObjectLock::obj_offset_in_bytes(), Rmonitors);  // store object
  __ lock_object(Rmonitors, A0);
}


//
// Generate a fixed interpreter frame. This is identical setup for interpreted
// methods and for native methods hence the shared code.

void InterpreterGenerator::generate_fixed_frame(bool native_call,
						bool synchronized) {

  __ verify_oop(Rmethod_call);

  const jint monitor_space = frame::interpreter_frame_monitor_size() * wordSize;

  // YYY not doing a stack overflow check!

  if (native_call) {
    Label not_static;

    // compute basic stack size
    jint fixed_size
      = frame::interpreter_frame_vm_native_locals_words * wordSize +
	(synchronized ? monitor_space : 0) +
        // one-element bytecode stack, just used for exception handling
        Interpreter::stackElementSize() +
        frame::GPR_saves_space + 
        frame::FPR_saves_space;
    fixed_size = ::round_to(fixed_size, 2 * wordSize);
    __ addu_long(Rscratch4, SP, - fixed_size);

    // Note: this may be an overestimate on _LP64, since longs take
    // two bytecode stack slots but only one C ABI slot.  That's ok.
    __ lhu(Rscratch2, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod_call);
    __ sll_long(Rscratch3, Rscratch2, LogBytesPerWord);
    __ lwu(Rscratch0, in_bytes(methodOopDesc::access_flags_offset()), Rmethod_call);
    __ andi(Rscratch0, Rscratch0, JVM_ACC_STATIC);
    __ beq(Rscratch0, R0, not_static);
    __ delayed()->
       addu_long(Rscratch3, Rscratch3, BytesPerWord);  // add for JNIEnv arg
    __ addu_long(Rscratch3, Rscratch3, BytesPerWord);  // add for mirror

    __ bind(not_static);
    __ round_to(Rscratch3, 2 * wordSize);              // keep size aligned

  } else {
    // compute stack offset
    __ lhu(Rscratch0, in_bytes(methodOopDesc::size_of_locals_offset()), Rmethod_call);
    __ sll_long(Rscratch0, Rscratch0, Interpreter::logStackElementSize());
    __ lhu(Rscratch1, in_bytes(methodOopDesc::max_stack_offset()), Rmethod_call);
    __ sll_long(Rscratch1, Rscratch1, Interpreter::logStackElementSize());
    __ addu_long(Rscratch2, SP,                     // Rlocals
		 - (frame::interpreter_frame_vm_locals_words + 1)*wordSize);
    __ subu_long(Rscratch3, Rscratch2, Rscratch0);  // Resp
    // monitor area is initially empty, unless method is synchronized
    if (synchronized) {
      // Lower Resp to make room.  Rmonitors is set from this later.
      __ addu_long(Rscratch3, Rscratch3, - monitor_space);
    }
    __ subu_long(Rscratch4, Rscratch3, Rscratch1);
    __ addu_long(Rscratch4, Rscratch4,
		 - (frame::GPR_saves_space + 
		    frame::FPR_saves_space +
		    frame::outgoing_args_space));
    __ round_down_to(Rscratch4, 2 * wordSize);
  }

  // Adjust SP (saving it first, so we can set FP below)
  __ move(Rscratch1, SP);
  __ move(SP, Rscratch4);

  // save regs
  __ save_registers(SP);

  // Set up local registers
  __ move(FP, Rscratch1);
  __ move(Rmethod, Rmethod_call);
  __ get_constant_pool_cache(RcpoolCache);
  if (native_call) {
    // Add the extra space for the outgoing parameters
    __ subu_long(SP, SP, Rscratch3);
    // Store so we can get it back easily after the JNI call
    // Note: we're going to update this after calling the signature handler
    __ subu_long(Rscratch1, Rscratch4, SP);
    __ st_long(Rscratch1, frame::interpreter_frame_args_size_offset * wordSize, FP);
    __ st_ptr(Rargs, frame::interpreter_frame_l_scratch_offset * wordSize, FP);
    // Cache original SP here, before params.  This is technically redundant,
    // since we could reconstruct it from the saved args_size.
    __ st_ptr(Rscratch4, frame::interpreter_frame_d_scratch_offset * wordSize, FP);
    // Point to the usual place just above the monitor area.  If this
    // method is synchronized, we will decrement it before using.
    __ addu_long(Rmonitors, FP, - frame::interpreter_frame_vm_native_locals_words * wordSize);
    // Set Rlocals to what would be the start of the locals area, since
    // get_top_most_monitor uses it to find the method monitor.
    __ addu_long(Rlocals, FP, - (frame::interpreter_frame_vm_native_locals_words + 1) * wordSize);
    __ addu_long(Resp, Rlocals, (synchronized ? monitor_space : 0));
    __ move(Rbcp, R0);
  } else {
    __ st_long(R0, frame::interpreter_frame_args_size_offset * wordSize, FP);
    __ move(Rlocals, Rscratch2);
    __ move(Resp, Rscratch3);
    __ addu_long(Rmonitors, Resp, wordSize + (synchronized ? monitor_space : 0));  // roughly, last local
    __ ld_ptr(Rbcp, in_bytes(methodOopDesc::const_offset()), Rmethod);
    __ addu_long(Rbcp, Rbcp, in_bytes(constMethodOopDesc::codes_offset()));
    // save static local var
    __ st_ptr(Rlocals, frame::interpreter_frame_Rlocals_offset * BytesPerWord, FP);
  }

  // save static locals
  __ st_ptr(Rmethod, frame::interpreter_frame_Rmethod_offset * BytesPerWord, FP);
  __ st_ptr(RcpoolCache, frame::interpreter_frame_RcpoolCache_offset * BytesPerWord, FP);

  // initial values, that will be overwritten in some cases
  __ st_ptr(R0, frame::interpreter_frame_Rbcp_offset * BytesPerWord, FP);
  __ st_ptr(Rmonitors, frame::interpreter_frame_Rmonitors_offset * BytesPerWord, FP);
  __ st_ptr(Resp, frame::interpreter_frame_Resp_offset * BytesPerWord, FP);
}

// End of helpers

// Various method entries

// Abstract method entry
// Attempt to execute abstract method. Throw exception
//
address InterpreterGenerator::generate_abstract_entry(void) {
  address entry = __ pc();
  // abstract method entry
  // throw exception
  __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_AbstractMethodError));
  // the call_VM checks for exception, so we should never return here.
  __ should_not_reach_here();
  return entry;
}

// Empty method, generate a very fast return.

address InterpreterGenerator::generate_empty_entry(void) {

  // A method that does nother but return...

  address entry = __ pc();
  Label slow_path;

  __ verify_oop(Rmethod_call);

  // do nothing for empty methods (do not even increment invocation counter)
  if ( UseFastEmptyMethods) {
    // If we need a safepoint check, generate full interpreter entry.
    __ li(Rscratch3, SafepointSynchronize::address_of_state());
    __ lw(Rscratch3, 0, Rscratch3);
    __ li(Rscratch1, SafepointSynchronize::_not_synchronized);
    __ bne(Rscratch3, Rscratch1, slow_path);
    __ delayed()->nop();

    // Code: _return
    __ j(RA);
    // YYY savedSP, if needed
    __ delayed()->nop();

    __ bind(slow_path);
    (void) generate_asm_interpreter_entry(false);

    return entry;
  }

  return NULL;
}

// Call an accessor method (assuming it is resolved, otherwise drop into
// vanilla (slow path) entry

// Generates code to elide accessor methods
address InterpreterGenerator::generate_accessor_entry(void) {

  // Code: _aload_0, _(i|a)getfield, _(i|a)return or any rewrites thereof;
  // parameter size = 1
  // Note: We can only use this code if the getfield has been resolved
  //       and if we don't have a null-pointer exception => check for
  //       these conditions first and use slow path if necessary.
  address entry = __ pc();
  Label slow_path;

  if ( UseFastAccessorMethods) {
    // Check if we need to reach a safepoint and generate full interpreter
    // frame if so.
    Unimplemented();		// YYY
  }
  return NULL;
}


//
// Interpreter stub for calling a native method. (asm interpreter)
// This sets up a somewhat different looking stack for calling the native method
// than the typical interpreter frame setup.
//

address InterpreterGenerator::generate_native_entry(bool synchronized) {
  address entry = __ pc();

  // the following temporary registers are used during frame creation
  const Register Rtmp1 = Rscratch3;
  const Register Rtmp2 = Rscratch1;
  bool inc_counter  = UseCompiler || CountCompiledCalls;

  // make sure registers are different!
  assert_different_registers(Rthread, Rmethod_call, Rargs, Rtmp1, Rtmp2);

  __ verify_oop(Rmethod_call);

  const Register Rlocals_size = Rscratch3;
  assert_different_registers(Rlocals_size, Rscratch4);

  // make sure method is native & not abstract
#ifdef ASSERT
  __ ld(Rtmp1, in_bytes(methodOopDesc::access_flags_offset()), Rmethod_call);
  {
    Label L;
    __ andi(Rtmp2, Rtmp1, JVM_ACC_NATIVE);
    __ bne(Rtmp2, R0, L);
    __ delayed()->nop();
    __ stop("tried to execute non-native method as native");
    __ bind(L);
  }
  { Label L;
    __ andi(Rtmp2, Rtmp1, JVM_ACC_ABSTRACT);
    __ beq(Rtmp2, R0, L);
    __ delayed()->nop();
    __ stop("tried to execute abstract method as non-abstract");
    __ bind(L);
  }
#endif // ASSERT

 // generate the code to allocate the interpreter stack frame
  generate_fixed_frame(true, synchronized);

  // this slot will be set later, we initialize it to null here just in
  // case we get a GC before the actual value is stored later
  __ st_ptr(R0, frame::interpreter_frame_oop_temp_offset*wordSize, FP);

  // Since at this point in the method invocation the exception handler
  // would try to exit the monitor of synchronized methods which hasn't
  // been entered yet, we set the thread local variable
  // _do_not_unlock_if_synchronized to true. If any exception was thrown by
  // runtime, exception handling i.e. unlock_if_synchronized_method will
  // check this thread local flag.
  // This flag has two effects, one is to force an unwind in the topmost
  // interpreter frame and not perform an unlock while doing so.
  __ li(Rtmp2, 1);
  __ sb(Rtmp2, in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);

  // increment invocation counter and check for overflow
  //
  // Note: checking for negative value instead of overflow
  //       so we have a 'sticky' overflow test (may be of
  //       importance as soon as we have true MT/MP)
  Label invocation_counter_overflow; 
  Label Lcontinue;
  if (inc_counter) {
    generate_counter_incr(&invocation_counter_overflow, NULL, NULL);

  }
  __ bind(Lcontinue);

  bang_stack_shadow_pages(true);

  // reset the _do_not_unlock_if_synchronized flag
  __ sb(R0, in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);

  // check for synchronized methods
  // Must happen AFTER invocation_counter check and stack overflow check,
  // so method is not locked if overflows.

  if (synchronized) {
    lock_method(true);
  } else {
#ifdef ASSERT
    { Label ok;
      __ lwu(A0, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);
      __ andi(Rtmp2, A0, JVM_ACC_SYNCHRONIZED);
      __ beq(Rtmp2, R0, ok);
      __ delayed()->nop();
      __ stop("method needs synchronization");
      __ bind(ok);
    }
#endif // ASSERT
  }

  // start execution
  __ verify_thread();

  // JVMTI support 
  __ notify_method_entry();

  // native call

  // (note that A0 is never an oop--at most it is a handle)
  // It is important not to smash any handles created by this call,
  // until any oop handle in A0 is dereferenced.

  // (note that the space for outgoing params is preallocated)
  // (in generate_fixed_frame)

  // get signature handler
  { Label L;
    __ ld_ptr(Rscratch3, in_bytes(methodOopDesc::signature_handler_offset()), Rmethod);
    __ bne(Rscratch3, R0, L);
    __ delayed()->nop();
    __ call_VM(noreg,
	       CAST_FROM_FN_PTR(address,
				InterpreterRuntime::prepare_native_call),
	       Rmethod);
    __ ld_ptr(Rscratch3, in_bytes(methodOopDesc::signature_handler_offset()), Rmethod);
    __ bind(L);
  }

  // sparc impl. adds a whole new frame here to protect the registers.
  // Rmethod etc are all callee-save, so looks like we don't need
  // this.

  // call signature handler
  // This is a generated routine, but it can lead to C code.
  __ jalr(Rscratch3);
  __ delayed()->nop();

  // Result handler is in Rsave0

  { Label not_static;

    __ lwu(A0, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);
    __ andi(A0, A0, JVM_ACC_STATIC);
    __ beq(A0, R0, not_static);
    __ delayed()->ld_ptr(Rcall, in_bytes(methodOopDesc::native_function_offset()), Rmethod);

    // for static methods insert the mirror argument
    const int mirror_offset = klassOopDesc::klass_part_offset_in_bytes() + Klass::java_mirror_offset_in_bytes();
    __ ld_ptr(A1, in_bytes(methodOopDesc:: constants_offset()), Rmethod);
    __ ld_ptr(A1, constantPoolOopDesc::pool_holder_offset_in_bytes(), A1);
    __ ld_ptr(A1, mirror_offset, A1);
#ifdef ASSERT
    if (!PrintSignatureHandlers)  // do not dirty the output with this
    { Label L;
      __ bne(A1, R0, L);
      __ delayed()->nop();
      __ stop("mirror is missing");
      __ bind(L);
    }
#endif // ASSERT
    __ addu_long(Rscratch2, FP, frame::interpreter_frame_oop_temp_offset*wordSize);
    __ st_ptr(A1, 0, Rscratch2);
    __ move(A1, Rscratch2);
    __ bind(not_static);
  }

  // At this point, arguments have been copied off of stack into
  // their JNI positions, which are A2, A3, ... and SP[...].
  // A1 holds the mirror.
  // Oops are boxed in-place on the stack, with handles copied to arguments.
  // The result handler is in Rsave0.  A0 will shortly hold the JNIEnv*.

#ifdef ASSERT
  { Label L;
    __ bne(Rcall, R0, L);
    __ delayed()->nop();
    __ stop("native entry point is missing");
    __ bind(L);
  }
#endif // ASSERT

  //
  // setup the frame anchor
  //
  // The scavenge function only needs to know that the PC of this frame is
  // in the interpreter method entry code, it doesn't need to know the exact
  // PC and hence we can use RA which points to the return address from the
  // previous call in the code stream (signature handler function)
  //

  // Sparc passes in FP for SP because of the extra frame they use,
  // but since we don't do that, just use SP as normal.

  __ set_last_Java_frame(SP, FP, RA);

  // Transition from _thread_in_Java to _thread_in_native. We are already safepoint ready.

#ifdef ASSERT
  { Label L;
    __ lwu(Rscratch3, in_bytes(JavaThread::thread_state_offset()), Rthread);
    __ li(Rscratch4, _thread_in_Java);
    __ beq(Rscratch3, Rscratch4, L);
    __ delayed()->nop();
    __ stop("Wrong thread state in native stub");
    __ bind(L);
  }
#endif // ASSERT
  __ li(Rscratch3, _thread_in_native);
  __ sw(Rscratch3, in_bytes(JavaThread::thread_state_offset()), Rthread);

  // Call the jni method, using the delay slot to set the JNIEnv* argument.
  __ jalr(Rcall);
  __ delayed()->addu_long(A0, Rthread, in_bytes(JavaThread::jni_environment_offset()));

  // Back from jni method

  // must we block?
  
  // Block, if necessary, before resuming in _thread_in_Java state.
  // In order for GC to work, don't clear the last_Java_sp until after blocking.
  { Label no_block;

    // Switch thread to "native transition" state before reading the synchronization state.
    // This additional state is necessary because reading and testing the synchronization
    // state is not atomic w.r.t. GC, as this scenario demonstrates:
    //     Java thread A, in _thread_in_native state, loads _not_synchronized and is preempted.
    //     VM thread changes sync state to synchronizing and suspends threads for GC.
    //     Thread A is resumed to finish this native method, but doesn't block here since it
    //     didn't see any synchronization is progress, and escapes.
    __ li(Rscratch3, _thread_in_native_trans);
    __ sw(Rscratch3, in_bytes(JavaThread::thread_state_offset()), Rthread);
    if(os::is_MP()) {
      __ membar();
    }
    __ li(Rscratch4, SafepointSynchronize::address_of_state());
    __ lwu(Rscratch3, 0, Rscratch4);
    __ li(Rscratch4, SafepointSynchronize::_not_synchronized);

    Label L;
    __ bne(Rscratch3, Rscratch4, L);
    __ delayed()->
       lwu(Rscratch3, in_bytes(JavaThread::suspend_flags_offset()), Rthread);
    __ beq(Rscratch3, R0, no_block);
    __ delayed()->nop();
    __ bind(L);

    // Block.  Save any potential method result value before the operation and
    // use a leaf call to leave the last_Java_frame setup undisturbed.
    __ save_return_value(ilgl, true);
    __ call_VM_leaf(Rthread,
		    CAST_FROM_FN_PTR(address, JavaThread::check_special_condition_for_native_trans),
		    Rthread);

    // Restore any method result value
    __ restore_return_value(ilgl, true);
    __ bind(no_block);
  }

  // Clear the frame anchor now

  __ reset_last_Java_frame();

  __ li(Rscratch3, _thread_in_Java);
  __ sw(Rscratch3, in_bytes(JavaThread::thread_state_offset()), Rthread);

  // reset handle block
  __ ld_ptr(Rscratch3, in_bytes(JavaThread::active_handles_offset()), Rthread);
  __ st_ptr(R0, JNIHandleBlock::top_offset_in_bytes(), Rscratch3);

  // If we have an oop result store it where it will be safe for any further gc
  // until we return now that we've released the handle it might be protected by
  {
    Label no_oop, store_result;

    __ li(Rscratch3, (intptr_t)AbstractInterpreter::result_handler(T_OBJECT));
    __ bne(Rscratch3, Rsave0, no_oop);
    __ delayed()->nop();
    __ beq(V0, R0, store_result);     // if result is not NULL
    __ delayed()->nop();
    __ ld_ptr(V0, 0, V0);             // unbox it

    __ bind(store_result);
    // Store it where gc will look for it and result handler expects it.
    __ st_ptr(V0, frame::interpreter_frame_oop_temp_offset*wordSize, FP);

    __ bind(no_oop);

  }


  // handle exceptions (exception handling will handle unlocking!)
  { Label L;
    __ ld_ptr(Rscratch1, in_bytes(Thread::pending_exception_offset()), Rthread);
    __ beq(Rscratch1, R0, L);
    __ delayed()->nop();
    // Note: This could be handled more efficiently since we know that the native
    //       method doesn't have an exception handler. We could directly return
    //       to the exception handler for the caller.
    __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_pending_exception));
    __ should_not_reach_here();
    __ bind(L);
  }

  // JVMTI support (preserves thread register) 
  __ notify_method_exit(true, ilgl, InterpreterMacroAssembler::NotifyJVMTI);  

  if (synchronized) {
    // save and restore any potential method result value around the unlocking operation
    __ save_return_value(ilgl, true);
    __ get_top_most_monitor(A1);
    __ unlock_object(A1);
    __ restore_return_value(ilgl, true);
  }

  // dispose of return address and remove activation
  __ jr(Rsave0);
  __ delayed()->nop();


  if (inc_counter) {
    // handle invocation counter overflow
    __ bind(invocation_counter_overflow);
    generate_counter_overflow(&Lcontinue);
  }



  return entry;
}

// Generic method entry to (asm) interpreter
//------------------------------------------------------------------------------------------------------------------------
// Entry points
//
// Arguments:

//----------------------------------------------------------------------------------------------------
// Entry points & stack frame layout
address InterpreterGenerator::generate_asm_interpreter_entry(bool synchronized) {
  address entry = __ pc();

  bool inc_counter  = UseCompiler || CountCompiledCalls;

  __ verify_oop(Rmethod_call);

  // make sure method is not native & not abstract
#ifdef ASSERT
  __ lwu(Rscratch1, in_bytes(methodOopDesc::access_flags_offset()), Rmethod_call);
  {
    Label L;
    __ andi(Rscratch0, Rscratch1, JVM_ACC_NATIVE);
    __ beq(Rscratch0, R0, L);
    __ delayed()->nop();
    __ stop("tried to execute native method as non-native");
    __ bind(L);
  }
  { Label L;
    __ andi(Rscratch0, Rscratch1, JVM_ACC_ABSTRACT);
    __ beq(Rscratch0, R0, L);
    __ delayed()->nop();
    __ stop("tried to execute abstract method as non-abstract");
    __ bind(L);
  }
#endif // ASSERT

  // generate the code to allocate the interpreter stack frame
  generate_fixed_frame(false, synchronized);

  // Zero locals and copy arguments.  
  // Locals space has been set up in generate_fixed_frame, while
  // incoming parameters are pointed to by Rargs.
  // In particular, the parameters are in caller's bytecode stack
  // and Rargs points one beyond the end of it.  This works out
  // because we are going to set everything in reverse order.

  // Code below doesn't make good use of delay slots.  Just trying to
  // be clear while I get it to work.  Optimize later YYY.

  Label zerolocals_entry, zerolocals_loop;
  Label copyargs_entry, copyargs_loop;

  __ lhu(Rscratch2, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod);
  __ sll_long(Rscratch2, Rscratch2, Interpreter::logStackElementSize());
  __ lhu(Rscratch3, in_bytes(methodOopDesc::size_of_locals_offset()), Rmethod);
  __ sll_long(Rscratch3, Rscratch3, Interpreter::logStackElementSize());
  __ subu_long(Rscratch2, Rlocals, Rscratch2);  // one below last param slot
  __ subu_long(Rscratch3, Rlocals, Rscratch3);  // one below last local slot
  __ b(zerolocals_entry);
  __ delayed()->nop();

  __ bind(zerolocals_loop);
  if (TaggedStackInterpreter) {
    __ st_long(R0, Interpreter::tag_offset_in_bytes(), Rscratch3);
  }
  __ st_long(R0, Interpreter::value_offset_in_bytes(), Rscratch3);

  __ bind(zerolocals_entry);
  __ addu_long(Rscratch3, Rscratch3, Interpreter::stackElementSize());
  __ bleu(Rscratch3, Rscratch2, zerolocals_loop);
  __ delayed()->nop();
  __ b(copyargs_entry);
  __ delayed()->nop();
  
  __ bind(copyargs_loop);
  __ ld_long(Rscratch1, Interpreter::value_offset_in_bytes(), Rargs);
  __ st_long(Rscratch1, Interpreter::value_offset_in_bytes(), Rscratch2);
  if (TaggedStackInterpreter) {
    __ ld_long(Rscratch1, Interpreter::tag_offset_in_bytes(), Rargs);
    __ st_long(Rscratch1, Interpreter::tag_offset_in_bytes(), Rscratch2);
  }

  __ bind(copyargs_entry);
  __ addu_long(Rscratch2, Rscratch2, Interpreter::stackElementSize());
  __ bleu(Rscratch2, Rlocals, copyargs_loop);
  __ delayed()->addu_long(Rargs, Rargs, Interpreter::stackElementSize());

#ifdef FAST_DISPATCH
  // set bytecode dispatch table base
#error unimplemented
#endif

  // Since at this point in the method invocation the exception handler
  // would try to exit the monitor of synchronized methods which hasn't
  // been entered yet, we set the thread local variable
  // _do_not_unlock_if_synchronized to true. If any exception was thrown by
  // runtime, exception handling i.e. unlock_if_synchronized_method will
  // check this thread local flag.
  __ li(Rscratch3, 1);
  __ sb(Rscratch3, in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);

  // increment invocation counter and check for overflow
  //
  // Note: checking for negative value instead of overflow
  //       so we have a 'sticky' overflow test (may be of
  //       importance as soon as we have true MT/MP)
  Label invocation_counter_overflow; 
  Label profile_method;
  Label profile_method_continue;
  Label Lcontinue;
  if (inc_counter) {
    generate_counter_incr(&invocation_counter_overflow, &profile_method, &profile_method_continue);
    if (ProfileInterpreter) {
      __ bind(profile_method_continue);
    }
  }
  __ bind(Lcontinue);

  bang_stack_shadow_pages(false);

  // reset the _do_not_unlock_if_synchronized flag
  __ sb(R0, in_bytes(JavaThread::do_not_unlock_if_synchronized_offset()), Rthread);

  // check for synchronized methods
  // Must happen AFTER invocation_counter check and stack overflow check,
  // so method is not locked if overflows.

  if (synchronized) {
    lock_method(false);
  } else {
#ifdef ASSERT
    { Label ok;
      __ lwu(A0, in_bytes(methodOopDesc::access_flags_offset()), Rmethod);
      __ andi(Rscratch3, A0, JVM_ACC_SYNCHRONIZED);
      __ beq(Rscratch3, R0, ok);
      __ delayed()->nop();
      __ stop("method needs synchronization");
      __ bind(ok);
    }
#endif // ASSERT
  }

  // start execution

  __ verify_thread();

  // jvmti support 
  __ notify_method_entry();

  // start executing instructions
  __ dispatch_next(vtos);


  if (inc_counter) {
    if (ProfileInterpreter) {
      // We have decided to profile this method in the interpreter
      __ bind(profile_method);

      __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::profile_method), Rbcp, true);

#ifdef ASSERT
      // should we verify V0==0?
#endif

      __ set_method_data_pointer();

      __ b(profile_method_continue);
      __ delayed()->nop();
    }

    // handle invocation counter overflow
    __ bind(invocation_counter_overflow);
    generate_counter_overflow(&Lcontinue);
  }


  return entry;
}

//----------------------------------------------------------------------------------------------------
// Entry points & stack frame layout
//
// Here we generate the various kind of entries into the interpreter.
// The two main entry type are generic bytecode methods and native call method.
// These both come in synchronized and non-synchronized versions but the
// frame layout they create is very similar. The other method entry
// types are really just special purpose entries that are really entry
// and interpretation all in one. These are for trivial methods like
// accessor, empty, or special math methods.
//
// When control flow reaches any of the entry types for the interpreter
// the following holds ->
// 
// C2 Calling Conventions:
//
// [...]

// Try this optimization:  Most method entries could live in a
// "one size fits all" stack frame without all the dynamic size
// calculations.  It might be profitable to do all this calculation
// statically and approximately for "small enough" methods.

//-----------------------------------------------------------------------------------------------

// C1 Calling conventions
//
// Upon method entry, the following registers are setup:
//
// [...]
//
//
// Stack:
//
// [...]
//
//
//
// AFTER FRAME HAS BEEN SETUP for method interpretation the stack looks like:
//
// [...]

address AbstractInterpreterGenerator::generate_method_entry(AbstractInterpreter::MethodKind kind) {
  // determine code generation flags
  bool synchronized = false;
  address entry_point = NULL;

  switch (kind) {    
    case Interpreter::zerolocals             :                                                                             break;
    case Interpreter::zerolocals_synchronized: synchronized = true;                                                        break;
    case Interpreter::native                 : entry_point = ((InterpreterGenerator*)this)->generate_native_entry(false);  break;
    case Interpreter::native_synchronized    : entry_point = ((InterpreterGenerator*)this)->generate_native_entry(true);   break;
    case Interpreter::empty                  : entry_point = ((InterpreterGenerator*)this)->generate_empty_entry();        break;
    case Interpreter::accessor               : entry_point = ((InterpreterGenerator*)this)->generate_accessor_entry();     break;
    case Interpreter::abstract               : entry_point = ((InterpreterGenerator*)this)->generate_abstract_entry();     break;
    case Interpreter::java_lang_math_sin     :                                                                             break;
    case Interpreter::java_lang_math_cos     :                                                                             break;
    case Interpreter::java_lang_math_tan     :                                                                             break;
    case Interpreter::java_lang_math_sqrt    :                                                                             break;
    case Interpreter::java_lang_math_abs     :                                                                             break;
    case Interpreter::java_lang_math_log     :                                                                             break;
    case Interpreter::java_lang_math_log10   :                                                                             break;
    default                                  : ShouldNotReachHere();                                                       break;
  }

  if (entry_point) return entry_point;

  return ((InterpreterGenerator*)this)->generate_asm_interpreter_entry(synchronized);
}


static int size_activation_helper(int vm_locals,
				  int callee_locals,
				  int max_stack,
				  int monitor_size)
{
  // callee_locals and max_stack are counts, not the size in frame.
  return
    round_to(callee_locals * Interpreter::stackElementWords()
	     + max_stack * Interpreter::stackElementWords()
	     + vm_locals * wordSize,
	     WordsPerLong)
    // already rounded
    + monitor_size;
}

// How much stack a method top interpreter activation needs in words.
int AbstractInterpreter::size_top_interpreter_activation(methodOop method) {
  // Save space for one monitor to get into the interpreted method in case
  // the method is synchronized
  int monitor_size    = method->is_synchronized() ?
                                1*frame::interpreter_frame_monitor_size() : 0;
  // See call_stub code
  int vm_locals = method->is_native() ?
    frame::interpreter_frame_vm_native_locals_words :
    frame::interpreter_frame_vm_locals_words;
  int call_stub_size  = round_to(19, WordsPerLong);  // register save area

  return
    size_activation_helper(vm_locals,
			   method->max_locals(),
			   method->max_stack(),
			   monitor_size)
    + call_stub_size;
}

// This method tells the deoptimizer how big an interpreted frame must be:
int AbstractInterpreter::size_activation(methodOop method,
					 int tempcount,
					 int popframe_extra_args,
					 int moncount,
					 int callee_param_count,
					 int callee_locals,
					 bool is_top_frame) {
  return layout_activation(method,
                           tempcount,
                           popframe_extra_args,
                           moncount,
                           callee_param_count,
                           callee_locals,
                           (frame*)NULL,
                           (frame*)NULL,
                           is_top_frame);
}


int AbstractInterpreter::layout_activation(methodOop method,
                                           int tempcount,
                                           int popframe_extra_args,
                                           int moncount,
                                           int callee_param_count,
                                           int callee_local_count,
                                           frame* caller,
                                           frame* interpreter_frame,
                                           bool is_top_frame) {
  // Note: This calculation must exactly parallel the frame setup
  // in InterpreterGenerator::generate_fixed_frame.
  Unimplemented();
  return 0;
}

void Deoptimization::unwind_callee_save_values(frame* f, vframeArray* vframe_array) {

  // This code is sort of the equivalent of C2IAdapter::setup_stack_frame back in
  // the days we had adapter frames. When we deoptimize a situation where a
  // compiled caller calls a compiled caller will have registers it expects
  // to survive the call to the callee. If we deoptimize the callee the only
  // way we can restore these registers is to have the oldest interpreter
  // frame that we create restore these values. That is what this routine
  // will accomplish.

  // At the moment we have modified c2 to not have any callee save registers
  // so this problem does not exist and this routine is just a place holder.

  assert(f->is_interpreted_frame(), "must be interpreted");
  Untested("sparc");
}


//----------------------------------------------------------------------------------------------------
// Exceptions
void AbstractInterpreterGenerator::generate_throw_exception() {

  // Entry point in previous activation (i.e., if the caller was interpreted)
  Interpreter::_rethrow_exception_entry = __ pc();
  // A0: exception

  // entry point for exceptions thrown within interpreter code
  Interpreter::_throw_exception_entry = __ pc();
  __ verify_thread();
  // expression stack is undefined here
  // Rbcp: exception bcx
  __ verify_oop(A0);

  // expression stack must be empty before entering the VM in case of an exception
  __ empty_expression_stack();
  // find exception handler address and preserve exception oop
  // call C routine to find handler and jump to it
  __ call_VM(V1, CAST_FROM_FN_PTR(address, InterpreterRuntime::exception_handler_for_exception), A0);
  __ push_ptr(V1);  // push exception for exception handler bytecodes
  __ jr(V0);
  __ delayed()->nop();

  // if the exception is not handled in the current frame
  // the frame is removed and the exception is rethrown
  // (i.e. exception continuation is _rethrow_exception)
  //
  // Note: At this point the bci is still the bxi for the instruction which caused
  //       the exception and the expression stack is empty. Thus, for any VM calls
  //       at this point, GC will find a legal oop map (with empty expression stack).

  // In current activation:
  // TOS: exception
  // Rbcp: exception bcp

  //
  // JVMTI PopFrame support
  //

  // ----------------------------------------------------------------
  Interpreter::_remove_activation_preserving_args_entry = __ pc();
  // Set the popframe_processing bit in popframe_condition indicating that we are
  // currently handling popframe, so that call_VMs that may happen later do not trigger new
  // popframe handling cycles.

  __ lw(Rscratch3, in_bytes(JavaThread::popframe_condition_offset()), Rthread);
  __ ori(Rscratch3, Rscratch3, JavaThread::popframe_processing_bit);
  __ sw(Rscratch3, in_bytes(JavaThread::popframe_condition_offset()), Rthread);

  // Empty the expression stack, as in normal exception handling
  __ empty_expression_stack();
  __ unlock_if_synchronized_method(vtos, /* throw_monitor_exception */ false, /* install_monitor_exception */ false);

  {
    // Check to see whether we are returning to a deoptimized frame.
    // (The PopFrame call ensures that the caller of the popped frame is
    // either interpreted or compiled and deoptimizes it if compiled.)
    // In this case, we can't call dispatch_next() after the frame is
    // popped, but instead must save the incoming arguments and restore
    // them after deoptimization has occurred.
    //
    // Note that we don't compare the return PC against the
    // deoptimization blob's unpack entry because of the presence of
    // adapter frames in C2.
    Label caller_not_deoptimized;

    // This bottoms out in StubQueue::contains and ultimately checks
    // if the faulting PC is in the interpreter code buffer.
    if (UseCompiler) {		// YYY unimplemented
    __ call_VM_leaf(Rthread,
		    CAST_FROM_FN_PTR(address,
				     InterpreterRuntime::interpreter_contains),
		    Rissuing_pc); // YYY ???
    __ bne(V0, R0, caller_not_deoptimized);
    __ delayed()->nop();

    __ UNIMPLEMENTED();
    }

    // Compute size of arguments for saving when returning to deoptimized caller

    // Save these arguments

    // Inform deoptimization that it is responsible for restoring these arguments

    // Return from the current method
    // The caller's SP was adjusted upon method entry to accomodate
    // the callee's non-argument locals. Undo that adjustment.

    __ bind(caller_not_deoptimized);
  }

  // Clear the popframe condition flag
  __ sw(R0, in_bytes(JavaThread::popframe_condition_offset()), Rthread);

  // Get out of the current method (how this is done depends on the particular compiler calling
  // convention that the interpreter currently follows)
  // YYY restore from savedSP?

  // The method data pointer was incremented already during
  // call profiling. We have to restore the mdp for the current bcp.
  if (ProfileInterpreter) {
    __ set_method_data_pointer_for_bcp();
  }

  // Resume bytecode interpretation at the current bcp
  __ dispatch_next(vtos);
  // end of JVMTI PopFrame support

  // ----------------------------------------------------------------
  Interpreter::_remove_activation_entry = __ pc();

  // preserve exception over this code sequence (remove activation calls the vm, but oopmaps are not correct here)
  __ pop_ptr(Rexception);                                  // get exception

  // Intel has the following comment:
  //// remove the activation (without doing throws on illegalMonitorExceptions)
  // They remove the activation without checking for bad monitor state.
  __ set_vm_result(Rexception);
  __ unlock_if_synchronized_method(vtos, /* throw_monitor_exception */ false);

  // jvmpi support (preserves thread register)
  // (jvmdi does not generate MethodExit on exception / popFrame)
  __ notify_method_exit(false, vtos, InterpreterMacroAssembler::SkipNotifyJVMTI);

  __ get_vm_result(Rexception);
  __ verify_oop(Rexception);
  __ push_ptr(Rexception);

  // We are done with this activation frame; find out where to go next.
  // The continuation point will be an exception handler, which expects
  // the following registers set up:
  //
  // Rexception: exception
  // Rissuing_pc: the local call that threw exception
  //
  // We do the required restore at the last possible moment, because we
  // need to preserve some state across a runtime call.
  // (Remember that the caller activation is unknown--it might not be
  // interpreted, so things like RscratchX are useless in the caller.)

  // But, we can at least get rid of any remaining extra argument
  // space, to make it easier to get at the saved regs.  Mainly, we
  // need the saved RA.
  __ ld_long(Rscratch0, frame::interpreter_frame_args_size_offset * wordSize, FP);
  __ addu_long(SP, SP, Rscratch0);
  __ ld_ptr(Rscratch0, RA->sp_offset_in_save_area(), SP);

  // Although the Intel version uses call_C, we can use the more
  // compact call_VM.  (The only real difference on MIPS is a
  // harmlessly ignored [re]set_last_Java_frame, compared with
  // the Intel code which lacks this.)
  __ super_call_VM_leaf(Rthread,
                        CAST_FROM_FN_PTR(address, SharedRuntime::exception_handler_for_return_address),
                        Rscratch0);
  // return address will have been returned from above call in V0

  __ pop_ptr(Rexception);
  __ ld_ptr(Rissuing_pc, RA->sp_offset_in_save_area(), SP);

  __ move(Rscratch0, FP);
  __ restore_registers(SP);
  __ jr(V0);                        // return from interpreter activation
  __ delayed()->move(SP, Rscratch0);

  // YYY restore from savedSP?
}


//
// JVMTI ForceEarlyReturn support
//

address AbstractInterpreterGenerator::generate_earlyret_entry_for(TosState state) {
  address entry = __ pc();

  __ empty_expression_stack();
  __ load_earlyret_value(state);

  __ ld_ptr(Rscratch3, in_bytes(JavaThread::jvmti_thread_state_offset()), Rthread);

  // Clear the earlyret state
  __ sw(R0 /* JvmtiThreadState::earlyret_inactive */,
	in_bytes(JvmtiThreadState::earlyret_state_offset()), Rscratch3);

  __ remove_activation(state,
                       /* throw_monitor_exception */ false,
                       /* install_monitor_exception */ false);

  __ jr(RA);
  // YYY restore savedSP?
  __ delayed()->move(SP, Rscratch5);

  return entry;
} // end of JVMTI ForceEarlyReturn support


//----------------------------------------------------------------
// Helper for vtos entry point generation

void AbstractInterpreterGenerator::set_vtos_entry_points(Template* t, address& bep, address& cep, address& sep, address& aep, address& iep, address& lep, address& fep, address& dep, address& vep) {
  assert(t->is_valid() && t->tos_in() == vtos, "illegal template");
  Label L;
  aep = __ pc(); __ push_ptr(); __ b(L); __ delayed()->nop(); 
  fep = __ pc(); __ push_f();   __ b(L); __ delayed()->nop();
  dep = __ pc(); __ push_d();   __ b(L); __ delayed()->nop();
  lep = __ pc(); __ push_l();   __ b(L); __ delayed()->nop();
  iep = __ pc(); __ push_i();          
  bep = cep = sep = iep;                        // there aren't any
  vep = __ pc();                                // fall through
  __ bind(L);  
  generate_and_dispatch(t);
}

// --------------------------------------------------------------------------------


InterpreterGenerator::InterpreterGenerator(StubQueue* code) 
 : AbstractInterpreterGenerator(code) {
   generate_all(); // down here so it can be "virtual"
}

// --------------------------------------------------------------------------------

// Non-product code
#ifndef PRODUCT
address AbstractInterpreterGenerator::generate_trace_code(TosState state) {
  address entry = __ pc();

  __ push(state);
  __ move(Rsave0, RA);  // protect return address within interpreter

  // Pass a 0 (not used in sparc) and the top of stack to the bytecode tracer

#ifdef _LP64
  __ call_VM(noreg, CAST_FROM_FN_PTR(address, SharedRuntime::trace_bytecode), R0, Rtos, R0);
#else
  __ move(Rscratch3, Rtos_l2);
  __ call_VM(noreg, CAST_FROM_FN_PTR(address, SharedRuntime::trace_bytecode), R0, Rtos_l1, Rscratch3);
#endif // _LP64
  __ move(RA, Rsave0);  // restore return address
  __ pop(state);
  __ jr(RA);
  __ delayed()->nop();

  return entry;
}


// helpers for generate_and_dispatch

void AbstractInterpreterGenerator::count_bytecode() { 
  __ li(Rscratch3, (address)&BytecodeCounter::_counter_value);
  __ lw(Rscratch4, 0, Rscratch3);
  __ addu_long(Rscratch4, Rscratch4, 1);
  __ sw(Rscratch4, 0, Rscratch3);
}


void AbstractInterpreterGenerator::histogram_bytecode(Template* t) { 
  __ li(Rscratch3, (address) &BytecodeHistogram::_counters[t->bytecode()]);
  __ lw(Rscratch4, 0, Rscratch3);
  __ addu_long(Rscratch4, Rscratch4, 1);
  __ sw(Rscratch4, 0, Rscratch3);
}


void AbstractInterpreterGenerator::histogram_bytecode_pair(Template* t) { 
  // get index, shift out old bytecode, bring in new bytecode, and store it
  // _index = (_index >> log2_number_of_codes) | 
  //          (bytecode << log2_number_of_codes);

  __ li(Rscratch2, (address)&BytecodePairHistogram::_index);
  __ lw(Rscratch4, 0, Rscratch2);
  __ srl_long(Rscratch4, Rscratch4, BytecodePairHistogram::log2_number_of_codes);
  __ li(Rscratch3, ((int)t->bytecode()) << BytecodePairHistogram::log2_number_of_codes);
  __ or_(Rscratch4, Rscratch4, Rscratch3);
  __ sw(Rscratch4, 0, Rscratch2);

  // bump bucket contents
  // _counters[_index] ++;

  __ li(Rscratch3, (address)&BytecodePairHistogram::_counters);
  __ sll_long(Rscratch4, Rscratch4, LogBytesPerWord); // Index is word address
  __ addu_long(Rscratch3, Rscratch3, Rscratch4);       // Add in index
  __ lw(Rscratch4, 0, Rscratch3);
  __ addu_long(Rscratch4, Rscratch4, 1);
  __ sw(Rscratch4, 0, Rscratch3);
}



void AbstractInterpreterGenerator::trace_bytecode(Template* t) {
  // Call a little run-time stub to avoid blow-up for each bytecode.
  // The run-time runtime saves the right registers, depending on
  // the tosca in-state for the given template.
  address entry = Interpreter::trace_code(t->tos_in());
  guarantee(entry != NULL, "entry must have been generated");
  __ jal_to_C(entry);
  __ delayed()->nop();
}


void AbstractInterpreterGenerator::stop_interpreter_at() {
  __ li(Rscratch3, (address)&BytecodeCounter::_counter_value);
  __ lw(Rscratch3, 0, Rscratch3);
  __ li(Rscratch4, (address)&StopInterpreterAt);
  __ ld_ptr(Rscratch4, 0, Rscratch4);
  __ teq(Rscratch3, Rscratch4);
}
#endif // not PRODUCT
