#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)jniFastGetField_sparc.cpp	1.10 07/05/05 17:04:29 JVM"
#endif
/*
 * Copyright 2004-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_jniFastGetField_mips.cpp.incl"

#define __ masm->

#define BUFFER_SIZE 30*wordSize

// Instead of issuing a SYNC for LoadLoad barrier, we create data dependency
// between loads, which is more efficient than lfence.

// Common register usage:
// A0        : env
// A1        : obj
// A2        : jfieldID
// A4        : offset (A2 >> 2)
// Rscratch4 : old safepoint counter

address JNI_FastGetField::generate_fast_get_int_field0(BasicType type) {
  const char *name;
  switch (type) {
    case T_BOOLEAN: name = "jni_fast_GetBooleanField"; break;
    case T_BYTE:    name = "jni_fast_GetByteField";    break;
    case T_CHAR:    name = "jni_fast_GetCharField";    break;
    case T_SHORT:   name = "jni_fast_GetShortField";   break;
    case T_INT:     name = "jni_fast_GetIntField";     break;
    case T_LONG:    name = "jni_fast_GetLongField";    break;
    default:        ShouldNotReachHere();
  }
  ResourceMark rm;
  BufferBlob* b = BufferBlob::create(name, BUFFER_SIZE);
  address fast_entry = b->instructions_begin();
  CodeBuffer cbuf(fast_entry, b->instructions_size());
  MacroAssembler* masm = new MacroAssembler(&cbuf);

  Label label1, label2;

  __ li(A3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(Rscratch4, 0, A3);
  __ andi(Rscratch0, Rscratch4, 1);
  __ bne(Rscratch0, R0, label1);
  __ delayed()->srl_long(A4, A2, 2);
  __ ld_ptr(A5, 0, A1);

  assert(count < LIST_CAPACITY, "LIST_CAPACITY too small");
  speculative_load_pclist[count] = __ pc();
  switch (type) {
    case T_BOOLEAN: __ lbu(Rscratch3, A5, A4);  break;
    case T_BYTE:    __ lb (Rscratch3, A5, A4);  break;
    case T_CHAR:    __ lhu(Rscratch3, A5, A4);  break;
    case T_SHORT:   __ lh (Rscratch3, A5, A4);  break;
    case T_INT:     __ lw (Rscratch3, A5, A4);  break;
    default:        ShouldNotReachHere();
  }

  __ li(A3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(A5, 0, A3);
  __ bne(A5, Rscratch4, label2);
  // YYY SPARC had a strange mov here I don't understand.
  // YYY It was O7, so it might be that they are saving RA.
  // YYY But they mov it to G1, is that safe across calls?
  __ delayed()->nop();
  __ jr(RA);
  __ delayed()->move(V0, Rscratch3);

  slowcase_entry_pclist[count++] = __ pc();
  __ bind(label1);
  // YYY sparc RA mov

  address slow_case_addr;
  switch (type) {
    case T_BOOLEAN: slow_case_addr = jni_GetBooleanField_addr(); break;
    case T_BYTE:    slow_case_addr = jni_GetByteField_addr();    break;
    case T_CHAR:    slow_case_addr = jni_GetCharField_addr();    break;
    case T_SHORT:   slow_case_addr = jni_GetShortField_addr();   break;
    case T_INT:     slow_case_addr = jni_GetIntField_addr();     break;
    default:        ShouldNotReachHere();
  }
  __ bind(label2);
  __ jal_to_C(slow_case_addr, relocInfo::none);
  // YYY sparc RA mov
  __ delayed()->nop();

  __ flush ();

  return fast_entry;
}

address JNI_FastGetField::generate_fast_get_boolean_field() {
  return generate_fast_get_int_field0(T_BOOLEAN);
}

address JNI_FastGetField::generate_fast_get_byte_field() {
  return generate_fast_get_int_field0(T_BYTE);
}

address JNI_FastGetField::generate_fast_get_char_field() {
  return generate_fast_get_int_field0(T_CHAR);
}

address JNI_FastGetField::generate_fast_get_short_field() {
  return generate_fast_get_int_field0(T_SHORT);
}

address JNI_FastGetField::generate_fast_get_int_field() {
  return generate_fast_get_int_field0(T_INT);
}

address JNI_FastGetField::generate_fast_get_long_field() {
  const char *name = "jni_fast_GetLongField";
  ResourceMark rm;
  BufferBlob* b = BufferBlob::create(name, BUFFER_SIZE*wordSize);
  address fast_entry = b->instructions_begin();
  CodeBuffer cbuf(fast_entry, b->instructions_size());
  MacroAssembler* masm = new MacroAssembler(&cbuf);

  Label label1, label2;

  __ li(Rscratch3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(Rscratch4, 0, Rscratch3);
  __ andi(Rscratch0, Rscratch4, 1);
  __ bne(Rscratch0, R0, label1);
  __ delayed()->srl_long(A4, A2, 2);
  __ ld_ptr(A5, 0, A1);
  __ addu_long(A5, A5, A4);

#ifndef _LP64
  assert(count < LIST_CAPACITY-1, "LIST_CAPACITY too small");
  speculative_load_pclist[count++] = __ pc();
  __ lw(Rscratch2, 0, A5);

  speculative_load_pclist[count] = __ pc();
  __ lw (A3, 4, A5);
#else
  assert(count < LIST_CAPACITY, "LIST_CAPACITY too small");
  speculative_load_pclist[count] = __ pc();
  __ ld(A3, 0, A5);
#endif

  __ li(Rscratch3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(Rscratch1, 0, Rscratch3);
  __ bne(Rscratch1, Rscratch4, label2);
  __ delayed()->move(Rscratch1, A7);

#ifndef _LP64
  __ move(V0, Rscratch2);
  __ jr(RA);
  __ delayed()->move(V1, A3);
#else
  __ jr(RA);
  __ delayed()->move(V0, A3);
#endif

#ifndef _LP64
  slowcase_entry_pclist[count-1] = __ pc();
  slowcase_entry_pclist[count++] = __ pc() ;
#else
  slowcase_entry_pclist[count++] = __ pc();
#endif

  __ bind(label1);
  // YYY sparc RA mov

  address slow_case_addr = jni_GetLongField_addr();
  __ bind(label2);
  __ jal_to_C(slow_case_addr, relocInfo::none);
  // YYY sparc RA mov
  __ delayed()->nop();

  __ flush ();

  return fast_entry;
}

address JNI_FastGetField::generate_fast_get_float_field0(BasicType type) {
  const char *name;
  switch (type) {
    case T_FLOAT:     name = "jni_fast_GetFloatField";     break;
    case T_DOUBLE:    name = "jni_fast_GetDoubleField";    break;
    default:          ShouldNotReachHere();
  }
  ResourceMark rm;
  BufferBlob* b = BufferBlob::create(name, BUFFER_SIZE);
  address fast_entry = b->instructions_begin();
  CodeBuffer cbuf(fast_entry, b->instructions_size());
  MacroAssembler* masm = new MacroAssembler(&cbuf);

  Label label1, label2;

  __ li(A3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(Rscratch4, 0, A3);
  __ andi(Rscratch0, Rscratch4, 1);
  __ bne(Rscratch0, R0, label1);
  __ delayed()->srl_long(A4, A2, 2);
  __ ld_ptr(A5, 0, A1);

  assert(count < LIST_CAPACITY, "LIST_CAPACITY too small");
  speculative_load_pclist[count] = __ pc();
  switch (type) {
    case T_FLOAT:  __ lwxc1(F0, A5, A4); break;
    case T_DOUBLE: __ ldxc1(F0, A5, A4); break;
    default:       ShouldNotReachHere();
  }

  __ li(A3, SafepointSynchronize::safepoint_counter_addr());
  __ lw(A5, 0, A3);
  __ bne(A5, Rscratch4, label2);
  __ delayed()->nop();
  // YYY sparc RA mov

  __ jr(RA);
  __ delayed()->nop();

  slowcase_entry_pclist[count++] = __ pc();
  __ bind(label1);
  // YYY sparc RA mov

  address slow_case_addr;
  switch (type) {
    case T_FLOAT:  slow_case_addr = jni_GetFloatField_addr();  break;
    case T_DOUBLE: slow_case_addr = jni_GetDoubleField_addr(); break;
    default:       ShouldNotReachHere();
  }
  __ bind (label2);
  __ jal_to_C(slow_case_addr, relocInfo::none);
  // YYY sparc RA mov
  __ delayed()->nop();

  __ flush ();

  return fast_entry;
}

address JNI_FastGetField::generate_fast_get_float_field() {
  return generate_fast_get_float_field0(T_FLOAT);
}

address JNI_FastGetField::generate_fast_get_double_field() {
  return generate_fast_get_float_field0(T_DOUBLE);
}

