#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)nativeInst_sparc.cpp	1.97 07/05/05 17:04:31 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_nativeInst_mips.cpp.incl"


void NativeInstruction::verify() {
  // not much we can do without expensive checking of all possible
  // patterns.
}

void NativeInstruction::print() {
  tty->print_cr(INTPTR_FORMAT ": 0x%x", addr_at(0), long_at(0));
}

void NativeInstruction::set_long_at(int offset, int i) {
  address addr = addr_at(offset);
  *(int*)addr = i;
  ICache::invalidate_word(addr);
}


// Similar to replace_mt_safe, but just changes the destination.  The
// important thing is that free-running threads are able to execute this
// call instruction at all times.  Thus, the displacement field must be
// instruction-word-aligned.  This is always true on SPARC.
//
// Used in the runtime linkage of calls; see class CompiledIC.
void NativeCall::set_destination_mt_safe(address dest) {
  assert(Patching_lock->is_locked() ||
         SafepointSynchronize::is_at_safepoint(), "concurrent code patching");
  Unimplemented();
  //// set_destination uses set_long_at which does the ICache::invalidate
  //set_destination(dest);
}

// Code for unit testing implementation of NativeCall class
// End code for unit testing implementation of NativeCall class

//-------------------------------------------------------------------

#ifdef _LP64

void NativeFarCall::set_destination(address dest) {
  int i0 = long_at(0);                     // lui
  int i1 = long_at(1 * BytesPerInstWord);  // ori
  int i2 = long_at(2 * BytesPerInstWord);  // dsll
  int i3 = long_at(3 * BytesPerInstWord);  // ori
  int i4 = long_at(4 * BytesPerInstWord);  // dsll
  int i5 = long_at(5 * BytesPerInstWord);  // ori
  int i6 = long_at(6 * BytesPerInstWord);  // jr or jalr
  verify();
  long desti = (long)dest;
  i0 = set_data32_imm16(i0, (desti >> 48) & 0xFFFF);
  i1 = set_data32_imm16(i0, (desti >> 32) & 0xFFFF);
  i3 = set_data32_imm16(i0, (desti >> 16) & 0xFFFF);
  i5 = set_data32_imm16(i0,  desti        & 0xFFFF);
  set_long_at(0,                    i0);
  set_long_at(1 * BytesPerInstWord, i1);
  set_long_at(3 * BytesPerInstWord, i3);
  set_long_at(5 * BytesPerInstWord, i5);
}

void NativeFarCall::verify() const {
  assert(is_call_at_helper(), "invalid NativeFarCall");
}

bool NativeFarCall::is_call_at(address instr) {
  return nativeFarCall_at(instr)->is_call_at_helper();
}

void NativeFarCall::print() {
  tty->print_cr(INTPTR_FORMAT ": call " INTPTR_FORMAT, addr_at(0), destination());
}

bool NativeFarCall::is_call_at_helper() const {
  int i0 = long_at(0);                     // lui
  int i1 = long_at(1 * BytesPerInstWord);  // ori
  int i2 = long_at(2 * BytesPerInstWord);  // dsll
  int i3 = long_at(3 * BytesPerInstWord);  // ori
  int i4 = long_at(4 * BytesPerInstWord);  // dsll
  int i5 = long_at(5 * BytesPerInstWord);  // ori
  int i6 = long_at(6 * BytesPerInstWord);  // jr or jalr
  return op(i0) == Assembler::op_lui &&
	 op(i1) == Assembler::op_ori &&
	 op(i2) == Assembler::op_special && funct(i2) == Assembler::func_dsll &&
	 op(i3) == Assembler::op_ori &&
	 op(i4) == Assembler::op_special && funct(i4) == Assembler::func_dsll &&
	 op(i5) == Assembler::op_ori &&
	 op(i6) == Assembler::op_special && (funct(i6) == Assembler::func_jr ||funct(i6) == Assembler::func_jalr) &&
	 rs(i0) == rs(i1) == rs(i2) == rs(i3) == rs(i4) == rs(i5) == rs(i6);
}

// Code for unit testing implementation of NativeFarCall class
// End code for unit testing implementation of NativeFarCall class

#else
#error unimplemented
#endif // _LP64

//-------------------------------------------------------------------


#ifdef _LP64
intptr_t NativeMovConstReg::data() const {
  Unimplemented();
  return 0;
}
#else
#error unimplemented
#endif


void NativeMovConstReg::set_data(intptr_t x) {
#ifdef _LP64
  Unimplemented();
#else
#error unimplemented
#endif

  // also store the value into an oop_Relocation cell, if any
  CodeBlob* nm = CodeCache::find_blob(instruction_address());
  if (nm != NULL) {
    if (UseCompiler) {		// YYY unimplemented
    RelocIterator iter(nm, instruction_address(), next_instruction_address());
    oop* oop_addr = NULL;
    while (iter.next()) {
      if (iter.type() == relocInfo::oop_type) {
	oop_Relocation *r = iter.oop_reloc();
	if (oop_addr == NULL) {
	  oop_addr = r->oop_addr();
	  *oop_addr = (oop)x;
	} else {
	  assert(oop_addr == r->oop_addr(), "must be only one set-oop here");
	}
      }
    }
    }
  }
}


// Code for unit testing implementation of NativeMovConstReg class
// End code for unit testing implementation of NativeMovConstReg class

//-------------------------------------------------------------------



// Code for unit testing implementation of NativeMovRegMem class
// End code for unit testing implementation of NativeMovRegMem class

//--------------------------------------------------------------------------------


// Code for unit testing implementation of NativeJump class
// End code for unit testing implementation of NativeJump class


// MT safe inserting of a jump over an unknown instruction sequence (used by nmethod::makeZombie)
// The problem: jump_to <dest> is a 3-word instruction (including its delay slot).
// Atomic write can be only with 1 word.
void NativeJump::patch_verified_entry(address entry, address verified_entry, address dest) {
  Unimplemented();
}


void NativeIllegalInstruction::insert(address code_pos) {
  NativeIllegalInstruction* nii = (NativeIllegalInstruction*) nativeInstruction_at(code_pos);
  nii->set_long_at(0, illegal_instruction());
}

static int illegal_instruction_bits = 0;

int NativeInstruction::illegal_instruction() {
  if (illegal_instruction_bits == 0) {
    ResourceMark rm;
    char buf[40];
    CodeBuffer cbuf((address)&buf[0], 20);
    MacroAssembler* a = new MacroAssembler(&cbuf);
    address ia = a->pc();
    a->break_();
    int bits = *(int*)ia;
// YYY: too many undefines in the following line: is_op3, trap_op3, arith_op
//    assert(is_op3(bits, Assembler::trap_op3, Assembler::arith_op), "bad instruction");
    illegal_instruction_bits = bits;
    assert(illegal_instruction_bits != 0, "oops");
  }
  return illegal_instruction_bits;
}


void NativeGeneralJump::verify() {
  Unimplemented();
  // assert(((NativeInstruction *)this)->is_branch(), "not a general jump instruction");
}


void NativeGeneralJump::insert_unconditional(address code_pos, address entry) {
  Unimplemented();
}


// MT-safe patching of a jmp instruction (and following word).
// First patches the second word, and then atomicly replaces
// the first word with the first new instruction word.
// Other processors might briefly see the old first word
// followed by the new second word.  This is OK if the old
// second word is harmless, and the new second word may be
// harmlessly executed in the delay slot of the call.
void NativeGeneralJump::replace_mt_safe(address instr_addr, address code_buffer) {
  Unimplemented();
}

