#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)nativeInst_sparc.hpp	1.89 07/05/05 17:04:31 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// We have interface for the following instructions:
// - NativeInstruction
// - - NativeCall
// - - NativeFarCall
// - - NativeMovConstReg
// - - NativeMovRegMem
// - - NativeJump
// - - NativeGeneralJump
// - - NativeIllegalInstruction
// The base class for different kinds of native instruction abstractions.
// Provides the primitive operations to manipulate code relative to this.
class NativeInstruction VALUE_OBJ_CLASS_SPEC {
  friend class Relocation;

 public:
  bool is_jump() {
    Unimplemented();
    return false;
  }

  bool is_safepoint_poll() {
#ifdef _LP64
    Unimplemented();
    return false;
#else
#error unimplemented
#endif
  }

  static int illegal_instruction();    // the output of __ breakpoint_trap()

 protected:
  address  addr_at(int offset) const    { return address(this) + offset; }
  int      long_at(int offset) const    { return *(int*)addr_at(offset); }
  void set_long_at(int offset, int i);      /* deals with I-cache */
  // Once again "long" is a big misnomer -- it's an int-sized thing.
  // Name only preserved for similarity with other platforms.

  static int set_data32_imm16(int imm16_insn, int imm16) {
    return (imm16_insn & ~ 0xFFFF) | imm16;
  }

  static int data32_imm16(int imm16_insn) {
    return imm16_insn & 0xFFFF;
  }

  static int op(int i)    { return Assembler::inv_u_field(i, 31, 26); }
  static int rs(int i)    { return Assembler::inv_u_field(i, 25, 21); }
  static int rt(int i)    { return Assembler::inv_u_field(i, 20, 16); }
  static int imm16(int i) { return Assembler::inv_u_field(i, 15, 0); }
  static int funct(int i) { return Assembler::inv_u_field(i, 5, 0); }

 public:
  void  verify();
  void  print();

  inline friend NativeInstruction* nativeInstruction_at(address address);
};

inline NativeInstruction* nativeInstruction_at(address address) {
    NativeInstruction* inst = (NativeInstruction*)address;
#ifdef ASSERT
      inst->verify();
#endif
    return inst;
}



//-----------------------------------------------------------------------------

// The NativeCall is an abstraction for accessing/manipulating native
// call immediate instructions.
// (used to manipulate inline caches, primitive & dll calls, etc.)
inline NativeCall* nativeCall_at(address instr);
inline NativeCall* nativeCall_overwriting_at(address instr,
                                             address destination);
inline NativeCall* nativeCall_before(address return_address);
class NativeCall: public NativeInstruction {
 public:
  enum {
    instruction_size = -1	// YYY unimplemented
  };
  address instruction_address() const       { return addr_at(0); }
  address next_instruction_address() const  { return addr_at(instruction_size); }
  address return_address() const            { Unimplemented(); return 0; }

  address destination() const               { Unimplemented(); return 0; }
  void set_destination_mt_safe(address dest);

  void  verify_alignment() {}	// do nothing on mips

  // Creation
  friend inline NativeCall* nativeCall_at(address instr);
  friend NativeCall* nativeCall_overwriting_at(address instr, address destination = NULL) {
    // insert a "blank" call:
    NativeCall* call = (NativeCall*)instr;
    Unimplemented();
    // check its structure now:
    assert(nativeCall_at(instr)->destination() == destination, "correct call destination");
    return call;
  }

  friend inline NativeCall* nativeCall_before(address return_address) {
    NativeCall* call;
    Unimplemented();
#ifdef ASSERT
    call->verify();
#endif
    return call;
  }

  static bool is_call_before(address instr) {
    Unimplemented();
    return false;
  }
};
inline NativeCall* nativeCall_at(address instr) {
  NativeCall* call = (NativeCall*)instr;
  Unimplemented();
#ifdef ASSERT
  call->verify();
#endif
  return call;
}

// The NativeFarCall is an abstraction for accessing/manipulating native call-anywhere
// instructions in the mips vm.  Used to call native methods which may be loaded
// anywhere in the address space, possibly out of reach of a call instruction.

#ifndef _LP64
#error unimplemented
#else

// The format of this extended-range call is:
//      jumpl_to addr, lreg
// That is, it is essentially the same as a NativeJump.
//
// This is the most complicated one.  A far call is a dli followed by
// j[al]r.  The dli expands to a bunch of instructions, so the end
// result looks like
//
//   lui  r, <imm>
//   ori  r, r, <imm>
//   dsll r, r, <imm>
//   ori  r, r, <imm>
//   dsll r, r, <imm>
//   ori  r, r, <imm>
//   jr   r
class NativeFarCall;
inline NativeFarCall* nativeFarCall_at(address instr);
class NativeFarCall: public NativeInstruction {
 public:
  enum {
    instruction_size = 7 * BytesPerInstWord, // see above
  };

  address destination() const {
    int i0 = long_at(0);                     // lui
    int i1 = long_at(1 * BytesPerInstWord);  // ori
    //int i2 = long_at(2 * BytesPerInstWord);  // dsll
    int i3 = long_at(3 * BytesPerInstWord);  // ori
    //int i4 = long_at(4 * BytesPerInstWord);  // dsll
    int i5 = long_at(5 * BytesPerInstWord);  // ori
    //int i6 = long_at(6 * BytesPerInstWord);  // jr or jalr
    verify();
    return address(
		   long(imm16(i0)) << 48 | long(imm16(i1)) << 32 |
		   long(imm16(i3)) << 16 | long(imm16(i5))
		   );
  }
  void set_destination(address dest);

  void  verify() const;
  void  print();

  // Creation
  friend inline NativeFarCall* nativeFarCall_at(address instr) {
    NativeFarCall* call = (NativeFarCall*)instr;
#ifdef ASSERT
      call->verify();
#endif
    return call;
  }

  static bool is_call_at(address instr);

 protected:
  bool is_call_at_helper() const;
};

#endif // _LP64

// An interface for accessing/manipulating native set_oop imm, reg instructions.
// (used to manipulate inlined data references, etc.)
//      set_oop imm, reg
class NativeMovConstReg;
inline NativeMovConstReg* nativeMovConstReg_at(address address);
class NativeMovConstReg: public NativeInstruction {
 public:
  enum {
#ifdef _LP64
    instruction_size = 6 * BytesPerInstWord, // lui; ori; shl; ori; shl; ori
#else
    instruction_size = 2 * BytesPerInstWord, // lui; ori
#endif
  };

  address instruction_address() const       { return addr_at(0); }
  address next_instruction_address() const  { return addr_at(instruction_size); }

  // (The [set_]data accessor respects oop_type relocs also.)
  intptr_t data() const;
  void set_data(intptr_t x);

  // Creation
  friend inline NativeMovConstReg* nativeMovConstReg_at(address address) {
    NativeMovConstReg* test = (NativeMovConstReg*)address;
    Unimplemented();
#ifdef ASSERT
    test->verify();
#endif
    return test;
  }
};


// An interface for accessing/manipulating native memory ops
class NativeMovRegMem;
inline NativeMovRegMem* nativeMovRegMem_at(address address);
class NativeMovRegMem: public NativeInstruction {
 public:
  intptr_t offset() const {
    Unimplemented();
    return 0;
  }
  void set_offset(intptr_t x) {
    Unimplemented();
  }

  void add_offset_in_bytes(intptr_t radd_offset) {
    Unimplemented();
  }

 private:
  friend inline NativeMovRegMem* nativeMovRegMem_at (address address) {
    NativeMovRegMem* test = (NativeMovRegMem*)address;
    Unimplemented();
#ifdef ASSERT
    test->verify();
#endif
    return test;
  }
};


// An interface for accessing/manipulating native jumps
//      jump_to addr
//      jumpl_to addr, lreg
class NativeJump;
inline NativeJump* nativeJump_at(address address);
class NativeJump: public NativeInstruction {
 public:
  enum {
#ifdef _LP64
    instruction_size = 0,	// YYY unimplemented
#else
#error unimplemented
#endif
  };

#ifdef _LP64
  address jump_destination() const {
    Unimplemented();
    return 0;
  }
  void set_jump_destination(address dest) {
    Unimplemented();
  }
#else
#error unimplemented
#endif

  // Creation
  friend inline NativeJump* nativeJump_at(address address) {
    NativeJump* jump = (NativeJump*)address;
    Unimplemented();
#ifdef ASSERT
    jump->verify();
#endif
    return jump;
  }

  // MT-safe insertion of native jump at verified method entry
  static void check_verified_entry_alignment(address entry, address verified_entry) {
    Unimplemented();
    // nothing to do for mips?
  }
  static void patch_verified_entry(address entry, address verified_entry, address dest);
};


// All of the local relative-offset branches.
// Despite the name, handles only simple branches.
class NativeGeneralJump;
inline NativeGeneralJump* nativeGeneralJump_at(address address);
class NativeGeneralJump: public NativeInstruction {
 public:
  enum {
    instruction_size = -1,	// YYY unimplemented
  };

  address jump_destination() const {
    return addr_at(0) + data32_imm16(long_at(0));
  }
  void set_jump_destination(address dest) {
    Unimplemented();
  }

  // Creation
  friend inline NativeGeneralJump* nativeGeneralJump_at(address address) {
    NativeGeneralJump* jump = (NativeGeneralJump*)(address);
#ifdef ASSERT
    jump->verify();
#endif
    return jump;
  }

  // Insertion of native general jump instruction
  static void insert_unconditional(address code_pos, address entry);
  static void replace_mt_safe(address instr_addr, address code_buffer);

  void verify();
};


class NativeIllegalInstruction: public NativeInstruction {
 public:
  enum {
    instruction_size = sizeof(NativeInstruction::illegal_instruction()),
  };

  // Insert illegal opcode as specific address
  static void insert(address code_pos);
};
