#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)register_sparc.cpp	1.16 07/05/05 17:04:30 JVM"
#endif
/*
 * Copyright 2000-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_register_mips.cpp.incl"

const int ConcreteRegisterImpl::max_gpr = RegisterImpl::number_of_registers;
const int ConcreteRegisterImpl::max_fpr = (ConcreteRegisterImpl::max_gpr +
					   FloatRegisterImpl::number_of_registers);

const char* RegisterImpl::name() const {
  const char* names[number_of_registers] = {
#if 0 // raw names are less useful in general than the convention ones
    "R0",  "R1",  "R2",  "R3",  "R4",  "R5",  "R6",  "R7",
    "R8",  "R9",  "R10", "R11", "R12", "R13", "R14", "R15",
    "R16", "R17", "R18", "R19", "R20", "R21", "R22", "R23",
    "R24", "R25", "R26", "R27", "R28", "R29", "R30", "R31"
#else
    "R0", "AT", "V0", "V1", "A0", "A1", "A2", "A3",
    "A4", "A5", "A6", "A7", "T0", "T1", "T2", "T3",
    "S0", "S1", "S2", "S3", "S4", "S5", "S6", "S7",
    "T8", "T9", "K0", "K1", "GP", "SP", "FP", "RA"
#endif
  };
  return is_valid() ? names[encoding()] : "noreg";
}


const char* FloatRegisterImpl::name() const {
  const char* names[number_of_registers] = {
    "F0",  "F1",   "F2",  "F3",   "F4",  "F5",   "F6",  "F7",   "F8",  "F9",
    "F10", "F11",  "F12", "F13",  "F14", "F15",  "F16", "F17",  "F18", "F19",
    "F20", "F21",  "F22", "F23",  "F24", "F25",  "F26", "F27",  "F28", "F29",
    "F30", "F31"
  };
  return is_valid() ? names[encoding()] : "fnoreg";
}

int RegisterImpl::sp_offset_in_save_area() {
  int offset = frame::GPR_saves_offset;
  if (this == RA) {
    return offset;
  } else if (this == GP) {
    return offset + BytesPerLong;
  } else if (encoding() >= S0->encoding() &&
	     encoding() <= S7->encoding()) {
    return
      offset +
      (2 * BytesPerLong) +
      ((encoding() - S0->encoding()) * BytesPerLong);
  } else if (this == FP) {
    return offset + (10 * BytesPerLong);
  }
  ShouldNotReachHere();
}

int FloatRegisterImpl::sp_offset_in_save_area() {
  int offset = frame::FPR_saves_offset;
  if (encoding() >= F24->encoding() &&
      encoding() <= F31->encoding()) {
    return (encoding() - F24->encoding()) * BytesPerLong;
  }
  ShouldNotReachHere();
}
