#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)register_sparc.hpp	1.27 07/05/05 17:04:30 JVM"
#endif
/*
 * Copyright 2000-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// forward declaration
class Address;
class VMRegImpl;
typedef VMRegImpl* VMReg;

// Use Register as shortcut
class RegisterImpl;
typedef RegisterImpl* Register;


// Conversion to the phony class type
inline Register as_Register(int encoding) {
  return (Register)(intptr_t) encoding;
}

// General-purpose registers on the MIPS architecture
class RegisterImpl: public AbstractRegisterImpl {
 public:
  enum {
    number_of_registers      = 32
    // There are also hi and lo, but only implicitly referenced
  };


  // construction
  friend Register as_Register(int encoding);

  VMReg as_VMReg();

  // accessors
  int   encoding() const				{ assert(is_valid(), "invalid register"); return value(); }
  const char* name() const;

  // testers
  bool is_valid() const {
    return 0 <= value() && value() < number_of_registers;
  }

  Register successor() const {
    return as_Register(encoding() + 1);
  }

  int sp_index_in_save_area() {
    return sp_offset_in_save_area() >> LogBytesPerWord;
  }

  int sp_offset_in_save_area();
};


// The integer registers of the MIPS architecture

CONSTANT_REGISTER_DECLARATION(Register, noreg , (-1));

CONSTANT_REGISTER_DECLARATION(Register, R0,  0);
CONSTANT_REGISTER_DECLARATION(Register, R1,  1);
CONSTANT_REGISTER_DECLARATION(Register, R2,  2);
CONSTANT_REGISTER_DECLARATION(Register, R3,  3);
CONSTANT_REGISTER_DECLARATION(Register, R4,  4);
CONSTANT_REGISTER_DECLARATION(Register, R5,  5);
CONSTANT_REGISTER_DECLARATION(Register, R6,  6);
CONSTANT_REGISTER_DECLARATION(Register, R7,  7);
CONSTANT_REGISTER_DECLARATION(Register, R8,  8);
CONSTANT_REGISTER_DECLARATION(Register, R9,  9);
CONSTANT_REGISTER_DECLARATION(Register, R10, 10);
CONSTANT_REGISTER_DECLARATION(Register, R11, 11);
CONSTANT_REGISTER_DECLARATION(Register, R12, 12);
CONSTANT_REGISTER_DECLARATION(Register, R13, 13);
CONSTANT_REGISTER_DECLARATION(Register, R14, 14);
CONSTANT_REGISTER_DECLARATION(Register, R15, 15);
CONSTANT_REGISTER_DECLARATION(Register, R16, 16);
CONSTANT_REGISTER_DECLARATION(Register, R17, 17);
CONSTANT_REGISTER_DECLARATION(Register, R18, 18);
CONSTANT_REGISTER_DECLARATION(Register, R19, 19);
CONSTANT_REGISTER_DECLARATION(Register, R20, 20);
CONSTANT_REGISTER_DECLARATION(Register, R21, 21);
CONSTANT_REGISTER_DECLARATION(Register, R22, 22);
CONSTANT_REGISTER_DECLARATION(Register, R23, 23);
CONSTANT_REGISTER_DECLARATION(Register, R24, 24);
CONSTANT_REGISTER_DECLARATION(Register, R25, 25);
CONSTANT_REGISTER_DECLARATION(Register, R26, 26);
CONSTANT_REGISTER_DECLARATION(Register, R27, 27);
CONSTANT_REGISTER_DECLARATION(Register, R28, 28);
CONSTANT_REGISTER_DECLARATION(Register, R29, 29);
CONSTANT_REGISTER_DECLARATION(Register, R30, 30);
CONSTANT_REGISTER_DECLARATION(Register, R31, 31);

#ifndef DONT_USE_REGISTER_DEFINES
// move here from other headers?
#endif // DONT_USE_REGISTER_DEFINES

REGISTER_DECLARATION(Register, AT, R1);
REGISTER_DECLARATION(Register, V0, R2);
REGISTER_DECLARATION(Register, V1, R3);
REGISTER_DECLARATION(Register, A0, R4);
REGISTER_DECLARATION(Register, A1, R5);
REGISTER_DECLARATION(Register, A2, R6);
REGISTER_DECLARATION(Register, A3, R7);
REGISTER_DECLARATION(Register, A4, R8);
REGISTER_DECLARATION(Register, A5, R9);
REGISTER_DECLARATION(Register, A6, R10);
REGISTER_DECLARATION(Register, A7, R11);
REGISTER_DECLARATION(Register, T0, R12);
REGISTER_DECLARATION(Register, T1, R13);
REGISTER_DECLARATION(Register, T2, R14);
REGISTER_DECLARATION(Register, T3, R15);
REGISTER_DECLARATION(Register, S0, R16);
REGISTER_DECLARATION(Register, S1, R17);
REGISTER_DECLARATION(Register, S2, R18);
REGISTER_DECLARATION(Register, S3, R19);
REGISTER_DECLARATION(Register, S4, R20);
REGISTER_DECLARATION(Register, S5, R21);
REGISTER_DECLARATION(Register, S6, R22);
REGISTER_DECLARATION(Register, S7, R23);
REGISTER_DECLARATION(Register, T8, R24);
REGISTER_DECLARATION(Register, T9, R25);
REGISTER_DECLARATION(Register, K0, R26);
REGISTER_DECLARATION(Register, K1, R27);
REGISTER_DECLARATION(Register, GP, R28);
REGISTER_DECLARATION(Register, SP, R29);
REGISTER_DECLARATION(Register, FP, R30);
REGISTER_DECLARATION(Register, RA, R31);

// Use FloatRegister as shortcut
class FloatRegisterImpl;
typedef FloatRegisterImpl* FloatRegister;


// construction
inline FloatRegister as_FloatRegister(int encoding) {
  return (FloatRegister)(intptr_t) encoding;
}

// The implementation of float registers for the MIPS architecture

class FloatRegisterImpl: public AbstractRegisterImpl {
 public:
  enum {
    number_of_registers = 32
  };

  // construction
  inline friend FloatRegister as_FloatRegister(int encoding);
  VMReg as_VMReg();

  // accessors
  int   encoding() const                               { assert(is_valid(), "invalid register"); return value(); }

 public:
  bool  is_valid() const                               { return 0 <= value() && value() < number_of_registers; }
  const char* name() const;

  FloatRegister successor() const                      { return as_FloatRegister(encoding() + 1); }

  int sp_index_in_save_area() {
    return sp_offset_in_save_area() >> LogBytesPerWord;
  }
  int sp_offset_in_save_area();
};


// The float registers of the MIPS architecture

CONSTANT_REGISTER_DECLARATION(FloatRegister, fnoreg , (-1));

CONSTANT_REGISTER_DECLARATION(FloatRegister, F0     , ( 0));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F1     , ( 1));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F2     , ( 2));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F3     , ( 3));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F4     , ( 4));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F5     , ( 5));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F6     , ( 6));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F7     , ( 7));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F8     , ( 8));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F9     , ( 9));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F10    , (10));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F11    , (11));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F12    , (12));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F13    , (13));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F14    , (14));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F15    , (15));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F16    , (16));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F17    , (17));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F18    , (18));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F19    , (19));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F20    , (20));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F21    , (21));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F22    , (22));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F23    , (23));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F24    , (24));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F25    , (25));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F26    , (26));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F27    , (27));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F28    , (28));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F29    , (29));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F30    , (30));
CONSTANT_REGISTER_DECLARATION(FloatRegister, F31    , (31));


#ifndef DONT_USE_REGISTER_DEFINES
// move here from other headers?
#endif // DONT_USE_REGISTER_DEFINES

// Need to know the total number of registers of all sorts for SharedInfo.
// Define a class that exports it.
class ConcreteRegisterImpl : public AbstractRegisterImpl {
 public:
  enum {
    // A big enough number for C2: all the registers plus flags
    // This number must be large enough to cover REG_COUNT (defined by c2) registers.
    // There is no requirement that any ordering here matches any ordering c2 gives
    // it's optoregs.
    number_of_registers = (RegisterImpl::number_of_registers +
			   FloatRegisterImpl::number_of_registers)
  };
  static const int max_gpr;
  static const int max_fpr;

};

