#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)stubGenerator_sparc.cpp	1.234 07/07/12 11:24:37 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *   
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *   
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *  
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_stubGenerator_mips.cpp.incl"

// Declaration and definition of StubGenerator (no .hpp file).
// For a more detailed description of the stub routine structure
// see the comment in stubRoutines.hpp.

#define __ _masm->

#ifdef PRODUCT
#define BLOCK_COMMENT(str) /* nothing */
#else
#define BLOCK_COMMENT(str) __ block_comment(str)
#endif

#define BIND(label) bind(label); BLOCK_COMMENT(#label ":")


// -------------------------------------------------------------------------------------------------------------------------
// Stub Code definitions

static address handle_unsafe_access() {
  JavaThread* thread = JavaThread::current();
  address pc  = thread->saved_exception_pc();
  address npc = Assembler::locate_next_instruction(pc);
  // pc is the instruction which we must emulate
  // doing a no-op is fine:  return garbage from the load

  // request an async exception
  thread->set_pending_unsafe_access_error();

  // return address of next instruction to execute
  return npc;
}


class StubGenerator: public StubCodeGenerator {
 private:

#ifdef PRODUCT
#define inc_counter_np(a,b,c) (0)
#else
  void inc_counter_np_(int& counter, Register t1, Register t2) {
    __ li(t2, counter);
    __ lw(t1, 0, t2);
    __ addu_long(t1, t1, 1);
    __ sw(t1, 0, t2);
  }
#define inc_counter_np(counter, t1, t2) \
  BLOCK_COMMENT("inc_counter " #counter); \
  inc_counter_np_(counter, t1, t2);
#endif

  //----------------------------------------------------------------------------------------------------
  // Call stubs are used to call Java from C

  address generate_call_stub(address& return_pc) {
    StubCodeMark mark(this, "StubRoutines", "call_stub");
    address start = __ pc();

    // Incoming arguments:
    //
    // a0         : call wrapper address
    // a1         : result (address)
    // a2         : result type
    // a3         : method
    // a4         : (interpreter) entry point
    // a5         : parameters (address)
    // a6         : parameter size (in words)
    // a7         : thread

    // note: if the link argument position changes, adjust
    //       the code in frame::entry_frame_call_wrapper()

    const Register link           = Argument(0).as_register(); // used only for GC
    const Register result         = Argument(1).as_register();
    const Register result_type    = Argument(2).as_register();
    const Register method         = Argument(3).as_register();
    const Register entry_point    = Argument(4).as_register();
    const Register parameters     = Argument(5).as_register();
    const Register parameter_size = Argument(6).as_register();
    const Register thread         = Argument(7).as_register();
    const Register parm_cnt       = Rscratch1;

    assert_different_registers(Rargs, entry_point, thread);

    // Push a frame with space for the arguments.
    // The frame needs to resemble an interpreter frame caller, i.e.
    // we will put the Java arguments in a form the same as what they
    // would look like in a bytecode stack of an interpreter frame.
    //
    // | [caller]   | <-- FP
    // +------------+
    // |0           |     (outgoing stack args size)
    // |link save   |
    // +------------+
    // |javaarg 1   |
    // | ...        |
    // |javaarg N   |
    // +------------+
    // | align pad  | <-- Resp, Rargs
    // +------------+
    // | GPR saves  |
    // | ...        |
    // +------------+
    // | FPR saves  |
    // | ...        | <-- SP
    // +------------+
    // 

    __ move(parm_cnt, parameter_size);
    __ sll_long(Rscratch0, Rscratch1, Interpreter::logStackElementSize());
    __ addu_long(Rscratch0, Rscratch0, frame::register_save_space + (frame::call_stub_locals_words*wordSize));
    __ round_to(Rscratch0, 2 * wordSize);

    __ subu_long(SP, SP, Rscratch0);
    __ save_registers();
    __ addu_long(FP, SP, Rscratch0);

#ifdef ASSERT
    // Save SP so when we come back from the call, we can verify it.
    __ move(Rsave0, SP);
#endif
    // Save arguments that we will need after the call.
    __ move(S1, result);
    __ move(S2, result_type);

    __ st_ptr(R0, frame::call_stub_args_size_offset * BytesPerWord, FP);

    // Store "link"; may be needed by GC
    __ st_ptr(link, frame::call_stub_wrapper_offset * BytesPerWord, FP);

    // Copy over the arguments
    Label exit, loop;
    const Register src = parameters;
    const Register dst = Resp;

    __ beq(parm_cnt, R0, exit);
    __ delayed()->addu_long(dst, FP, -((frame::call_stub_locals_words + 1) * BytesPerWord));  // set up dst

    __ bind(loop);
    if (TaggedStackInterpreter) {
      __ ld_ptr(Rscratch0, 0, src);
      __ addu_long(src, src, BytesPerWord);
      __ st_ptr(Rscratch0, Interpreter::tag_offset_in_bytes(), dst);
    }
    // Store parameter value itself
    __ ld_ptr(Rscratch0, 0, src);
    __ addu_long(src, src, BytesPerWord);
    __ st_ptr(Rscratch0, Interpreter::value_offset_in_bytes(), dst);
    __ addu_long(parm_cnt, parm_cnt, -1);
    __ bne(parm_cnt, R0, loop);
    __ delayed()->addu_long(dst, dst, - Interpreter::stackElementSize());

    __ bind(exit);
    // At the end, dst will be pointing beyond the end of the params
    // on the stack.  That's expected by the entry stub.
    __ st_ptr(dst, frame::call_stub_Resp_offset * BytesPerWord, FP);

    // setup thread register
    __ move(Rthread, thread);

#ifdef ASSERT
    // make sure we have no pending exceptions
    {
      const Register t = Rscratch3;
      Label L;
      __ ld_ptr(t, in_bytes(Thread::pending_exception_offset()), Rthread);
      __ beq(t, R0, L);
      __ delayed()->nop();
      __ stop("StubRoutines::call_stub: entered with pending exception");
      __ bind(L);
    }
#endif

    __ move(Rargs, dst);
    __ jalr(entry_point);
    __ delayed()->move(Rmethod_call, method);

    return_pc = __ pc();

#ifdef ASSERT
    {
      Label ok;
      __ beq(Rsave0, SP, ok);
      __ delayed()->nop();
      __ stop("StubRoutines::call_stub: SP invalid coming back from entry call");
      __ bind(ok);
    }
#endif
    // Restore saved arguments
    __ move(result, S1);
    __ move(result_type, S2);

    // restore the frame
    __ move(Rscratch0, FP);
    __ restore_registers();
    __ move(SP, Rscratch0);

    {
      Label ret_float, exit;

#ifndef _LP64
#error code below does not handle longs
#endif

      __ li(Rscratch0, T_FLOAT);
      __ beq(result_type, Rscratch0, ret_float);  // dmfc1 to avoid sign-extend
      __ delayed()->li(Rscratch0, T_DOUBLE);
      __ beq(result_type, Rscratch0, ret_float);
      __ delayed()->nop();

      __ b(exit);
      __ delayed()->st_long(V0, 0, result);

      __ bind(ret_float);
      __ b(exit);
      __ delayed()->sdc1(F0, 0, result);

      __ bind(exit);
    }

    __ j(RA);
    __ delayed()->nop();

    return start;
  }


  //----------------------------------------------------------------------------------------------------
  // Return point for a Java call if there's an exception thrown in Java code.
  // The exception is caught and transformed into a pending exception stored in
  // JavaThread that can be tested from within the VM.
  //
  // Rexception: exception oop

  address generate_catch_exception() {
    StubCodeMark mark(this, "StubRoutines", "catch_exception");

    address start = __ pc();
    // verify that thread corresponds
    __ verify_thread();

    const Register tempreg = Rscratch0;

    // set pending exception
    __ verify_oop(Rexception);
    __ st_ptr(Rexception, in_bytes(Thread::pending_exception_offset()), Rthread);
    __ li(tempreg, (intptr_t)__FILE__);
    __ st_ptr(tempreg, in_bytes(Thread::exception_file_offset()), Rthread);
    __ li(tempreg, (intptr_t)__LINE__);
    __ st_ptr(tempreg, in_bytes(Thread::exception_line_offset()), Rthread);

    // complete return to VM
    assert(StubRoutines::_call_stub_return_address != NULL, "must have been generated before");

    __ li(tempreg, StubRoutines::_call_stub_return_address);
    __ j(tempreg);
    __ delayed()->nop();

    return start;
  }


  //----------------------------------------------------------------------------------------------------
  // Continuation point for runtime calls returning with a pending exception
  // The pending exception check happened in the runtime or native call stub
  // The pending exception in Thread is converted into a Java-level exception
  //
  // Contract with Java-level exception handler:
  //   A0 = exception
  //   A1 = throwing pc

  address generate_forward_exception() {
    StubCodeMark mark(this, "StubRoutines", "forward_exception");
    address start = __ pc();

    // Upon entry, RA has the return address returning into Java
    // (interpreted or compiled) code; i.e. the return address
    // becomes the throwing pc.

    const Register handler_reg = Rscratch0;
    const int pending_exception_offset
      = in_bytes(Thread::pending_exception_offset());

    __ get_thread();
    __ ld_ptr(Rexception, pending_exception_offset, Rthread);

#ifdef ASSERT
    // make sure that this code is only executed if there is a pending exception
    { Label L;
      __ bne(Rexception, R0, L);
      __ delayed()->nop();
      __ stop("StubRoutines::forward exception: no pending exception (1)");
      __ bind(L);
    }
#endif

    // compute exception handler into handler_reg
    __ verify_oop(Rexception);

    __ save_minimal_frame();
    BLOCK_COMMENT("call exception_handler_for_return_address");
    __ call_VM_leaf(noreg, CAST_FROM_FN_PTR(address, SharedRuntime::exception_handler_for_return_address), RA);
    __ move(handler_reg, V0);
    __ restore_minimal_frame();

    __ ld_ptr(Rexception, pending_exception_offset, Rthread);
    __ move(Rissuing_pc, RA);

#ifdef ASSERT
    // make sure exception is set
    { Label L;
      __ bne(Rexception, R0, L);
      __ delayed()->nop();
      __ stop("StubRoutines::forward exception: no pending exception (2)");
      __ bind(L);
    }
#endif

    // jump to exception handler
    __ j(handler_reg);
    // clear pending exception
    __ delayed()->st_ptr(R0, pending_exception_offset, Rthread);

    return start;
  }


  //------------------------------------------------------------------------------------------------------------------------
  // Continuation point for throwing of implicit exceptions that are not handled in
  // the current activation. Fabricates an exception oop and initiates normal
  // exception dispatching in this frame. Since we need to preserve callee-saved
  // values (currently only for C2, but done for C1 as well) we need a callee-saved
  // oop map and therefore have to make these stubs into RuntimeStubs
  // rather than BufferBlobs.
  // If the compiler needs all registers to be preserved between the fault
  // point and the exception handler then it must assume responsibility for that in
  // AbstractCompiler::continuation_for_implicit_null_exception or
  // continuation_for_implicit_division_by_zero_exception. All other implicit
  // exceptions (e.g., NullPointerException or AbstractMethodError on entry) are
  // either at call sites or otherwise assume that stack unwinding will be initiated,
  // so caller saved registers were assumed volatile in the compiler.

  // Since we generate this stub into a RuntimeStub (as described
  // above, because it needs to be properly traversed and ignored
  // during GC), we change the meaning of the "__" macro within this
  // method (only).
#undef __
#define __ masm->

  address generate_throw_exception(const char* name, address runtime_entry, bool restore_saved_exception_pc) {
#ifdef ASSERT
    int insts_size = VerifyThread ? 1 * K : 600;
#else
    int insts_size = VerifyThread ? 1 * K : 256;
#endif /* ASSERT */
    int locs_size  = 32;

    CodeBuffer      code(name, insts_size, locs_size);
    MacroAssembler* masm = new MacroAssembler(&code);

    __ verify_thread();

    // This is an inlined and slightly modified version of call_VM
    // which has the ability to fetch the return PC out of thread-local storage
    __ assert_not_delayed();

    // Note that we always push a frame because for all of our
    // implicit exception kinds at call
    // sites, the implicit exception is taken before the callee frame
    // is pushed.
    // YYY Verify that the above is the case on MIPS too.
    // YYY We might be able to get away with a minimal frame here, instead
    // YYY of the standard full one.
    __ save_frame();

    int frame_complete = __ offset();

    if (restore_saved_exception_pc) {
      __ ld_ptr(Rscratch1, in_bytes(JavaThread::saved_exception_pc_offset()), Rthread);
    }

    // Note that we always have a runtime stub frame on the top of stack by this point
    Register last_java_sp = SP;
    __ set_last_Java_frame(last_java_sp, FP, R0);
    __ li(Rscratch0, runtime_entry);
    __ jalr(Rscratch0, relocInfo::runtime_call_type);
    __ delayed()->move(A0, Rthread);

    // check for pending exceptions.
#ifdef ASSERT
    Label L;

    const int pending_exception_offset =
      in_bytes(Thread::pending_exception_offset());
    __ ld_ptr(Rscratch0, pending_exception_offset, Rthread);
    __ bne(Rscratch0, R0, L);
    __ delayed()->nop();
    __ should_not_reach_here();
    __ bind(L);
#endif // ASSERT
    BLOCK_COMMENT("call forward_exception_entry");
    __ restore_frame();
    __ li(Rscratch0, StubRoutines::forward_exception_entry());
    __ jalr(Rscratch0, relocInfo::runtime_call_type);
    __ delayed()->nop();
    
    RuntimeStub* stub = RuntimeStub::new_runtime_stub(name, &code, frame_complete, masm->total_frame_size_in_bytes(0), NULL, false);
    return stub->entry_point();
  }

#undef __
#define __ _masm->


  // Generate a routine that sets all the registers so we
  // can tell if the stop routine prints them correctly.
  address generate_test_stop() {
    StubCodeMark mark(this, "StubRoutines", "test_stop");
    address start = __ pc();

    int i;

    __ save_frame();

    static double zero = 0.0, one = 1.0;
    __ li(Rscratch0, (address)&zero);
    __ ldc1(F0, 0, Rscratch0);
    __ li(Rscratch0, (address)&one);
    __ ldc1(F1, 0, Rscratch0);

    for (i = 2; i <= 31; i++) {
      __ add_d(as_FloatRegister(i), as_FloatRegister(i-1), F1);
    }

    // Note: we set and use AT here.  The below instructions are
    // selected to not use AT.
    __ li(AT, 1);
    for (i = 2; i <= 25; i++) {
      __ addu_long(as_Register(i), as_Register(i-1), 1);
    }

    __ stop("testing stop");


    __ restore_frame();
    __ j(RA);
    __ delayed()->nop();

    return start;
  }

  address generate_stop_subroutine() {
    StubCodeMark mark(this, "StubRoutines", "stop_subroutine");
    address start = __ pc();

    __ stop_subroutine();

    return start;
  }


  // YYY
  // YYY Support for Atomic:: goes here.
  // YYY


  // Support for void OrderAccess::fence().
  //
  address generate_fence() {
    StubCodeMark mark(this, "StubRoutines", "fence");
    address start = __ pc();

    __ sync();
    __ j(RA);
    __ delayed()->nop();

    return start;
  }


  //------------------------------------------------------------------------------------------------------------------------
  // The following routine generates a subroutine to throw an asynchronous
  // UnknownError when an unsafe access gets a fault that could not be
  // reasonably prevented by the programmer.  (Example: SIGBUS/OBJERR.)
  //
  // Arguments :
  //
  //      trapping PC: RA
  //
  // Results:
  //     posts an asynchronous exception, skips the trapping instruction
  //
  address generate_handler_for_unsafe_access() {
    StubCodeMark mark(this, "StubRoutines", "handler_for_unsafe_access");
    address start = __ pc();

    __ save_frame();

    __ jal_to_C(CAST_FROM_FN_PTR(address, handle_unsafe_access),
		relocInfo::runtime_call_type);
    __ delayed()->nop();

    __ restore_frame_and_return();

    return start;
  }

  // Called from MacroAssembler::verify_oop
  //
  address generate_verify_oop_subroutine() {
    StubCodeMark mark(this, "StubRoutines", "verify_oop_stub");
     
    address start = __ pc();
    
    __ verify_oop_subroutine();
    
    return start;
  }


  // YYY
  // YYY lots of specialized arraycopy stub generators go here.
  // YYY

  //
  //  Generate generic array copy stubs
  //
  //  Input:
  //    A0    -  src oop
  //    A1    -  src_pos
  //    A2    -  dst oop
  //    A3    -  dst_pos
  //    A4    -  element count
  //
  //  Output:
  //    V0 ==  0  -  success
  //    V0 == -1  -  need to call System.arraycopy
  //
  address generate_generic_copy(const char *name) {

    Label L_failed, L_objArray;

    // Input registers
    __ UNIMPLEMENTED();

    // registers used as temp

    StubCodeMark mark(this, "StubRoutines", name);
    address start = __ pc();

    BLOCK_COMMENT("arraycopy initial argument checks");

    BLOCK_COMMENT("arraycopy argument klass checks");

#ifdef ASSERT
    //  assert(src->klass() != NULL);
    BLOCK_COMMENT("assert klasses not null");
    { Label L_a, L_b;
      BLOCK_COMMENT("assert done");
    }
#endif

    // Load layout helper
    //

    BLOCK_COMMENT("scale indexes to element size");

    BLOCK_COMMENT("choose copy loop based on element size");

#ifdef ASSERT
    { Label L;
      __ stop("must be long copy, but elsize is wrong");
      __ bind(L);
    }
#endif

    // objArrayKlass
  __ BIND(L_objArray); 

    Label L_plain_copy, L_checkcast_copy;
    //  test array classes for subtyping

    // Identically typed arrays can be copied without element-wise checks.

  __ BIND(L_plain_copy);

  __ BIND(L_failed);
    __ jr(RA);
    __ delayed()->li(V0, -1);	// return -1
    return start;
  }

  void generate_arraycopy_stubs() {

    // Note:  the disjoint stubs must be generated first, some of
    //        the conjoint stubs use them.
    // YYY to be filled in
    StubRoutines::_jbyte_disjoint_arraycopy = NULL;
    StubRoutines::_jshort_disjoint_arraycopy = NULL;
    StubRoutines::_jint_disjoint_arraycopy   = NULL;
    StubRoutines::_jlong_disjoint_arraycopy  = NULL;
    StubRoutines::_oop_disjoint_arraycopy    = NULL;
    StubRoutines::_arrayof_jbyte_disjoint_arraycopy  = NULL;
    StubRoutines::_arrayof_jshort_disjoint_arraycopy = NULL;
    StubRoutines::_arrayof_jint_disjoint_arraycopy   = NULL;
    StubRoutines::_arrayof_jlong_disjoint_arraycopy  = NULL;
    StubRoutines::_arrayof_oop_disjoint_arraycopy    = NULL;

    StubRoutines::_jbyte_arraycopy  = NULL;
    StubRoutines::_jshort_arraycopy = NULL;
    StubRoutines::_jint_arraycopy   = NULL;
    StubRoutines::_jlong_arraycopy  = NULL;
    StubRoutines::_oop_arraycopy    = NULL;
    StubRoutines::_arrayof_jbyte_arraycopy  = NULL;
    StubRoutines::_arrayof_jshort_arraycopy = NULL;
    StubRoutines::_arrayof_jint_arraycopy = NULL;
    StubRoutines::_arrayof_jlong_arraycopy = NULL;
    StubRoutines::_arrayof_oop_arraycopy   = NULL;

    StubRoutines::_generic_arraycopy = generate_generic_copy("generic_arraycopy");
  }

  void generate_initial() {
    // Generates all stubs and initializes the entry points

    //------------------------------------------------------------------------------------------------------------------------
    // entry points that exist in all platforms
    // Note: This is code that could be shared among different platforms - however the benefit seems to be smaller than
    //       the disadvantage of having a much more complicated generator structure. See also comment in stubRoutines.hpp.

    StubRoutines::_forward_exception_entry                 = generate_forward_exception();

    StubRoutines::_call_stub_entry                         = generate_call_stub(StubRoutines::_call_stub_return_address);    
    StubRoutines::_catch_exception_entry                   = generate_catch_exception();

    //------------------------------------------------------------------------------------------------------------------------
    // entry points that are platform specific
    StubRoutines::Mips::_test_stop_entry                   = generate_test_stop();

    StubRoutines::Mips::_stop_subroutine_entry             = generate_stop_subroutine();
  }


  void generate_all() {
    // Generates all stubs and initializes the entry points

    // These entry points require SharedInfo::stack0 to be set up in non-core builds
    StubRoutines::_throw_AbstractMethodError_entry         = generate_throw_exception("AbstractMethodError throw_exception",          CAST_FROM_FN_PTR(address, SharedRuntime::throw_AbstractMethodError),  false);
    StubRoutines::_throw_ArithmeticException_entry         = generate_throw_exception("ArithmeticException throw_exception",          CAST_FROM_FN_PTR(address, SharedRuntime::throw_ArithmeticException),  true);
    StubRoutines::_throw_NullPointerException_entry        = generate_throw_exception("NullPointerException throw_exception",         CAST_FROM_FN_PTR(address, SharedRuntime::throw_NullPointerException), true);
    StubRoutines::_throw_NullPointerException_at_call_entry= generate_throw_exception("NullPointerException at call throw_exception", CAST_FROM_FN_PTR(address, SharedRuntime::throw_NullPointerException_at_call), false);
    StubRoutines::_throw_StackOverflowError_entry          = generate_throw_exception("StackOverflowError throw_exception",           CAST_FROM_FN_PTR(address, SharedRuntime::throw_StackOverflowError),   false);

    StubRoutines::_handler_for_unsafe_access_entry =
      generate_handler_for_unsafe_access();

    // support for verify_oop (must happen after universe_init)
    StubRoutines::_verify_oop_subroutine_entry	   = generate_verify_oop_subroutine();

    // arraycopy stubs used by compilers
    if (UseCompiler) {		// YYY unimplemented
    generate_arraycopy_stubs();
    }
  }

 public:
  StubGenerator(CodeBuffer* code, bool all) : StubCodeGenerator(code) { 
    // replace the standard masm with a special one:
    _masm = new MacroAssembler(code);

    _stub_count = !all ? 0x100 : 0x200;
    if (all) {
      generate_all();
    } else {
      generate_initial();
    }
  }


 private:
  int _stub_count;
  void stub_prolog(StubCodeDesc* cdesc) {
    # ifdef ASSERT
      // put extra information in the stub code, to make it more readable
#ifdef _LP64
// Write the high part of the address
// [RGV] Check if there is a dependency on the size of this prolog
      __ emit_rdat((intptr_t)cdesc >> 32, relocInfo::none);
#endif
      __ emit_rdat((intptr_t)cdesc, relocInfo::none);
      __ emit_rdat(++_stub_count, relocInfo::none);
    # endif
      // YYY On sparc they align to the cache line size.  Not sure if
      // this is useful on mips.
      //align(true);
  }

  void stub_epilog(StubCodeDesc* cdesc) {
#ifndef PRODUCT
    if (PrintInterpreter) {
      tty->print_cr("StubCodeDesc: 0x%lx - 0x%lx %s", cdesc->begin(), cdesc->end(), cdesc->name());
    }
#endif
  }

}; // end class declaration

void StubGenerator_generate(CodeBuffer* code, bool all) {
  StubGenerator g(code, all);
}
