#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)templateTable_sparc.cpp	1.261 07/05/05 17:04:33 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

#include "incls/_precompiled.incl"
#include "incls/_templateTable_mips.cpp.incl"

#define __ _masm->


//----------------------------------------------------------------------------------------------------
// Platform-dependent initialization

void TemplateTable::pd_initialize() {
  // (none)
}


//----------------------------------------------------------------------------------------------------
// Condition conversion
Assembler::Condition ccNot(TemplateTable::Condition cc) {
  switch (cc) {
    case TemplateTable::equal        : return Assembler::notEqual;
    case TemplateTable::not_equal    : return Assembler::equal;
    case TemplateTable::less         : return Assembler::greaterEqual;
    case TemplateTable::less_equal   : return Assembler::greater;
    case TemplateTable::greater      : return Assembler::lessEqual;
    case TemplateTable::greater_equal: return Assembler::less;
  }
  ShouldNotReachHere();
  return Assembler::never; // arbitrary value
}

//----------------------------------------------------------------------------------------------------
// Miscelaneous helper routines 


// offset in bytes
Address TemplateTable::at_bcp(int offset) {
  assert(_desc->uses_bcp(), "inconsistent uses_bcp information");
  return Address(Rbcp, offset);
}


void TemplateTable::patch_bytecode(Bytecodes::Code bc, Register Rbyte_code,
                                   Register Rscratch,
                                   bool load_bc_into_scratch /*=true*/) {
  // Note: we ignore load_bc_into_scratch
  if (!RewriteBytecodes) return;
  Label patch_done;
  if (JvmtiExport::can_post_breakpoint()) {
    Label fast_patch;
    __ lbu(Rbyte_code, 0, Rbcp);
    __ li(Rscratch, Bytecodes::_breakpoint);
    __ bne(Rbyte_code, Rscratch, fast_patch);
    __ delayed()->li(Rbyte_code, bc);
    // perform the quickening, slowly, in the bowels of the breakpoint table
    __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::set_original_bytecode_at), Rmethod, Rbcp, Rbyte_code);
    __ b(patch_done);
    __ delayed()->nop();
    __ bind(fast_patch);
  }
#ifdef ASSERT
  Bytecodes::Code orig_bytecode = Bytecodes::java_code(bc);
  Label okay;
  __ lbu(Rscratch, 0, Rbcp);
  __ li(Rbyte_code, orig_bytecode);
  __ beq(Rscratch, Rbyte_code, okay);
  __ delayed()->nop();
  __ li(Rbyte_code, bc);
  __ beq(Rscratch, Rbyte_code, okay);
  __ delayed()->nop();
  __ stop("Rewriting wrong bytecode location");
  __ bind(okay);
#endif
  __ li(Rbyte_code, bc);
  __ sb(Rbyte_code, 0, Rbcp);
  __ bind(patch_done);
}

//----------------------------------------------------------------------------------------------------
// Individual instructions

void TemplateTable::nop() {
  transition(vtos, vtos);
  // nothing to do
}

void TemplateTable::shouldnotreachhere() {
  transition(vtos, vtos);
  __ stop("shouldnotreachhere bytecode");
}

void TemplateTable::aconst_null() {
  transition(vtos, atos);
  __ move(Rtos, R0);
}


void TemplateTable::iconst(int value) {
  transition(vtos, itos);
  if (value == 0) {
    __ move(Rtos, R0);
  } else {
    __ li(Rtos, value);
  }
}


void TemplateTable::lconst(int value) {
  transition(vtos, ltos);
  if (value == 0) {
    __ move(Rtos, R0);
  } else {
    __ li(Rtos, value);
  }
}


void TemplateTable::fconst(int value) {
  transition(vtos, ftos);
  static float zero = 0.0, one = 1.0, two = 2.0;
  float* p;
  switch( value ) {
   default: ShouldNotReachHere();
   case 0:  p = &zero;  break;
   case 1:  p = &one;   break;
   case 2:  p = &two;   break;
  }
  __ li(Rscratch3, (address)p);
  __ lwc1(Ftos, 0, Rscratch3);
}


void TemplateTable::dconst(int value) {
  transition(vtos, dtos);
  static double zero = 0.0, one = 1.0;
  double* p;
  switch( value ) {
   default: ShouldNotReachHere();
   case 0:  p = &zero;  break;
   case 1:  p = &one;   break;
  }
  __ li(Rscratch3, (address)p);
  __ ldc1(Ftos, 0, Rscratch3);
}


// %%%%% Should factore most snippet templates across platforms 

void TemplateTable::bipush() {
  transition(vtos, itos);
  __ lb(Rtos, 1, Rbcp);
  // could instead use at_bcp(1)
}

void TemplateTable::sipush() {
  transition(vtos, itos);
  __ get_2_byte_integer_at_bcp(1, Rscratch3, Rtos, true);

  // Can't use lh(Rtos, 1, Rbcp): it may not be aligned, plus it's not
  // endian-safe.
}

void TemplateTable::ldc(bool wide) {
  transition(vtos, vtos);
  Label call_ldc, notInt, notString, notClass, exit;

  if (wide) {
    __ get_2_byte_integer_at_bcp(1, Rscratch3, A1, false);
  } else {
    __ lbu(A1, 1, Rbcp);
  }
  __ get_cpool_and_tags(A0, A2);

  const int base_offset = constantPoolOopDesc::header_size() * wordSize;
  const int tags_offset = typeArrayOopDesc::header_size(T_BYTE) * wordSize;

  // get type from tags
  __ addu_long(A2, A2, A1);
  __ lbu(A2, tags_offset, A2);      // type in A2

  // check for unresolved string
  __ li(Rscratch3, JVM_CONSTANT_UnresolvedString);
  __ beq(A2, Rscratch3, call_ldc);
  __ delayed()->nop();

  // check for unresolved class
  __ li(Rscratch3, JVM_CONSTANT_UnresolvedClass);
  __ beq(A2, Rscratch3, call_ldc);
  __ delayed()->nop();

  // check for unresolved class in error state
  __ li(Rscratch3, JVM_CONSTANT_UnresolvedClassInError);
  __ beq(A2, Rscratch3, call_ldc);
  __ delayed()->nop();

  // if not a class, need to call vm to get java mirror of class
  __ li(Rscratch3, JVM_CONSTANT_Class);
  __ bne(A2, Rscratch3, notClass);
  __ delayed()->addu_long(A0, A0, base_offset);  // base_offset small enough

  __ bind(call_ldc);
  __ li(A1, wide);
  call_VM(Rtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::ldc), A1);
  assert(V0 == Rtos, "expect return value in Rtos");
  __ push(atos);
  __ b(exit);
  __ delayed()->nop();

  __ bind(notClass);
  __ sll_long(A1, A1, LogBytesPerWord);
  __ li(Rscratch3, JVM_CONSTANT_Integer);
  __ bne(A2, Rscratch3, notInt);
  __ delayed()->nop();
  __ lw(Rtos, A1, A0);
  __ push(itos);
  __ b(exit);
  __ delayed()->nop();

  __ bind(notInt);
  __ li(Rscratch3, JVM_CONSTANT_String);
  __ bne(A2, Rscratch3, notString);
  __ delayed()->nop();
  __ ld_ptr(Rtos, A1, A0);
  __ verify_oop(Rtos);
  __ push(atos);
  __ b(exit);
  __ delayed()->nop();

  __ bind(notString);
  __ lwxc1(Ftos, A1, A0);
  __ push(ftos);

  __ bind(exit);
}

void TemplateTable::ldc2_w() {
  transition(vtos, vtos);
  Label retry, resolved, Long, exit;

  __ bind(retry);
  __ get_2_byte_integer_at_bcp(1, Rscratch3, Rscratch1, false);
  __ get_cpool_and_tags(Rscratch0, Rscratch2);

  const int base_offset = constantPoolOopDesc::header_size() * wordSize;
  const int tags_offset = typeArrayOopDesc::header_size(T_BYTE) * wordSize;
  // get type from tags
  __ addu_long(Rscratch2, Rscratch2, Rscratch1);  // add index; base added below
  __ lbu(Rscratch2, tags_offset, Rscratch2);      // type in scratch2
  
  __ sll_long(Rscratch1, Rscratch1, LogBytesPerWord);
  __ addu_long(Rscratch0, Rscratch0, Rscratch1);  // cp entry in scratch0

  __ li(Rscratch3, JVM_CONSTANT_Double); 
  __ bne(Rscratch2, Rscratch3, Long);
  __ delayed()->nop();

#ifdef _LP64
  __ ldc1(Ftos, base_offset, Rscratch0);
#else
  assert(0, "must be more precise about bit width of FP support");
#endif
  __ push(dtos);
  __ b(exit);
  __ delayed()->nop();

  __ bind(Long);
#ifdef _LP64
  __ ld(Rtos, base_offset, Rscratch0);
#else
  assert(0, "must be more precise about bit width of FP support");
#endif
  __ push(ltos);

  __ bind(exit);
}


void TemplateTable::locals_index(Register reg, int offset) {
  __ lbu(reg, offset, Rbcp);
}


void TemplateTable::locals_index_wide(Register reg) {
  // offset is 2, not 1, because Rbcp points to wide prefix code // was: Lbcp
  __ get_2_byte_integer_at_bcp(2, Rscratch4, reg, false);
}

void TemplateTable::iload() {
  transition(vtos, itos);

  // Rewrite iload,iload  pair into fast_iload2
  //         iload,caload pair into fast_icaload
  if (RewriteFrequentPairs) {
    Label rewrite, done;

    // get next byte
    __ lbu(Rscratch0, Bytecodes::length_for(Bytecodes::_iload), Rbcp);

    // if _iload, wait to rewrite to iload2.  We only want to rewrite the
    // last two iloads in a pair.  Comparing against fast_iload means that
    // the next bytecode is neither an iload or a caload, and therefore
    // an iload pair.
    __ li(Rscratch1, Bytecodes::_iload);
    __ beq(Rscratch0, Rscratch1, done);
    __ delayed()->nop();

    __ li(Rscratch1, Bytecodes::_fast_iload);
    __ beq(Rscratch0, Rscratch1, rewrite);
    __ delayed()->li(Rscratch1, Bytecodes::_fast_iload2);

    __ li(Rscratch1, Bytecodes::_caload);
    __ beq(Rscratch0, Rscratch1, rewrite);
    __ delayed()->li(Rscratch1, Bytecodes::_fast_icaload);

    __ li(Rscratch1, Bytecodes::_fast_iload); // don't check again
    // rewrite
    // Rscratch1: fast bytecode
    __ bind(rewrite);
    patch_bytecode(Bytecodes::_iload, Rscratch1, Rscratch0, false);
    __ bind(done);
  }

  // Get the local value into tos
  locals_index(Rscratch0);
  __ access_local_int(Rtos, Rscratch0);
}

void TemplateTable::fast_iload2() {
  transition(vtos, itos);
  __ stop("fast_iload2 not implemented");
}

void TemplateTable::fast_iload() {
  transition(vtos, itos);
  __ stop("fast_iload not implemented");
}

void TemplateTable::lload() {
  transition(vtos, ltos);
  locals_index(Rscratch0);
  __ access_local_long(Rtos, Rscratch0);
}


void TemplateTable::fload() {
  transition(vtos, ftos);
  locals_index(Rscratch0);
  __ access_local_float(Ftos, Rscratch0);
}


void TemplateTable::dload() {
  transition(vtos, dtos);
  locals_index(Rscratch0);
  __ access_local_double(Ftos, Rscratch0);
}


void TemplateTable::aload() {
  transition(vtos, atos);
  locals_index(Rscratch0);
  __ access_local_ptr(Rtos, Rscratch0);
}


void TemplateTable::wide_iload() {
  transition(vtos, itos);
  locals_index_wide(Rscratch0);
  __ access_local_long(Rtos, Rscratch0);
}


void TemplateTable::wide_lload() {
  transition(vtos, ltos);
  locals_index_wide(Rscratch0);
  __ access_local_long(Rtos, Rscratch0);
}


void TemplateTable::wide_fload() {
  transition(vtos, ftos);
  locals_index_wide(Rscratch0);
  __ access_local_float(Ftos, Rscratch0);
}


void TemplateTable::wide_dload() {
  transition(vtos, dtos);
  locals_index_wide(Rscratch0);
  __ access_local_float(Ftos, Rscratch0);
}


void TemplateTable::wide_aload() {
  transition(vtos, atos);
  locals_index_wide(Rscratch0);
  __ access_local_ptr(Rtos, Rscratch0);
}


void TemplateTable::iaload() {
  transition(itos, itos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerInt, Rscratch2, Rscratch1);
  __ lw(Rtos, arrayOopDesc::base_offset_in_bytes(T_INT), Rscratch1);
}


void TemplateTable::laload() {
  transition(itos, ltos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerLong, Rscratch2, Rscratch1);
  __ ld(Rtos, arrayOopDesc::base_offset_in_bytes(T_LONG), Rscratch1);
}


void TemplateTable::faload() {
  transition(itos, ftos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerInt, Rscratch2, Rscratch1);
  __ lwc1(Ftos, arrayOopDesc::base_offset_in_bytes(T_FLOAT), Rscratch1);
}


void TemplateTable::daload() {
  transition(itos, dtos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerLong, Rscratch2, Rscratch1);
  __ ldc1(Ftos, arrayOopDesc::base_offset_in_bytes(T_DOUBLE), Rscratch1);
}


void TemplateTable::aaload() {
  transition(itos, atos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerWord, Rscratch2, Rscratch1);
  __ ld_ptr(Rtos, arrayOopDesc::base_offset_in_bytes(T_OBJECT), Rscratch1);
}


void TemplateTable::baload() {
  transition(itos, itos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, 0, Rscratch2, Rscratch1);
  __ lb(Rtos, arrayOopDesc::base_offset_in_bytes(T_BYTE), Rscratch1);
}


void TemplateTable::caload() {
  transition(itos, itos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerShort, Rscratch2, Rscratch1);
  __ lhu(Rtos, arrayOopDesc::base_offset_in_bytes(T_CHAR), Rscratch1);
}

void TemplateTable::fast_icaload() {
  __ stop("fast_icaload not implemented");
}


void TemplateTable::saload() {
  transition(itos, itos);
  // Rtos=index -> Rtos=value
  // tos=array
  __ index_check(Rscratch0, Rtos, LogBytesPerShort, Rscratch2, Rscratch1);
  __ lh(Rtos, arrayOopDesc::base_offset_in_bytes(T_SHORT), Rscratch1);
}


void TemplateTable::iload(int n) {
  transition(vtos, itos);
  debug_only(__ verify_local_tag(frame::TagValue, Rlocals, Rtos, n));
  __ lw(Rtos, Interpreter::local_offset_in_bytes(n), Rlocals);
}


void TemplateTable::lload(int n) {
  transition(vtos, ltos);
  assert(n+1 < Argument::n_register_parameters, "would need more code");
  debug_only(__ verify_local_tag(frame::TagCategory2, Rlocals, Rtos, n));
  __ load_unaligned_long(Rtos, Interpreter::local_offset_in_bytes(n+1), Rlocals);
}


void TemplateTable::fload(int n) {
  transition(vtos, ftos);
  assert(n < Argument::n_register_parameters, "would need more code");
  debug_only(__ verify_local_tag(frame::TagValue, Rlocals, Rscratch0, n));
  __ lwc1(Ftos, Interpreter::local_offset_in_bytes(n), Rlocals);
}


void TemplateTable::dload(int n) {
  transition(vtos, dtos);
  debug_only(__ verify_local_tag(frame::TagCategory2, Rlocals, Rscratch0, n));
  __ load_unaligned_double(Ftos, Interpreter::local_offset_in_bytes(n+1), Rlocals);
}


void TemplateTable::aload(int n) {
  transition(vtos, atos);
  debug_only(__ verify_local_tag(frame::TagReference, Rlocals, Rtos, n));
  __ ld_ptr(Rtos, Interpreter::local_offset_in_bytes(n), Rlocals);
}


void TemplateTable::aload_0() {
  transition(vtos, atos);

  // According to bytecode histograms, the pairs:
  //
  // _aload_0, _fast_igetfield (itos)
  // _aload_0, _fast_agetfield (atos)
  // _aload_0, _fast_fgetfield (ftos)
  //
  // occur frequently. If RewriteFrequentPairs is set, the (slow) _aload_0
  // bytecode checks the next bytecode and then rewrites the current
  // bytecode into a pair bytecode; otherwise it rewrites the current
  // bytecode into _fast_aload_0 that doesn't do the pair check anymore.
  //
  if (RewriteFrequentPairs) {
    // YYY implement for efficiency

    Label rewrite, done;

    // get next byte

    // do actual aload_0
    aload(0);

    // if _getfield then wait with rewrite

    // if _igetfield then rewrite to _fast_iaccess_0

    // if _agetfield then rewrite to _fast_aaccess_0

    // if _fgetfield then rewrite to _fast_faccess_0

    // else rewrite to _fast_aload0

    // rewrite
    // G4_scratch: fast bytecode
    __ bind(rewrite);
    //patch_bytecode(Bytecodes::_aload_0, G4_scratch, G3_scratch, false);
    __ bind(done);
  } else {
    aload(0);
  }
}


void TemplateTable::istore() {
  transition(itos, vtos);
  locals_index(Rscratch0);
  __ store_local_int(Rtos, Rscratch0);
}


void TemplateTable::lstore() {
  transition(ltos, vtos);
  locals_index(Rscratch0);
  __ store_local_long(Rtos, Rscratch0);
}


void TemplateTable::fstore() {
  transition(ftos, vtos);
  locals_index(Rscratch0);
  __ store_local_float(Ftos, Rscratch0);
}


void TemplateTable::dstore() {
  transition(dtos, vtos);
  locals_index(Rscratch0);
  __ store_local_double(Ftos, Rscratch0);
}


void TemplateTable::astore() {
  transition(vtos, vtos);
  // astore tos can also be a returnAddress, so load and store the tag too
  __ load_ptr_and_tag(0, Rscratch0, Rscratch1);
  __ addu_long(Resp, Resp, Interpreter::stackElementSize());
  __ verify_oop_or_return_address(Rscratch0, Rscratch3);
  locals_index(Rscratch3);
  __ store_local_ptr(Rscratch0, Rscratch1, Rscratch3);
}


void TemplateTable::wide_istore() {
  transition(vtos, vtos);
  __ pop_i();
  locals_index_wide(Rscratch3);
  __ store_local_int(Rtos, Rscratch3);
}


void TemplateTable::wide_lstore() {
  transition(vtos, vtos);
  __ pop_l();
  locals_index_wide(Rscratch3);
  __ store_local_long(Rtos, Rscratch3);
}


void TemplateTable::wide_fstore() {
  transition(vtos, vtos);
  __ pop_f();
  locals_index_wide(Rscratch3);
  __ store_local_float(Ftos, Rscratch3);
}


void TemplateTable::wide_dstore() {
  transition(vtos, vtos);
  __ pop_d();
  locals_index_wide(Rscratch3);
  __ store_local_double(Ftos, Rscratch3);
}


void TemplateTable::wide_astore() {
  transition(vtos, vtos);
  // astore tos can also be a returnAddress, so load and store the tag too
  __ load_ptr_and_tag(0, Rscratch0, Rscratch1);
  __ addu_long(Resp, Resp, Interpreter::stackElementSize());
  __ verify_oop_or_return_address(Rscratch0, Rscratch3);
  locals_index_wide(Rscratch3);
  __ store_local_ptr(Rscratch0, Rscratch1, Rscratch3);
}


void TemplateTable::iastore() {
  transition(itos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, LogBytesPerInt, Rscratch3, Rscratch2);
  __ sw(Rtos, arrayOopDesc::base_offset_in_bytes(T_INT), Rscratch2);
}


void TemplateTable::lastore() {
  transition(ltos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, LogBytesPerLong, Rscratch3, Rscratch2);
  __ sd(Rtos, arrayOopDesc::base_offset_in_bytes(T_LONG), Rscratch2);
}


void TemplateTable::fastore() {
  transition(ftos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, LogBytesPerInt, Rscratch3, Rscratch2);
  __ swc1(Ftos, arrayOopDesc::base_offset_in_bytes(T_FLOAT), Rscratch2);
}


void TemplateTable::dastore() {
  transition(dtos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, LogBytesPerLong, Rscratch3, Rscratch2);
  __ sdc1(Ftos, arrayOopDesc::base_offset_in_bytes(T_FLOAT), Rscratch2);
}


void TemplateTable::aastore() {
  Label store_ok, is_null, done;
  transition(vtos, vtos);
  // ...,array,index,value -> ...

  // REGISTER USAGE
  // Rtos     :  value
  // Rsave0   :  tmp
  // Rscratch0:  tmp
  // Rscratch0:  array         == O3
  // Rscratch1:  element addr  == O1
  // Rscratch2:  array index   == O2
  // Rscratch3:  tmp
  // Rscratch4:  array element klass
  // Rscratch5:  value klass

  __ ld_ptr(Rtos, Interpreter::expr_offset_in_bytes(0), Resp);       // value
  __ ld_long(Rscratch2, Interpreter::expr_offset_in_bytes(1), Resp); // index
  __ ld_ptr(Rscratch0, Interpreter::expr_offset_in_bytes(2), Resp);     // array
  __ verify_oop(Rtos);
  __ index_check_without_pop(Rscratch0, Rscratch2, LogBytesPerWord, Rscratch3, Rscratch1);

  // array store check: null value always ok
  __ beq(Rtos, R0, is_null);
  __ delayed()->ld_ptr(Rscratch4, oopDesc::klass_offset_in_bytes(), Rscratch0); // array klass

  // array store check: fast instanceof cache test
  __ ld_ptr(Rscratch5, oopDesc::klass_offset_in_bytes(), Rtos);  // value klass
  __ ld_ptr(Rscratch4, sizeof(oopDesc) + objArrayKlass::element_klass_offset_in_bytes(), Rscratch4);  // array klass -> array element klass
  __ gen_subtype_check(Rscratch5, Rscratch4, Rscratch0, Rsave0, Rscratch3, store_ok);

  // End of gen_subtype_check is the failure case
  __ jump_to(Interpreter::_throw_ArrayStoreException_entry);
  __ delayed()->nop();

  // Store is OK.
  __ bind(store_ok);

  // Do the store.
  __ st_ptr(Rtos, arrayOopDesc::base_offset_in_bytes(T_OBJECT), Rscratch1);

  // Do the write barrier.
  // Quote from rememberedSet.hpp: For objArrays, the precise card
  // corresponding to the pointer store is dirtied so we don't need to
  // scavenge the entire array.

  // First though, we have to add the array header size, because it was
  // never actually done in index_check_without_pop.
  // Is there any efficiency gain from delaying this add?
  // Otherwise it seems better to go back and fix this contract;
  // it's just asking for trouble.
  __ addu_long(Rscratch1, Rscratch1, arrayOopDesc::base_offset_in_bytes(T_OBJECT));

  __ store_check(Rscratch3, Rscratch1);
  __ b(done);
  __ delayed()->addu_long(Resp, Resp, 3 * Interpreter::stackElementSize());

  __ bind(is_null);
  __ st_ptr(Rtos, arrayOopDesc::base_offset_in_bytes(T_OBJECT), Rscratch1);
  __ profile_null_seen(Rscratch3);
  __ addu_long(Resp, Resp, 3 * Interpreter::stackElementSize());  // pop stack
  __ bind(done);
}


void TemplateTable::bastore() {
  transition(itos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, 0, Rscratch3, Rscratch2);
  __ sb(Rtos, arrayOopDesc::base_offset_in_bytes(T_INT), Rscratch2);
}


void TemplateTable::castore() {
  transition(itos, vtos);
  // ...,array,index, Rtos=value -> ...
  __ pop_i(Rscratch2); // index
  __ index_check(Rscratch0, Rscratch2, LogBytesPerShort, Rscratch3, Rscratch2);
  __ sh(Rtos, arrayOopDesc::base_offset_in_bytes(T_INT), Rscratch2);
}


void TemplateTable::sastore() {
  // %%%%% Factor across platform
  castore();
}


void TemplateTable::istore(int n) {
  transition(itos, vtos);
  __ tag_local(frame::TagValue, Rlocals, Rscratch3, n);
  __ sw(Rtos, Interpreter::local_offset_in_bytes(n), Rlocals);
}


void TemplateTable::lstore(int n) {
  transition(ltos, vtos);
  __ tag_local(frame::TagCategory2, Rlocals, Rscratch3, n);
  __ store_unaligned_long(Rtos, Interpreter::local_offset_in_bytes(n+1), Rlocals);
  // n+1: coming down from above, we need to add 1 to go from low to high.
}


void TemplateTable::fstore(int n) {
  transition(ftos, vtos);
  __ tag_local(frame::TagValue, Rlocals, Rscratch3, n);
  __ swc1(Ftos, Interpreter::local_offset_in_bytes(n), Rlocals);
}


void TemplateTable::dstore(int n) {
  transition(dtos, vtos);
  __ tag_local(frame::TagCategory2, Rlocals, Rscratch3, n);
  __ store_unaligned_double(Ftos, Interpreter::local_offset_in_bytes(n+1), Rlocals);
  // n+1: coming down from above, we need to add 1 to go from low to high.
}


void TemplateTable::astore(int n) {
  transition(vtos, vtos);
  // astore tos can also be a returnAddress, so load and store the tag too
  __ load_ptr_and_tag(0, Rscratch0, Rscratch1);
  __ addu_long(Resp, Resp, Interpreter::stackElementSize());
  __ verify_oop_or_return_address(Rscratch0, Rscratch3);
  __ store_local_ptr(Rscratch0, Rscratch1, n);
}


void TemplateTable::pop() {
  transition(vtos, vtos);
  __ addu_long(Resp, Resp, Interpreter::stackElementSize());
}


void TemplateTable::pop2() {
  transition(vtos, vtos);
  __ addu_long(Resp, Resp, 2 * Interpreter::stackElementSize());
}


void TemplateTable::dup() {
  transition(vtos, vtos);
  // stack: ..., a
  // load a and tag
  __ load_ptr_and_tag(0, Rtos, Rtos2);
  __ push_ptr(Rtos, Rtos2);
  // stack: ..., a, a
}


void TemplateTable::dup_x1() {
  transition(vtos, vtos);
  // stack: ..., a, b
  __ load_ptr_and_tag(1, Rscratch0, Rscratch1);  // get a
  __ load_ptr_and_tag(0, Rtos, Rtos2);           // get b
  __ store_ptr_and_tag(1, Rtos, Rtos2);          // put b
  __ store_ptr_and_tag(0, Rscratch0, Rscratch1); // put a - like swap
  __ push_ptr(Rtos, Rtos2);                      // push b
  // stack: ..., b, a, b
}


void TemplateTable::dup_x2() {
  transition(vtos, vtos);
  // stack: ..., a, b, c
  // get c and push on stack, reuse registers
  __ load_ptr_and_tag(0, Rscratch0, Rscratch1);  // get c
  __ push_ptr(Rscratch0, Rscratch1);             // push c with tag
  // stack: ..., a, b, c, c  (c in reg)  (Rsp - 4)
  // (stack offsets n+1 now)
  __ load_ptr_and_tag(3, Rtos, Rtos2);           // get a
  __ store_ptr_and_tag(3, Rscratch0, Rscratch1); // put c at 3
  // stack: ..., c, b, c, c  (a in reg)
  __ load_ptr_and_tag(2, Rscratch0, Rscratch1);  // get b
  __ store_ptr_and_tag(2, Rtos, Rtos2);          // put a at 2
  // stack: ..., c, a, c, c  (b in reg)
  __ store_ptr_and_tag(1, Rscratch0, Rscratch1); // put b at 1
  // stack: ..., c, a, b, c 
}


void TemplateTable::dup2() {
  transition(vtos, vtos);
  __ load_ptr_and_tag(1, Rscratch0, Rscratch1);  // get a
  __ load_ptr_and_tag(0, Rtos, Rtos2);           // get b
  __ push_ptr(Rscratch0, Rscratch1);             // push a
  __ push_ptr(Rtos, Rtos2);                      // push b
  // stack: ..., a, b, a, b
}


void TemplateTable::dup2_x1() {
  transition(vtos, vtos);
  // stack: ..., a, b, c
  __ load_ptr_and_tag(1, Rscratch2, Rscratch3);  // get b
  __ load_ptr_and_tag(2, Rtos, Rtos2);           // get a
  __ store_ptr_and_tag(2, Rscratch2, Rscratch3); // put b at a
  // stack: ..., b, b, c
  __ load_ptr_and_tag(0, Rscratch0, Rscratch1);  // get c
  __ store_ptr_and_tag(1, Rscratch0, Rscratch1); // put c at b
  // stack: ..., b, c, c
  __ store_ptr_and_tag(0, Rtos, Rtos2);          // put a at c
  // stack: ..., b, c, a
  __ push_ptr(Rscratch2, Rscratch3);             // push b
  __ push_ptr(Rscratch0, Rscratch1);             // push c
  // stack: ..., b, c, a, b, c
}

// The spec says that these types can be a mixture of category 1 (1 word)
// types and/or category 2 types (long and doubles)
void TemplateTable::dup2_x2() {
  transition(vtos, vtos);
  // stack: ..., a, b, c, d
  __ load_ptr_and_tag(1, Rscratch2, Rscratch3);  // get c
  __ load_ptr_and_tag(3, Rtos, Rtos2);           // get a
  __ store_ptr_and_tag(3, Rscratch2, Rscratch3); // put c at 3
  __ store_ptr_and_tag(1, Rtos, Rtos2);          // put a at 1
  // stack: ..., c, b, a, d
  __ load_ptr_and_tag(2, Rscratch0, Rscratch1);  // get b
  __ load_ptr_and_tag(0, Rtos, Rtos2);           // get d
  __ store_ptr_and_tag(0, Rscratch0, Rscratch1); // put b at 0
  __ store_ptr_and_tag(2, Rtos, Rtos2);          // put d at 2
  // stack: ..., c, d, a, b
  __ push_ptr(Rscratch2, Rscratch3);             // push c
  __ push_ptr(Rtos, Rtos2);                      // push d
  // stack: ..., c, d, a, b, c, d
}


void TemplateTable::swap() {
  transition(vtos, vtos);
  // stack: ..., a, b
  __ load_ptr_and_tag(1, Rscratch0, Rscratch1);
  __ load_ptr_and_tag(0, Rtos, Rtos2);
  __ store_ptr_and_tag(0, Rscratch0, Rscratch1);
  __ store_ptr_and_tag(1, Rtos, Rtos2);
  // stack: ..., b, a
}


void TemplateTable::iop2(Operation op) {
  transition(itos, itos);
  __ pop_i(Rscratch1);
  switch (op) {
  case  add:  __ addu(Rtos, Rscratch1, Rtos); break;
  case  sub:  __ subu(Rtos, Rscratch1, Rtos); break;
  case  mul:  __ mul( Rtos, Rscratch1, Rtos); break;
  case _and:  __ and_(Rtos, Rscratch1, Rtos); break;
  case  _or:  __ or_( Rtos, Rscratch1, Rtos); break;
  case _xor:  __ xor_(Rtos, Rscratch1, Rtos); break;
  case  shl:  __ sllv(Rtos, Rscratch1, Rtos); break;
  case  shr:  __ srav(Rtos, Rscratch1, Rtos); break;
  case ushr:  __ srlv(Rtos, Rscratch1, Rtos); break;
   default: ShouldNotReachHere();
  }
}


void TemplateTable::lop2(Operation op) {
  transition(ltos, ltos);
  __ pop_l(Rscratch1);
  switch(op) {
#ifdef _LP64
  case  add:  __ daddu(Rtos, Rscratch1, Rtos); break;
  case  sub:  __ dsubu(Rtos, Rscratch1, Rtos); break;
  case _and:  __ and_( Rtos, Rscratch1, Rtos); break;
  case  _or:  __ or_(  Rtos, Rscratch1, Rtos); break;
  case _xor:  __ xor_( Rtos, Rscratch1, Rtos); break;
#else
#error unimplemented
#endif
  default: ShouldNotReachHere();
  }
}


void TemplateTable::idiv() {
  Label zero, normal, done;
  transition(itos, itos);
  __ pop_i(Rscratch1); // get numerator (demoninator is in Rtos)

  // Start by doing the operation; this helps hide the latency.  This
  // also helps with the implementation of irem, which expects the
  // remainder to be in HI.
  __ div(Rscratch1, Rtos);

  // test for zero divisor
  __ beq(Rtos, R0, zero);
  __ delayed()->nop();

  // Test for special case of -1 divisor and MAXNEGINT dividend.
  // Not sure this special case is actually necessary; check
  // the definition of "div" used in the MIPS64 spec.
  const int min_int = 0x80000000;
  bnei(Rscratch1, min_int, normal, Rscratch3);
  __ delayed()->nop();
  bnei(Rtos, -1, normal, Rscratch3);
  __ delayed()->nop();
  __ move(Rtos, Rscratch1);
  __ b(done);
  __ delayed()->nop();

  __ bind(zero);
  __ jump_to(Interpreter::_throw_ArithmeticException_entry);
  __ delayed()->nop();

  // common case
  __ bind(normal);
  __ mflo(Rtos);

  __ bind(done);
}


void TemplateTable::irem() {
  transition(itos, itos);
  idiv();
  __ mfhi(Rtos);
}


void TemplateTable::lmul() {
  transition(ltos, ltos);
  __ pop_l(A2);
#ifdef _LP64
  __ dmult(Rtos, A2);
  __ mflo(Rtos);
#else
  __ call_VM_leaf(Rscratch0, CAST_FROM_FN_PTR(address, SharedRuntime::lmul));
#endif

}


// Same as idiv except for the use of ddiv
void TemplateTable::ldiv() {
  Label zero, normal, done;
  transition(ltos, ltos);
  __ pop_l(Rscratch1); // get dividend

  // Start by doing the operation; this helps hide the latency.  This
  // also helps with the implementation of irem, which expects the
  // remainder to be in HI.
  __ ddiv(Rscratch1, Rtos);

  // test for zero divisor
  __ beq(Rtos, R0, zero);
  __ delayed()->nop();

  // Test for special case of -1 divisor and MAXNEGINT dividend.
  // Not sure this special case is actually necessary; check
  // the definition of "div" used in the MIPS64 spec.
  const long min_int = 0x8000000000000000;
  __ li(Rscratch2, min_int);
  __ bne(Rscratch1, Rscratch2, normal);
  __ delayed()->nop();
  bnei(Rtos, -1, normal, Rscratch3);
  __ delayed()->nop();
  __ move(Rtos, Rscratch1);
  __ b(done);
  __ delayed()->nop();

  __ bind(zero);
  __ jump_to(Interpreter::_throw_ArithmeticException_entry);
  __ delayed()->nop();

  // common case
  __ bind(normal);
  __ mflo(Rtos);

  __ bind(done);
}


void TemplateTable::lrem() {
  transition(ltos, ltos);
  ldiv();
  __ mfhi(Rtos);
}


void TemplateTable::lshl() {
  transition(itos, ltos);
  __ pop_l(Rscratch1);
  __ dsllv(Rtos, Rscratch1, Rtos);
}


void TemplateTable::lshr() {
  transition(itos, ltos);
  __ pop_l(Rscratch1);
  __ dsrav(Rtos, Rscratch1, Rtos);
}



void TemplateTable::lushr() {
  transition(itos, ltos);
  __ pop_l(Rscratch1);
  __ dsrlv(Rtos, Rscratch1, Rtos);
}


void TemplateTable::fop2(Operation op) {
  transition(ftos, ftos);
  switch (op) {
   case  add:  
     __ pop_f(Fscratch0);
     __ add_s(Ftos, Fscratch0, Ftos);
     break;
   case  sub:
     __ pop_f(Fscratch0);
     __ sub_s(Ftos, Fscratch0, Ftos);
     break;
   case  mul:
     __ pop_f(Fscratch0);
     __ mul_s(Ftos, Fscratch0, Ftos);
     break;
   case  div:
     __ pop_f(Fscratch0);
     __ div_s(Ftos, Fscratch0, Ftos);
     break;
   case  rem:
     __ pop_f(F12);
     __ mov_s(F13, Ftos);
     __ call_VM_leaf(Rscratch0, CAST_FROM_FN_PTR(address, SharedRuntime::frem));
     assert(Ftos == F0, "return expected in Ftos");
     break;

   default: ShouldNotReachHere();
  }
}

void TemplateTable::dop2(Operation op) {
  transition(dtos, dtos);
  switch (op) {
   case  add:  
     __ pop_d(Fscratch0);
     __ add_d(Ftos, Fscratch0, Ftos);
     break;
   case  sub:
     __ pop_d(Fscratch0);
     __ sub_d(Ftos, Fscratch0, Ftos);
     break;
   case  mul:
     __ pop_d(Fscratch0);
     __ mul_d(Ftos, Fscratch0, Ftos);
     break;
   case  div:
     __ pop_d(Fscratch0);
     __ div_d(Ftos, Fscratch0, Ftos);
     break;
   case  rem:
     __ pop_d(F12);
     __ mov_d(F13, Ftos);
     __ call_VM_leaf(Rscratch0, CAST_FROM_FN_PTR(address, SharedRuntime::drem));
     assert(Ftos == F0, "return expected in Ftos");
     break;

   default: ShouldNotReachHere();
  }
}


void TemplateTable::ineg() {
  transition(itos, itos);
  __ subu(Rtos, R0, Rtos);
}


void TemplateTable::lneg() {
  transition(ltos, ltos);
  __ dsubu(Rtos, R0, Rtos);
}


void TemplateTable::fneg() {
  transition(ftos, ftos);
  __ neg_s(Ftos, Ftos);
}


void TemplateTable::dneg() {
  transition(dtos, dtos);
  __ neg_d(Ftos, Ftos);
}


void TemplateTable::iinc() {
  transition(vtos, vtos);
  locals_index(Rscratch3);
  __ lb(Rscratch2, 2, Rbcp);  // load inc constant
  __ access_local_int(Rtos, Rscratch3);
  // Rscratch3 now holds the effective address to write back to
  __ addu(Rtos, Rtos, Rscratch2);
  __ sw(Rtos, Interpreter::value_offset_in_bytes(), Rscratch3);
}


void TemplateTable::wide_iinc() {
  transition(vtos, vtos);
  locals_index_wide(Rscratch3);
  // wide iinc is special; the constant is wide as well
  __ get_2_byte_integer_at_bcp(4, Rscratch1, Rscratch2, true);
  __ access_local_int(Rtos, Rscratch3);
  // Rscratch3 now holds the effective address to write back to
  __ addu(Rtos, Rtos, Rscratch2);
  __ sw(Rtos, Interpreter::value_offset_in_bytes(), Rscratch3);
}


void TemplateTable::convert() {
// %%%%% Factor this first part across platforms 
  #ifdef ASSERT
    TosState tos_in  = ilgl;
    TosState tos_out = ilgl;
    switch (bytecode()) {
      case Bytecodes::_i2l: // fall through
      case Bytecodes::_i2f: // fall through
      case Bytecodes::_i2d: // fall through
      case Bytecodes::_i2b: // fall through
      case Bytecodes::_i2c: // fall through
      case Bytecodes::_i2s: tos_in = itos; break;
      case Bytecodes::_l2i: // fall through
      case Bytecodes::_l2f: // fall through
      case Bytecodes::_l2d: tos_in = ltos; break;
      case Bytecodes::_f2i: // fall through
      case Bytecodes::_f2l: // fall through
      case Bytecodes::_f2d: tos_in = ftos; break;
      case Bytecodes::_d2i: // fall through
      case Bytecodes::_d2l: // fall through
      case Bytecodes::_d2f: tos_in = dtos; break;
      default             : ShouldNotReachHere();
    }
    switch (bytecode()) {
      case Bytecodes::_l2i: // fall through
      case Bytecodes::_f2i: // fall through
      case Bytecodes::_d2i: // fall through
      case Bytecodes::_i2b: // fall through
      case Bytecodes::_i2c: // fall through
      case Bytecodes::_i2s: tos_out = itos; break;
      case Bytecodes::_i2l: // fall through
      case Bytecodes::_f2l: // fall through
      case Bytecodes::_d2l: tos_out = ltos; break;
      case Bytecodes::_i2f: // fall through
      case Bytecodes::_l2f: // fall through
      case Bytecodes::_d2f: tos_out = ftos; break;
      case Bytecodes::_i2d: // fall through
      case Bytecodes::_l2d: // fall through
      case Bytecodes::_f2d: tos_out = dtos; break;
      default             : ShouldNotReachHere();
    }
    transition(tos_in, tos_out);
  #endif

  // Conversion
  Label done;
  switch (bytecode()) {
  case Bytecodes::_i2l: 
#ifdef _LP64
    __ sra(Rtos, Rtos, 0);
#else
#error unimplemented
#endif
    break;

  case Bytecodes::_i2f:
    __ mtc1(Rtos, Fscratch0);
    __ cvt_s_w(Ftos, Fscratch0);
    break;

  case Bytecodes::_i2d:
    __ mtc1(Rtos, Fscratch0);
    __ cvt_d_w(Ftos, Fscratch0);
    break;

  case Bytecodes::_i2b:
    __ sll(Rtos, Rtos, 24);
    __ sra(Rtos, Rtos, 24);
    break;

  case Bytecodes::_i2c:
    __ sll(Rtos, Rtos, 16);
    __ srl(Rtos, Rtos, 16);
    break;

  case Bytecodes::_i2s:
    __ sll(Rtos, Rtos, 16);
    __ sra(Rtos, Rtos, 16);
    break;

  case Bytecodes::_l2i:
#ifdef _LP64
    __ dsll32(Rtos, Rtos, 0);
    __ dsra32(Rtos, Rtos, 0);
#else
#error unimplemented
#endif
    break;

  case Bytecodes::_l2f:
    __ dmtc1(Rtos, Fscratch0);
    __ cvt_s_l(Ftos, Fscratch0);
    break;

  case Bytecodes::_l2d:
    __ dmtc1(Rtos, Fscratch0);
    __ cvt_d_l(Ftos, Fscratch0);
    break;

  case Bytecodes::_f2i:
    // trunc doesn't do the right thing in the exceptional cases, but
    // it's right for FP_NORMAL.
    __ trunc_w_s(Fscratch0, Ftos);
    __ cfc1(Rscratch0, Assembler::cp1_fcsr);
    __ andi(Rscratch0, Rscratch0, Assembler::fcsr_cause_invalidop);
    __ beq(Rscratch0, R0, done);
    __ delayed()->mfc1(Rtos, Fscratch0);
    __ call_VM_leaf(Rscratch0, CAST_FROM_FN_PTR(address, SharedRuntime::f2i));
    break;

  case Bytecodes::_f2l:
    __ trunc_l_s(Fscratch0, Ftos);
    __ cfc1(Rscratch0, Assembler::cp1_fcsr);
    __ andi(Rscratch0, Rscratch0, Assembler::fcsr_cause_invalidop);
    __ beq(Rscratch0, R0, done);
    __ delayed()->mfc1(Rtos, Fscratch0);
    __ call_VM_leaf(CAST_FROM_FN_PTR(address, SharedRuntime::f2l));
    break;

  case Bytecodes::_f2d:
    __ cvt_d_s(Ftos, Ftos);
    break;

  case Bytecodes::_d2i:
    __ trunc_w_d(Fscratch0, Ftos);
    __ cfc1(Rscratch0, Assembler::cp1_fcsr);
    __ andi(Rscratch0, Rscratch0, Assembler::fcsr_cause_invalidop);
    __ beq(Rscratch0, R0, done);
    __ delayed()->mfc1(Rtos, Fscratch0);
    __ call_VM_leaf(CAST_FROM_FN_PTR(address, SharedRuntime::d2i));
    break;

  case Bytecodes::_d2l:
    __ trunc_l_d(Fscratch0, Ftos);
    __ cfc1(Rscratch0, Assembler::cp1_fcsr);
    __ andi(Rscratch0, Rscratch0, Assembler::fcsr_cause_invalidop);
    __ beq(Rscratch0, R0, done);
    __ delayed()->mfc1(Rtos, Fscratch0);
    __ call_VM_leaf(CAST_FROM_FN_PTR(address, SharedRuntime::d2l));
    break;

  case Bytecodes::_d2f:
    __ cvt_s_d(Ftos, Ftos);
    break;

  default: ShouldNotReachHere();
  }
  __ bind(done);
}


void TemplateTable::lcmp() {
  transition(ltos, itos);

#ifdef _LP64
  __ pop_l(Rscratch1);
  __ move(Rscratch3, R0);
  __ slt(Rscratch3, Rtos, Rscratch1);   // Rscratch3 = 1 if Rtos < Rscratch1
  __ move(Rscratch4, R0);
  __ slt(Rscratch4, Rscratch1, Rtos);
  __ li(Rscratch5, -1);
  __ movn(Rscratch4, Rscratch5, Rscratch4);  // Rscratch4 = -1 if Rscratch1 < Rtos
  __ or_(Rscratch3, Rscratch3, Rscratch4);
  // used to have this, but actually a move is just fine.
  //__ xor_(Rscratch5, Rscratch1, Rtos);   // Rscratch5 = 0 if Rscratch1 = Rtos
  __ move(Rtos, R0);
  __ movn(Rtos, Rscratch3, Rscratch3);
#else
#error unimplemented
#endif
}


void TemplateTable::float_cmp(bool is_float, int unordered_result) {
  if (is_float) {
    fcmp(unordered_result);
  } else {
    dcmp(unordered_result);
  }
}

void TemplateTable::fcmp(int unordered_result) {
  Label done;

  __ pop_f(Fscratch1);  // value1 = Fscratch1/F3, value2 = Ftos/F0

  // test for value1 > value2
  __ c_cond_s(Assembler::cond_ule, Fscratch1, Ftos);
  __ bc1f(done);
  __ delayed()->li(Rtos, 1);

  // test for value1 == value2
  __ c_cond_s(Assembler::cond_eq, Fscratch1, Ftos);
  __ bc1t(done);
  __ delayed()->move(Rtos, R0);

  // test for value1 < value2
  __ c_cond_s(Assembler::cond_ult, Fscratch1, Ftos);
  __ bc1t(done);
  __ delayed()->li(Rtos, -1);

  // unordered
  __ li(Rtos, unordered_result);

  __ bind(done);
}

void TemplateTable::dcmp(int unordered_result) {
  Label done;

  __ pop_d(Fscratch1);

  // test for value1 > value2
  __ c_cond_d(Assembler::cond_ule, Fscratch1, Ftos);
  __ bc1f(done);
  __ delayed()->li(Rtos, 1);

  // test for value1 == value2
  __ c_cond_d(Assembler::cond_eq, Fscratch1, Ftos);
  __ bc1t(done);
  __ delayed()->move(Rtos, R0);

  // test for value1 < value2
  __ c_cond_d(Assembler::cond_ult, Fscratch1, Ftos);
  __ bc1t(done);
  __ delayed()->li(Rtos, -1);

  // unordered
  __ li(Rtos, unordered_result);

  __ bind(done);
}

void TemplateTable::branch(bool is_jsr, bool is_wide) {
  __ verify_oop(Rmethod);
  __ verify_thread();

  // Rscratch0 : current bcp
  // Rscratch1 : displacement
  // Rscratch2 : bumped count
  // Rscratch3 : temp
  // Rscratch4 : temp

  __ profile_taken_branch(Rscratch3, Rscratch2);

  if (is_wide)  __ get_4_byte_integer_at_bcp(1, Rscratch4, Rscratch1);
  else          __ get_2_byte_integer_at_bcp(1, Rscratch4, Rscratch1, true);

  // Handle all the JSR stuff here, then exit.
  // It's much shorter and cleaner than intermingling with the
  // non-JSR normal-branch stuff occuring below.
  if( is_jsr ) {
    // compute return address as bci in Rtos
    __ ld_ptr(Rscratch3, in_bytes(methodOopDesc::const_offset()), Rmethod);
    __ subu_long(Rscratch3, Rbcp, Rscratch3);
    __ li32(Rscratch4, in_bytes(constMethodOopDesc::codes_offset()) - (is_wide ? 5 : 3) );
    __ subu_long(Rtos, Rscratch3, Rscratch4);

    // Bump Rbcp to target of JSR
    __ addu_long(Rbcp, Rbcp, Rscratch1);
    // Push returnAddress for "ret" on stack
    __ push_ptr(Rtos, R0); // push ptr sized thing plus 0 for tag.
    // And away we go!
    __ dispatch_next(vtos);
    return;
  }

  // Normal (non-jsr) branch handling

  // Save the current Rbcp
  __ move(Rscratch0, Rbcp);

  bool increment_invocation_counter_for_backward_branches = UseCompiler && UseLoopCounter;
  if ( increment_invocation_counter_for_backward_branches ) {
    Label Lforward;
    // check branch direction
    __ bgt(Rscratch1, R0, Lforward);
    // Bump bytecode pointer by displacement (take the branch)
    __ delayed()->addu_long(Rbcp, Rbcp, Rscratch1); // add to bc addr

    // Update Backedge branch separately from invocations
    // YYY need to distinguish invoke_ctr from bumped_count; see sparc
    __ increment_backedge_counter(Rscratch2, Rscratch3);
    if (ProfileInterpreter) {
      __ test_invocation_counter_for_mdp(Rscratch2, Rbcp, Rscratch3, Lforward);
      if (UseOnStackReplacement) {
        __ test_backedge_count_for_osr(Rscratch2, Rscratch0, Rscratch3);
      }
    } else {
      if (UseOnStackReplacement) {
        __ test_backedge_count_for_osr(Rscratch2, Rscratch0, Rscratch3);
      }
    }

    __ bind(Lforward);
  } else
    // Bump bytecode pointer by displacement (take the branch)
    __ addu_long(Rbcp, Rbcp, Rscratch1); // add to bc addr

  // continue with bytecode @ target
  // %%%%% Like Intel, could speed things up by moving bytecode fetch to code above,
  // %%%%% and changing dispatch_next to dispatch_only
  __ dispatch_next(vtos);
}


// Note Condition in argument is TemplateTable::Condition
// arg scope is within class scope

void TemplateTable::if_0cmp(Condition cc) {
  // no pointers, integer only!
  transition(itos, vtos);
  // assume branch is more often taken than not (loops use backward branches)
  Label not_taken;
  __ bx(ccNot(cc), Rtos, R0, not_taken);
  __ delayed()->nop();
  branch(false, false);
  __ bind(not_taken);
  __ profile_not_taken_branch(Rscratch3);
}


void TemplateTable::if_icmp(Condition cc) {
  transition(itos, vtos);
  Label not_taken;
  __ pop_i(Rscratch1);
  __ bx(ccNot(cc), Rscratch1, Rtos, not_taken);
  __ delayed()->nop();
  branch(false, false);
  __ bind(not_taken);
  __ profile_not_taken_branch(Rscratch3);
}


void TemplateTable::if_nullcmp(Condition cc) {
  transition(atos, vtos);
  Label not_taken;
  __ bx(ccNot(cc), Rtos, R0, not_taken);
  __ delayed()->nop();
  branch(false, false);
  __ bind(not_taken);
  __ profile_not_taken_branch(Rscratch3);
}


void TemplateTable::if_acmp(Condition cc) {
  transition(atos, vtos);
  Label not_taken;  
  __ pop_ptr(Rscratch1);
  __ verify_oop(Rscratch1);
  __ verify_oop(Rtos);
  __ bx(ccNot(cc), Rscratch1, Rtos, not_taken);
  __ delayed()->nop();
  branch(false, false);
  __ bind(not_taken);
  __ profile_not_taken_branch(Rscratch3);
}



void TemplateTable::ret() {
  transition(vtos, vtos);
  locals_index(Rscratch3);
  __ access_local_returnAddress(Rtos, Rscratch3);
  // Rtos contains the bci, compute the bcp from that

  __ profile_ret(vtos, Rtos, Rscratch4);

  __ ld_ptr(Rscratch3, in_bytes(methodOopDesc::const_offset()), Rmethod);
  __ addu_long(Rscratch3, Rtos, Rscratch3);
  __ addu_long(Rbcp, Rscratch3, in_bytes(constMethodOopDesc::codes_offset()));
  __ dispatch_next(vtos);
}


void TemplateTable::wide_ret() {
  transition(vtos, vtos);
  locals_index_wide(Rscratch3);
  __ access_local_returnAddress(Rtos, Rscratch3);
  // Rtos contains the bci, compute the bcp from that

  __ profile_ret(vtos, Rtos, Rscratch4);

  __ ld_ptr(Rscratch3, in_bytes(methodOopDesc::const_offset()), Rmethod);
  __ addu_long(Rscratch3, Rtos, Rscratch3);
  __ addu_long(Rbcp, Rscratch3, in_bytes(constMethodOopDesc::codes_offset()));
  __ dispatch_next(vtos);
}

void TemplateTable::tableswitch() {
  transition(itos, vtos);
  Label default_case, continue_execution;

  // Rscratch1 <- aligned bcp
  __ addu_long(Rscratch1, Rbcp, BytesPerInt);
  __ andi(Rscratch1, Rscratch1, -BytesPerInt);

  // Rscratch2 <- lo
  __ get_4_byte_integer_at_base(Rscratch1, 1 * BytesPerInt, Rscratch0, Rscratch2, true);

  // Rscratch3 <- hi
  __ get_4_byte_integer_at_base(Rscratch1, 2 * BytesPerInt, Rscratch0, Rscratch3, true);

  // check against lo & hi
  __ blt(Rtos, Rscratch2, default_case);
  __ delayed()->nop();
  __ bgt(Rtos, Rscratch3, default_case);
  // lookup dispatch offset
  // Rscratch2 <- offset into jump table
  __ delayed()->sub(Rscratch2, Rtos, Rscratch2);
  __ profile_switch_case(Rscratch2, Rscratch3, Rscratch4, Rscratch5);
  __ sll_long(Rscratch2, Rscratch2, LogBytesPerInt);
  __ addu_long(Rscratch2, Rscratch2, 3 * BytesPerInt);
  // Rscratch2 <- switch arm target offset
  __ get_4_byte_integer(Rscratch1, Rscratch2, Rscratch3, Rscratch4, Rscratch2, true);
  __ b(continue_execution);
  __ delayed()->nop();
  // handle default
  __ bind(default_case);
  __ profile_switch_default(Rscratch3);
  // Rscratch2 <- switch default target offset
  __ get_4_byte_integer_at_base(Rscratch1, 0, Rscratch3, Rscratch2, true);
  // continue execution
  __ bind(continue_execution);
  __ addu_long(Rbcp, Rbcp, Rscratch2);
  __ dispatch_next(vtos);
}

void TemplateTable::lookupswitch() {
  transition(itos, itos);
  __ stop("lookupswitch bytecode should have been rewritten");
}  

void TemplateTable::fast_linearswitch() {
  transition(itos, vtos);

  // Rtos : key value
  // A1   : aligned start of tableswitch data
  // A2   : pointer just beyond last pair
  // A3   : search pointer
  // A4   : test pair key value

  Label loop_entry, loop, found, continue_execution;
  Label done;
  // align bcp
  __ addu_long(A1, Rbcp, BytesPerInt);
  __ andi(A1, A1, -BytesPerInt);
  // set counter
  __ get_4_byte_integer_at_base(A1, BytesPerInt, Rscratch1, A2, true);
  __ sll_long(A2, A2, LogBytesPerInt + 1);
  __ addu_long(A3, A1, 2 * BytesPerInt);
  __ b(loop_entry);
  __ delayed()->addu_long(A2, A3, A2);

  // table search
  __ bind(loop);
  __ get_4_byte_integer_at_base(A3, 0, Rscratch1, A4, true);
  __ beq(A4, Rtos, found);
  __ delayed()->nop();
  __ addu_long(A3, A3, 2 * BytesPerInt);

  __ bind(loop_entry);
  __ bgeu(A2, A3, loop);
  __ delayed()->nop();

  // default case
  // get default offset
  __ get_4_byte_integer_at_base(A1, 0, Rscratch1, A4, true);
  if (ProfileInterpreter) {
    __ profile_switch_default(A3);
    __ b(continue_execution);
    __ delayed()->nop();
  }
  __ b(done);
  __ delayed()->nop();

  // entry found -> get offset
  __ bind(found);
  __ get_4_byte_integer_at_base(A3, BytesPerInt, Rscratch1, A4, true);

  __ bind(done);
  if (ProfileInterpreter) {
    __ subu_long(A3, A3, A1);
    __ addu_long(A3, A3, -2*BytesPerInt);
    __ srl_long(A3, A3, LogBytesPerInt + 1); // in word-pairs
    __ profile_switch_case(A3, A1, A2, Rscratch3);
    
    __ bind(continue_execution);
  }
  __ addu_long(Rbcp, Rbcp, A4);
  __ dispatch_next(vtos);
}

void TemplateTable::fast_binaryswitch() {
  transition(itos, vtos);
  // Implementation using the following core algorithm: (copied from Intel)
  //
  // int binary_search(int key, LookupswitchPair* array, int n) {
  //   // Binary search according to "Methodik des Programmierens" by
  //   // Edsger W. Dijkstra and W.H.J. Feijen, Addison Wesley Germany 1985.
  //   int i = 0;
  //   int j = n;
  //   while (i+1 < j) {
  //     // invariant P: 0 <= i < j <= n and (a[i] <= key < a[j] or Q)
  //     // with      Q: for all i: 0 <= i < n: key < a[i]
  //     // where a stands for the array and assuming that the (inexisting)
  //     // element a[n] is infinitely big.
  //     int h = (i + j) >> 1;
  //     // i < h < j
  //     if (key < array[h].fast_match()) {
  //       j = h;
  //     } else {
  //       i = h;
  //     }
  //   }
  //   // R: a[i] <= key < a[i+1] or Q
  //   // (i.e., if key is within array, i is the correct index)
  //   return i;
  // }

  // register allocation
  const Register Rkey     = Rtos; // already set (tosca)
  const Register Rarray   = A1;
  const Register Ri       = A2;
  const Register Rj       = A3;
  const Register Rh       = A4;
  const Register Rscratch = A5;

  const int log_entry_size = 3;
  const int entry_size = 1 << log_entry_size;
  
  Label found;
  // Find Array start
  __ addu_long(Rarray, Rbcp, 3 * BytesPerInt);
  __ andi(Rarray, Rarray, -BytesPerInt);
  // initialize i & j
  __ move(Ri, R0);
  __ get_4_byte_integer_at_base(Rarray, -BytesPerInt, Rscratch1, Rj, true);

  // and start
  Label entry;
  __ b(entry);
  __ delayed()->nop();

  // binary search loop
  { Label loop;
    __ bind(loop);
    // int h = (i + j) >> 1;
    __ addu_long(Rh, Ri, Rj);
    __ sra_long(Rh, Rh, 1);
    // if (key < array[h].fast_match()) {
    //   j = h;
    // } else {
    //   i = h;
    // }
    __ sll_long(Rscratch, Rh, log_entry_size);
    __ addu_long(Rscratch2, Rarray, Rscratch);
    __ get_4_byte_integer_at_base(Rscratch2, 0, Rscratch1, Rscratch, true);
    __ slt(Rscratch2, Rkey, Rscratch);
    __ movn(Rj, Rh, Rscratch2);
    __ movz(Ri, Rh, Rscratch2);

    // while (i+1 < j)
    __ bind(entry);
    __ addu_long(Rscratch, Ri, 1);
    __ blt(Rscratch, Rj, loop);
    __ delayed()->nop();
  }

  // end of binary search, result index is i (must check again!)
  Label default_case;
  Label continue_execution;
  if (ProfileInterpreter) {
    __ move(Rh, Ri);		// Save index in i for profiling
  }
  __ sll_long(Ri, Ri, log_entry_size);
  __ addu_long(Rscratch2, Rarray, Ri);
  __ get_4_byte_integer_at_base(Rscratch2, 0, Rscratch1, Rscratch, true);
  __ bne(Rkey, Rscratch, default_case);
  __ delayed()->nop();

  // entry found -> j = offset
  __ addu_long(Ri, Ri, BytesPerInt);
  __ profile_switch_case(Rh, Rj, Rscratch, Rkey);
  __ addu_long(Rscratch2, Rarray, Ri);
  __ get_4_byte_integer_at_base(Rscratch2, 0, Rscratch1, Rj, true);

  __ b(continue_execution);
  __ delayed()->nop();

  __ bind(default_case); // fall through (if not profiling)
  __ get_4_byte_integer_at_base(Rarray, -2 * BytesPerInt, Rscratch1, Rj, true);
  __ profile_switch_default(Ri);

  __ bind(continue_execution);
  __ addu_long(Rbcp, Rbcp, Rj);
  __ dispatch_next( vtos );
}


void TemplateTable::_return(TosState state) {
  transition(state, state);
  assert(Rtos == V0, "expects top of stack to be return value reg");
  assert(_desc->calls_vm(), "inconsistent calls_vm information");

  if (_desc->bytecode() == Bytecodes::_return_register_finalizer) {
    assert(state == vtos, "only valid state");
    __ move(Rscratch3, R0);
    __ access_local_ptr(Rtos, Rscratch3);  // "this" pointer
    __ ld_ptr(Rscratch2, oopDesc::klass_offset_in_bytes(), Rtos);
    __ lw(Rscratch2, Klass::access_flags_offset_in_bytes() + sizeof(oopDesc), Rscratch2);
    __ andi(Rscratch2, Rscratch2, JVM_ACC_HAS_FINALIZER);
    Label skip_register_finalizer;
    __ beq(Rscratch2, R0, skip_register_finalizer);
    __ delayed()->nop();

    // Call out to do finalizer registration
    __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::register_finalizer), Rtos);

    __ bind(skip_register_finalizer);
  }

  __ remove_activation(state, /* throw_monitor_exception */ true);

  __ j(RA);
  __ delayed()->move(SP, Rscratch5);
}


// ----------------------------------------------------------------------------
// Volatile variables demand their effects be made known to all CPU's in
// order.  Store buffers on most chips allow reads & writes to reorder; the
// JMM's ReadAfterWrite.java test fails in -Xint mode without some kind of
// memory barrier (i.e., it's not sufficient that the interpreter does not
// reorder volatile references, the hardware also must not reorder them).
// 
// According to the new Java Memory Model (JMM):
// (1) All volatiles are serialized wrt to each other.  
// ALSO reads & writes act as aquire & release, so:
// (2) A read cannot let unrelated NON-volatile memory refs that happen after
// the read float up to before the read.  It's OK for non-volatile memory refs
// that happen before the volatile read to float down below it.
// (3) Similar a volatile write cannot let unrelated NON-volatile memory refs
// that happen BEFORE the write float down to after the write.  It's OK for
// non-volatile memory refs that happen after the volatile write to float up
// before it.
//
// We only put in barriers around volatile refs (they are expensive), not
// _between_ memory refs (that would require us to track the flavor of the
// previous memory refs).  Requirements (2) and (3) require some barriers
// before volatile stores and after volatile loads.  These nearly cover
// requirement (1) but miss the volatile-store-volatile-load case.  This final
// case is placed after volatile-stores although it could just as well go
// before volatile-loads.
void TemplateTable::volatile_barrier() {
  __ sync();
}

// ----------------------------------------------------------------------------
void TemplateTable::resolve_cache_and_index(int byte_no, Register Rcache, Register index) {
  assert(byte_no == 1 || byte_no == 2, "byte_no out of range");
  // Depends on cpCacheOop layout!
  const int shift_count = (1 + byte_no)*BitsPerByte;
  Label resolved;

  __ get_cache_and_index_at_bcp(Rcache, index, 1);
  __ ld_ptr(Rbyte_code,
	    in_bytes(constantPoolCacheOopDesc::base_offset() +
		     ConstantPoolCacheEntry::indices_offset()),
	    Rcache);

  __ srl(Rbyte_code, Rbyte_code, shift_count);
  __ andi(Rbyte_code, Rbyte_code, 0xFF);
  __ li(A1, (int)bytecode());
  __ beq(Rbyte_code, A1, resolved);
  __ delayed()->nop();

  address entry;
  switch (bytecode()) {
    case Bytecodes::_getstatic      : // fall through
    case Bytecodes::_putstatic      : // fall through
    case Bytecodes::_getfield       : // fall through
    case Bytecodes::_putfield       : entry = CAST_FROM_FN_PTR(address, InterpreterRuntime::resolve_get_put); break;
    case Bytecodes::_invokevirtual  : // fall through
    case Bytecodes::_invokespecial  : // fall through
    case Bytecodes::_invokestatic   : // fall through
    case Bytecodes::_invokeinterface: entry = CAST_FROM_FN_PTR(address, InterpreterRuntime::resolve_invoke);  break;
    default                         : ShouldNotReachHere();                                 break;
  }
  // first time invocation - must resolve first
  __ call_VM(noreg, entry, A1); 
  // Update registers with resolved info
  __ get_cache_and_index_at_bcp(Rcache, index, 1);
  __ bind(resolved);
}

void TemplateTable::load_invoke_cp_cache_entry(int byte_no,
                                               Register method,
                                               Register Ritable_index,
                                               Register Rflags,
                                               bool is_invokevirtual,
                                               bool is_invokevfinal) {
  Register Rcache = Rscratch3;
  Register Rscratch = Rscratch4;
  assert_different_registers(Rcache, method, Ritable_index);

  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  // determine constant pool cache field offsets
  const int method_offset = in_bytes(
    cp_base_offset +
      (is_invokevirtual
       ? ConstantPoolCacheEntry::f2_offset()
       : ConstantPoolCacheEntry::f1_offset()
      )
    );
  const int flags_offset = in_bytes(cp_base_offset +
                                    ConstantPoolCacheEntry::flags_offset());
  // access constant pool cache fields
  const int index_offset = in_bytes(cp_base_offset + 
                                    ConstantPoolCacheEntry::f2_offset());

  if (is_invokevfinal) {
    __ get_cache_and_index_at_bcp(Rcache, Rscratch, 1);
  } else {
    resolve_cache_and_index(byte_no, Rcache, Rscratch);
  }

  __ ld_ptr(method, method_offset, Rcache);
  if (Ritable_index != noreg) {
    __ ld_ptr(Ritable_index, index_offset, Rcache);
  }
#if 0
  // SPARC does a ld_ptr here, but watch out -- the _flags field is a
  // long storing a 32-bit quantity, and if tos bits are "vtos" (8),
  // the high bit will be set and will have been sign-extended between
  // as_flags and set_flags!  So an ld will get us all those bits.
  // This is due to f and as_flags being not declared unsigned.
  // If that bug is fixed, we don't need this change.
  __ lwu(Rflags, flags_offset, Rcache);
#else
  __ ld_ptr(Rflags, flags_offset, Rcache);
#endif
}

// The Rcache register must be set before call
void TemplateTable::load_field_cp_cache_entry(Register Robj,
                                              Register Rcache,
                                              Register index,
                                              Register Roffset,
                                              Register Rflags,
                                              bool is_static) {
  assert_different_registers(Rcache, Rflags, Roffset);

  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  __ ld_ptr(Rflags,
	    in_bytes(cp_base_offset + ConstantPoolCacheEntry::flags_offset()),
	    Rcache);
  __ ld_ptr(Roffset,
	    in_bytes(cp_base_offset + ConstantPoolCacheEntry::f2_offset()),
	    Rcache);
  if (is_static) {
    __ ld_ptr(Robj,
	      in_bytes(cp_base_offset + ConstantPoolCacheEntry::f1_offset()),
	      Rcache);
  }
}

// The registers Rcache and index expected to be set before call.
// Correct values of the Rcache and index registers are preserved.
void TemplateTable::jvmti_post_field_access(Register Rcache,
                                            Register index,
                                            bool is_static,
                                            bool has_tos) {
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  if (JvmtiExport::can_post_field_access()) { 
    assert(0, "unimplemented");
  }
}

void TemplateTable::getfield_or_static(int byte_no, bool is_static) {
  transition(vtos, vtos);

  // We barely fit in the 6 scratch registers.
  // I am marking their uses with numbers so we can follow our "reg alloc".

  Register R0flags     = Rscratch0;
  Register R1volatile  = Rscratch1;
  Register R2fieldaddr = Rscratch2;
  Register R3class     = Rscratch3;
  Register R4offset    = Rscratch4;
  // These are used where R3class, R4offset are dead,
  // i.e. right before b(checkVolatile):
  Register R3scratch   = R3class;
  Register R4scratch   = R4offset;

  {
  Register R1cache  = R1volatile;  // R1volatile is not in use yet
  Register R4index  = R4offset;
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  resolve_cache_and_index(byte_no, R1cache, R4index);
  jvmti_post_field_access(R1cache, noreg /*R4index*/, is_static, false);
  load_field_cp_cache_entry(R3class, R1cache, noreg /*R4index*/, R4offset,
			    R0flags, is_static);
  }
  // R1cache and R4index are no longer needed

  if (!is_static) {
    pop_and_check_object(R3class);
  } else {
    __ verify_oop(R3class);
  }

  Label exit;

  {
    // Get volatile flag
    __ andi(R1volatile, R0flags, (1 << ConstantPoolCacheEntry::volatileField));
  }

  Label checkVolatile;

  // compute field type
  Label notByte, notInt, notShort, notChar, notLong, notFloat, notObj;
  __ srl(R0flags, R0flags, ConstantPoolCacheEntry::tosBits);
  // Make sure we don't need to mask R0flags for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();

  // Check atos before itos for getstatic, more likely (in Queens at least)
  bnei(R0flags, atos, notObj, Rscratch5);
  __ delayed()->addu_long(R2fieldaddr, R3class, R4offset);

  // atos
  __ ld_ptr(Rtos, 0, R2fieldaddr);
  __ verify_oop(Rtos);
  __ push(atos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_agetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notObj);

  // cmp(Rflags, itos);
  bnei(R0flags, itos, notInt, Rscratch5);
  __ delayed()->nop();

  // itos
  __ lw(Rtos, 0, R2fieldaddr);
  __ push(itos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_igetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notInt);

  // cmp(Rflags, ltos);
  bnei(R0flags, ltos, notLong, Rscratch5);
  __ delayed()->nop();

  // ltos
  // load must be atomic
  __ ld_long(Rtos, 0, R2fieldaddr);
  __ push(ltos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_lgetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notLong);

  // cmp(Rflags, btos);
  bnei(R0flags, btos, notByte, Rscratch5);
  __ delayed()->nop();

  // btos
  __ lb(Rtos, 0, R2fieldaddr);
  __ push(itos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_bgetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notByte);

  // cmp(Rflags, ctos);
  bnei(R0flags, ctos, notChar, Rscratch5);
  __ delayed()->nop();

  // ctos
  __ lhu(Rtos, 0, R2fieldaddr);
  __ push(itos);
  if (! is_static) {
    // patch_bytecode(Bytecodes::_fast_cgetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notChar);

  // cmp(Rflags, stos);
  bnei(R0flags, stos, notShort, Rscratch5);
  __ delayed()->nop();

  // stos
  __ lh(Rtos, 0, R2fieldaddr);
  __ push(itos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_sgetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notShort);


  // cmp(Rflags, ftos);
  bnei(R0flags, ftos, notFloat, Rscratch5);
  __ delayed()->nop();

  // ftos
  __ lwc1(Ftos, 0, R2fieldaddr);
  __ push(ftos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_fgetfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notFloat);


  // dtos
  __ ldc1(Ftos, 0, R2fieldaddr);
  __ push(dtos);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_dgetfield, R3scratch, R4scratch);
  }

  __ bind(checkVolatile);
  {
    __ beq(R1volatile, R0, exit);
    __ delayed()->nop();
    volatile_barrier();
  }

  __ bind(exit);
}

void TemplateTable::getfield(int byte_no) {
  getfield_or_static(byte_no, false);
}

void TemplateTable::getstatic(int byte_no) {
  getfield_or_static(byte_no, true);
}


void TemplateTable::fast_accessfield(TosState state) {
  transition(atos, state);
  Register Rcache  = Rscratch0;
  Register index   = Rscratch1;
  Register Roffset = Rscratch1;
  Register Rflags  = Rcache;
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  __ get_cache_and_index_at_bcp(Rcache, index, 1);
  jvmti_post_field_access(Rcache, index, /*is_static*/false, /*has_tos*/true);

  __ ld_ptr(Roffset, in_bytes(cp_base_offset + ConstantPoolCacheEntry::f2_offset()), Rcache);

  __ null_check(Rtos);
  __ verify_oop(Rtos);

  Label exit;

  {
    // Get volatile flag
    __ ld_ptr(Rflags,
	      in_bytes(cp_base_offset + ConstantPoolCacheEntry::f2_offset()),
	      Rcache);
    __ li(Rscratch2, 1 << ConstantPoolCacheEntry::volatileField);
  }

  switch (bytecode()) {
    case Bytecodes::_fast_bgetfield: 
      __ lb(Rtos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_cgetfield: 
      __ lhu(Rtos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_sgetfield: 
      __ lh(Rtos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_igetfield: 
      __ lw(Rtos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_lgetfield: 
      __ ld_long(Rtos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_fgetfield: 
      __ lwxc1(Ftos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_dgetfield: 
      __ ldxc1(Ftos, Roffset, Rtos);
      break;
    case Bytecodes::_fast_agetfield:
      __ ld_ptr(Rtos, Roffset, Rtos);
      break;
    default:
      ShouldNotReachHere();
  }
  
  {
    __ and_(Rscratch2, Rscratch2, Rflags);
    __ beq(Rscratch2, R0, exit);
    __ delayed()->nop();
    volatile_barrier();
    __ bind(exit);
  }

  if (state == atos) {
    __ verify_oop(Rtos);	// does not blow flags!
  }
}

void TemplateTable::jvmti_post_fast_field_mod() {
  if (JvmtiExport::can_post_field_modification()) { 
    // YYY unimplemented
    ShouldNotReachHere();
  }
}

// The registers Rcache and index expected to be set before call.
// The function may destroy various registers, just not the Rcache and index registers.
void TemplateTable::jvmti_post_field_mod(Register Rcache, Register index, bool is_static) {
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  if (JvmtiExport::can_post_field_modification()) {
    // YYY unimplemented
    ShouldNotReachHere();
  }
}

void TemplateTable::pop_and_check_object(Register r) {
  __ pop_ptr(r);
  __ null_check(r);  // for field access must check obj.
  __ verify_oop(r);
}

void TemplateTable::putfield_or_static(int byte_no, bool is_static) {
  transition(vtos, vtos);

  // tight control over scratch registers, like in getfield_or_static()
  Register R0flags     = Rscratch0;
  Register R1volatile  = Rscratch1;
  Register R3class     = Rscratch3;
  Register R4offset    = Rscratch4;
  // these are used where R3class, R4offset are dead:
  Register R3scratch   = R3class;
  Register R4scratch   = R4offset;

  {
  Register R1cache  = R1volatile;  // R1volatile is not in use yet
  Register R4index  = R4offset;
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  resolve_cache_and_index(byte_no, R1cache, R4index);
  jvmti_post_field_mod(R1cache, noreg /*R4index*/, is_static);
  load_field_cp_cache_entry(R3class, R1cache, noreg /*R4index*/, R4offset,
			    R0flags, is_static);
  }
  // R1cache and R4index are no longer needed

  Label notVolatile, checkVolatile, exit;
  {
    // Get volatile flag
    __ andi(R1volatile, R0flags, (1 << ConstantPoolCacheEntry::volatileField));
    
    {
      // check need for barrier before write
      __ beq(R1volatile, R0, notVolatile);
      __ delayed()->nop();
      volatile_barrier();
      __ bind(notVolatile);
    }
  }

  __ srl(R0flags, R0flags, ConstantPoolCacheEntry::tosBits);
  // Make sure we don't need to mask R0flags for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();

  // compute field type
  Label notInt, notShort, notChar, notObj, notByte, notLong, notFloat;

  // Check atos before itos for getstatic, more likely (in Queens at least)
  // cmp(Rflags, atos ); 
  bnei(R0flags, atos, notObj, Rscratch5);
  __ delayed()->nop();

  // atos
  __ pop_ptr();
  if (! is_static) pop_and_check_object(R3class);
  __ verify_oop(Rtos);
  __ st_ptr(Rtos, R4offset, R3class);
  __ store_check(Rscratch5 /*or Rscratch0*/, R3class, R4offset);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_aputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notObj);

  // cmp(Rflags, itos );
  bnei(R0flags, itos, notInt, Rscratch5);
  __ delayed()->nop();

  // itos
  __ pop_i();
  if (! is_static) pop_and_check_object(R3class);
  __ sw(Rtos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_iputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notInt);

  // cmp(Rflags, ltos );
  bnei(R0flags, ltos, notLong, Rscratch5);
  __ delayed()->nop();

  // ltos
  __ pop_l();
  if (! is_static) pop_and_check_object(R3class);
  __ st_long(Rtos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_lputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notLong);

  // cmp(Rflags, btos );
  bnei(R0flags, btos, notByte, Rscratch5);
  __ delayed()->nop();

  // btos
  __ pop_i();
  if (! is_static) pop_and_check_object(R3class);
  __ sb(Rtos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_bputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notByte);

  // cmp(Rflags, ctos );
  bnei(R0flags, ctos, notChar, Rscratch5);
  __ delayed()->nop();

  // ctos (char)
  __ pop_i();
  if (! is_static) pop_and_check_object(R3class);
  __ sh(Rtos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_cputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notChar);
  // cmp(Rflags, stos ); 
  bnei(R0flags, stos, notShort, Rscratch5);
  __ delayed()->nop();

  // stos (char)
  __ pop_i();
  if (! is_static) pop_and_check_object(R3class);
  __ sh(Rtos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_sputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notShort);
  // cmp(Rflags, ftos );
  bnei(R0flags, ftos, notFloat, Rscratch5);
  __ delayed()->nop();

  // ftos
  __ pop_f();
  if (! is_static) pop_and_check_object(R3class);
  __ swc1(Ftos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_fputfield, R3scratch, R4scratch);
  }
  __ b(checkVolatile);
  __ delayed()->nop();

  __ bind(notFloat);

  // dtos
  __ pop_d();
  if (! is_static) pop_and_check_object(R3class);
  __ sdc1(Ftos, R4offset, R3class);
  if (! is_static) {
    patch_bytecode(Bytecodes::_fast_dputfield, R3scratch, R4scratch);
  }

  __ bind(checkVolatile);

  {
    __ beq(R1volatile, R0, exit);
    __ delayed()->nop();
    volatile_barrier();
    __ bind(exit);
  }
}

void TemplateTable::fast_storefield(TosState state) {
  transition(state, vtos);
  Register Rcache = Rscratch0;
  Register Rclass = Rcache;
  Register Roffset= Rscratch1;
  Register Rflags = Rscratch2;
  ByteSize cp_base_offset = constantPoolCacheOopDesc::base_offset();

  jvmti_post_fast_field_mod();

  __ get_cache_and_index_at_bcp(Rcache, Rscratch3, 1);

  Label notVolatile, checkVolatile, exit;
  {
    __ ld_ptr(Rflags,
	      in_bytes(cp_base_offset +
		       ConstantPoolCacheEntry::flags_offset()),
	      Rcache);
    __ andi(Rscratch3, Rflags, 1 << ConstantPoolCacheEntry::volatileField);
    {
      __ beq(Rscratch3, R0, notVolatile);
      __ delayed()->nop();
      volatile_barrier();
      __ bind(notVolatile);
    }
  }

  __ ld_ptr(Roffset,
	    in_bytes(cp_base_offset +
		     ConstantPoolCacheEntry::f2_offset()),
	    Rcache);
  pop_and_check_object(Rclass);

  switch (bytecode()) {
    case Bytecodes::_fast_bputfield:
      __ sb(Rtos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_cputfield: /* fall through */
    case Bytecodes::_fast_sputfield:
      __ sh(Rtos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_iputfield:
      __ sw(Rtos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_lputfield:
      __ st_long(Rtos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_fputfield:
      __ swc1(Ftos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_dputfield:
      __ sdc1(Ftos, Roffset, Rclass);
      break;
    case Bytecodes::_fast_aputfield:
      __ st_ptr(Rtos, Roffset, Rclass);
      __ store_check(Rscratch3, Rclass, Roffset);
      break;
    default:
      ShouldNotReachHere();
  }

  {
    __ beq(Rscratch3, R0, exit);
    __ delayed()->nop();
    volatile_barrier();
    __ bind(exit);
  }
}


void TemplateTable::putfield(int byte_no) {
  putfield_or_static(byte_no, false);
}

void TemplateTable::putstatic(int byte_no) {
  putfield_or_static(byte_no, true);
}


void TemplateTable::fast_xaccess(TosState state) {
  transition(vtos, state);
  Register Rcache = Rscratch0;
  Register Roffset = Rscratch1;
  Register Rflags = Rscratch1;
  Register Rreceiver = Rscratch2;

  __ ld_ptr(Rreceiver, Interpreter::value_offset_in_bytes(), Rlocals);

  // access constant pool cache  (is resolved)
  __ get_cache_and_index_at_bcp(Rcache, Rscratch1, 2);
  __ ld_ptr(Roffset,
	    in_bytes(constantPoolCacheOopDesc::base_offset() +
		     ConstantPoolCacheEntry::f2_offset()),
	    Rcache);
  __ daddiu(Rbcp, Rbcp, 1); // needed to report exception at the correct bcp

  __ verify_oop(Rreceiver);
  __ null_check(Rreceiver);
  if (state == atos) { 
    __ ld_ptr(Rtos, Roffset, Rreceiver);
  } else if (state == itos) {
    __ lw(Rtos, Roffset, Rreceiver);
  } else if (state == ftos) {
    __ lwxc1(Ftos, Roffset, Rreceiver);
  } else {
    ShouldNotReachHere();
  }

  {

    // Get is_volatile value in Rflags and check if membar is needed
    __ ld_ptr(Rflags,
	      in_bytes(constantPoolCacheOopDesc::base_offset() +
		       ConstantPoolCacheEntry::flags_offset()),
	      Rcache);

    // Test volatile
    Label notVolatile;
    __ andi(Rscratch3, Rflags, 1 << ConstantPoolCacheEntry::volatileField);
    __ beq(Rscratch3, R0, notVolatile);
    __ delayed()->nop();
    volatile_barrier();
    __ bind(notVolatile);
  }

  __ interp_verify_oop(Rtos, state, __FILE__, __LINE__);
  __ addu_long(Rbcp, Rbcp, -1);
}

//----------------------------------------------------------------------------------------------------
// Calls

void TemplateTable::count_calls(Register method, Register temp) {
  // implemented elsewhere
  ShouldNotReachHere();
}

void TemplateTable::generate_vtable_call(Register Rrecv, Register Rindex, Register Rret) {
  Register Rtemp = Rscratch4;
  assert_different_registers(Rtemp, Rindex, Rmethod_call, Rret);

  // get target methodOop & entry point
  const int base = instanceKlass::vtable_start_offset() * wordSize;
  if (vtableEntry::size() % 3 == 0) {
    // scale the vtable index by 12:
    int one_third = vtableEntry::size() / 3;
    __ sll_long(Rtemp, Rindex, exact_log2(one_third * 1 * wordSize));
    __ sll_long(Rindex, Rindex, exact_log2(one_third * 2 * wordSize));
    __ addu_long(Rindex, Rindex, Rtemp);
  } else {
    // scale the vtable index by 8:
    __ sll_long(Rindex, Rindex, exact_log2(vtableEntry::size() * wordSize));
  }

  __ addu_long(Rrecv, Rindex, Rrecv);
  __ ld_ptr(Rrecv, base + vtableEntry::method_offset_in_bytes(), Rmethod_call);

  __ call_from_interpreter(Rindex, Rscratch0, Rret);
}

void TemplateTable::invokevirtual(int byte_no) {
  transition(vtos, vtos);

  Register Rscratch = Rscratch3;
  Register Rtemp = Rscratch4;
  Register Rret = Rscratch2;
  Register Rrecv = Rmethod_call;
  Label notFinal;

  load_invoke_cp_cache_entry(byte_no, Rmethod_call, noreg, Rret, true);
  // YYY _savedSP
  // I don't think we need to save SP here at all.  This is just a
  // placeholder in case I'm wrong and we have to go back and do it.

  // Check for vfinal
  __ li(Rscratch4, (1 << ConstantPoolCacheEntry::vfinalMethod));
  __ and_(Rscratch4, Rret, Rscratch4);
  __ beq(Rscratch4, R0, notFinal);
  __ delayed()->andi(Rscratch4, Rret, 0xFF);  // gets number of parameters

  patch_bytecode(Bytecodes::_fast_invokevfinal, Rscratch, Rtemp);

  invokevfinal_helper(Rscratch, Rret);

  __ bind(notFinal);

  __ move(Rscratch, Rmethod_call);
  __ load_receiver(A0, Rscratch4);  // gets receiverOop
  __ verify_oop(A0);

  // get return address
  __ li(Rtemp, (address)Interpreter::return_3_addrs_by_index_table());
  __ srl_long(Rret, Rret, ConstantPoolCacheEntry::tosBits);  // get return type
  // Make sure we don't need to mask Rret for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();
  __ sll_long(Rret, Rret, LogBytesPerWord);
  __ ld_ptr(Rret, Rtemp, Rret);	   // get return address

  // get receiver klass
  __ null_check(A0, oopDesc::klass_offset_in_bytes());
  __ ld_ptr(Rrecv, oopDesc::klass_offset_in_bytes(), A0);
  __ verify_oop(Rrecv);

  __ profile_virtual_call(Rrecv, A4);

  generate_vtable_call(Rrecv, Rscratch, Rret);
}

void TemplateTable::fast_invokevfinal(int byte_no) {
  transition(vtos, vtos);

  load_invoke_cp_cache_entry(byte_no, Rmethod_call, noreg, Rscratch0, true,
                             /*is_invokevfinal*/true);
  // YYY _savedSP
  invokevfinal_helper(Rscratch3, Rsave0);
}

void TemplateTable::invokevfinal_helper(Register Rscratch, Register Rret) {
  Register Rtemp = Rscratch4;
  assert_different_registers(Rscratch, Rtemp, Rscratch0);

  __ verify_oop(Rmethod_call);

  // Load receiver from stack slot
  __ lhu(Rtemp, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod_call);
  __ load_receiver(A0, Rtemp);

  // receiver NULL check
  __ null_check(A0);

  __ profile_final_call(Rscratch0);

  // get return address
  __ li(Rtemp, (address)Interpreter::return_3_addrs_by_index_table());
  __ srl_long(Rret, Rret, ConstantPoolCacheEntry::tosBits);  // get return type
  // Make sure we don't need to mask Rret for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();
  __ sll_long(Rret, Rret, LogBytesPerWord);
  __ ld_ptr(Rret, Rtemp, Rret);	   // get return address

  // do the call
  __ call_from_interpreter(Rscratch, Rscratch0, Rret);
}

void TemplateTable::invokespecial(int byte_no) {
  transition(vtos, vtos);

  Register Rscratch = Rscratch3;
  Register Rtemp = Rscratch4;
  Register Rret = Rsave0;

  load_invoke_cp_cache_entry(byte_no, Rmethod_call, noreg, Rret, false);
  // YYY _savedSP

  __ verify_oop(Rmethod_call);

  __ lhu(Rtemp, in_bytes(methodOopDesc::size_of_parameters_offset()), Rmethod_call);
  __ load_receiver(A0, Rtemp);

  // receiver NULL check
  __ null_check(A0);

  // get return address
  __ li(Rtemp, (address)Interpreter::return_3_addrs_by_index_table());
  __ srl_long(Rret, Rret, ConstantPoolCacheEntry::tosBits);  // get return type
  // Make sure we don't need to mask Rret for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();
  __ sll_long(Rret, Rret, LogBytesPerWord);
  __ ld_ptr(Rret, Rret, Rtemp);  // get return address

  // do the call
  __ call_from_interpreter(Rscratch, Rscratch0, Rret);
}

void TemplateTable::invokestatic(int byte_no) {
  transition(vtos, vtos);

  Register Rscratch = Rscratch3;
  Register Rtemp = Rscratch4;
  Register Rret = Rsave0;

  load_invoke_cp_cache_entry(byte_no, Rmethod_call, noreg, Rret, false);
  // YYY _savedSP

  __ verify_oop(Rmethod_call);

  __ profile_call(Rscratch0);

  // get return address
  __ li(Rtemp, (address)Interpreter::return_3_addrs_by_index_table());
  __ srl_long(Rret, Rret, ConstantPoolCacheEntry::tosBits);  // get return type
  // Make sure we don't need to mask Rret for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();
  __ sll_long(Rret, Rret, LogBytesPerWord);
  __ ld_ptr(Rret, Rret, Rtemp);  // get return address

  // do the call
  __ call_from_interpreter(Rscratch, Rscratch0, Rret);
}


void TemplateTable::invokeinterface_object_method(Register RklassOop,
                                                  Register Rinterface,
                                                  Register Rret,
                                                  Register Rflags) {
  Register Rscratch = Rscratch4;
  Register Rindex = Rsave0;

  assert_different_registers(Rscratch, Rindex, Rret);

  Label notFinal;

  // Check for vfinal 
  __ li(Rscratch, (1 << ConstantPoolCacheEntry::vfinalMethod));
  __ and_(Rscratch, Rscratch, Rflags);
  __ beq(Rscratch, R0, notFinal);
  __ delayed()->nop();

  __ profile_final_call(Rscratch0);

  // do the call
  assert_different_registers(Rmethod_call, Rscratch0, Rinterface);
  __ move(Rmethod_call, Rindex);
  __ call_from_interpreter(Rinterface, Rscratch0, Rret);

  __ bind(notFinal);

  __ profile_virtual_call(RklassOop, Rscratch0);
  generate_vtable_call(RklassOop, Rindex, Rret);
}

void TemplateTable::invokeinterface(int byte_no) {
  transition(vtos, vtos);

  Register Rscratch = Rscratch4;
  Register Rret = Rscratch3;
  Register Rindex = Rscratch2;
  Register Rinterface = Rscratch1;
  Register RklassOop = Rmethod_call;
  Register Rflags = A1;
  assert_different_registers(Rscratch, Rmethod_call);

  load_invoke_cp_cache_entry(byte_no, Rinterface, Rindex, Rflags, false);
  // YYY _savedSP

  // get receiver
  __ andi(Rscratch, Rflags, 0xFF);
  __ load_receiver(A0, Rscratch);
  __ verify_oop(A0);

  __ move(Rret, Rflags);

  // get return address
  __ li(Rscratch, (address)Interpreter::return_5_addrs_by_index_table());
  __ srl_long(Rret, Rret, ConstantPoolCacheEntry::tosBits);  // get return type
  // Make sure we don't need to mask Rret for tosBits after the above shift
  ConstantPoolCacheEntry::verify_tosBits();
  __ sll_long(Rret, Rret, LogBytesPerWord);
  __ ld_ptr(Rret, Rret, Rscratch);  // get return address

  // get receiver klass
  __ null_check(A0, oopDesc::klass_offset_in_bytes());
  __ ld_ptr(RklassOop, oopDesc::klass_offset_in_bytes(), A0);
  __ verify_oop(RklassOop);

  // Special case of invokeinterface called for virtual method of
  // java.lang.Object.  See cpCacheOop.cpp for details.
  // This code isn't produced by javac, but could be produced by
  // another compliant java compiler.
  Label notMethod;
  __ li(Rscratch, (1 << ConstantPoolCacheEntry::methodInterface));
  __ and_(Rscratch, Rscratch, Rflags);
  __ beq(Rscratch, R0, notMethod);
  __ delayed()->nop();

  invokeinterface_object_method(RklassOop, Rinterface, Rret, Rflags);

  __ bind(notMethod);

  __ profile_virtual_call(RklassOop, Rscratch0);

  //
  // find entry point to call
  //

  // compute start of first itableOffsetEntry (which is at end of vtable)
  const int base = instanceKlass::vtable_start_offset() * wordSize;
  Label search, found;
  Register Rtemp = Rflags;

  __ lwu(Rtemp, instanceKlass::vtable_length_offset() * wordSize, RklassOop);
  if (align_object_offset(1) > 1) {
    __ round_to(Rtemp, align_object_offset(1));
  }
  __ sll_long(Rtemp, Rtemp, LogBytesPerWord);  // Rscratch *= 4
  __ addu_long(Rtemp, Rtemp, base);
  __ addu_long(Rscratch, RklassOop, Rtemp);

  __ bind(search);

  __ ld_ptr(Rtemp, itableOffsetEntry::interface_offset_in_bytes(), Rscratch);
  {
    Label ok;

    // Check that entry is non-null.  Null entries are probably a bytecode
    // problem.  If the interface isn't implemented by the reciever class,
    // the VM should throw IncompatibleClassChangeError.  linkResolver checks
    // this too but that's only if the entry isn't already resolved, so we
    // need to check again.
    __ bne(Rtemp, R0, ok);
    __ delayed()->nop();
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_IncompatibleClassChangeError));
    __ should_not_reach_here();

    __ bind(ok);
    __ verify_oop(Rtemp);
  }

  __ verify_oop(Rinterface);

  __ beq(Rinterface, Rtemp, found);
  __ delayed()->nop();

  __ b(search);
  __ delayed()->addu_long(Rscratch, Rscratch, itableOffsetEntry::size() * wordSize);

  __ bind(found);
  // entry found and Rscratch points to it
  __ ld_long(Rscratch, itableOffsetEntry::offset_offset_in_bytes(), Rscratch);

  assert(itableMethodEntry::method_offset_in_bytes() == 0, "adjust instruction below");
  __ sll_long(Rindex, Rindex, exact_log2(itableMethodEntry::size() * wordSize));  // Rindex *= 8;
  __ addu_long(Rscratch, Rindex, Rscratch);
  __ ld_ptr(Rmethod_call, RklassOop, Rscratch);

  // Check for abstract method error.
  {
    Label ok;
    __ bne(Rmethod_call, R0, ok);
    __ delayed()->nop();
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_AbstractMethodError));
    __ should_not_reach_here();
    __ bind(ok);
  }

  assert_different_registers(Rinterface, Rmethod_call, Rscratch0, Rret);

  __ verify_oop(Rmethod_call);
  __ call_from_interpreter(Rinterface, Rscratch0, Rret);
}


//----------------------------------------------------------------------------------------------------
// Allocation

void TemplateTable::_new() {
  transition(vtos, atos);

  Label slow_case;
  Label done;
  Label initialize_header;
  Label initialize_object;  // including clearing the fields

  Register RallocatedObject = Rtos;
  Register RinstanceKlass = A1;
  Register Roffset = A3;
  Register Rscratch = A4;

  __ get_2_byte_integer_at_bcp(1, Rscratch, Roffset, false);
  __ get_cpool_and_tags(Rscratch, Rscratch3);
  // make sure the class we're about to instantiate has been resolved
  __ addu_long(Rscratch3, Rscratch3, typeArrayOopDesc::header_size(T_BYTE) * wordSize);
  __ lbu(Rscratch3, Roffset, Rscratch3);
  __ li(Rscratch0, JVM_CONSTANT_Class);
  __ bne(Rscratch3, Rscratch0, slow_case);
  __ delayed()->sll_long(Roffset, Roffset, LogBytesPerWord);

  __ addu_long(Roffset, Roffset, sizeof(constantPoolOopDesc));
  __ ld_ptr(RinstanceKlass, Roffset, Rscratch);

  // make sure klass is fully initialized:
  __ ld_ptr(Rscratch3, instanceKlass::init_state_offset_in_bytes() + sizeof(oopDesc), RinstanceKlass);
  __ li(Rscratch0, instanceKlass::fully_initialized);
  __ bne(Rscratch3, Rscratch0, slow_case);
  __ delayed()->lw(Roffset, Klass::layout_helper_offset_in_bytes() + sizeof(oopDesc), RinstanceKlass);
  // [above] get instance_size in instanceKlass (already aligned)

  // make sure klass does not have has_finalizer, or is abstract, or interface or java/lang/Class
  __ andi(Rscratch0, Roffset, Klass::_lh_instance_slow_path_bit);
  __ bne(Rscratch0, R0, slow_case);
  __ delayed()->nop();

  // allocate the instance
  // 1) Try to allocate in the TLAB
  // 2) if fail, and the TLAB is not full enough to discard, allocate in the shared Eden
  // 3) if the above fails (or is not applicable), go to a slow case
  // (creates a new TLAB, etc.)

  const bool allow_shared_alloc =
    Universe::heap()->supports_inline_contig_alloc() && !CMSIncrementalMode;

  if(UseTLAB) {
    Register RoldTopValue = RallocatedObject;
    Register RtopAddr = Rscratch3, RtlabWasteLimitValue = Rscratch3;
    Register RnewTopValue = Rscratch1;
    Register RendValue = Rscratch;
    Register RfreeValue = RnewTopValue; 

    // check if we can allocate in the TLAB
    __ ld_ptr(RoldTopValue, in_bytes(JavaThread::tlab_top_offset()), Rthread); // sets up RallocatedObject
    __ ld_ptr(RendValue, in_bytes(JavaThread::tlab_end_offset()), Rthread);
    __ addu_long(RnewTopValue, RoldTopValue, Roffset);

    // if there is enough space, we do not CAS and do not clear
    if (ZeroTLAB) {
      // the fields have already been cleared
      __ bleu(RnewTopValue, RendValue, initialize_header);
    } else {
      // initialize both the header and fields
      __ bleu(RnewTopValue, RendValue, initialize_object);
    }
    __ delayed()->st_ptr(RnewTopValue, in_bytes(JavaThread::tlab_top_offset()), Rthread);

    if (allow_shared_alloc) {
    // Check if tlab should be discarded (refill_waste_limit >= free)
      __ ld_ptr(RtlabWasteLimitValue, in_bytes(JavaThread::tlab_refill_waste_limit_offset()), Rthread);
      __ subu_long(RfreeValue, RendValue, RoldTopValue);
      __ srl_long(RfreeValue, RfreeValue, LogHeapWordSize);
      __ bgeu(RtlabWasteLimitValue, RfreeValue, slow_case); // tlab waste is small
      __ delayed()->nop();

      // increment waste limit to prevent getting stuck on this slow path
      __ addu_long(RtlabWasteLimitValue, RtlabWasteLimitValue, ThreadLocalAllocBuffer::refill_waste_limit_increment());
      __ st_ptr(RtlabWasteLimitValue, in_bytes(JavaThread::tlab_refill_waste_limit_offset()), Rthread);
    } else {
      // No allocation in the shared eden.
      __ b(slow_case);
      __ delayed()->nop();
    }
  }

  // Allocation in the shared Eden
  if (allow_shared_alloc) {
    Register RoldTopValue = Rscratch1;
    Register RtopAddr = Rscratch3;
    Register RnewTopValue = RallocatedObject;
    Register RendValue = Rscratch;

    __ li(RtopAddr, (intptr_t)Universe::heap()->top_addr());

    Label retry;
    __ bind(retry);
    __ li(RendValue, (intptr_t)Universe::heap()->end_addr());
    __ ld_ptr(RendValue, 0, RendValue);
    __ ll_long(RoldTopValue, 0, RtopAddr);
    __ addu_long(RnewTopValue, RoldTopValue, Roffset);

    // RnewTopValue contains the top address after the new object
    // has been allocated.
    __ bgtu(RnewTopValue, RendValue, slow_case);
    __ delayed()->nop();

    __ sc_long(RnewTopValue, 0, RtopAddr);

    // if someone beat us on the allocation, try again, otherwise continue
    __ beq(RnewTopValue, R0, retry);
    __ delayed()->nop();

    __ move(RallocatedObject, RoldTopValue);  // could move into delay slot
  }

  if (UseTLAB || Universe::heap()->supports_inline_contig_alloc()) {
    // clear object fields
    __ bind(initialize_object);
    __ addu_long(Roffset, Roffset, - sizeof(oopDesc));
    __ beq(Roffset, R0, initialize_header);
    __ delayed()->addu_long(Rscratch3, RallocatedObject, sizeof(oopDesc));

    // initialize remaining object fields
    {
      // [YYY/opt] Should redo this loop to dec the pointer itself
      // instead of using an offset.  This avoids an extra add in the
      // expansion of st_ptr.
      Label loop;
      __ addu_long(Roffset, Roffset, - wordSize);

      __ bind(loop);
      __ st_ptr(R0, Roffset, Rscratch3);
      __ bgtu(Roffset, R0, loop);
      __ delayed()->addu_long(Roffset, Roffset, - wordSize);
    }
    __ b(initialize_header);
    __ delayed()->nop();
  }

  // slow case
  __ bind(slow_case);
  __ get_2_byte_integer_at_bcp(1, Rscratch3, A2, false);
  __ get_constant_pool(A1);

  call_VM(Rtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::_new), A1, A2);

  __ b(done);
  __ delayed()->nop();


  // Initialize the header: mark, klass
  __ bind(initialize_header);

  if (UseBiasedLocking) {
    __ ld_ptr(Rscratch4, Klass::prototype_header_offset_in_bytes() + sizeof(oopDesc), RinstanceKlass);
  } else {
    __ li(Rscratch4, (intptr_t)markOopDesc::prototype());
  }
  __ st_ptr(Rscratch4, oopDesc::mark_offset_in_bytes(), RallocatedObject);       // mark
  __ st_ptr(RinstanceKlass, oopDesc::klass_offset_in_bytes(), RallocatedObject); // klass

  {
    SkipIfEqual skip_if(
      _masm, Rscratch4, R0, &DTraceAllocProbes, Assembler::equal);
    // Trigger dtrace event
    __ push(atos);
    __ call_VM_leaf(noreg, 
       CAST_FROM_FN_PTR(address, SharedRuntime::dtrace_object_alloc), A0);
    __ pop(atos);
  }

  // continue
  __ bind(done);
}



void TemplateTable::newarray() {
  transition(itos, atos);
  __ lbu(A1, 1, Rbcp);
     call_VM(Rtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::newarray), A1, Rtos);
}


void TemplateTable::anewarray() {
  transition(itos, atos);
  __ get_constant_pool(A1);
  __ get_2_byte_integer_at_bcp(1, Rscratch4, A2, false);
     call_VM(Rtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::anewarray), A1, A2, Rtos);
}


void TemplateTable::arraylength() {
  transition(atos, itos);
  Label ok;
  __ verify_oop(Rtos);
  // Explicit null-pointer check
  __ bne(Rtos, R0, ok);
  __ delayed()->lw(Rtos, arrayOopDesc::length_offset_in_bytes(), Rtos);

  __ jump_to(Interpreter::_throw_NullPointerException_entry);
  __ delayed()->nop();

  __ bind(ok);
}


void TemplateTable::checkcast() {
  transition(atos, atos);
  Label done, is_null, quicked, cast_ok, resolved;
  Register Roffset = Rscratch1;
  Register RobjKlass = A5;
  Register RspecifiedKlass = A4;

  // Check for casting a NULL
  __ beq(Rtos, R0, is_null);
  __ delayed()->nop();

  // Get value klass in RobjKlass
  __ ld_ptr(RobjKlass, oopDesc::klass_offset_in_bytes(), Rtos);

  // Get constant pool tag
  __ get_2_byte_integer_at_bcp(1, Rscratch0, Roffset, false);

  // See if the checkcast has been quickened
  __ get_cpool_and_tags(Rscratch0, Rscratch3);
  __ addu_long(Rscratch3, Rscratch3, typeArrayOopDesc::header_size(T_BYTE) * wordSize);
  __ lbu(Rscratch3, Roffset, Rscratch3);
  __ li(Rscratch2, JVM_CONSTANT_Class);
  __ beq(Rscratch3, Rscratch2, quicked);
  __ delayed()->sll_long(Roffset, Roffset, LogBytesPerWord);

  __ push_ptr(); // save receiver for result, and for GC
  call_VM(RspecifiedKlass, CAST_FROM_FN_PTR(address, InterpreterRuntime::quicken_io_cc) );
  __ pop_ptr(Rtos, Rscratch3); // restore receiver

  __ b(resolved);
  __ delayed()->ld_ptr(RobjKlass, oopDesc::klass_offset_in_bytes(), Rtos); // get value klass

  // Extract target class from constant pool
  __ bind(quicked);
  __ addu_long(Roffset, Roffset, sizeof(constantPoolOopDesc));
  __ ld_ptr(RspecifiedKlass, Roffset, Rscratch0);
  __ bind(resolved);

  // Generate a fast subtype check.  Branch to cast_ok if no
  // failure.  Throw exception if failure.
  __ gen_subtype_check(RobjKlass, RspecifiedKlass, Rscratch3, Rscratch4, Rscratch1, cast_ok );

  // Not a subtype; so must throw exception
  __ jump_to(Interpreter::_throw_ClassCastException_entry);
  __ delayed()->nop();

  __ bind(cast_ok);
  if (ProfileInterpreter) {
    __ b(done);
    __ delayed()->nop();
  }
  __ bind(is_null);
  __ profile_null_seen(Rscratch3);
  __ bind(done);
}

void TemplateTable::instanceof() {
  Label done, is_null, quicked, resolved;
  transition(atos, itos);
  Register Roffset = Rscratch1;
  Register RobjKlass = A5;
  Register RspecifiedKlass = A4;

  // Check for casting a NULL
  __ beq(Rtos, R0, is_null);
  __ delayed()->nop();

  // Get value klass in RobjKlass
  __ ld_ptr(RobjKlass, oopDesc::klass_offset_in_bytes(), Rtos); // get value klass

  // Get constant pool tag
  __ get_2_byte_integer_at_bcp(1, Rscratch0, Roffset, false);

  // See if the instanceof has been quickened
  __ get_cpool_and_tags(Rscratch0, Rscratch3);
  __ addu_long(Rscratch3, Rscratch3, typeArrayOopDesc::header_size(T_BYTE) * wordSize);
  __ lbu(Rscratch3, Roffset, Rscratch3);
  __ li(Rscratch2, JVM_CONSTANT_Class);
  __ beq(Rscratch3, Rscratch2, quicked);
  __ delayed()->sll_long(Roffset, Roffset, LogBytesPerWord);

  __ push_ptr(); // save receiver for result, and for GC
  call_VM(RspecifiedKlass, CAST_FROM_FN_PTR(address, InterpreterRuntime::quicken_io_cc) );
  __ pop_ptr(Rtos, Rscratch3); // restore receiver

  __ b(resolved);
  __ delayed()->ld_ptr(RobjKlass, oopDesc::klass_offset_in_bytes(), Rtos); // get value klass

  // Extract target class from constant pool
  __ bind(quicked);
  __ addu_long(Roffset, Roffset, sizeof(constantPoolOopDesc));
  __ get_constant_pool(Rscratch0);
  __ ld_ptr(RspecifiedKlass, Roffset, Rscratch0);
  __ bind(resolved);

  // Generate a fast subtype check.  Return 1 if no failure.
  // Return 0 if failure.
  __ li32(Rtos, 1);      // set result assuming quick tests succeed
  __ gen_subtype_check(RobjKlass, RspecifiedKlass, Rscratch3, Rscratch4, Rscratch1, done);
  // Not a subtype; return 0;
  __ li32(Rtos, 0);

  if (ProfileInterpreter) {
    __ b(done);
    __ delayed()->nop();
  }
  __ bind(is_null);
  __ profile_null_seen(Rscratch3);
  __ bind(done);
}

void TemplateTable::_breakpoint() {

   // Note: We get here even if we are single stepping..
   // jbug inists on setting breakpoints at every bytecode 
   // even if we are in single step mode.  
 
   transition(vtos, vtos);
   // get the unpatched byte code
   __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::get_original_bytecode_at), Rmethod, Rbcp);
   __ move(A0, Rbyte_code);

   // post the breakpoint event
   __ call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::_breakpoint), Rmethod, Rbcp);

   // complete the execution of original bytecode
   __ dispatch_normal(vtos);
}


//----------------------------------------------------------------------------------------------------
// Exceptions

void TemplateTable::athrow() {
  transition(atos, vtos);

  // Exception is on top of the stack,
  // but throw_exception_entry expects it in Rexception (A0).

  __ verify_oop(Rtos);
  __ null_check(Rtos);
  __ move(Rexception, Rtos);
  __ jump_to(Interpreter::throw_exception_entry());
  __ delayed()->nop();
}


//----------------------------------------------------------------------------------------------------
// Synchronization


void TemplateTable::monitorenter() {
  transition(atos, vtos);
  __ verify_oop(Rtos);
  // Try to acquire a lock on the object.
  // Repeat until succeeded (i.e., until monitorenter returns true).

  {
    Label ok;
    __ bne(Rtos, R0, ok);
    __ delayed()->move(Rscratch0, Rtos);  // save obj
    __ jump_to(Interpreter::_throw_NullPointerException_entry);
    __ delayed()->nop();
    __ bind(ok);
  }

  // find a free slot in the monitor block

  // initialize entry pointer
  __ move(A1, R0);

  {
    Label entry, loop, exit;
    __ get_top_most_monitor(A2);  // last one to check
    __ b(entry);
    __ delayed()->move(A3, Rmonitors);      // first one to check

    __ bind(loop);
    __ verify_oop(A4);              // verify each monitor's oop
    __ movz(A1, A3, A4);            // A1 <- A3 if A4==0
    __ beq(A4, Rtos, exit);  // check if current entry is for same object
    __ delayed()->addu_long(A3, A3, frame::interpreter_frame_monitor_size() * wordSize ); // check next one

    __ bind(entry);
    __ bleu(A3, A2, loop);
    __ delayed()->ld_ptr(A4, BasicObjectLock::obj_offset_in_bytes(), A3);

    __ bind(exit);
  }

  {
    Label allocated;

    // found free slot?
    __ bne(A1, R0, allocated);
    __ delayed()->nop();

    __ add_monitor_to_stack(A2, A3);
    __ move(A1, Rmonitors);

    __ bind(allocated);
  }

  // Increment bcp to point to the next bytecode, so exception handling for async. exceptions work correctly.
  // The object has already been popped from the stack, so the expression stack looks correct.
  __ addu_long(Rbcp, Rbcp, 1);

  __ st_ptr(Rtos, BasicObjectLock::obj_offset_in_bytes(), A1);
  __ lock_object(A1, Rtos);

  // check if there's enough space on the stack for the monitors after locking
  __ generate_stack_overflow_check(0);

  // The bcp has already been incremented. Just need to dispatch to next instruction.
  __ dispatch_next(vtos); 
}


void TemplateTable::monitorexit() {
  transition(atos, vtos);
  __ verify_oop(Rtos);

  Label ok1;
  __ bne(Rtos, R0, ok1);
  __ delayed()->nop();
  __ jump_to(Interpreter::_throw_NullPointerException_entry);
  __ delayed()->nop();
  __ bind(ok1);

  {
    Label entry, loop, found;
    __ get_top_most_monitor(A2);  // last one to check
    __ b(entry);
    __ delayed()->move(Rsave0, Rmonitors); 
    // Use Rsave0 to hold monitor elem to check, start with most recent monitor.

    __ bind(loop);
    __ verify_oop(A4);	     // verify each monitor's oop
    __ beq(Rtos, A4, found); // check if current entry is for desired object
    __ delayed()->move(A1, Rsave0); // pass found entry as argument to monitorexit

    __ addu_long(Rsave0, Rsave0, frame::interpreter_frame_monitor_size() * wordSize ); // advance to next

    __ bind(entry);
    __ bleu(Rsave0, A2, loop);
    __ delayed()->ld_ptr(A4, BasicObjectLock::obj_offset_in_bytes(), Rsave0);

    // ran out of monitors to check
    call_VM(noreg, CAST_FROM_FN_PTR(address, InterpreterRuntime::throw_illegal_monitor_state_exception));
    __ should_not_reach_here();

    __ bind(found);
  }
  __ unlock_object(A1);
}


//----------------------------------------------------------------------------------------------------
// Wide instructions

void TemplateTable::wide() {
  transition(vtos, vtos);
  __ lbu(Rscratch3, 1, Rbcp);  // get next bytecode
  __ sll_long(Rscratch3, Rscratch3, LogBytesPerWord);
  __ li(Rscratch4, (address)Interpreter::_wentry_point);
  __ ld_ptr(Rscratch3, Rscratch3, Rscratch4);
  __ jr(Rscratch3);
  __ delayed()->nop();
  // Note: the Rbcp increment step is part of the individual wide bytecode implementations
}


//----------------------------------------------------------------------------------------------------
// Multi arrays

void TemplateTable::multianewarray() {
  transition(vtos, atos);
     // put ndims * wordSize into Rsave0
  __ lbu(Rsave0, 3, Rbcp);
  __ sll_long(Rsave0, Rsave0, Interpreter::logStackElementSize());
     // Resp points past last_dim, so set A1 to first_dim address
  __ addu_long(A1, Resp, Rsave0); 
     call_VM(Rtos, CAST_FROM_FN_PTR(address, InterpreterRuntime::multianewarray), A1);
  __ addu_long(Resp, Resp, Rsave0); // pop all dimensions off the stack
}
