#ifdef USE_PRAGMA_IDENT_SRC
#pragma ident "@(#)vm_version_sparc.cpp	1.56 07/07/02 18:40:59 JVM"
#endif
/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

# include "incls/_precompiled.incl"
# include "incls/_vm_version_mips.cpp.incl"

VM_Version::ProcInfo VM_Version::_proc_info;
const char*          VM_Version::_features_str = "";

static BufferBlob* stub_blob;
static const int stub_size = 300;  // YYY find real size and adjust

extern "C" {
  typedef void (*getPsrInfo_stub_t)(void*);
}
static getPsrInfo_stub_t getPsrInfo_stub = NULL;


class VM_Version_StubGenerator: public StubCodeGenerator {
 public:

  VM_Version_StubGenerator(CodeBuffer *c) : StubCodeGenerator(c) {}

  address generate_getPsrInfo() {
    StubCodeMark mark(this, "VM_Version", "getPsrInfo_stub");
#   define __ _masm->

    address start = __ pc();

    // A0 should be &_proc_info
#if 0
    // Unfortuantely, mfc0 is not available in user mode.
    __ mfc0(Rscratch1, 15, 0);
    __ sw(Rscratch1, in_bytes(VM_Version::procinfo_proc_reg_offset()), A0);
    __ mfc0(Rscratch1, 16, 0);
    __ sw(Rscratch1, in_bytes(VM_Version::procinfo_config_reg_offset()), A0);
    __ mfc0(Rscratch1, 16, 1);
    __ sw(Rscratch1, in_bytes(VM_Version::procinfo_config1_reg_offset()), A0);
#endif
    __ j(RA);
    __ delayed()->nop();

    return start;

#   undef __
  };
};


void VM_Version::get_processor_features() {
  getPsrInfo_stub(&_proc_info);

  char buf[256];
  jio_snprintf(buf, sizeof(buf), "processor type %u revision %u (company %u, options 0x%x) arch %u rev %u, %s-endian, C1 %s, C2 %s",
	       _proc_info.processor_reg.bits.processorID,
	       _proc_info.processor_reg.bits.revision,
	       _proc_info.processor_reg.bits.companyID,
	       _proc_info.processor_reg.bits.companyOptions,
	       _proc_info.config_reg.bits.AT_,
	       _proc_info.config_reg.bits.AR,
	       (_proc_info.config_reg.bits.BE ? "big" : "little"),
	       (_proc_info.config1_reg.bits.FP_ ? "present" : "absent"),
	       (_proc_info.config1_reg.bits.C2 ? "present" : "absent"));
  _features_str = strdup(buf);

  // Set other VM flags that are processor-dependent
}

void VM_Version::initialize() {
  ResourceMark rm;
  // Making this stub must be FIRST use of assembler

  stub_blob = BufferBlob::create("getPsrInfo_stub", stub_size);
  if (stub_blob == NULL) {
    vm_exit_during_initialization("Unable to allocate getPsrInfo_stub");
  }
  CodeBuffer c(stub_blob->instructions_begin(),
               stub_blob->instructions_size());
  VM_Version_StubGenerator g(&c);
  getPsrInfo_stub = CAST_TO_FN_PTR(getPsrInfo_stub_t,
                                   g.generate_getPsrInfo());

  get_processor_features();
}
