#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)atomic_solaris_sparc.inline.hpp	1.22 07/05/05 17:04:53 JVM"
#endif
/*
 * Copyright 1999-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * Copyright 2007-2008 Reservoir Labs, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 *  
 */

// Implementation of class atomic

inline void Atomic::store    (jbyte    store_value, jbyte*    dest) { *dest = store_value; }
inline void Atomic::store    (jshort   store_value, jshort*   dest) { *dest = store_value; }
inline void Atomic::store    (jint     store_value, jint*     dest) { *dest = store_value; }
inline void Atomic::store    (jlong    store_value, jlong*    dest) { *dest = store_value; }
inline void Atomic::store_ptr(intptr_t store_value, intptr_t* dest) { *dest = store_value; }
inline void Atomic::store_ptr(void*    store_value, void*     dest) { *(void**) dest = store_value; }

inline void Atomic::store    (jbyte    store_value, volatile jbyte*    dest) { *dest = store_value; }
inline void Atomic::store    (jshort   store_value, volatile jshort*   dest) { *dest = store_value; }
inline void Atomic::store    (jint     store_value, volatile jint*     dest) { *dest = store_value; }
inline void Atomic::store    (jlong    store_value, volatile jlong*    dest) { *dest = store_value; }
inline void Atomic::store_ptr(intptr_t store_value, volatile intptr_t* dest) { *dest = store_value; }
inline void Atomic::store_ptr(void*    store_value, volatile void*     dest) { *(void* volatile *) dest = store_value; }

inline void Atomic::inc    (volatile jint*     dest) { (void)add    (1, dest); }
inline void Atomic::inc_ptr(volatile intptr_t* dest) { (void)add_ptr(1, dest); }
inline void Atomic::inc_ptr(volatile void*     dest) { (void)add_ptr(1, dest); }

inline void Atomic::dec    (volatile jint*     dest) { (void)add    (-1, dest); }
inline void Atomic::dec_ptr(volatile intptr_t* dest) { (void)add_ptr(-1, dest); }
inline void Atomic::dec_ptr(volatile void*     dest) { (void)add_ptr(-1, dest); }

#ifdef _GNU_SOURCE

#define ASM(reg) reg
#define t0 "$8"

inline jint Atomic::add(jint add_value, volatile jint* dest) {
  intptr_t rv;
  __asm__ volatile(
    "1: ll     "ASM(t0)", 0(%2)    \n\t"
#ifdef SICORTEX_ERRATA
    "   ll     "ASM(t0)", 0(%2)    \n\t"
#endif
    "   addu   "ASM(t0)", "ASM(t0)", %1   \n\t"
    "   move   %0, "ASM(t0)"       \n\t"
    "   sc     "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"
    : "=&r" (rv)
    : "r" (add_value), "r" (dest)
    : "memory", "t0");
  return rv;
}

inline intptr_t Atomic::add_ptr(intptr_t add_value, volatile intptr_t* dest) {
  intptr_t rv;
  __asm__ volatile(
#ifdef _LP64

    "1: lld    "ASM(t0)", 0(%2)    \n\t"
#ifdef SICORTEX_ERRATA
    "   lld    "ASM(t0)", 0(%2)    \n\t"
#endif
    "   addu   "ASM(t0)", "ASM(t0)", %1 \n\t"
    "   move   %0, "ASM(t0)"       \n\t"
    "   scd    "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"

#else

    "1: ll     "ASM(t0)", 0(%2)    \n\t"
# ifdef SICORTEX_ERRATA
    "   ll     "ASM(t0)", 0(%2)    \n\t"
# endif
    "   addu   "ASM(t0)", "ASM(t0)", %1 \n\t"
    "   move   %0, "ASM(t0)"       \n\t"
    "   sc     "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"

#endif  // _LP64

    : "=&r" (rv)
    : "r" (add_value), "r" (dest)
    : "memory", "t0"); // asm volatile

  return rv;
}

inline void* Atomic::add_ptr(intptr_t add_value, volatile void* dest) {
  return (void*)add_ptr((intptr_t)add_value, (volatile intptr_t*)dest);
}


inline jint Atomic::xchg(jint exchange_value, volatile jint* dest) {
  intptr_t rv;
  __asm__ volatile(
    "1: ll     %0, 0(%2)      \n\t"
#ifdef SICORTEX_ERRATA
    "   ll     %0, 0(%2)      \n\t"
#endif
    "   move   "ASM(t0)", %1       \n\t"
    "   sc     "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"
    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest)
    : "memory", "t0");
  return rv;
}

inline intptr_t Atomic::xchg_ptr(intptr_t exchange_value, volatile intptr_t* dest) {
  intptr_t rv;
  __asm__ volatile(
#ifdef _LP64

    "1: lld    %0, 0(%2)      \n\t"
#ifdef SICORTEX_ERRATA
    "   lld    %0, 0(%2)      \n\t"
#endif
    "   move   "ASM(t0)", %1       \n\t"
    "   scd    "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"

#else

    "1: ll     %0, 0(%2)      \n\t"
#ifdef SICORTEX_ERRATA
    "   ll     %0, 0(%2)      \n\t"
#endif
    "   move   "ASM(t0)", %1       \n\t"
    "   sc     "ASM(t0)", 0(%2)    \n\t"
    "   beqz   "ASM(t0)", 1b       \n\t"
    "   nop                        \n\t"

#endif  // _LP64

    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest)
    : "memory", "t0"); // asm volatile

  return rv;
}

inline void* Atomic::xchg_ptr(void* exchange_value, volatile void* dest) {
  return (void*)xchg_ptr((intptr_t)exchange_value, (volatile intptr_t*)dest);
}

inline jint Atomic::cmpxchg(jint exchange_value, volatile jint* dest, jint compare_value) {
  jint rv;
  __asm__ volatile(
    "1: ll %0, 0(%2)            \n\t"
#ifdef SICORTEX_ERRATA
    "   ll %0, 0(%2)            \n\t"
#endif
    "   bne %0, %3, 2f          \n\t"
    "   move "ASM(t0)", %1      \n\t"
    "   sc   "ASM(t0)", 0(%2)   \n\t"
    "   beqz "ASM(t0)", 1b      \n\t"
    "2:  nop                    \n\t"
    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest), "r" (compare_value)
    : "memory", "t0");
  return rv;
}

inline jlong Atomic::cmpxchg(jlong exchange_value, volatile jlong* dest, jlong compare_value) {
#ifdef _LP64
  jlong rv;
  __asm__ volatile(
    "1: lld %0, 0(%2)           \n\t"
#ifdef SICORTEX_ERRATA
    "   lld %0, 0(%2)           \n\t"
#endif
    "   bne %0, %3, 2f          \n\t"
    "   move "ASM(t0)", %1      \n\t"
    "   scd  "ASM(t0)", 0(%2)   \n\t"
    "   beqz "ASM(t0)", 1b      \n\t"
    "2:  nop                    \n\t"
    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest), "r" (compare_value)
    : "memory", "t0");
  return rv;
#else
#error unimplemented
#endif // _LP64
}

inline intptr_t Atomic::cmpxchg_ptr(intptr_t exchange_value, volatile intptr_t* dest, intptr_t compare_value) {
  intptr_t rv;
  __asm__ volatile(
#ifdef _LP64

    "1: lld %0, 0(%2)           \n\t"
#ifdef SICORTEX_ERRATA
    "   lld %0, 0(%2)           \n\t"
#endif
    "   bne %0, %3, 2f          \n\t"
    "   move "ASM(t0)", %1      \n\t"
    "   scd  "ASM(t0)", 0(%2)    \n\t"
    "   beqz "ASM(t0)", 1b      \n\t"
    "2:  nop                    \n\t"
    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest), "r" (compare_value)
    : "memory", "t0");

#else

    "1: ll  %0, 0(%2)           \n\t"
#ifdef SICORTEX_ERRATA
    "   ll  %0, 0(%2)           \n\t"
#endif
    "   bne %0, %3, 2f          \n\t"
    "   move "ASM(t0)", %1      \n\t"
    "   sc   "ASM(t0)", 0(%2)   \n\t"
    "   beqz "ASM(t0)", 1b      \n\t"
    "2:  nop                    \n\t"
    : "=&r" (rv)
    : "r" (exchange_value), "r" (dest), "r" (compare_value)
    : "memory", "t0");

#endif  // _LP64
  return rv;
}

#undef t0
#undef ASM

inline void* Atomic::cmpxchg_ptr(void* exchange_value, volatile void* dest, void* compare_value) {
  return (void*)cmpxchg_ptr((intptr_t)exchange_value, (volatile intptr_t*)dest, (intptr_t)compare_value);
}

#endif // _GNU_SOURCE

